# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
from toardb.contacts.models import Person, Organisation, Contact
# Required imports 'create_test_database'
from toardb.test_base import (
    client,
    get_test_db,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)

class TestApps:
    def setup(self):
        self.application_url = "/contacts/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        # id_seq will not be reset automatically between tests!
        _db_conn = get_test_engine()
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        fake_cur.execute("ALTER SEQUENCE persons_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE organisations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE contacts_id_seq RESTART WITH 1")
        fake_conn.commit()
        infilename = "tests/fixtures/contacts/persons.json"
        with open(infilename, encoding="utf-8") as f:
            metajson=json.load(f)
            for entry in metajson:
                new_person = Person(**entry)
                db.add(new_person)
                db.commit()
                db.refresh(new_person)
        infilename = "tests/fixtures/contacts/organisations.json"
        with open(infilename, encoding="utf-8") as f:
            metajson=json.load(f)
            for entry in metajson:
                new_organisation = Organisation(**entry)
                db.add(new_organisation)
                db.commit()
                db.refresh(new_organisation)
        infilename = "tests/fixtures/contacts/contacts.json"
        with open(infilename, encoding="utf-8") as f:
            metajson=json.load(f)
            for entry in metajson:
                new_contact = Contact(**entry)
                db.add(new_contact)
                db.commit()
                db.refresh(new_contact)

    def test_get_organisations(self, client, db):
        response = client.get("/contacts/organisations/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{"id":0,"name":"","longname":"",
                          "kind":"other","city":"","postcode":"",
                          "street_address":"","country":"N/A",
                          "homepage":"", "contact_url": "" },
                         {"id":1,"name":"UBA","longname":"Umweltbundesamt",
                          "kind":"government","city":"Dessau-Roßlau","postcode":"06844",
                          "street_address":"Wörlitzer Platz 1","country":"Germany",
                          "homepage":"https://www.umweltbundesamt.de",
                          "contact_url": "mailto:immission@uba.de"},
                         {"id":2,"name":"FZJ","longname":"Forschungszentrum Jülich",
                          "kind":"research","city":"Jülich","postcode":"52425",
                          "street_address":"Wilhelm-Johnen-Straße","country":"Germany",
                          "homepage":"https://www.fz-juelich.de",
                          "contact_url": "mailto:toar-data@fz-juelich.de"}]
        assert response.json() == expected_resp


    def test_get_all_organisations(self, client, db):
        response = client.get("/contacts/organisations/?limit=None")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{"id":0,"name":"","longname":"",
                          "kind":"other","city":"","postcode":"",
                          "street_address":"","country":"N/A",
                          "homepage":"", "contact_url": "" },
                         {"id":1,"name":"UBA","longname":"Umweltbundesamt",
                          "kind":"government","city":"Dessau-Roßlau","postcode":"06844",
                          "street_address":"Wörlitzer Platz 1","country":"Germany",
                          "homepage":"https://www.umweltbundesamt.de",
                          "contact_url": "mailto:immission@uba.de"},
                         {"id":2,"name":"FZJ","longname":"Forschungszentrum Jülich",
                          "kind":"research","city":"Jülich","postcode":"52425",
                          "street_address":"Wilhelm-Johnen-Straße","country":"Germany",
                          "homepage":"https://www.fz-juelich.de",
                          "contact_url": "mailto:toar-data@fz-juelich.de"}]
        assert response.json() == expected_resp


    def test_get_special_organisation(self, client, db):
        response = client.get("/contacts/organisations/id/2")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {"id":2,"name":"FZJ","longname":"Forschungszentrum Jülich",
                         "kind":"research","city":"Jülich","postcode":"52425",
                         "street_address":"Wilhelm-Johnen-Straße","country":"Germany",
                         "homepage":"https://www.fz-juelich.de",
                         "contact_url": "mailto:toar-data@fz-juelich.de"}
        assert response.json() == expected_resp


    def test_get_special_organisation_out_of_index(self, client, db):
        response = client.get("/contacts/organisations/id/99")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail":"Organisation not found."}
        assert response.json() == expected_resp


    def test_get_special_organisation_by_name(self, client, db):
        response = client.get("/contacts/organisations/FZJ")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {"id":2,"name":"FZJ","longname":"Forschungszentrum Jülich",
                         "kind":"research","city":"Jülich","postcode":"52425",
                         "street_address":"Wilhelm-Johnen-Straße","country":"Germany",
                         "homepage":"https://www.fz-juelich.de",
                         "contact_url": "mailto:toar-data@fz-juelich.de"}
        assert response.json() == expected_resp


#   def test_insert_new_organisation_without_credits(self):
#?      response = client.post("/contacts/organisations/")
#       expected_status_code=401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp



#   def test_insert_new_organisation_wrong_credits(self):
#?      response = client.post("/contacts/organisations/")
#       expected_status_code = 401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp


    def test_insert_new_organisation(self, client, db):
        response = client.post("/contacts/organisations/",
                json= {"organisation":
                          {"name": "FZJ2", "longname": "Forschungszentrum Test",
                           "kind": "Research", "city": "Jülich", "postcode": "52425",
                           "street_address": "Wilhelm-Johnen-Straße",
                           "country": "DE",
                           "homepage": "https://www.fz-juelich.de",
                           "contact_url": "mailto:toar-data2@fz-juelich.de"}
                      }
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {"id":3,"name":"FZJ2","longname":"Forschungszentrum Test",
                         "kind":"research","city":"Jülich","postcode":"52425",
                         "street_address":"Wilhelm-Johnen-Straße","country":"Germany",
                         "homepage":"https://www.fz-juelich.de",
                         "contact_url": "mailto:toar-data2@fz-juelich.de"}
        assert response.json() == expected_resp
 
                                                    
    def test_insert_duplicate_organisation(self, client, db):
        response = client.post("/contacts/organisations/",
                json= {"organisation":
                          {"name": "FZJ", "longname": "Forschungszentrum Jülich",
                           "kind": "Research", "city": "Jülich", "postcode": "52425",
                           "street_address": "Wilhelm-Johnen-Straße",
                           "country": "DE",
                           "homepage": "https://www.fz-juelich.de",
                           "contact_url": "mailto:toar-data2@fz-juelich.de"}
                      }
                   )
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Organisation already registered.'}
        assert response.json() == expected_resp


    def test_get_persons(self, client, db):
        response = client.get("/contacts/persons/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                          'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False}]
        assert response.json() == expected_resp


    def test_get_special_person(self, client, db):
        response = client.get("/contacts/persons/?id=3")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                         'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False}]
        assert response.json() == expected_resp


    def test_get_special_person_out_of_index(self, client, db):
        response = client.get("/contacts/persons/id/3")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                         'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False}
        assert response.json() == expected_resp


    def test_get_special_person_out_of_index_not_found(self, client, db):
        response = client.get("/contacts/persons/id/99")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail":"Person not found."}
        assert response.json() == expected_resp


    def test_get_special_person_by_name_url(self, client, db):
        response = client.get("/contacts/persons/Sabine Schröder")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                         'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False}
        assert response.json() == expected_resp


    def test_get_special_person_by_name(self, client, db):
        response = client.get("/contacts/persons/?name=Sabine Schröder")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                         'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False}]
        assert response.json() == expected_resp


    def test_get_special_person_by_name_using_fields(self, client, db):
        response = client.get("/contacts/persons/?name=Sabine Schröder&fields=orcid")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'orcid': '0000-0002-0309-8010'}]
        assert response.json() == expected_resp


    def test_get_special_person_using_wrong_fieldname(self, client, db):
        response = client.get("/contacts/persons/?name=Sabine Schröder&fields=orcid,bla")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "Wrong field given: bla"
        assert response.json() == expected_resp


#   def test_insert_new_person_without_credits(self):
#?      response = client.post("/contacts/persons/")
#       expected_status_code=401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp



#   def test_insert_new_person_wrong_credits(self):
#?      response = client.post("/contacts/persons/")
#       expected_status_code = 401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp


    def test_insert_new_person(self, client, db):
        response = client.post("/contacts/persons/",
                json= {'person':
                          {'name': 'Martin Schultz', 'email': 'm.schultz@fz-juelich.de',
                           'phone': '+49-2461-61-96870', 'orcid': '0000-0003-3455-774X', 'isprivate': True}
                      }
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 4, 'name': 'Martin Schultz', 'email': 'm.schultz@fz-juelich.de',
                         'phone': '+49-2461-61-96870', 'orcid': '0000-0003-3455-774X', 'isprivate': True}
        assert response.json() == expected_resp

                                                    
    def test_insert_duplicate_person(self, client, db):
        response = client.post("/contacts/persons/",
                json= {'person':
                          {'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                           'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': True}
                      }
                   )
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Person already registered.'}
        assert response.json() == expected_resp


## also test requiring all contacts (a contact is either an organisation or a person)
    def test_get_contacts(self, client, db):
        response = client.get("/contacts/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 3,
                          'person': {'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                          'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False},
                          'organisation': None},
                         {'id': 4,
                          'person': None,
                          'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt', 'kind': 'government',
                          'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                          'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de',
                          'contact_url': 'mailto:immission@uba.de'}},
                         {'id': 5,
                          'person': None,
                          'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich', 'kind': 'research',
                          'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                          'country': 'Germany', 'homepage': 'https://www.fz-juelich.de',
                          'contact_url': 'mailto:toar-data@fz-juelich.de'}}]
        assert response.json() == expected_resp


    def test_get_all_contacts(self, client, db):
        response = client.get("/contacts/?limit=None")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 3,
                          'person': {'id': 3, 'name': 'Sabine Schröder', 'email': 's.schroeder@fz-juelich.de',
                          'phone': '+49-2461-61-6397', 'orcid': '0000-0002-0309-8010', 'isprivate': False},
                          'organisation': None},
                         {'id': 4,
                          'person': None,
                          'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt', 'kind': 'government',
                          'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                          'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de',
                          'contact_url': 'mailto:immission@uba.de'}},
                         {'id': 5,
                          'person': None,
                          'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich', 'kind': 'research',
                          'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                          'country': 'Germany', 'homepage': 'https://www.fz-juelich.de',
                          'contact_url': 'mailto:toar-data@fz-juelich.de'}}]
        assert response.json() == expected_resp


    def test_get_person_not_found(self, client, db):
        response = client.get("/contacts/persons/Wrong Name")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Person not found."}
        assert response.json() == expected_resp


    def test_get_contact_by_person_name(self, client, db):
        response = client.get("/contacts/person_name/Sabine Schröder")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 3,
                         'person': {'id': 3, 'name':
                                    'Sabine Schröder',
                                    'email': 's.schroeder@fz-juelich.de',
                                    'phone': '+49-2461-61-6397',
                                    'orcid': '0000-0002-0309-8010',
                                    'isprivate': False
                                   },
                         'organisation': None}
        assert response.json() == expected_resp


    def test_get_contact_by_person_email(self, client, db):
        response = client.get("/contacts/person_email/s.schroeder@fz-juelich.de")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 3,
                         'person': {'id': 3, 'name':
                                    'Sabine Schröder',
                                    'email': 's.schroeder@fz-juelich.de',
                                    'phone': '+49-2461-61-6397',
                                    'orcid': '0000-0002-0309-8010',
                                    'isprivate': False
                                   },
                         'organisation': None}
        assert response.json() == expected_resp


    def test_get_organisation_not_found(self, client, db):
        response = client.get("/contacts/organisations/Wrong Organisation")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Organisation not found."}
        assert response.json() == expected_resp


    def test_get_contact_by_id(self, client, db):
        response = client.get("/contacts/id/3")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {
            'id': 3,
            'name': 'Sabine Schröder',
            'email': 's.schroeder@fz-juelich.de',
            'phone': '+49-2461-61-6397',
            'orcid': '0000-0002-0309-8010',
            'isprivate': False
        }
        assert response.json() == expected_resp


    def test_get_all_contacts_by_orga_name(self, client, db):
        response = client.get("/contacts/orga_name/Umweltbundesamt")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {
            'id': 4,
            'person': None,
            'organisation':
                {'id': 1,
                 'name': 'UBA',
                 'longname': 'Umweltbundesamt',
                 'kind': 'government',
                 'city': 'Dessau-Roßlau',
                 'postcode': '06844',
                 'street_address': 'Wörlitzer Platz 1',
                 'country': 'Germany',
                 'homepage': 'https://www.umweltbundesamt.de',
                 'contact_url': 'mailto:immission@uba.de'
                 }
        }
        assert response.json() == expected_resp


    def test_get_contact_not_found(self, client, db):
        response = client.get("/contacts/orga_name/Luftbundesamt")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Contact not found.'}
        assert response.json() == expected_resp
