# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
import pandas as pd
from sqlalchemy import insert
from fastapi import Request
from toardb.toardb import app
from toardb.data.models import Data
from toardb.timeseries.models import Timeseries, timeseries_timeseries_roles_table
from toardb.timeseries.models_programme import TimeseriesProgramme
from toardb.timeseries.models_role import TimeseriesRole
from toardb.stationmeta.models import StationmetaCore
from toardb.stationmeta.schemas import get_geom_from_coordinates, Coordinates
from toardb.variables.models import Variable
from toardb.contacts.models import Person, Organisation, Contact
from toardb.auth_user.models import AuthUser
from toardb.test_base import (
    client,
    get_test_db,
    override_dependency,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)
# for mocking datetime.now(timezone.utc)
from datetime import datetime
from unittest.mock import patch

# only datetime.now needs to be overridden because otherwise daterange-arguments would be provided as MagicMock-objects!
class FixedDatetime(datetime):
    @classmethod
    def now(cls, tz=None):
        return datetime(2023, 7, 28, 12, 0, 0)


class TestApps:
    def setup(self):
        self.application_url = "/data/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        _db_conn = get_test_engine()
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        # id_seq will not be reset automatically between tests!
        fake_cur.execute("ALTER SEQUENCE auth_user_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE variables_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_core_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_global_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_roles_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_doc_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_image_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_url_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE persons_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE organisations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE contacts_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_roles_id_seq RESTART WITH 3")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_programmes_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_id_seq RESTART WITH 1")
        fake_conn.commit()
        infilename = "tests/fixtures/auth_user/auth.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_auth_user = AuthUser(**entry)
                db.add(new_auth_user)
                db.commit()
                db.refresh(new_auth_user)
        infilename = "tests/fixtures/contacts/persons.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_person = Person(**entry)
                db.add(new_person)
                db.commit()
                db.refresh(new_person)
        infilename = "tests/fixtures/contacts/organisations.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_organisation = Organisation(**entry)
                db.add(new_organisation)
                db.commit()
                db.refresh(new_organisation)
        infilename = "tests/fixtures/contacts/contacts.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_contact = Contact(**entry)
                db.add(new_contact)
                db.commit()
                db.refresh(new_contact)
        infilename = "tests/fixtures/variables/variables.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_variable = Variable(**entry)
                db.add(new_variable)
                db.commit()
                db.refresh(new_variable)
        infilename = "tests/fixtures/stationmeta/stationmeta_core.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_core = StationmetaCore(**entry)
                # there's a mismatch with coordinates --> how to automatically switch back and forth?!
                new_stationmeta_core.coordinates = get_geom_from_coordinates(Coordinates(**new_stationmeta_core.coordinates))
                db.add(new_stationmeta_core)
                db.commit()
                db.refresh(new_stationmeta_core)
        infilename = "tests/fixtures/timeseries/timeseries_programmes.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_programme = TimeseriesProgramme(**entry)
                db.add(new_timeseries_programme)
                db.commit()
                db.refresh(new_timeseries_programme)
        infilename = "tests/fixtures/timeseries/timeseries.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries = Timeseries(**entry)
                db.add(new_timeseries)
                db.commit()
                db.refresh(new_timeseries)
        infilename = "tests/fixtures/timeseries/timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_role = TimeseriesRole(**entry)
                db.add(new_timeseries_role)
                db.commit()
                db.refresh(new_timeseries_role)
        infilename = "tests/fixtures/timeseries/timeseries_timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                db.execute(insert(timeseries_timeseries_roles_table).values(timeseries_id=entry["timeseries_id"], role_id=entry["role_id"]))
                db.execute("COMMIT")
        infilename = "tests/fixtures/data/data.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_data = Data (**entry)
                db.add(new_data)
                db.commit()
                db.refresh(new_data)
        infilename = "tests/fixtures/data/staging_data.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_data = Data (**entry)
                fake_cur.execute(("INSERT INTO staging.data(datetime, value, flags, timeseries_id, version) "
                                  f" VALUES('{new_data.datetime}', {new_data.value}, {new_data.flags}, "
                                  f"{new_data.timeseries_id}, '{new_data.version}');"))
                fake_conn.commit()


    def test_get_data(self, client, db):
        response = client.get("/data/timeseries/?limit=4")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                         {'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified', 'timeseries_id': 2, 'version': '1.0'},
                         {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                         {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 2, 'version': '1.0'}]
        assert response.json() == expected_resp
 

    def test_get_special(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)
        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/data/timeseries/1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'metadata': {'id': 1, 'label': 'CMA', 'order': 1, 'sampling_frequency': 'hourly',
                                      'aggregation': 'mean',
                                      'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/',
                                      'doi': '',
                                      'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                                      'data_origin': 'instrument', 'data_origin_type': 'measurement', 'provider_version': 'N/A',
                                      'sampling_height': 7.0, 'additional_metadata': {'original_units': 'ppb'},
                                      'station': {'id': 2, 'codes': ['SDZ54421'], 'name': 'Shangdianzi', 'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                                  'coordinate_validation_status': 'not checked', 'country': 'China', 'state': 'Beijing Shi', 'type': 'unknown',
                                                  'type_of_area': 'unknown', 'timezone': 'Asia/Shanghai',
                                                  'additional_metadata': {'add_type': 'nature reservation'},
                                                  'aux_images': [], 'aux_docs': [], 'aux_urls': [], 'changelog': []},
                                      'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                                   'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C7H8', 'id': 7},
                                      'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''},
                                      'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                                 'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                                                       'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                                                       'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}},
                                                {'id': 3, 'role': 'principal investigator', 'status': 'active',
                                                 'contact': {'id': 3, 'person': {'email': 's.schroeder@fz-juelich.de','id': 3, 'isprivate': False,
                                                                                 'name': 'Sabine Schröder', 'orcid': '0000-0002-0309-8010', 'phone': '+49-2461-61-6397'}}}
                                               ],
                                      'citation': 'Sabine Schröder: time series of toluene at Shangdianzi, accessed from the TOAR database on 2023-07-28 12:00:00'},
                         'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'OK validated QC passed', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated QC passed', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated modified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62,  'flags': 'OK validated QC passed', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T05:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified', 'version': '1.0', 'timeseries_id': 1},
                                  {'datetime': '2012-12-17T06:00:00+00:00', 'value':  5.886, 'flags': 'OK validated verified', 'version': '1.0', 'timeseries_id': 1}]
                        }
        assert response.json() == expected_resp


    # the data/map-endpoint is a special need of the analysis service
    def test_get_map_data(self, client, db):
        response = client.get("/data/map/?variable_id=7&daterange=2012-12-16T21:00,2012-12-17T06:00")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'timeseries_id': 1, 'value': 21.581},
                         {'timeseries_id': 1, 'value': 13.734},
                         {'timeseries_id': 1, 'value': 13.734},
                         {'timeseries_id': 1, 'value': 7.848},
                         {'timeseries_id': 1, 'value': 15.696},
                         {'timeseries_id': 1, 'value': 11.772},
                         {'timeseries_id': 1, 'value': 13.734},
                         {'timeseries_id': 1, 'value': 19.62},
                         {'timeseries_id': 1, 'value': 15.696},
                         {'timeseries_id': 1, 'value': 5.886}]
        assert response.json() == expected_resp
  

    def test_get_data_with_fields(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)
        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/data/timeseries/1?fields=datetime,value")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'metadata': {'id': 1,
                                      'label': 'CMA',
                                      'order': 1,
                                      'sampling_frequency': 'hourly',
                                      'aggregation': 'mean',
                                      'data_start_date': '2003-09-07T15:30:00+00:00',
                                      'data_end_date': '2016-12-31T14:30:00+00:00',
                                      'data_origin': 'instrument',
                                      'data_origin_type': 'measurement',
                                      'provider_version': 'N/A',
                                      'sampling_height': 7.0,
                                      'additional_metadata': {'original_units': 'ppb'},
                                      'doi': '',
                                      'coverage': -1.0,
                                      'station': {'id': 2,
                                                  'codes': ['SDZ54421'],
                                                  'name': 'Shangdianzi',
                                                  'coordinates': {'lat': 40.65,
                                                                  'lng': 117.106,
                                                                  'alt': 293.9},
                                                  'coordinate_validation_status': 'not checked',
                                                  'country': 'China',
                                                  'state': 'Beijing Shi',
                                                  'type': 'unknown',
                                                  'type_of_area': 'unknown',
                                                  'timezone': 'Asia/Shanghai',
                                                  'additional_metadata': {'add_type': 'nature reservation'},
                                                  'aux_images': [],
                                                  'aux_docs': [],
                                                  'aux_urls': [],
                                                  'changelog': []},
                                      'variable': {'name': 'toluene',
                                                   'longname': 'toluene',
                                                   'displayname': 'Toluene',
                                                   'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                                   'units': 'nmol mol-1',
                                                   'chemical_formula': 'C7H8',
                                                   'id': 7},
                                      'programme': {'id': 0,
                                                    'name': '',
                                                    'longname': '',
                                                    'homepage': '',
                                                    'description': ''},
                                      'roles': [{'id': 2,
                                                 'role': 'resource provider',
                                                 'status': 'active',
                                                 'contact': {'id': 4,
                                                             'organisation': {'id': 1,
                                                                              'name': 'UBA',
                                                                              'longname': 'Umweltbundesamt',
                                                                              'kind': 'government',
                                                                              'city': 'Dessau-Roßlau',
                                                                              'postcode': '06844',
                                                                              'street_address': 'Wörlitzer Platz 1',
                                                                              'country': 'Germany',
                                                                              'homepage': 'https://www.umweltbundesamt.de',
                                                                              'contact_url': 'mailto:immission@uba.de'}}},
                                                {'id': 3,
                                                 'role': 'principal investigator',
                                                 'status': 'active',
                                                 'contact': {'id': 3,
                                                             'person': {'email': 's.schroeder@fz-juelich.de',
                                                                        'id': 3,
                                                                        'isprivate': False,
                                                                        'name': 'Sabine Schröder',
                                                                        'orcid': '0000-0002-0309-8010',
                                                                        'phone': '+49-2461-61-6397'}}}],
                                      'citation': 'Sabine Schröder: time series of toluene at '
                                                  'Shangdianzi, accessed from the TOAR database on '
                                                  '2023-07-28 12:00:00',
                                      'license': 'This data is published under a Creative Commons '
                                                 'Attribution 4.0 International (CC BY 4.0). '
                                                 'https://creativecommons.org/licenses/by/4.0/'},
                         'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581},
                                  {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734},
                                  {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734},
                                  {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848},
                                  {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696},
                                  {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772},
                                  {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734},
                                  {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62},
                                  {'datetime': '2012-12-17T05:00:00+00:00', 'value': 15.696},
                                  {'datetime': '2012-12-17T06:00:00+00:00', 'value':  5.886}]
                         }
        assert response.json() == expected_resp


    def test_get_data_with_fields_limited(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)
        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/data/timeseries/1?fields=datetime,value&limit=4")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'metadata': {'id': 1,
                                      'label': 'CMA',
                                      'order': 1,
                                      'sampling_frequency': 'hourly',
                                      'aggregation': 'mean',
                                      'data_start_date': '2003-09-07T15:30:00+00:00',
                                      'data_end_date': '2016-12-31T14:30:00+00:00',
                                      'data_origin': 'instrument',
                                      'data_origin_type': 'measurement',
                                      'provider_version': 'N/A',
                                      'sampling_height': 7.0,
                                      'additional_metadata': {'original_units': 'ppb'},
                                      'doi': '',
                                      'coverage': -1.0,
                                      'station': {'id': 2,
                                                  'codes': ['SDZ54421'],
                                                  'name': 'Shangdianzi',
                                                  'coordinates': {'lat': 40.65,
                                                                  'lng': 117.106,
                                                                  'alt': 293.9},
                                                  'coordinate_validation_status': 'not checked',
                                                  'country': 'China',
                                                  'state': 'Beijing Shi',
                                                  'type': 'unknown',
                                                  'type_of_area': 'unknown',
                                                  'timezone': 'Asia/Shanghai',
                                                  'additional_metadata': {'add_type': 'nature reservation'},
                                                  'aux_images': [],
                                                  'aux_docs': [],
                                                  'aux_urls': [],
                                                  'changelog': []},
                                      'variable': {'name': 'toluene',
                                                   'longname': 'toluene',
                                                   'displayname': 'Toluene',
                                                   'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                                   'units': 'nmol mol-1',
                                                   'chemical_formula': 'C7H8',
                                                   'id': 7},
                                      'programme': {'id': 0,
                                                    'name': '',
                                                    'longname': '',
                                                    'homepage': '',
                                                    'description': ''},
                                      'roles': [{'id': 2,
                                                 'role': 'resource provider',
                                                 'status': 'active',
                                                 'contact': {'id': 4,
                                                             'organisation': {'id': 1,
                                                                              'name': 'UBA',
                                                                              'longname': 'Umweltbundesamt',
                                                                              'kind': 'government',
                                                                              'city': 'Dessau-Roßlau',
                                                                              'postcode': '06844',
                                                                              'street_address': 'Wörlitzer Platz 1',
                                                                              'country': 'Germany',
                                                                              'homepage': 'https://www.umweltbundesamt.de',
                                                                              'contact_url': 'mailto:immission@uba.de'}}},
                                                {'id': 3,
                                                 'role': 'principal investigator',
                                                 'status': 'active',
                                                 'contact': {'id': 3, 'person': {'email': 's.schroeder@fz-juelich.de',
                                                                                 'id': 3,
                                                                                 'isprivate': False,
                                                                                 'name': 'Sabine Schröder',
                                                                                 'orcid': '0000-0002-0309-8010',
                                                                                 'phone': '+49-2461-61-6397'}}}],
                                      'citation': 'Sabine Schröder: time series of toluene at '
                                                  'Shangdianzi, accessed from the TOAR database on '
                                                  '2023-07-28 12:00:00',
                                      'license': 'This data is published under a Creative Commons '
                                                 'Attribution 4.0 International (CC BY 4.0). '
                                                 'https://creativecommons.org/licenses/by/4.0/'},
                         'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581},
                                  {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734},
                                  {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734},
                                  {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848}]
                         }
        assert response.json() == expected_resp
 

    def test_get_data_as_csv_with_fields(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)
        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/data/timeseries/1?format=csv&fields=datetime,value")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = ''.join(['#{\n',
                        '#    "id": 1,\n',
                        '#    "label": "CMA",\n',
                        '#    "order": 1,\n',
                        '#    "sampling_frequency": "hourly",\n',
                        '#    "aggregation": "mean",\n',
                        '#    "data_start_date": "2003-09-07T15:30:00+00:00",\n',
                        '#    "data_end_date": "2016-12-31T14:30:00+00:00",\n',
                        '#    "data_origin": "instrument",\n',
                        '#    "data_origin_type": "measurement",\n',
                        '#    "provider_version": "N/A",\n',
                        '#    "sampling_height": 7.0,\n',
                        '#    "additional_metadata": {\n',
                        '#        "original_units": "ppb"\n',
                        '#    },\n',
                        '#    "data_license_accepted": null,\n',
                        '#    "dataset_approved_by_provider": null,\n',
                        '#    "doi": "",\n',
                        '#    "coverage": -1.0,\n',
                        '#    "station": {\n',
                        '#        "id": 2,\n',
                        '#        "codes": [\n',
                        '#            "SDZ54421"\n',
                        '#        ],\n',
                        '#        "name": "Shangdianzi",\n',
                        '#        "coordinates": {\n',
                        '#            "lat": 40.65,\n',
                        '#            "lng": 117.106,\n',
                        '#            "alt": 293.9\n',
                        '#        },\n',
                        '#        "coordinate_validation_status": "not checked",\n',
                        '#        "country": "China",\n',
                        '#        "state": "Beijing Shi",\n',
                        '#        "type": "unknown",\n',
                        '#        "type_of_area": "unknown",\n',
                        '#        "timezone": "Asia/Shanghai",\n',
                        '#        "additional_metadata": {\n',
                        '#            "add_type": "nature reservation"\n',
                        '#        },\n',
                        '#        "roles": null,\n',
                        '#        "annotations": null,\n',
                        '#        "aux_images": [],\n',
                        '#        "aux_docs": [],\n',
                        '#        "aux_urls": [],\n',
                        '#        "globalmeta": null,\n',
                        '#        "changelog": []\n',
                        '#    },\n',
                        '#    "variable": {\n',
                        '#        "name": "toluene",\n',
                        '#        "longname": "toluene",\n',
                        '#        "displayname": "Toluene",\n',
                        '#        "cf_standardname": "mole_fraction_of_toluene_in_air",\n',
                        '#        "units": "nmol mol-1",\n',
                        '#        "chemical_formula": "C7H8",\n',
                        '#        "id": 7\n',
                        '#    },\n',
                        '#    "programme": {\n',
                        '#        "id": 0,\n',
                        '#        "name": "",\n',
                        '#        "longname": "",\n',
                        '#        "homepage": "",\n',
                        '#        "description": ""\n',
                        '#    },\n',
                        '#    "roles": [\n',
                        '#        {\n',
                        '#            "id": 2,\n',
                        '#            "role": "resource provider",\n',
                        '#            "status": "active",\n',
                        '#            "contact": {\n',
                        '#                "id": 4,\n',
                        '#                "person": null,\n',
                        '#                "organisation": {\n',
                        '#                    "id": 1,\n',
                        '#                    "name": "UBA",\n',
                        '#                    "longname": "Umweltbundesamt",\n',
                        '#                    "kind": "government",\n',
                        '#                    "city": "Dessau-Roßlau",\n',
                        '#                    "postcode": "06844",\n',
                        '#                    "street_address": "Wörlitzer Platz 1",\n',
                        '#                    "country": "Germany",\n',
                        '#                    "homepage": "https://www.umweltbundesamt.de",\n',
                        '#                    "contact_url": "mailto:immission@uba.de"\n',
                        '#                }\n',
                        '#            }\n',
                        '#        },\n',
                        '#        {\n',
                        '#            "id": 3,\n',
                        '#            "role": "principal investigator",\n',
                        '#            "status": "active",\n',
                        '#            "contact": {\n',
                        '#                "id": 3,\n',
                        '#                "person": {\n',
                        '#                    "id": 3,\n',
                        '#                    "name": "Sabine Schröder",\n',
                        '#                    "email": "s.schroeder@fz-juelich.de",\n',
                        '#                    "phone": "+49-2461-61-6397",\n',
                        '#                    "orcid": "0000-0002-0309-8010",\n',
                        '#                    "isprivate": false\n',
                        '#                },\n',
                        '#                "organisation": null\n',
                        '#            }\n',
                        '#        }\n',
                        '#    ],\n',
                        '#    "annotations": null,\n',
                        '#    "changelog": null,\n',
                        '#    "citation": "Sabine Schröder: time series of toluene at Shangdianzi, accessed from the TOAR database on 2023-07-28 12:00:00",\n',
                        '#    "attribution": null,\n',
                        '#    "license": "This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/"\n',
                        '#}\n',
                        'datetime,value\n',
                        '2012-12-16 21:00:00+00:00,21.581\n',
                        '2012-12-16 22:00:00+00:00,13.734\n',
                        '2012-12-16 23:00:00+00:00,13.734\n',
                        '2012-12-17 00:00:00+00:00,7.848\n',
                        '2012-12-17 01:00:00+00:00,15.696\n',
                        '2012-12-17 02:00:00+00:00,11.772\n',
                        '2012-12-17 03:00:00+00:00,13.734\n',
                        '2012-12-17 04:00:00+00:00,19.62\n',
                        '2012-12-17 05:00:00+00:00,15.696\n',
                        '2012-12-17 06:00:00+00:00,5.886'])
        assert response.text == expected_resp


    def test_get_no_data_with_variable_and_timerage(self, client, db):
        # see: https://gitlab.jsc.fz-juelich.de/esde/toar-data/toardb_fastapi/-/issues/171
        response = client.get("/data/map/?variable_id=25&daterange=2012-12-16T21:00,2012-12-17T06:00")
#       expected_status_code = 404
        expected_status_code = 200
        assert response.status_code == expected_status_code
#       expected_response = {'detail': 'Data not found.'}
        expected_response = []
        assert response.json() == expected_response


    def test_insert_new_wrong_credentials(self, client, db):
        response = client.post("/data/timeseries/?toarqc_config_type=standard",
                               files={"file": open("tests/fixtures/data/toluene_SDZ54421_2013_2013_v1-0.dat", "rb")},
                               headers={"email": "j.doe@fz-juelich.de"})
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new_without_credentials(self, client, db):
        response = client.post("/data/timeseries/?toarqc_config_type=standard",
                               files={"file": open("tests/fixtures/data/toluene_SDZ54421_2013_2013_v1-0.dat", "rb")})
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new(self, client, db):
        response = client.post("/data/timeseries/?toarqc_config_type=standard",
                               files={"file": open("tests/fixtures/data/toluene_SDZ54421_2013_2013_v1-0.dat", "rb")},
                               headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'Data successfully inserted.'}}
        assert response.json() == expected_resp
  
                                                    
    def test_insert_duplicate(self, client, db):
        response = client.post("/data/timeseries/?toarqc_config_type=standard",
                               files={"file": open("tests/fixtures/data/toluene_SDZ54421_2012_2012_v1-0.dat", "rb")},
                               headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Data for timeseries already registered.'}
        assert response.json() == expected_resp


    def test_insert_new_as_bulk(self, client, db):
        df = pd.read_csv("tests/fixtures/data/toluene_SDZ54421_2013_2013_v1-0.dat",
                         delimiter=';', comment='#', header=None, names=["datetime", "value", "flags"], parse_dates=["datetime"])
        df['datetime'] = df['datetime'].dt.tz_localize('UTC')
        df['version'] = '000001.000000.00000000000000'
        df['timeseries_id'] = 2
        df_json=df.to_json(orient='records', date_format='iso')
        response = client.post("/data/timeseries/bulk/", data=df_json, headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'Data successfully inserted.'}}
        assert response.json() == expected_resp
 

    """def test_insert_missing_id(self, client, db):
        response = client.post("/data/timeseries/", files={"file": open("tests/fixtures/data/o3_SDZ54421_2012_2012_v1-0.dat", "rb")})
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = {'detail': '"Timeseries not found for station SDZ54421, variable o3, label CMA"'}
        assert response.json() == expected_resp"""


    def test_get_data_not_found(self, client, db):
        response = client.get("/data/timeseries/-1")
        expected_repsonse_code = 404
        assert response.status_code == expected_repsonse_code
        expected_response = {"detail": "Data not found."}
        assert response.json() == expected_response


    def test_get_data_invalid_type(self, client, db):
        response = client.get("/data/timeseries/A")
        expected_repsonse_code = 422
        assert response.status_code == expected_repsonse_code
        """expected_response = {"detail": "Data not found."}
        assert response.json() == expected_response"""


    def test_get_data2(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)
        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/data/timeseries/id/1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'metadata': {'id': 1, 'label': 'CMA', 'order': 1, 'sampling_frequency': 'hourly', 'aggregation': 'mean',
                                      'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/',
                                      'doi': '',
                                      'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0, 'data_origin': 'instrument',
                                      'data_origin_type': 'measurement', 'provider_version': 'N/A', 'sampling_height': 7.0, 'additional_metadata': {'original_units': 'ppb'},
                                      'station': {'id': 2, 'codes': ['SDZ54421'], 'name': 'Shangdianzi', 'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                                  'coordinate_validation_status': 'not checked', 'country': 'China', 'state': 'Beijing Shi', 'type': 'unknown',
                                                  'type_of_area': 'unknown', 'timezone': 'Asia/Shanghai',
                                                  'additional_metadata': {'add_type': 'nature reservation'},
                                                  'aux_images': [], 'aux_docs': [], 'aux_urls': [], 'changelog': []},
                                      'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                                   'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C7H8', 'id': 7},
                                      'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''},
                                      'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                                 'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                                                       'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                                                       'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}},
                                                {'id': 3, 'role': 'principal investigator', 'status': 'active',
                                                 'contact': {'id': 3, 'person': {'email': 's.schroeder@fz-juelich.de','id': 3, 'isprivate': False,
                                                                                 'name': 'Sabine Schröder', 'orcid': '0000-0002-0309-8010', 'phone': '+49-2461-61-6397'}}}],
                                      'citation': 'Sabine Schröder: time series of toluene at Shangdianzi, accessed from the TOAR database on 2023-07-28 12:00:00'},
                         'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated modified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62,  'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T05:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified', 'timeseries_id': 1, 'version': '1.0'},
                                  {'datetime': '2012-12-17T06:00:00+00:00', 'value':  5.886, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'}]}
        assert response.json() == expected_response


    def test_get_data2_not_found(self, client, db):
        response = client.get("/data/timeseries/id/-1")
        expected_repsonse_code = 404
        assert response.status_code == expected_repsonse_code
        expected_response = {"detail": "Data not found."}
        assert response.json() == expected_response

    def test_get_data2_invalid_type(self, client, db):
        response = client.get("/data/timeseries/id/A")
        expected_repsonse_code = 422
        assert response.status_code == expected_repsonse_code
        """expected_response = {"detail": "Data not found."}
        assert response.json() == expected_response"""


    def test_get_data_with_daterange(self, client, db):
        with patch('toardb.timeseries.crud.dt.datetime', FixedDatetime):
            response = client.get("/data/timeseries/id/2?flags=AllOK&daterange=2012-12-16%2023:00,%202012-12-17%2006:00")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'metadata': {'id': 2,
                                          'label': 'CMA',
                                          'order': 1,
                                          'sampling_frequency': 'hourly',
                                          'aggregation': 'mean',
                                          'data_start_date': '2003-09-07T15:30:00+00:00',
                                          'data_end_date': '2016-12-31T14:30:00+00:00',
                                          'data_origin': 'instrument',
                                          'data_origin_type': 'measurement',
                                          'provider_version': 'N/A',
                                          'sampling_height': 7.0,
                                          'additional_metadata': {'original_units': {'since_19740101000000': 'nmol/mol'},
                                                                  'measurement_method': 'uv_abs',
                                                                  'absorption_cross_section': 0,
                                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA', 'Data level': '2', 'Frameworks': 'GAW-WDCRG NOAA-ESRL', 'Station code': 'XXX', 'Station name': 'Secret'}},
                                          'doi': '',
                                          'coverage': -1.0,
                                          'station': {'id': 3,
                                                      'codes': ['China_test8'],
                                                      'name': 'Test_China',
                                                      'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                                      'coordinate_validation_status': 'not checked',
                                                      'country': 'China',
                                                      'state': 'Shandong Sheng',
                                                      'type': 'unknown',
                                                      'type_of_area': 'unknown',
                                                      'timezone': 'Asia/Shanghai',
                                                      'additional_metadata': {},
                                                      'aux_images': [],
                                                      'aux_docs': [],
                                                      'aux_urls': [],
                                                      'changelog': []},
                                          'variable': {'name': 'o3',
                                                       'longname': 'ozone',
                                                       'displayname': 'Ozone',
                                                       'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                                       'units': 'nmol mol-1',
                                                       'chemical_formula': 'O3',
                                                       'id': 5},
                                          'programme': {'id': 0,
                                                        'name': '',
                                                        'longname': '',
                                                        'homepage': '',
                                                        'description': ''},
                                          'roles': [{'id': 1,
                                                     'role': 'resource provider',
                                                     'status': 'active',
                                                     'contact': {'id': 5,
                                                                 'organisation': {'id': 2,
                                                                                  'name': 'FZJ',
                                                                                  'longname': 'Forschungszentrum Jülich',
                                                                                  'kind': 'research',
                                                                                  'city': 'Jülich',
                                                                                  'postcode': '52425',
                                                                                  'street_address': 'Wilhelm-Johnen-Straße',
                                                                                  'country': 'Germany',
                                                                                  'homepage': 'https://www.fz-juelich.de',
                                                                                  'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                                          'citation': 'Forschungszentrum Jülich: time series of o3 at Test_China, accessed from the TOAR database on 2023-07-28 12:00:00',
                                          'attribution': 'Test-Attributions to be announced',
                                          'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/'},
                             'data': [{'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'OK validated modified', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62, 'flags': 'OK validated verified', 'timeseries_id': 2, 'version': '1.0'}]}
        assert response.json() == expected_response


    def test_get_data_with_specific_flags(self, client, db):
        with patch('toardb.timeseries.crud.dt.datetime', FixedDatetime):
            response = client.get("/data/timeseries/id/2?flags=OKValidatedVerified,OKValidatedQCPassed&daterange=2012-12-16%2023:00,%202012-12-17%2006:00")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'metadata': {'id': 2,
                                          'label': 'CMA',
                                          'order': 1,
                                          'sampling_frequency': 'hourly',
                                          'aggregation': 'mean',
                                          'data_start_date': '2003-09-07T15:30:00+00:00',
                                          'data_end_date': '2016-12-31T14:30:00+00:00',
                                          'data_origin': 'instrument',
                                          'data_origin_type': 'measurement',
                                          'provider_version': 'N/A',
                                          'sampling_height': 7.0,
                                          'additional_metadata': {'original_units': {'since_19740101000000': 'nmol/mol'},
                                                                  'measurement_method': 'uv_abs',
                                                                  'absorption_cross_section': 0,
                                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA', 'Data level': '2', 'Frameworks': 'GAW-WDCRG NOAA-ESRL', 'Station code': 'XXX', 'Station name': 'Secret'}},
                                          'doi': '',
                                          'coverage': -1.0,
                                          'station': {'id': 3,
                                                      'codes': ['China_test8'],
                                                      'name': 'Test_China',
                                                      'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                                      'coordinate_validation_status': 'not checked',
                                                      'country': 'China',
                                                      'state': 'Shandong Sheng',
                                                      'type': 'unknown',
                                                      'type_of_area': 'unknown',
                                                      'timezone': 'Asia/Shanghai',
                                                      'additional_metadata': {},
                                                      'aux_images': [],
                                                      'aux_docs': [],
                                                      'aux_urls': [],
                                                      'changelog': []},
                                          'variable': {'name': 'o3',
                                                       'longname': 'ozone',
                                                       'displayname': 'Ozone',
                                                       'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                                       'units': 'nmol mol-1',
                                                       'chemical_formula': 'O3',
                                                       'id': 5},
                                          'programme': {'id': 0,
                                                        'name': '',
                                                        'longname': '',
                                                        'homepage': '',
                                                        'description': ''},
                                          'roles': [{'id': 1,
                                                     'role': 'resource provider',
                                                     'status': 'active',
                                                     'contact': {'id': 5,
                                                                 'organisation': {'id': 2,
                                                                                  'name': 'FZJ',
                                                                                  'longname': 'Forschungszentrum Jülich',
                                                                                  'kind': 'research',
                                                                                  'city': 'Jülich',
                                                                                  'postcode': '52425',
                                                                                  'street_address': 'Wilhelm-Johnen-Straße',
                                                                                  'country': 'Germany',
                                                                                  'homepage': 'https://www.fz-juelich.de',
                                                                                  'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                                          'citation': 'Forschungszentrum Jülich: time series of o3 at Test_China, accessed from the TOAR database on 2023-07-28 12:00:00',
                                          'attribution': 'Test-Attributions to be announced',
                                          'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/'},
                             'data': [{'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62, 'flags': 'OK validated verified', 'timeseries_id': 2, 'version': '1.0'}]}
        assert response.json() == expected_response


    def test_get_data_with_staging(self, client, db):
        with patch('toardb.timeseries.crud.dt.datetime', FixedDatetime):
            response = client.get("/data/timeseries_with_staging/id/2")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'metadata': {'id': 2,
                                          'label': 'CMA',
                                          'order': 1,
                                          'sampling_frequency': 'hourly',
                                          'aggregation': 'mean',
                                          'data_start_date': '2003-09-07T15:30:00+00:00',
                                          'data_end_date': '2016-12-31T14:30:00+00:00',
                                          'data_origin': 'instrument',
                                          'data_origin_type': 'measurement',
                                          'provider_version': 'N/A',
                                          'sampling_height': 7.0,
                                          'additional_metadata':
                                               {'original_units': {'since_19740101000000': 'nmol/mol'},
                                                'measurement_method': 'uv_abs',
                                                'absorption_cross_section': 0,
                                                'ebas_metadata_19740101000000_29y':
                                                     {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, '
                                                                   'some long division name, SHORT, , '
                                                                   '111 Streetname, , zipcode, Boulder, CO, USA',
                                                      'Data level': '2',
                                                      'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                      'Station code': 'XXX',
                                                      'Station name': 'Secret'}},
                                          'doi': '',
                                          'coverage': -1.0,
                                          'station': {'id': 3,
                                                      'codes': ['China_test8'],
                                                      'name': 'Test_China',
                                                      'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                                      'coordinate_validation_status': 'not checked',
                                                      'country': 'China',
                                                      'state': 'Shandong Sheng',
                                                      'type': 'unknown',
                                                      'type_of_area': 'unknown',
                                                      'timezone': 'Asia/Shanghai',
                                                      'additional_metadata': {},
                                                      'aux_images': [],
                                                      'aux_docs': [],
                                                      'aux_urls': [],
                                                      'changelog': []},
                                          'variable': {'name': 'o3',
                                                       'longname': 'ozone',
                                                       'displayname': 'Ozone',
                                                       'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                                       'units': 'nmol mol-1',
                                                       'chemical_formula': 'O3',
                                                       'id': 5},
                                          'programme': {'id': 0,
                                                        'name': '',
                                                        'longname': '',
                                                        'homepage': '',
                                                        'description': ''},
                                          'roles': [{'id': 1,
                                                     'role': 'resource provider',
                                                     'status': 'active',
                                                     'contact': {'id': 5,
                                                                 'organisation':
                                                                     {'id': 2,
                                                                      'name': 'FZJ',
                                                                      'longname': 'Forschungszentrum Jülich',
                                                                      'kind': 'research',
                                                                      'city': 'Jülich',
                                                                      'postcode': '52425',
                                                                      'street_address': 'Wilhelm-Johnen-Straße',
                                                                      'country': 'Germany',
                                                                      'homepage': 'https://www.fz-juelich.de',
                                                                      'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                                          'citation': 'Forschungszentrum Jülich: time series of o3 at '
                                                      'Test_China, accessed from the TOAR database on '
                                                      '2023-07-28 12:00:00',
                                          'attribution': 'Test-Attributions to be announced',
                                          'license': 'This data is published under a Creative Commons '
                                                     'Attribution 4.0 International (CC BY 4.0). '
                                                     'https://creativecommons.org/licenses/by/4.0/'
                            },
                             'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified',              'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed',             'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed',             'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'OK validated modified',              'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified',              'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified',              'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed',             'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62,  'flags': 'OK validated verified',              'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'questionable validated confirmed',   'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'questionable validated confirmed',   'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'questionable validated confirmed',   'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'questionable validated unconfirmed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'questionable validated unconfirmed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'questionable validated confirmed',   'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'questionable validated flagged',     'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T04:00:00+00:00', 'value': 19.62,  'flags': 'questionable validated unconfirmed', 'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T05:00:00+00:00', 'value': 15.696, 'flags': 'questionable validated flagged',     'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2013-12-17T06:00:00+00:00', 'value':  5.886, 'flags': 'questionable validated confirmed',   'timeseries_id': 2, 'version': '1.0'},
                                      {'datetime': '2014-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed',             'timeseries_id': 2, 'version': '0.0.20141217235502'},
                                      {'datetime': '2014-12-17T00:00:00+00:00', 'value':  7.848, 'flags': 'OK validated modified',              'timeseries_id': 2, 'version': '0.0.20141217235502'},
                                      {'datetime': '2014-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified',              'timeseries_id': 2, 'version': '0.0.20141217235502'},
                                      {'datetime': '2014-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated verified',              'timeseries_id': 2, 'version': '0.0.20141217235502'},
                                      {'datetime': '2014-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed',             'timeseries_id': 2, 'version': '0.0.20141217235502'},
                                      {'datetime': '2014-12-17T04:00:00+00:00', 'value': 19.62,  'flags': 'OK validated verified',              'timeseries_id': 2, 'version': '0.0.20141217235502'}]
                            }
        assert response.json() == expected_response


    def test_get_no_data_with_staging(self, client, db):
        response = client.get("/data/timeseries_with_staging/id/5")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_response = {'detail': 'Data not found.'}
        assert response.json() == expected_response


    def test_create_data_record(self, client, db):
        response = client.post("/data/timeseries/record/?series_id=2&datetime=2021-08-23%2015:00:00&value=67.3&flag=OK&version=000001.000001.00000000000000",
                               headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = 'Data successfully inserted.'
        assert response.json() == expected_response
        patched_response = client.get("/timeseries/2")
        assert patched_response.status_code == expected_status_code
        patched_response_changelog = patched_response.json()["changelog"][-1]
        patched_response_changelog["datetime"] = ""
        expected_changelog = {'datetime': '', 'description': 'data record created', 'old_value': '', 'new_value': '',
                              'timeseries_id': 2, 'author_id': 1, 'type_of_change': 'created', 'period_start': '2021-08-23T15:00:00+00:00',
                              'period_end': '2021-08-23T15:00:00+00:00', 'version': '000001.000001.00000000000000'}
        assert patched_response_changelog == expected_changelog


    def test_patch_data(self, client, db):
        response = client.patch("/data/timeseries/?description=test patch&version=000002.000000.00000000000000",
                                files={"file": open("tests/fixtures/data/toluene_SDZ54421_2012_2012_v2-0.dat", "rb")},
                                headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'detail': {'message': 'Data successfully inserted.'}}
        assert response.json() == expected_response
        patched_response = client.get("/timeseries/1")
        assert patched_response.status_code == expected_status_code
        patched_response_changelog = patched_response.json()["changelog"][-1]
        patched_response_changelog["datetime"] = ""
        expected_changelog = {'datetime': '', 'description': 'test patch', 'old_value': '', 'new_value': '',
                              'timeseries_id': 1, 'author_id': 1, 'type_of_change': 'unspecified data value corrections', 'period_start': '2012-12-15T14:00:00+00:00',
                              'period_end': '2012-12-17T17:00:00+00:00', 'version': '000002.000000.00000000000000'}
        assert patched_response_changelog == expected_changelog


    def test_patch_bulk_data(self, client, db):
        # show version of the original time series
        version_response = client.get("/data/timeseries/next_version/1")
        expected_status_code = 200
        assert version_response.status_code == expected_status_code
        expected_response = '000001.000001.00000000000000'
        assert version_response.json() == expected_response
        version_response = client.get("/data/timeseries/next_version/1?major=True")
        expected_status_code = 200
        assert version_response.status_code == expected_status_code
        expected_response = '000002.000000.00000000000000'
        assert version_response.json() == expected_response
        response = client.patch("/data/timeseries/bulk/?description=test patch bulk data&version=000002.000000.00000000000000",
                data='''[{"datetime": "2012-12-17T00:00:00+00:00", "value": 7.848,  "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T01:00:00+00:00", "value": 15.696, "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T02:00:00+00:00", "value": 11.772, "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T03:00:00+00:00", "value": 13.734, "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T04:00:00+00:00", "value": 19.62,  "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"}]''',
                headers={"email": "s.schroeder@fz-juelich.de"} )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'detail': {'message': 'Data successfully inserted.'}}
        assert response.json() == expected_response
        # check, what has been inserted!
        response = client.get("/data/timeseries/id/1")
        expected_status_code = 200
        expected_response = {'metadata': {'id': 1,
                                          'label': 'CMA',
                                          'order': 1,
                                          'sampling_frequency': 'hourly',
                                          'aggregation': 'mean',
                                          'data_start_date': '2003-09-07T15:30:00+00:00',
                                          'data_end_date': '2016-12-31T14:30:00+00:00',
                                          'coverage': -1.0,
                                          'data_origin': 'instrument',
                                          'data_origin_type': 'measurement',
                                          'provider_version': 'N/A',
                                          'sampling_height': 7.0,
                                          'additional_metadata': {'original_units': 'ppb'},
                                          'station': {'id': 2,
                                                      'codes': ['SDZ54421'],
                                                      'name': 'Shangdianzi',
                                                      'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                                      'coordinate_validation_status': 'not checked',
                                                      'country': 'China',
                                                      'state': 'Beijing Shi',
                                                      'type': 'unknown',
                                                      'type_of_area': 'unknown',
                                                      'timezone': 'Asia/Shanghai',
                                                      'additional_metadata': {'add_type': 'nature reservation'},
                                                      'aux_images': [],
                                                      'aux_docs': [],
                                                      'aux_urls': [],
                                                      'changelog': []
                                                     },
                                          'variable': {'name': 'toluene',
                                                       'longname': 'toluene',
                                                       'displayname': 'Toluene',
                                                       'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                                       'units': 'nmol mol-1',
                                                       'chemical_formula': 'C7H8',
                                                       'id': 7
                                                      },
                                          'programme': {'id': 0,
                                                        'name': '',
                                                        'longname': '',
                                                        'homepage': '',
                                                        'description': ''
                                                       },
                                          'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                                     'contact': {'id': 4,
                                                                 'organisation': {'id': 1,
                                                                                  'name': 'UBA',
                                                                                  'longname': 'Umweltbundesamt',
                                                                                  'kind': 'government',
                                                                                  'city': 'Dessau-Roßlau',
                                                                                  'postcode': '06844',
                                                                                  'street_address': 'Wörlitzer Platz 1',
                                                                                  'country': 'Germany',
                                                                                  'homepage': 'https://www.umweltbundesamt.de',
                                                                                  'contact_url': 'mailto:immission@uba.de',
                                                                                 }
                                                                 }
                                                     },
                                                     {'id': 3, 'role': 'principal investigator', 'status': 'active',
                                                      'contact': {'id': 3,
                                                                  'person': {'email': 's.schroeder@fz-juelich.de',
                                                                             'id': 3,
                                                                             'isprivate': False,
                                                                             'name': 'Sabine Schröder',
                                                                             'orcid': '0000-0002-0309-8010',
                                                                             'phone': '+49-2461-61-6397'
                                                                            }
                                                                 }
                                                     }],
                                          'changelog': [{'datetime': '',
                                                         'description': 'Sabine Schröder (s.schroeder@fz-juelich.de): test patch bulk data',
                                                         'old_value': '',
                                                         'new_value': '',
                                                         'timeseries_id': 1,
                                                         'author_id': 1,
                                                         'type_of_change': 'replaced data with a new version',
                                                         'period_start': '2012-12-17T00:00:00+00:00',
                                                         'period_end': '2012-12-17T04:00:00+00:00',
                                                         'version': '000002.000000.00000000000000'
                                                         }],
                                          'citation': 'Sabine Schröder: time series of toluene at Shangdianzi, accessed from the TOAR database',
                                          'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/',
                                          'doi': ''},
                             'data': [{'datetime': '2012-12-16T21:00:00+00:00', 'value': 21.581, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                      {'datetime': '2012-12-16T22:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                      {'datetime': '2012-12-16T23:00:00+00:00', 'value': 13.734, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'},
                                      {'datetime': '2012-12-17T00:00:00+00:00', 'value': 7.848, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '2.0'},
                                      {'datetime': '2012-12-17T01:00:00+00:00', 'value': 15.696, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '2.0'},
                                      {'datetime': '2012-12-17T02:00:00+00:00', 'value': 11.772, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '2.0'},
                                      {'datetime': '2012-12-17T03:00:00+00:00', 'value': 13.734, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '2.0'},
                                      {'datetime': '2012-12-17T04:00:00+00:00', 'value': 19.62, 'flags': 'OK validated QC passed', 'timeseries_id': 1, 'version': '2.0'},
                                      {'datetime': '2012-12-17T05:00:00+00:00', 'value': 15.696, 'flags': 'OK validated modified', 'timeseries_id': 1, 'version': '1.0'},
                                      {'datetime': '2012-12-17T06:00:00+00:00', 'value': 5.886, 'flags': 'OK validated verified', 'timeseries_id': 1, 'version': '1.0'}]
                            }
        patched_response = response.json()
        patched_response['metadata']['changelog'][0]['datetime'] = ""
        dindex = patched_response['metadata']['citation'].index(" on 2")
        patched_response['metadata']['citation'] = patched_response['metadata']['citation'][:dindex]
        assert patched_response == expected_response
        # also show that next_version has changed by this patch
        version_response = client.get("/data/timeseries/next_version/1")
        expected_status_code = 200
        assert version_response.status_code == expected_status_code
        expected_response = '000002.000001.00000000000000'
        assert version_response.json() == expected_response
        version_response = client.get("/data/timeseries/next_version/1?major=True")
        expected_status_code = 200
        assert version_response.status_code == expected_status_code
        expected_response = '000003.000000.00000000000000'
        assert version_response.json() == expected_response


    def test_patch_bulk_data2(self, client, db):
        response = client.patch("/data/timeseries/bulk/?toarqc_config_type=realtime&no_archive=True&description=NOLOG&version=000000.000001.20230208144921&force=True",
                data='''[{"datetime": "2012-12-17T00:00:00+00:00", "value": 7.848,  "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T01:00:00+00:00", "value": 15.696, "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T02:00:00+00:00", "value": 11.772, "flags": 0, "timeseries_id": 1, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T03:00:00+00:00", "value": 13.734, "flags": 0, "timeseries_id": 2, "version": "000002.000000.00000000000000"},
                         {"datetime": "2012-12-17T04:00:00+00:00", "value": 19.62,  "flags": 0, "timeseries_id": 2, "version": "000002.000000.00000000000000"}]''',
                headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'detail': {'message': 'Data successfully inserted.'}}
        assert response.json() == expected_response
