# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
from sqlalchemy import insert
from toardb.timeseries.models import Timeseries, timeseries_timeseries_roles_table
from toardb.timeseries.models_programme import TimeseriesProgramme
from toardb.timeseries.models_role import TimeseriesRole
from toardb.stationmeta.models import (
        StationmetaCore,
        StationmetaGlobal,
        StationmetaChangelog
)
from toardb.stationmeta.schemas import get_geom_from_coordinates, Coordinates
from toardb.variables.models import Variable
from toardb.contacts.models import Person, Organisation, Contact
from toardb.auth_user.models import AuthUser
# Required imports 'create_test_database'
from toardb.test_base import (
    client,
    get_test_db,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)

class TestApps:
    def setup(self):
        self.application_url = "/timeseries/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        _db_conn = get_test_engine()
        # id_seq will not be reset automatically between tests!
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        fake_cur.execute("ALTER SEQUENCE auth_user_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE variables_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_core_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_global_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_roles_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_doc_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_image_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_url_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE persons_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE organisations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE contacts_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_roles_id_seq RESTART WITH 3")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_programmes_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_id_seq RESTART WITH 1")
        fake_conn.commit()
        infilename = "tests/fixtures/auth_user/auth.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_auth_user = AuthUser(**entry)
                db.add(new_auth_user)
                db.commit()
                db.refresh(new_auth_user)
        infilename = "tests/fixtures/contacts/persons.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_person = Person(**entry)
                db.add(new_person)
                db.commit()
                db.refresh(new_person)
        infilename = "tests/fixtures/contacts/organisations.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_organisation = Organisation(**entry)
                db.add(new_organisation)
                db.commit()
                db.refresh(new_organisation)
        infilename = "tests/fixtures/contacts/contacts.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_contact = Contact(**entry)
                db.add(new_contact)
                db.commit()
                db.refresh(new_contact)
        infilename = "tests/fixtures/variables/variables.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_variable = Variable(**entry)
                db.add(new_variable)
                db.commit()
                db.refresh(new_variable)
        infilename = "tests/fixtures/stationmeta/stationmeta_core.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_core = StationmetaCore(**entry)
                # there's a mismatch with coordinates --> how to automatically switch back and forth?!
                tmp_coordinates = new_stationmeta_core.coordinates
                new_stationmeta_core.coordinates = get_geom_from_coordinates(Coordinates(**new_stationmeta_core.coordinates))
                db.add(new_stationmeta_core)
                db.commit()
                db.refresh(new_stationmeta_core)
        infilename = "tests/fixtures/stationmeta/stationmeta_changelog.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_changelog = StationmetaChangelog(**entry)
                db.add(new_stationmeta_changelog)
                db.commit()
                db.refresh(new_stationmeta_changelog)
        infilename = "tests/fixtures/stationmeta/stationmeta_global.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_global = StationmetaGlobal(**entry)
                db.add(new_stationmeta_global)
                db.commit()
                db.refresh(new_stationmeta_global)
        infilename = "tests/fixtures/timeseries/timeseries_programmes.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_programme = TimeseriesProgramme(**entry)
                db.add(new_timeseries_programme)
                db.commit()
                db.refresh(new_timeseries_programme)
        infilename = "tests/fixtures/timeseries/timeseries.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries = Timeseries(**entry)
                db.add(new_timeseries)
                db.commit()
                db.refresh(new_timeseries)
        infilename = "tests/fixtures/timeseries/timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_role = TimeseriesRole(**entry)
                db.add(new_timeseries_role)
                db.commit()
                db.refresh(new_timeseries_role)
        infilename = "tests/fixtures/timeseries/timeseries_timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                db.execute(insert(timeseries_timeseries_roles_table).values(timeseries_id=entry["timeseries_id"], role_id=entry["role_id"]))
                db.execute("COMMIT")


    def test_search_no_arguments(self, client, db):
        response = client.get("/search/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 1,
                          'label': 'CMA',
                          'order': 1,
                          'sampling_frequency': 'hourly',
                          'aggregation': 'mean',
                          'data_start_date': '2003-09-07T15:30:00+00:00',
                          'data_end_date': '2016-12-31T14:30:00+00:00',
                          'data_origin': 'instrument',
                          'data_origin_type': 'measurement',
                          'provider_version': 'N/A',
                          'sampling_height': 7.0,
                          'additional_metadata': {"original_units": "ppb"},
                          'doi': '',
                          'coverage': -1.0,
                          'station': {'id': 2,
                                      'codes': ['SDZ54421'],
                                      'name': 'Shangdianzi',
                                      'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China',
                                      'state': 'Beijing Shi',
                                      'type': 'unknown',
                                      'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai',
                                      'additional_metadata': {'add_type': 'nature reservation'},
                                      'aux_images': [],
                                      'aux_docs': [],
                                      'aux_urls': [],
                                      'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                                     'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                                     'landcover_description_25km_year2012': '',
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-07-15T19:27:09.463245+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 2,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'}]},
                          'variable': {'name': 'toluene',
                                       'longname': 'toluene',
                                       'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                       'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8',
                                       'id': 7},
                          'programme': {'id': 0,
                                        'name': '',
                                        'longname': '',
                                        'homepage': '',
                                        'description': ''},
                          'roles': [{'id': 2,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 4,
                                                 'organisation': {'id': 1,
                                                                  'name': 'UBA',
                                                                  'longname': 'Umweltbundesamt',
                                                                  'kind': 'government',
                                                                  'city': 'Dessau-Roßlau',
                                                                  'postcode': '06844',
                                                                  'street_address': 'Wörlitzer Platz 1',
                                                                  'country': 'Germany',
                                                                  'homepage': 'https://www.umweltbundesamt.de',
                                                                  'contact_url': 'mailto:immission@uba.de'}}},
                                    {'id': 3,
                                     'role': 'principal investigator',
                                     'status': 'active',
                                     'contact': {'id': 3,
                                                 'person': {'email': 's.schroeder@fz-juelich.de',
                                                            'id': 3,
                                                            'isprivate': False,
                                                            'name': 'Sabine Schröder',
                                                            'orcid': '0000-0002-0309-8010',
                                                            'phone': '+49-2461-61-6397'}}}]},
                         {'id': 2,
                          'label': 'CMA',
                          'order': 1,
                          'sampling_frequency': 'hourly',
                          'aggregation': 'mean',
                          'data_start_date': '2003-09-07T15:30:00+00:00',
                          'data_end_date': '2016-12-31T14:30:00+00:00',
                          'data_origin': 'instrument',
                          'data_origin_type': 'measurement',
                          'provider_version': 'N/A',
                          'sampling_height': 7.0,
                          'additional_metadata': {'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'measurement_method': 'uv_abs',
                                                  'absorption_cross_section': 'Hearn 1961',
                                                  'ebas_metadata_19740101000000_29y':
                                                      {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                       'Data level': '2',
                                                       'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                       'Station code': 'XXX',
                                                       'Station name': 'Secret'}},
                          'doi': '',
                          'coverage': -1.0,
                          'station': {'id': 3,
                                  'codes': ['China_test8'],
                                      'name': 'Test_China',
                                      'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China',
                                      'state': 'Shandong Sheng',
                                      'type': 'unknown',
                                      'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai',
                                      'additional_metadata': {},
                                      'aux_images': [],
                                      'aux_docs': [],
                                      'aux_urls': [],
                                      'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'landcover_description_25km_year2012': '',
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'}]},
                          'variable': {'name': 'o3',
                                       'longname': 'ozone',
                                       'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                       'units': 'nmol mol-1',
                                       'chemical_formula': 'O3',
                                       'id': 5},
                          'programme': {'id': 0,
                                        'name': '',
                                        'longname': '',
                                        'homepage': '',
                                        'description': ''},
                          'roles': [{'id': 1,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 5,
                                                 'organisation':
                                                     {'id': 2,
                                                      'name': 'FZJ',
                                                      'longname': 'Forschungszentrum Jülich',
                                                      'kind': 'research',
                                                      'city': 'Jülich',
                                                      'postcode': '52425',
                                                      'street_address': 'Wilhelm-Johnen-Straße',
                                                      'country': 'Germany',
                                                      'homepage': 'https://www.fz-juelich.de',
                                                      'contact_url': 'mailto:toar-data@fz-juelich.de'}}}]},
                       {'id': 3,
                        'additional_metadata': {},
                        'aggregation': 'mean',
                        'coverage': -1.0,
                        'data_end_date': '2024-12-31T14:30:00+00:00',
                        'data_origin': 'instrument',
                        'data_origin_type': 'measurement',
                        'data_start_date': '2023-09-07T15:30:00+00:00',
                        'doi': '',
                        'label': 'CMA',
                        'order': 3,
                        'programme': {
                            'description': '',
                            'homepage': '',
                            'id': 0,
                            'longname': '',
                            'name': '',
                        },
                        'provider_version': 'N/A',
                        'roles': [
                            {
                                'contact': {
                                    'id': 4,
                                    'organisation': {
                                        'city': 'Dessau-Roßlau',
                                        'contact_url': 'mailto:immission@uba.de',
                                        'country': 'Germany',
                                        'homepage': 'https://www.umweltbundesamt.de',
                                        'id': 1,
                                        'kind': 'government',
                                        'longname': 'Umweltbundesamt',
                                        'name': 'UBA',
                                        'postcode': '06844',
                                        'street_address': 'Wörlitzer Platz 1',
                                    },
                                },
                                'id': 2,
                                'role': 'resource provider',
                                'status': 'active',
                            },
                        ],
                        'sampling_frequency': 'hourly',
                        'sampling_height': 7.0,
                        'station': {
                            'additional_metadata': {},
                            'aux_docs': [],
                            'aux_images': [],
                            'aux_urls': [],
                            'changelog': [
                                {
                                    'author_id': 1,
                                    'datetime': '2023-08-15T21:16:20.596545+00:00',
                                    'description': 'station created',
                                    'new_value': '',
                                    'old_value': '',
                                    'station_id': 3,
                                    'type_of_change': 'created',
                                },
                            ],
                            'codes': [
                                'China_test8',
                            ],
                            'coordinate_validation_status': 'not checked',
                            'coordinates': {
                                'alt': 1534.0,
                                'lat': 36.256,
                                'lng': 117.106,
                            },
                            'country': 'China',
                            'globalmeta': {
                                'climatic_zone_year2016': '6 (warm temperate dry)',
                                'distance_to_major_road_year2020': -999.0,
                                'dominant_ecoregion_year2017': '-1 (undefined)',
                                'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                'ecoregion_description_25km_year2017': '',
                                'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                'landcover_description_25km_year2012': '',
                                'max_population_density_25km_year1990': -1.0,
                                'max_population_density_25km_year2015': -1.0,
                                'max_stable_nightlights_25km_year1992': -999.0,
                                'max_stable_nightlights_25km_year2013': -999.0,
                                'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                'mean_nox_emissions_10km_year2000': -999.0,
                                'mean_nox_emissions_10km_year2015': -999.0,
                                'mean_population_density_250m_year1990': -1.0,
                                'mean_population_density_250m_year2015': -1.0,
                                'mean_population_density_5km_year1990': -1.0,
                                'mean_population_density_5km_year2015': -1.0,
                                'mean_stable_nightlights_1km_year2013': -999.0,
                                'mean_stable_nightlights_5km_year2013': -999.0,
                                'mean_topography_srtm_alt_1km_year1994': -999.0,
                                'mean_topography_srtm_alt_90m_year1994': -999.0,
                                'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                'toar1_category': 'unclassified',
                                'toar2_category': 'suburban',
                            },
                            'id': 3,
                            'name': 'Test_China',
                            'state': 'Shandong Sheng',
                            'timezone': 'Asia/Shanghai',
                            'type': 'unknown',
                            'type_of_area': 'unknown',
                        },
                        'variable': {
                            'cf_standardname': 'mole_fraction_of_toluene_in_air',
                            'chemical_formula': 'C7H8',
                            'displayname': 'Toluene',
                            'id': 7,
                            'longname': 'toluene',
                            'name': 'toluene',
                            'units': 'nmol mol-1' }}]
        assert response.json() == expected_resp


    def test_search_variable_wrong_syntax(self, client, db):
        response = client.get("/search/?variable_id=ozone")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "Wrong value (not int) given: variable_id"
        assert response.json() == expected_resp


    def test_search_with_variable_id(self, client, db):
        response = client.get("/search/?variable_id=5")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai',
                                      'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp
 

    def test_search_with_codes(self, client, db):
        response = client.get("/search/?codes=China_test8")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}},
                         {'id': 3, 'label': 'CMA', 'order': 3,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2023-09-07T15:30:00+00:00', 'data_end_date': '2024-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {},
                          'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                 'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                 'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}}],
                          'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8', 'id': 7},
                          'station': {'id': 3, 'codes': [ 'China_test8' ], 'name': 'Test_China',
                                      'coordinates': { 'alt': 1534.0, 'lat': 36.256, 'lng': 117.106 },
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                              'globalmeta': {
                                  'climatic_zone_year2016': '6 (warm temperate dry)',
                                  'distance_to_major_road_year2020': -999.0,
                                  'dominant_ecoregion_year2017': '-1 (undefined)',
                                  'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                  'ecoregion_description_25km_year2017': '',
                                  'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                  'landcover_description_25km_year2012': '',
                                  'max_population_density_25km_year1990': -1.0,
                                  'max_population_density_25km_year2015': -1.0,
                                  'max_stable_nightlights_25km_year1992': -999.0,
                                  'max_stable_nightlights_25km_year2013': -999.0,
                                  'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'mean_nox_emissions_10km_year2000': -999.0,
                                  'mean_nox_emissions_10km_year2015': -999.0,
                                  'mean_population_density_250m_year1990': -1.0,
                                  'mean_population_density_250m_year2015': -1.0,
                                  'mean_population_density_5km_year1990': -1.0,
                                  'mean_population_density_5km_year2015': -1.0,
                                  'mean_stable_nightlights_1km_year2013': -999.0,
                                  'mean_stable_nightlights_5km_year2013': -999.0,
                                  'mean_topography_srtm_alt_1km_year1994': -999.0,
                                  'mean_topography_srtm_alt_90m_year1994': -999.0,
                                  'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'toar1_category': 'unclassified',
                                  'toar2_category': 'suburban'},
                              'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                             'description': 'station created',
                                             'old_value': '',
                                             'new_value': '',
                                             'station_id': 3,
                                             'author_id': 1,
                                             'type_of_change': 'created'
                                           }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_with_wrong_bounding_box(self, client, db):
        response = client.get("/search/?bounding_box=Asia")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "not enough values to unpack (expected 4, got 1)"
        assert response.json() == expected_resp


    def test_search_empty_bounding_box(self, client, db):
        response = client.get("/search/?bounding_box=117,36,118,37")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = []
        assert response.json() == expected_resp


    def test_search_empty_bounding_box_and_fields(self, client, db):
        response = client.get("/search/?bounding_box=117,36,118,37&fields=station_id")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = []
        assert response.json() == expected_resp


    def test_search_with_bounding_box(self, client, db):
        response = client.get("/search/?bounding_box=36,117,37,118")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}},
                         {'id': 3,
                          'additional_metadata': {},
                          'aggregation': 'mean',
                          'coverage': -1.0,
                          'data_end_date': '2024-12-31T14:30:00+00:00',
                          'data_origin': 'instrument',
                          'data_origin_type': 'measurement',
                          'data_start_date': '2023-09-07T15:30:00+00:00',
                          'doi': '',
                          'label': 'CMA',
                          'order': 3,
                          'programme': {
                              'description': '',
                              'homepage': '',
                              'id': 0,
                              'longname': '',
                              'name': '',
                          },
                          'provider_version': 'N/A',
                          'roles': [
                              {
                                  'contact': {
                                      'id': 4,
                                      'organisation': {
                                          'city': 'Dessau-Roßlau',
                                          'contact_url': 'mailto:immission@uba.de',
                                          'country': 'Germany',
                                          'homepage': 'https://www.umweltbundesamt.de',
                                          'id': 1,
                                          'kind': 'government',
                                          'longname': 'Umweltbundesamt',
                                          'name': 'UBA',
                                          'postcode': '06844',
                                          'street_address': 'Wörlitzer Platz 1',
                                      },
                                  },
                                  'id': 2,
                                  'role': 'resource provider',
                                  'status': 'active',
                              },
                          ],
                          'sampling_frequency': 'hourly',
                          'sampling_height': 7.0,
                          'station': {
                              'additional_metadata': {},
                              'aux_docs': [],
                              'aux_images': [],
                              'aux_urls': [],
                              'changelog': [
                                  {
                                      'author_id': 1,
                                      'datetime': '2023-08-15T21:16:20.596545+00:00',
                                      'description': 'station created',
                                      'new_value': '',
                                      'old_value': '',
                                      'station_id': 3,
                                      'type_of_change': 'created',
                                  },
                              ],
                              'codes': [
                                  'China_test8',
                              ],
                              'coordinate_validation_status': 'not checked',
                              'coordinates': {
                                  'alt': 1534.0,
                                  'lat': 36.256,
                                  'lng': 117.106,
                              },
                              'country': 'China',
                              'globalmeta': {
                                  'climatic_zone_year2016': '6 (warm temperate dry)',
                                  'distance_to_major_road_year2020': -999.0,
                                  'dominant_ecoregion_year2017': '-1 (undefined)',
                                  'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                  'ecoregion_description_25km_year2017': '',
                                  'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                  'landcover_description_25km_year2012': '',
                                  'max_population_density_25km_year1990': -1.0,
                                  'max_population_density_25km_year2015': -1.0,
                                  'max_stable_nightlights_25km_year1992': -999.0,
                                  'max_stable_nightlights_25km_year2013': -999.0,
                                  'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'mean_nox_emissions_10km_year2000': -999.0,
                                  'mean_nox_emissions_10km_year2015': -999.0,
                                  'mean_population_density_250m_year1990': -1.0,
                                  'mean_population_density_250m_year2015': -1.0,
                                  'mean_population_density_5km_year1990': -1.0,
                                  'mean_population_density_5km_year2015': -1.0,
                                  'mean_stable_nightlights_1km_year2013': -999.0,
                                  'mean_stable_nightlights_5km_year2013': -999.0,
                                  'mean_topography_srtm_alt_1km_year1994': -999.0,
                                  'mean_topography_srtm_alt_90m_year1994': -999.0,
                                  'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'toar1_category': 'unclassified',
                                  'toar2_category': 'suburban',
                              },
                              'id': 3,
                              'name': 'Test_China',
                              'state': 'Shandong Sheng',
                              'timezone': 'Asia/Shanghai',
                              'type': 'unknown',
                              'type_of_area': 'unknown',
                          },
                          'variable': {
                              'cf_standardname': 'mole_fraction_of_toluene_in_air',
                              'chemical_formula': 'C7H8',
                              'displayname': 'Toluene',
                              'id': 7,
                              'longname': 'toluene',
                              'name': 'toluene',
                              'units': 'nmol mol-1' }}]
        assert response.json() == expected_resp


    def test_search_with_time(self, client, db):
        response = client.get("/search/?variable_id=5&data_start_date=2003-09-07T15:30:00+00:00")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_wrong_time_syntax(self, client, db):
        response = client.get("/search/?variable_id=5&data_start_date=2050-01-01T00:00:00Z")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "Wrong value for time given: 2050-01-01T00:00:00Z"
        assert response.json() == expected_resp


    def test_search_with_fields(self, client, db):
        response = client.get("/search/?fields=id,role,order,additional_metadata")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'additional_metadata': {'original_units': 'ppb'},
                         'id': 1,
                         'order': 1,
                         'roles': {'contact_id': 3, 'role': 'principal investigator', 'status': 'active'}
                         },
                         {'id': 1,
                          'order': 1,
                          'additional_metadata': {"original_units": "ppb"},
                          'roles': {'role': 'resource provider', 'status': 'active', 'contact_id': 4}
                         },
                         {'id': 2,
                          'order': 1,
                          'additional_metadata':{'original_units': {'since_19740101000000': 'nmol/mol'},
                                                 'measurement_method': 'uv_abs',
                                                 'absorption_cross_section': 'Hearn 1961',
                                                 'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                                                      'Data level': '2',
                                                                                      'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                                                      'Station code': 'XXX',
                                                                                      'Station name': 'Secret'}
                                                },
                          'roles': {'role': 'resource provider', 'status': 'active', 'contact_id': 5}
                         },
                         {'id': 3,
                          'order': 3,
                          'additional_metadata': {},
                          'roles': {'role': 'resource provider', 'status': 'active', 'contact_id': 4}
                         }]
        assert response.json() == expected_resp


    def test_search_with_fields_unique_station_id(self, client, db):
        response = client.get("/search/?fields=station_id")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{"station": {"id": 2}},
                         {"station": {"id": 3}}]
        assert response.json() == expected_resp


    def test_search_with_variable_id_in_fields(self, client, db):
        response = client.get("/search/?fields=variable_id")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'variable': {'id': 7}},
                         {'variable': {'id': 5}}]
        assert response.json() == expected_resp


    def test_search_with_whole_variable_in_fields(self, client, db):
        response = client.get("/search/?fields=variable")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "Wrong field given: variable"
        assert response.json() == expected_resp


    def test_search_with_stationname(self, client, db):
        response = client.get("/search/?name=Test_China")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}},
                         {'id': 3, 'label': 'CMA', 'order': 3,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2023-09-07T15:30:00+00:00', 'data_end_date': '2024-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {},
                          'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                 'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                 'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}}],
                          'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8', 'id': 7},
                          'station': {'id': 3, 'codes': [ 'China_test8' ], 'name': 'Test_China',
                                      'coordinates': { 'alt': 1534.0, 'lat': 36.256, 'lng': 117.106 },
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                              'globalmeta': {
                                  'climatic_zone_year2016': '6 (warm temperate dry)',
                                  'distance_to_major_road_year2020': -999.0,
                                  'dominant_ecoregion_year2017': '-1 (undefined)',
                                  'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                  'ecoregion_description_25km_year2017': '',
                                  'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                  'landcover_description_25km_year2012': '',
                                  'max_population_density_25km_year1990': -1.0,
                                  'max_population_density_25km_year2015': -1.0,
                                  'max_stable_nightlights_25km_year1992': -999.0,
                                  'max_stable_nightlights_25km_year2013': -999.0,
                                  'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'mean_nox_emissions_10km_year2000': -999.0,
                                  'mean_nox_emissions_10km_year2015': -999.0,
                                  'mean_population_density_250m_year1990': -1.0,
                                  'mean_population_density_250m_year2015': -1.0,
                                  'mean_population_density_5km_year1990': -1.0,
                                  'mean_population_density_5km_year2015': -1.0,
                                  'mean_stable_nightlights_1km_year2013': -999.0,
                                  'mean_stable_nightlights_5km_year2013': -999.0,
                                  'mean_topography_srtm_alt_1km_year1994': -999.0,
                                  'mean_topography_srtm_alt_90m_year1994': -999.0,
                                  'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'toar1_category': 'unclassified',
                                  'toar2_category': 'suburban'},
                              'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                             'description': 'station created',
                                             'old_value': '',
                                             'new_value': '',
                                             'station_id': 3,
                                             'author_id': 1,
                                             'type_of_change': 'created'
                                           }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_with_global_attributes(self, client, db):
        response = client.get("/search/?climatic_zone_year2016=WarmTemperateDry&htap_region_tier1_year2010=HTAPTier1SAF")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}},
                         {'id': 3, 'label': 'CMA', 'order': 3,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2023-09-07T15:30:00+00:00', 'data_end_date': '2024-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {},
                          'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                 'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                 'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}}],
                          'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8', 'id': 7},
                          'station': {'id': 3, 'codes': [ 'China_test8' ], 'name': 'Test_China',
                                      'coordinates': { 'alt': 1534.0, 'lat': 36.256, 'lng': 117.106 },
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                              'globalmeta': {
                                  'climatic_zone_year2016': '6 (warm temperate dry)',
                                  'distance_to_major_road_year2020': -999.0,
                                  'dominant_ecoregion_year2017': '-1 (undefined)',
                                  'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                  'ecoregion_description_25km_year2017': '',
                                  'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                  'landcover_description_25km_year2012': '',
                                  'max_population_density_25km_year1990': -1.0,
                                  'max_population_density_25km_year2015': -1.0,
                                  'max_stable_nightlights_25km_year1992': -999.0,
                                  'max_stable_nightlights_25km_year2013': -999.0,
                                  'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'mean_nox_emissions_10km_year2000': -999.0,
                                  'mean_nox_emissions_10km_year2015': -999.0,
                                  'mean_population_density_250m_year1990': -1.0,
                                  'mean_population_density_250m_year2015': -1.0,
                                  'mean_population_density_5km_year1990': -1.0,
                                  'mean_population_density_5km_year2015': -1.0,
                                  'mean_stable_nightlights_1km_year2013': -999.0,
                                  'mean_stable_nightlights_5km_year2013': -999.0,
                                  'mean_topography_srtm_alt_1km_year1994': -999.0,
                                  'mean_topography_srtm_alt_90m_year1994': -999.0,
                                  'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'toar1_category': 'unclassified',
                                  'toar2_category': 'suburban'},
                              'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                             'description': 'station created',
                                             'old_value': '',
                                             'new_value': '',
                                             'station_id': 3,
                                             'author_id': 1,
                                             'type_of_change': 'created'
                                           }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_with_global_attributes2(self, client, db):
        response = client.get("/search/?dominant_landcover_year2012=CroplandRainfed&dominant_ecoregion_year2017=Undefined")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}},
                         {'id': 3, 'label': 'CMA', 'order': 3,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2023-09-07T15:30:00+00:00', 'data_end_date': '2024-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {},
                          'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                 'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                 'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}}],
                          'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8', 'id': 7},
                          'station': {'id': 3, 'codes': [ 'China_test8' ], 'name': 'Test_China',
                                      'coordinates': { 'alt': 1534.0, 'lat': 36.256, 'lng': 117.106 },
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                              'globalmeta': {
                                  'climatic_zone_year2016': '6 (warm temperate dry)',
                                  'distance_to_major_road_year2020': -999.0,
                                  'dominant_ecoregion_year2017': '-1 (undefined)',
                                  'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                  'ecoregion_description_25km_year2017': '',
                                  'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                  'landcover_description_25km_year2012': '',
                                  'max_population_density_25km_year1990': -1.0,
                                  'max_population_density_25km_year2015': -1.0,
                                  'max_stable_nightlights_25km_year1992': -999.0,
                                  'max_stable_nightlights_25km_year2013': -999.0,
                                  'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'mean_nox_emissions_10km_year2000': -999.0,
                                  'mean_nox_emissions_10km_year2015': -999.0,
                                  'mean_population_density_250m_year1990': -1.0,
                                  'mean_population_density_250m_year2015': -1.0,
                                  'mean_population_density_5km_year1990': -1.0,
                                  'mean_population_density_5km_year2015': -1.0,
                                  'mean_stable_nightlights_1km_year2013': -999.0,
                                  'mean_stable_nightlights_5km_year2013': -999.0,
                                  'mean_topography_srtm_alt_1km_year1994': -999.0,
                                  'mean_topography_srtm_alt_90m_year1994': -999.0,
                                  'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                  'toar1_category': 'unclassified',
                                  'toar2_category': 'suburban'},
                              'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                             'description': 'station created',
                                             'old_value': '',
                                             'new_value': '',
                                             'station_id': 3,
                                             'author_id': 1,
                                             'type_of_change': 'created'
                                           }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_with_additional_metadata(self, client, db):
        response = client.get("/search/?additional_metadata->'absorption_cross_section'=Hearn1961")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 2, 'label': 'CMA', 'order': 1,
                          'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                          'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                          'data_origin': 'instrument', 'sampling_height': 7.0,
                          'provider_version': 'N/A',
                          'doi': '',
                          'additional_metadata': {'absorption_cross_section': 'Hearn 1961',
                                                  'measurement_method': 'uv_abs',
                                                  'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA',
                                                          'Data level': '2',
                                                          'Frameworks': 'GAW-WDCRG NOAA-ESRL',
                                                          'Station code': 'XXX',
                                                          'Station name': 'Secret' } },
                          'roles': [{'id': 1, 'role': 'resource provider', 'status': 'active',
                                     'contact': {'id': 5, 'organisation': {'id': 2, 'name': 'FZJ', 'longname': 'Forschungszentrum Jülich',
                                                                           'kind': 'research', 'city': 'Jülich', 'postcode': '52425', 'street_address': 'Wilhelm-Johnen-Straße',
                                                                           'country': 'Germany', 'homepage': 'https://www.fz-juelich.de', 'contact_url': 'mailto:toar-data@fz-juelich.de'}}}],
                          'variable': {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                                       'chemical_formula': 'O3', 'id': 5},
                          'station': {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                                      'coordinates': {'alt': 1534.0, 'lat': 36.256, 'lng': 117.106},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China', 'state': 'Shandong Sheng',
                                      'type': 'unknown', 'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai', 
                                      'additional_metadata': {},
                                      'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                      'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                     'landcover_description_25km_year2012': '',
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                                     'description': 'station created',
                                                     'old_value': '',
                                                     'new_value': '',
                                                     'station_id': 3,
                                                     'author_id': 1,
                                                     'type_of_change': 'created'
                                                    }]},
                          'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}]
        assert response.json() == expected_resp


    def test_search_with_additional_metadata2(self, client, db):
        response = client.get("/search/?additional_metadata->'absorption_cross_section'=Hearn1961"+
                                      "&additional_metadata->'sampling_type'=Continuous"+
                                      "&additional_metadata->'calibration_type'=Automatic")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = []
        assert response.json() == expected_response


    def test_search_with_additional_metadata3(self, client, db):
        response = client.get("/search/?additional_metadata->'original_units'=ppb")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [{'id': 1,
                              'label': 'CMA',
                              'order': 1,
                              'sampling_frequency': 'hourly',
                              'aggregation': 'mean',
                              'data_start_date': '2003-09-07T15:30:00+00:00',
                              'data_end_date': '2016-12-31T14:30:00+00:00',
                              'data_origin': 'instrument',
                              'data_origin_type': 'measurement',
                              'provider_version': 'N/A',
                              'sampling_height': 7.0,
                              'additional_metadata': {"original_units": "ppb"},
                              'doi': '',
                              'coverage': -1.0,
                              'station': {'id': 2,
                                          'codes': ['SDZ54421'],
                                          'name': 'Shangdianzi',
                                          'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                          'coordinate_validation_status': 'not checked',
                                          'country': 'China',
                                          'state': 'Beijing Shi',
                                          'type': 'unknown',
                                          'type_of_area': 'unknown',
                                          'timezone': 'Asia/Shanghai',
                                          'additional_metadata': {'add_type': 'nature reservation'},
                                          'aux_images': [],
                                          'aux_docs': [],
                                          'aux_urls': [],
                                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                                         'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                                         'landcover_description_25km_year2012': '',
                                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                                         'ecoregion_description_25km_year2017': '',
                                                         'distance_to_major_road_year2020': -999.0,
                                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year1992': -999.0,
                                                         'mean_population_density_250m_year2015': -1.0,
                                                         'mean_population_density_5km_year2015': -1.0,
                                                         'max_population_density_25km_year2015': -1.0,
                                                         'mean_population_density_250m_year1990': -1.0,
                                                         'mean_population_density_5km_year1990': -1.0,
                                                         'max_population_density_25km_year1990': -1.0,
                                                         'mean_nox_emissions_10km_year2015': -999.0,
                                                         'mean_nox_emissions_10km_year2000': -999.0,
                                                         'toar1_category': 'unclassified',
                                                         'toar2_category': 'suburban'},
                                          'changelog': [{'datetime': '2023-07-15T19:27:09.463245+00:00', 'description': 'station created', 'old_value': '', 'new_value': '', 'station_id': 2, 'author_id': 1, 'type_of_change': 'created'}]},
                              'variable': {'name': 'toluene',
                                           'longname': 'toluene',
                                           'displayname': 'Toluene',
                                           'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                           'units': 'nmol mol-1',
                                           'chemical_formula': 'C7H8',
                                           'id': 7},
                              'programme': {'id': 0,
                                            'name': '',
                                            'longname': '',
                                            'homepage': '',
                                            'description': ''},
                              'roles': [{'id': 2,
                                         'role': 'resource provider',
                                         'status': 'active',
                                         'contact': {'id': 4,
                                                     'organisation': {'id': 1,
                                                                      'name': 'UBA',
                                                                      'longname': 'Umweltbundesamt',
                                                                      'kind': 'government',
                                                                      'city': 'Dessau-Roßlau',
                                                                      'postcode': '06844',
                                                                      'street_address': 'Wörlitzer Platz 1',
                                                                      'country': 'Germany',
                                                                      'homepage': 'https://www.umweltbundesamt.de',
                                                                      'contact_url': 'mailto:immission@uba.de'}}},
                                        {'id': 3,
                                         'role': 'principal investigator',
                                         'status': 'active',
                                         'contact': {'id': 3,
                                                     'person': {'email': 's.schroeder@fz-juelich.de',
                                                                'id': 3,
                                                                'isprivate': False,
                                                                'name': 'Sabine Schröder',
                                                                'orcid': '0000-0002-0309-8010',
                                                                'phone': '+49-2461-61-6397'
                                                               }
                                                    }
                                        }]
                               }]
        assert response.json() == expected_response


    def test_search_with_additional_metadata_unknown(self, client, db):
        response = client.get("/search/?additional_metadata->'not_yet_defined'=42")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = []
        assert response.json() == expected_response


    def test_search_with_additional_metadata_station(self, client, db):
        response = client.get("/search/?station_additional_metadata->'add_type'=nature reservation")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [{'id': 1,
                              'label': 'CMA',
                              'order': 1,
                              'sampling_frequency': 'hourly',
                              'aggregation': 'mean',
                              'data_start_date': '2003-09-07T15:30:00+00:00',
                              'data_end_date': '2016-12-31T14:30:00+00:00',
                              'data_origin': 'instrument',
                              'data_origin_type': 'measurement',
                              'provider_version': 'N/A',
                              'sampling_height': 7.0,
                              'additional_metadata': {"original_units": "ppb"},
                              'doi': '',
                              'coverage': -1.0,
                              'station': {'id': 2,
                                          'codes': ['SDZ54421'],
                                          'name': 'Shangdianzi',
                                          'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                          'coordinate_validation_status': 'not checked',
                                          'country': 'China',
                                          'state': 'Beijing Shi',
                                          'type': 'unknown',
                                          'type_of_area': 'unknown',
                                          'timezone': 'Asia/Shanghai',
                                          'additional_metadata': {'add_type': 'nature reservation'},
                                          'aux_images': [],
                                          'aux_docs': [],
                                          'aux_urls': [],
                                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                                         'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                                         'landcover_description_25km_year2012': '',
                                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                                         'ecoregion_description_25km_year2017': '',
                                                         'distance_to_major_road_year2020': -999.0,
                                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year1992': -999.0,
                                                         'mean_population_density_250m_year2015': -1.0,
                                                         'mean_population_density_5km_year2015': -1.0,
                                                         'max_population_density_25km_year2015': -1.0,
                                                         'mean_population_density_250m_year1990': -1.0,
                                                         'mean_population_density_5km_year1990': -1.0,
                                                         'max_population_density_25km_year1990': -1.0,
                                                         'mean_nox_emissions_10km_year2015': -999.0,
                                                         'mean_nox_emissions_10km_year2000': -999.0,
                                                         'toar1_category': 'unclassified',
                                                         'toar2_category': 'suburban'},
                                          'changelog': [{'datetime': '2023-07-15T19:27:09.463245+00:00', 'description': 'station created', 'old_value': '', 'new_value': '', 'station_id': 2, 'author_id': 1, 'type_of_change': 'created'}]},
                              'variable': {'name': 'toluene',
                                           'longname': 'toluene',
                                           'displayname': 'Toluene',
                                           'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                           'units': 'nmol mol-1',
                                           'chemical_formula': 'C7H8',
                                           'id': 7},
                              'programme': {'id': 0,
                                            'name': '',
                                            'longname': '',
                                            'homepage': '',
                                            'description': ''},
                              'roles': [{'id': 2,
                                         'role': 'resource provider',
                                         'status': 'active',
                                         'contact': {'id': 4,
                                                     'organisation': {'id': 1,
                                                                      'name': 'UBA',
                                                                      'longname': 'Umweltbundesamt',
                                                                      'kind': 'government',
                                                                      'city': 'Dessau-Roßlau',
                                                                      'postcode': '06844',
                                                                      'street_address': 'Wörlitzer Platz 1',
                                                                      'country': 'Germany',
                                                                      'homepage': 'https://www.umweltbundesamt.de',
                                                                      'contact_url': 'mailto:immission@uba.de'}}},
                                        {'id': 3,
                                         'role': 'principal investigator',
                                         'status': 'active',
                                         'contact': {'id': 3,
                                                     'person': {'email': 's.schroeder@fz-juelich.de',
                                                                'id': 3,
                                                                'isprivate': False,
                                                                'name': 'Sabine Schröder',
                                                                'orcid': '0000-0002-0309-8010',
                                                                'phone': '+49-2461-61-6397'
                                                               }
                                                    }
                                        }]
                               }]
        assert response.json() == expected_response
