# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
from fastapi import Request
from sqlalchemy import insert
from toardb.stationmeta.models import (
        StationmetaCore,
        StationmetaGlobal,
        StationmetaGlobalService,
        StationmetaRole,
        StationmetaChangelog,
        stationmeta_core_stationmeta_roles_table
)
from toardb.toardb import app
from toardb.stationmeta import crud
from toardb.stationmeta.schemas import (
        get_geom_from_coordinates,
        Coordinates,
        StationmetaCreate
)
from toardb.auth_user.models import AuthUser
from toardb.contacts.models import Person, Organisation, Contact
from toardb.test_base import (
    client,
    get_test_db,
    override_dependency,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)


class TestApps:
    def setup(self):
        self.application_url = "/stationmeta/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        # id_seq will not be reset automatically between tests!
        _db_conn = get_test_engine()
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        fake_cur.execute("ALTER SEQUENCE auth_user_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_core_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_global_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_roles_id_seq RESTART WITH 3")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_doc_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_image_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_url_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE persons_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE organisations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE contacts_id_seq RESTART WITH 1")
        fake_conn.commit()
        infilename = "tests/fixtures/auth_user/auth.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_auth_user= AuthUser(**entry)
                db.add(new_auth_user)
                db.commit()
                db.refresh(new_auth_user)
        infilename = "tests/fixtures/contacts/persons.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_person = Person(**entry)
                db.add(new_person)
                db.commit()
                db.refresh(new_person)
        infilename = "tests/fixtures/contacts/organisations.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_organisation = Organisation(**entry)
                db.add(new_organisation)
                db.commit()
                db.refresh(new_organisation)
        infilename = "tests/fixtures/contacts/contacts.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_contact = Contact(**entry)
                db.add(new_contact)
                db.commit()
                db.refresh(new_contact)
        # I also need to upload tests with nested data!!!
        infilename = "tests/fixtures/stationmeta/stationmeta_core.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_core = StationmetaCore(**entry)
                # there's a mismatch with coordinates --> how to automatically switch back and forth?!
                tmp_coordinates = new_stationmeta_core.coordinates
                new_stationmeta_core.coordinates = get_geom_from_coordinates(Coordinates(**new_stationmeta_core.coordinates))
                # there's also a mismatch with additional_metadata --> BUT: this should not be switched back!
                # in upload command, we have now: "additional_metadata": "{}"
                # but return from this method gives: "additional_metadata": {}
                # ==> there is a mismatch between model(JSONB) and schema(JSON)
                new_stationmeta_core.additional_metadata = str(new_stationmeta_core.additional_metadata)
                db.add(new_stationmeta_core)
                db.commit()
                db.refresh(new_stationmeta_core)
        infilename = "tests/fixtures/stationmeta/stationmeta_changelog.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_changelog = StationmetaChangelog(**entry)
                db.add(new_stationmeta_changelog)
                db.commit()
                db.refresh(new_stationmeta_changelog)
        infilename = "tests/fixtures/stationmeta/stationmeta_global.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_global = StationmetaGlobal(**entry)
                db.add(new_stationmeta_global)
                db.commit()
                db.refresh(new_stationmeta_global)
        infilename = "tests/fixtures/stationmeta/stationmeta_global_services.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_global_service = StationmetaGlobalService(**entry)
                db.add(new_stationmeta_global_service)
                db.commit()
                db.refresh(new_stationmeta_global_service)
        infilename = "tests/fixtures/stationmeta/stationmeta_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_role = StationmetaRole(**entry)
                db.add(new_stationmeta_role)
                db.commit()
                db.refresh(new_stationmeta_role)
        infilename = "tests/fixtures/stationmeta/stationmeta_core_stationmeta_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                db.execute(insert(stationmeta_core_stationmeta_roles_table).values(station_id=entry["station_id"], role_id=entry["role_id"]))
                db.execute("COMMIT")


    # 1. tests retrieving station metadata

    def test_get_all(self, client, db):
        response = client.get("/stationmeta/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 1,
                          'codes': ['China11'],
                          'name': 'Mount Tai',
                          'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                          'coordinate_validation_status': 'not checked',
                          'country': 'China',
                          'state': 'Shandong Sheng',
                          'type': 'unknown',
                          'type_of_area': 'unknown',
                          'timezone': 'Asia/Shanghai',
                          'additional_metadata': {'dummy_info': 'Here is some more information about the station'},
                          'roles': [{'id': 2,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 1,
                                                 'name': 'UBA',
                                                 'longname': 'Umweltbundesamt',
                                                 'kind': 'government',
                                                 'city': 'Dessau-Roßlau',
                                                 'postcode': '06844',
                                                 'street_address': 'Wörlitzer Platz 1',
                                                 'country': 'Germany',
                                                 'homepage': 'https://www.umweltbundesamt.de',
                                                 'contact_url': 'mailto:immission@uba.de'}}],
                          'aux_images': [],
                          'aux_docs': [],
                          'aux_urls': [],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. '
                                                                       'Arabia, Oman, etc, Iran, Iraq)',
                                         'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                         'landcover_description_25km_year2012': '11 (Cropland, '
                                                                                'rainfed, herbaceous '
                                                                                'cover): 30.7 %, 12 '
                                                                                '(Cropland, rainfed, '
                                                                                'tree or shrub cover): '
                                                                                '25.0 %, 210 (Water '
                                                                                'bodies): 16.9 %, 130 '
                                                                                '(Grassland): 8.6 %, '
                                                                                '190 (Urban areas): '
                                                                                '5.8 %, 100 (Mosaic '
                                                                                'tree and shrub (>50%) '
                                                                                '/ herbaceous cover '
                                                                                '(<50%)): 3.5 %, 40 '
                                                                                '(Mosaic natural '
                                                                                'vegetation (tree, '
                                                                                'shrub, herbaceous '
                                                                                'cover) (>50%) / '
                                                                                'cropland (<50%)): 2.6 '
                                                                                '%, 10 (Cropland, '
                                                                                'rainfed): 2.0 %, 30 '
                                                                                '(Mosaic cropland '
                                                                                '(>50%) / natural '
                                                                                'vegetation (tree, '
                                                                                'shrub, herbaceous '
                                                                                'cover) (<50%)): 1.9 %',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'urban'},
                          'changelog': [{'datetime': '2023-07-05T08:23:04.551645+00:00',
                                         'description': 'station created',
                                         'old_value': '',
                                         'new_value': '',
                                         'station_id': 1,
                                         'author_id': 1,
                                         'type_of_change': 'created'}]},
                         {'id': 2,
                          'codes': ['SDZ54421'],
                          'name': 'Shangdianzi',
                          'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                          'coordinate_validation_status': 'not checked',
                          'country': 'China',
                          'state': 'Beijing Shi',
                          'type': 'unknown',
                          'type_of_area': 'unknown',
                          'timezone': 'Asia/Shanghai',
                          'additional_metadata': {'add_type': 'nature reservation'},
                          'roles': [{'id': 1,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 2,
                                                 'name': 'FZJ',
                                                 'longname': 'Forschungszentrum Jülich',
                                                 'kind': 'research',
                                                 'city': 'Jülich',
                                                 'postcode': '52425',
                                                 'street_address': 'Wilhelm-Johnen-Straße',
                                                 'country': 'Germany',
                                                 'homepage': 'https://www.fz-juelich.de',
                                                 'contact_url': 'mailto:toar-data@fz-juelich.de'}}],
                          'aux_images': [],
                          'aux_docs': [],
                          'aux_urls': [],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. '
                                                                       'Arabia, Oman, etc, Iran, Iraq)',
                                         'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                         'landcover_description_25km_year2012': '',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'suburban'},
                          'changelog': [{'datetime': '2023-07-15T19:27:09.463245+00:00',
                                         'description': 'station created',
                                         'old_value': '',
                                         'new_value': '',
                                         'station_id': 2,
                                         'author_id': 1,
                                         'type_of_change': 'created'}]},
                         {'id': 3,
                          'codes': ['China_test8'],
                          'name': 'Test_China',
                          'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                          'coordinate_validation_status': 'not checked',
                          'country': 'China',
                          'state': 'Shandong Sheng',
                          'type': 'unknown',
                          'type_of_area': 'unknown',
                          'timezone': 'Asia/Shanghai',
                          'additional_metadata': {},
                          'aux_images': [],
                          'aux_docs': [],
                          'aux_urls': [],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel '
                                                                       'Africa)',
                                         'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                         'landcover_description_25km_year2012': '',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'suburban'},
                          'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                         'description': 'station created',
                                         'old_value': '',
                                         'new_value': '',
                                         'station_id': 3,
                                         'author_id': 1,
                                         'type_of_change': 'created'}]}]
        assert response.json() == expected_resp


    def test_get_special_nested(self, client, db):
        response = client.get("/stationmeta/China_test8")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 3, 'codes': ['China_test8'], 'name': 'Test_China',
                         'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                         'coordinate_validation_status': 'not checked',
                         'country': 'China', 'state': 'Shandong Sheng',
                         'type': 'unknown',
                         'type_of_area': 'unknown', 'timezone': 'Asia/Shanghai',
                         'additional_metadata': {},
                         'aux_images': [], 'aux_docs': [],
                         'aux_urls': [],
                         'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                        'distance_to_major_road_year2020': -999.0,
                                        'dominant_ecoregion_year2017': '-1 (undefined)',
                                        'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                        'ecoregion_description_25km_year2017': '',
                                        'landcover_description_25km_year2012': '',
                                        'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                        'max_stable_nightlights_25km_year1992': -999.0,
                                        'max_stable_nightlights_25km_year2013': -999.0,
                                        'max_population_density_25km_year1990': -1.0,
                                        'max_population_density_25km_year2015': -1.0,
                                        'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'mean_stable_nightlights_1km_year2013': -999.0,
                                        'mean_stable_nightlights_5km_year2013': -999.0,
                                        'mean_nox_emissions_10km_year2000': -999.0,
                                        'mean_nox_emissions_10km_year2015': -999.0,
                                        'mean_population_density_250m_year1990': -1.0,
                                        'mean_population_density_250m_year2015': -1.0,
                                        'mean_population_density_5km_year1990': -1.0,
                                        'mean_population_density_5km_year2015': -1.0,
                                        'mean_topography_srtm_alt_1km_year1994': -999.0,
                                        'mean_topography_srtm_alt_90m_year1994': -999.0,
                                        'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'toar1_category': 'unclassified',
                                        'toar2_category': 'suburban'},
                          'changelog': [{'datetime': '2023-08-15T21:16:20.596545+00:00',
                                         'description': 'station created',
                                         'old_value': '',
                                         'new_value': '',
                                         'station_id': 3,
                                         'author_id': 1,
                                         'type_of_change': 'created'
                                        }]}
        assert response.json() == expected_resp


    def test_get_special_with_fields(self, client, db):
        response = client.get("/stationmeta/China11?fields=codes,globalmeta")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'codes': ['China11'],
                         'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                        'mean_topography_srtm_alt_1km_year1994': -999.0,
                                        'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'climatic_zone_year2016': '6 (warm temperate dry)',
                                        'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                        'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                        'landcover_description_25km_year2012': '11 (Cropland, rainfed, herbaceous cover): 30.7 %, 12 (Cropland, rainfed, tree or shrub cover): 25.0 %, 210 (Water bodies): 16.9 %, 130 (Grassland): 8.6 %, 190 (Urban areas): 5.8 %, 100 (Mosaic tree and shrub (>50%) / herbaceous cover (<50%)): 3.5 %, 40 (Mosaic natural vegetation (tree, shrub, herbaceous cover) (>50%) / cropland (<50%)): 2.6 %, 10 (Cropland, rainfed): 2.0 %, 30 (Mosaic cropland (>50%) / natural vegetation (tree, shrub, herbaceous cover) (<50%)): 1.9 %',
                                        'dominant_ecoregion_year2017': '-1 (undefined)',
                                        'ecoregion_description_25km_year2017': '',
                                        'distance_to_major_road_year2020': -999.0,
                                        'mean_stable_nightlights_1km_year2013': -999.0,
                                        'mean_stable_nightlights_5km_year2013': -999.0,
                                        'max_stable_nightlights_25km_year2013': -999.0,
                                        'max_stable_nightlights_25km_year1992': -999.0,
                                        'mean_population_density_250m_year2015': -1.0,
                                        'mean_population_density_5km_year2015': -1.0,
                                        'max_population_density_25km_year2015': -1.0,
                                        'mean_population_density_250m_year1990': -1.0,
                                        'mean_population_density_5km_year1990': -1.0,
                                        'max_population_density_25km_year1990': -1.0,
                                        'mean_nox_emissions_10km_year2015': -999.0,
                                        'mean_nox_emissions_10km_year2000': -999.0,
                                        'toar1_category': 'unclassified',
                                        'toar2_category': 'urban'
                                       }
                        }
        assert response.json() == expected_resp


    def test_get_special_with_ambiguous_field(self, client, db):
        response = client.get("/stationmeta/China11?fields=id")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 1}
        assert response.json() == expected_resp


    def test_get_special_with_fields_station_not_found(self, client, db):
        response = client.get("/stationmeta/DENW078?fields=codes,globalmeta")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail":"Metadata for station 'DENW078' not found."}
        assert response.json() == expected_resp

    def test_get_all_with_fields(self, client, db):
        response = client.get("/stationmeta/?fields=codes,globalmeta")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'codes': ['China11'],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                         'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                         'landcover_description_25km_year2012': '11 (Cropland, rainfed, herbaceous cover): 30.7 %, 12 (Cropland, rainfed, tree or shrub cover): 25.0 %, 210 (Water bodies): 16.9 %, 130 (Grassland): 8.6 %, 190 (Urban areas): 5.8 %, 100 (Mosaic tree and shrub (>50%) / herbaceous cover (<50%)): 3.5 %, 40 (Mosaic natural vegetation (tree, shrub, herbaceous cover) (>50%) / cropland (<50%)): 2.6 %, 10 (Cropland, rainfed): 2.0 %, 30 (Mosaic cropland (>50%) / natural vegetation (tree, shrub, herbaceous cover) (<50%)): 1.9 %',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'urban'
                                        }
                         },
                         {'codes': ['SDZ54421'],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                         'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                         'landcover_description_25km_year2012': '',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'suburban'
                                        }
                         },
                         {'codes': ['China_test8'],
                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                         'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                         'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                         'landcover_description_25km_year2012': '',
                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                         'ecoregion_description_25km_year2017': '',
                                         'distance_to_major_road_year2020': -999.0,
                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year2013': -999.0,
                                         'max_stable_nightlights_25km_year1992': -999.0,
                                         'mean_population_density_250m_year2015': -1.0,
                                         'mean_population_density_5km_year2015': -1.0,
                                         'max_population_density_25km_year2015': -1.0,
                                         'mean_population_density_250m_year1990': -1.0,
                                         'mean_population_density_5km_year1990': -1.0,
                                         'max_population_density_25km_year1990': -1.0,
                                         'mean_nox_emissions_10km_year2015': -999.0,
                                         'mean_nox_emissions_10km_year2000': -999.0,
                                         'toar1_category': 'unclassified',
                                         'toar2_category': 'suburban'
                                        }
                         }]
        assert response.json() == expected_resp


    def test_get_all_wrong_params(self, client, db):
        response = client.get("/stationmeta/?altitude=2000&limit=None")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "'An unknown argument was received: altitude.'"
        assert response.json() == expected_resp


    def test_get_stationmeta_not_found(self, client, db):
        response = client.get("/stationmeta/-1")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Metadata for station '-1' not found."}
        assert response.json() == expected_resp


    def test_get_stationmeta_by_id_not_found(self, client, db):
        response = client.get("/stationmeta/id/-1")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Metadata for station with ID '-1' not found."}
        assert response.json() == expected_resp


    def test_get_stationmeta_by_id(self, client, db):
        response = client.get("/stationmeta/id/1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 1, 'codes': ['China11'], 'name': 'Mount Tai',
                         'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                         'coordinate_validation_status': 'not checked',
                         'country': 'China', 'state': 'Shandong Sheng',
                         'type': 'unknown', 'type_of_area': 'unknown',
                         'timezone': 'Asia/Shanghai', 'additional_metadata': {},
                         'roles': [{ 'contact': {
                                     'city': 'Dessau-Roßlau',
                                     'contact_url': 'mailto:immission@uba.de',
                                     'country': 'Germany',
                                     'homepage': 'https://www.umweltbundesamt.de',
                                     'id': 1,
                                     'kind': 'government',
                                     'longname': 'Umweltbundesamt',
                                     'name': 'UBA',
                                     'postcode': '06844',
                                     'street_address': 'Wörlitzer Platz 1',
                                 },
                                 'id': 2,
                                 'role': 'resource provider',
                                 'status': 'active' }],
                         'additional_metadata': {'dummy_info': 'Here is some more information about the station'},
                         'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                         'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                        'distance_to_major_road_year2020': -999.0,
                                        'dominant_ecoregion_year2017': '-1 (undefined)',
                                        'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                        'ecoregion_description_25km_year2017': '',
                                        'landcover_description_25km_year2012': '11 (Cropland, rainfed, '
                                                                         +'herbaceous cover): '
                                                                         +'30.7 %, 12 (Cropland, '
                                                                         +'rainfed, tree or shrub '
                                                                         +'cover): 25.0 %, 210 '
                                                                         +'(Water bodies): 16.9 '
                                                                         +'%, 130 (Grassland): '
                                                                         +'8.6 %, 190 (Urban '
                                                                         +'areas): 5.8 %, 100 '
                                                                         +'(Mosaic tree and shrub '
                                                                         +'(>50%) / herbaceous '
                                                                         +'cover (<50%)): 3.5 %, '
                                                                         +'40 (Mosaic natural '
                                                                         +'vegetation (tree, '
                                                                         +'shrub, herbaceous '
                                                                         +'cover) (>50%) / '
                                                                         +'cropland (<50%)): 2.6 '
                                                                         +'%, 10 (Cropland, '
                                                                         +'rainfed): 2.0 %, 30 '
                                                                         +'(Mosaic cropland '
                                                                         +'(>50%) / natural '
                                                                         +'vegetation (tree, '
                                                                         +'shrub, herbaceous '
                                                                         +'cover) (<50%)): 1.9 %',
                                        'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                        'max_stable_nightlights_25km_year1992': -999.0,
                                        'max_stable_nightlights_25km_year2013': -999.0,
                                        'max_population_density_25km_year1990': -1.0,
                                        'max_population_density_25km_year2015': -1.0,
                                        'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'mean_stable_nightlights_1km_year2013': -999.0,
                                        'mean_stable_nightlights_5km_year2013': -999.0,
                                        'mean_nox_emissions_10km_year2000': -999.0,
                                        'mean_nox_emissions_10km_year2015': -999.0,
                                        'mean_population_density_250m_year1990': -1.0,
                                        'mean_population_density_250m_year2015': -1.0,
                                        'mean_population_density_5km_year1990': -1.0,
                                        'mean_population_density_5km_year2015': -1.0,
                                        'mean_topography_srtm_alt_1km_year1994': -999.0,
                                        'mean_topography_srtm_alt_90m_year1994': -999.0,
                                        'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                        'toar1_category': 'unclassified',
                                        'toar2_category': 'urban'},
                         'changelog': [{'datetime': '2023-07-05T08:23:04.551645+00:00',
                                        'description': 'station created',
                                        'old_value': '',
                                        'new_value': '',
                                        'station_id': 1,
                                        'author_id': 1,
                                        'type_of_change': 'created'
                                       }]}
        assert response.json() == expected_resp


    def test_get_stationmeta_changelog(self, client, db):
        response = client.get("/stationmeta_changelog/1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp =  [{ 'datetime': '2023-07-05T08:23:04.551645+00:00',
                                         'description': 'station created',
                                         'old_value': '',
                                         'new_value': '',
                                         'station_id': 1,
                                         'author_id': 1,
                                         'type_of_change': 'created'
                                       }]
        assert response.json() == expected_resp

    # 2. tests creating station metadata


    def test_insert_new_wrong_credentials(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":37.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "additional_metadata": "{}" }
                     },
                headers={"email": "j.doe@fz-juelich.de"}
                   )
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new_without_credentials(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":37.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "additional_metadata": "{}" }
                     })
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":37.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "additional_metadata": "{}" }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': "new station: ['ttt3', 'ttt4'],Test_China,{'lat': 37.256, 'lng': 117.106, 'alt': 1534.0}",
                                    'station_id': 4}}
        assert response.json() == expected_resp


    def test_insert_new_with_roles(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":37.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"}],
                           "additional_metadata": "{}" }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': "new station: ['ttt3', 'ttt4'],Test_China,{'lat': 37.256, 'lng': 117.106, 'alt': 1534.0}",
                                    'station_id': 4}}
        assert response.json() == expected_resp


    def test_insert_new_with_annotations(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":37.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "annotations": [{"kind": "User",
                                            "text": "some foo",
                                            "date_added":"2021-07-27 00:00",
                                            "approved": True,
                                            "contributor_id":1}],
                           "additional_metadata": "{}" }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': "new station: ['ttt3', 'ttt4'],Test_China,{'lat': 37.256, 'lng': 117.106, 'alt': 1534.0}",
                                    'station_id': 4}}
        assert response.json() == expected_resp


    def test_insert_new_same_coordinates(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["ttt3","ttt4"],
                           "name":"Test_China","coordinates":{"lat":36.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "additional_metadata":"{}"},
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 444
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'more than one station falls within the given radius!\nchoose which station record to patch (add station code)',
                                    'station_ids': [1, 3],
                                    'station_records': [[['China11'], 'lat=36.256 lng=117.106 alt=1534.0', 'Shandong Sheng', 48],
                                                        [['China_test8'], 'lat=36.256 lng=117.106 alt=1534.0', 'Shandong Sheng', 48]]}}
        assert response.json() == expected_resp


    def test_insert_duplicate(self, client, db):
        response = client.post("/stationmeta/",
                json={"stationmeta":
                          {"codes":["China11"],
                           "name":"Test_China","coordinates":{"lat":36.256,"lng":117.106,"alt":1534.0},
                           "coordinate_validation_status": "NotChecked",
                           "country":"CN","state":"Shandong Sheng",
                           "type":"Unknown","type_of_area":"Unknown","timezone":"Asia/Shanghai",
                           "additional_metadata":"{}"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 443
        assert response.status_code == expected_status_code
        expected_resp = {'detail' : {'message': 'Station already registered!',
                                     'station_id': 1}}
        assert response.json() == expected_resp


    # 3. tests updating station metadata

    def test_patch_stationmeta_no_description(self, client, db):
        response = client.patch("/stationmeta/-1",
                json={"stationmeta":
                          {"name":"TTTT95TTTT"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "description text ist missing."}
        assert response.json() == expected_resp


    def test_patch_stationmeta_not_found1(self, client, db):
        response = client.patch("/stationmeta/-1?description=changing station name",
                json={"stationmeta":
                          {"name":"TTTT95TTTT"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Station for patching not found."}
        assert response.json() == expected_resp


    def test_patch_stationmeta_name(self, client, db):
        response = client.patch("/stationmeta/SDZ54421?description=changing station name",
                json={"stationmeta":
                          {"name":"TTTT95TTTT"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'message': 'patched stationmeta record for station_id 2', 'station_id': 2}
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/stationmeta/id/{response_json['station_id']}")
        response_json = response.json()
        # just check special changes
        assert response_json['name'] == 'TTTT95TTTT'
        assert response_json['changelog'][1]['old_value'] == "{'name': 'Shangdianzi'}"
        assert response_json['changelog'][1]['new_value'] == "{'name': 'TTTT95TTTT'}"
        assert response_json['changelog'][1]['author_id'] == 1
        assert response_json['changelog'][1]['type_of_change'] == 'single value correction in metadata'


    def test_patch_single_stationmeta_global(self, client, db):
        response = client.patch("/stationmeta/SDZ54421?description=changing global metadata",
                json={"stationmeta": 
                          {"globalmeta": {"climatic_zone_year2016": "WarmTemperateMoist"}}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'message': 'patched stationmeta record for station_id 2', 'station_id': 2}
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/stationmeta/id/{response_json['station_id']}")
        response_json = response.json()
        # just check special changes
        assert response_json['name'] == 'Shangdianzi'
        assert response_json['changelog'][1]['old_value'] == "{'climatic_zone_year2016': 'WarmTemperateDry'}"
        assert response_json['changelog'][1]['new_value'] == "{'climatic_zone_year2016': 'WarmTemperateMoist'}"
        assert response_json['changelog'][1]['author_id'] == 1
        assert response_json['changelog'][1]['type_of_change'] == 'single value correction in metadata'


    def test_patch_multiple_stationmeta_global(self, client, db):
        response = client.patch("/stationmeta/SDZ54421?description=changing global metadata",
                json={"stationmeta": 
                          {"globalmeta": {"climatic_zone_year2016": "WarmTemperateMoist",
                                          "toar1_category": "RuralLowElevation",
                                          "toar2_category": "Urban",
                                          "htap_region_tier1_year2010": "HTAPTier1PAN",
                                          "dominant_landcover_year2012": "TreeNeedleleavedEvergreenClosedToOpen",
                                          "landcover_description_25km_year2012": "TreeNeedleleavedEvergreenClosedToOpen: 100 %",
                                          "dominant_ecoregion_year2017": "Guianansavanna", 
                                          "ecoregion_description_25km_year2017": "Guianansavanna: 90 %, Miskitopineforests: 10 %"}}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'message': 'patched stationmeta record for station_id 2', 'station_id': 2}
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/stationmeta/id/{response_json['station_id']}")
        response_json = response.json()
        # just check special changes
        assert response_json['name'] == 'Shangdianzi'
        assert response_json['changelog'][1]['old_value'] == (
                    "{'climatic_zone_year2016': 'WarmTemperateDry', "
                    "'htap_region_tier1_year2010': 'HTAPTier1MDE', "
                    "'dominant_landcover_year2012': 'CroplandRainfedHerbaceousCover', "
                    "'landcover_description_25km_year2012': '', "
                    "'dominant_ecoregion_year2017': 'Undefined', "
                    "'ecoregion_description_25km_year2017': '', "
                    "'toar1_category': 'Unclassified', " 
                    "'toar2_category': 'Suburban'}" )
        assert response_json['changelog'][1]['new_value'] == (
                    "{'climatic_zone_year2016': 'WarmTemperateMoist', "
                    "'htap_region_tier1_year2010': 'HTAPTier1PAN', "
                    "'dominant_landcover_year2012': 'TreeNeedleleavedEvergreenClosedToOpen', "
                    "'landcover_description_25km_year2012': 'TreeNeedleleavedEvergreenClosedToOpen: 100 %', "
                    "'dominant_ecoregion_year2017': 'Guianansavanna', "
                    "'ecoregion_description_25km_year2017': 'Guianansavanna: 90 %, Miskitopineforests: 10 %'"
                    ", 'toar1_category': 'RuralLowElevation'" 
                    ", 'toar2_category': 'Urban'}" )
        assert response_json['changelog'][1]['author_id'] == 1
        assert response_json['changelog'][1]['type_of_change'] == 'single value correction in metadata'


    def test_patch_stationmeta_roles_and_annotations(self, client, db):
        response = client.patch("/stationmeta/SDZ54421?description=adding annotation text",
                json={"stationmeta":
                          {"roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"}],
                           "annotations": [{"kind": "User",
                                            "text": "some annotation text",
                                            "date_added": "2025-02-10 17:00",
                                            "approved": True,
                                            "contributor_id":1}]
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'message': 'patched stationmeta record for station_id 2', 'station_id': 2}
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/stationmeta/id/{response_json['station_id']}")
        response_json = response.json()
        assert response_json['annotations'] == [{'id': 1, 'kind': 'user comment', 'text': 'some annotation text', 'date_added': '2025-02-10T17:00:00+00:00', 'approved': True, 'contributor_id': 1}]
        assert response_json['changelog'][1]['old_value'] == "{'roles': {{'role': 'ResourceProvider', 'status': 'Active', 'contact_id': 5},}"
        assert response_json['changelog'][1]['new_value'] == (
                    "{'roles': [{'role': 'PointOfContact', 'contact_id': 3, 'status': 'Active'}, "
                               "{'role': 'Originator', 'contact_id': 1, 'status': 'Active'}], "
                     "'annotations': [{'kind': 'User', 'text': 'some annotation text', 'date_added': '2025-02-10 17:00', 'approved': True, 'contributor_id': 1}]}" )
        assert response_json['changelog'][1]['author_id'] == 1
        assert response_json['changelog'][1]['type_of_change'] == 'comprehensive metadata revision'


    def test_delete_roles_from_stationmeta(self, client, db):
        response = client.patch("/stationmeta/delete_field/China11?field=roles",
                                headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        # Database defaults cannot be patched within pytest
        patched_response = response.json()
        patched_response["changelog"][-1]["datetime"] = ""
        expected_resp = {'id': 1,
                         'codes': ['China11'],
                         'name': 'Mount Tai',
                         'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                         'coordinate_validation_status': 'not checked',
                         'country': 'China',
                         'state': 'Shandong Sheng',
                         'type': 'unknown',
                         'type_of_area': 'unknown',
                         'timezone': 'Asia/Shanghai',
                         'additional_metadata': {'dummy_info': 'Here is some more information about the station'},
                         'aux_images': [],
                         'aux_docs': [],
                         'aux_urls': [],
                         'changelog': [
                                      {
                                          'author_id': 1,
                                          'datetime': '2023-07-05T08:23:04.551645+00:00',
                                          'description': 'station created',
                                          'new_value': '',
                                          'old_value': '',
                                          'station_id': 1,
                                          'type_of_change': 'created',
                                      },
                                      {
                                          'author_id': 1,
                                          'datetime': '',
                                          'description': 'delete field roles',
                                          'new_value': "'roles': []",
                                          'old_value': "'roles': '[]'",
                                          'station_id': 1,
                                          'type_of_change': 'single value correction in metadata',
                                      },
                                  ],
                              }
        assert patched_response == expected_resp


    def test_delete_field_station_not_found(self, client, db):
        response = client.patch("/stationmeta/delete_field/China22?field=timezone",
                                headers={"email": "s.schroeder@fz-juelich.de"})
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Station for deleting field not found.'}
        assert response.json() == expected_resp


    def test_test(self, client, db):
        response = client.patch("/stationmeta/SDY54421?description=changing global metadata",
                json={"stationmeta":
                          {"globalmeta": { "toar2_category": "Rural"}}
                          },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Station for patching not found.'}
        assert response.json() == expected_resp
