# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
from sqlalchemy import insert
from toardb.timeseries.models import Timeseries, timeseries_timeseries_roles_table, \
                                     s1_contributors_table
from toardb.timeseries.models_programme import TimeseriesProgramme
from toardb.timeseries.models_role import TimeseriesRole
from toardb.stationmeta.models import StationmetaCore, StationmetaGlobal
from toardb.stationmeta.schemas import get_geom_from_coordinates, Coordinates
from toardb.variables.models import Variable
from toardb.contacts.models import Person, Organisation, Contact
from toardb.auth_user.models import AuthUser
# Required imports 'create_test_database'
from toardb.test_base import (
    client,
    get_test_db,
    override_dependency,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)
# for mocking datetime.now(timezone.utc)
from datetime import datetime
from unittest.mock import patch


class TestApps:
    def setup(self):
        self.application_url = "/timeseries/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        _db_conn = get_test_engine()
        # id_seq will not be reset automatically between tests!
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        fake_cur.execute("ALTER SEQUENCE auth_user_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE variables_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_core_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_global_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_roles_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_doc_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_image_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE stationmeta_aux_url_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE persons_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE organisations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE contacts_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_annotations_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_roles_id_seq RESTART WITH 4")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_programmes_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER SEQUENCE timeseries_id_seq RESTART WITH 1")
        fake_conn.commit()
        fake_cur.execute("ALTER TABLE timeseries_changelog ALTER COLUMN datetime SET DEFAULT '2023-07-28 12:00:00'")
        fake_conn.commit()
        infilename = "tests/fixtures/auth_user/auth.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_auth_user = AuthUser(**entry)
                db.add(new_auth_user)
                db.commit()
                db.refresh(new_auth_user)
        infilename = "tests/fixtures/contacts/persons.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_person = Person(**entry)
                db.add(new_person)
                db.commit()
                db.refresh(new_person)
        infilename = "tests/fixtures/contacts/organisations.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_organisation = Organisation(**entry)
                db.add(new_organisation)
                db.commit()
                db.refresh(new_organisation)
        infilename = "tests/fixtures/contacts/contacts.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_contact = Contact(**entry)
                db.add(new_contact)
                db.commit()
                db.refresh(new_contact)
        infilename = "tests/fixtures/variables/variables.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_variable = Variable(**entry)
                db.add(new_variable)
                db.commit()
                db.refresh(new_variable)
        infilename = "tests/fixtures/stationmeta/stationmeta_core.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_core = StationmetaCore(**entry)
                # there's a mismatch with coordinates --> how to automatically switch back and forth?!
                tmp_coordinates = new_stationmeta_core.coordinates
                new_stationmeta_core.coordinates = get_geom_from_coordinates(Coordinates(**new_stationmeta_core.coordinates))
                db.add(new_stationmeta_core)
                db.commit()
                db.refresh(new_stationmeta_core)
        infilename = "tests/fixtures/stationmeta/stationmeta_global.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_stationmeta_global = StationmetaGlobal(**entry)
                db.add(new_stationmeta_global)
                db.commit()
                db.refresh(new_stationmeta_global)
        infilename = "tests/fixtures/timeseries/timeseries_programmes.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_programme = TimeseriesProgramme(**entry)
                db.add(new_timeseries_programme)
                db.commit()
                db.refresh(new_timeseries_programme)
        infilename = "tests/fixtures/timeseries/timeseries.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries = Timeseries(**entry)
                db.add(new_timeseries)
                db.commit()
                db.refresh(new_timeseries)
        infilename = "tests/fixtures/timeseries/timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                new_timeseries_role = TimeseriesRole(**entry)
                db.add(new_timeseries_role)
                db.commit()
                db.refresh(new_timeseries_role)
        infilename = "tests/fixtures/timeseries/timeseries_timeseries_roles.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                db.execute(insert(timeseries_timeseries_roles_table).values(timeseries_id=entry["timeseries_id"], role_id=entry["role_id"]))
                db.execute("COMMIT")
        infilename = "tests/fixtures/timeseries/timeseries_contributors.json"
        with open(infilename) as f:
            metajson=json.load(f)
            for entry in metajson:
                db.execute(insert(s1_contributors_table).values(request_id=entry["request_id"], timeseries_ids=entry["timeseries_ids"]))
                db.execute("COMMIT")


    def test_get_timeseries(self, client, db):
        response = client.get("/timeseries/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'id': 1,
                          'label': 'CMA',
                          'order': 1,
                          'sampling_frequency': 'hourly',
                          'aggregation': 'mean',
                          'data_start_date': '2003-09-07T15:30:00+00:00',
                          'data_end_date': '2016-12-31T14:30:00+00:00',
                          'data_origin': 'instrument',
                          'data_origin_type': 'measurement',
                          'provider_version': 'N/A',
                          'sampling_height': 7.0,
                          'additional_metadata': {'original_units': 'ppb'},
                          'doi': '',
                          'coverage': -1.0,
                          'station': {'id': 2,
                                      'codes': ['SDZ54421'],
                                      'name': 'Shangdianzi',
                                      'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China',
                                      'state': 'Beijing Shi',
                                      'type': 'unknown',
                                      'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai',
                                      'additional_metadata': {'add_type': 'nature reservation'},
                                      'aux_images': [],
                                      'aux_docs': [],
                                      'aux_urls': [],
                                      'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'htap_region_tier1_year2010': '11 (MDE Middle '
                                                                                   'East: S. Arabia, '
                                                                                   'Oman, etc, Iran, '
                                                                                   'Iraq)',
                                                     'dominant_landcover_year2012': '11 (Cropland, '
                                                                                    'rainfed, herbaceous cover)',
                                                     'landcover_description_25km_year2012': '',
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': []},
                          'variable': {'name': 'toluene',
                                       'longname': 'toluene',
                                       'displayname': 'Toluene',
                                       'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                       'units': 'nmol mol-1',
                                       'chemical_formula': 'C7H8',
                                       'id': 7},
                          'programme': {'id': 0,
                                        'name': '',
                                        'longname': '',
                                        'homepage': '',
                                        'description': ''},
                          'roles': [{'id': 2,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 4,
                                                 'organisation': {'id': 1,
                                                                  'name': 'UBA',
                                                                  'longname': 'Umweltbundesamt',
                                                                  'kind': 'government',
                                                                  'city': 'Dessau-Roßlau',
                                                                  'postcode': '06844',
                                                                  'street_address': 'Wörlitzer Platz 1',
                                                                  'country': 'Germany',
                                                                  'homepage': 'https://www.umweltbundesamt.de',
                                                                  'contact_url': 'mailto:immission@uba.de'}}},
                                    {'id': 3,
                                     'role': 'principal investigator',
                                     'status': 'active',
                                     'contact': {'id': 3,
                                                 'person': {'email': 's.schroeder@fz-juelich.de',
                                                            'id': 3,
                                                            'isprivate': False,
                                                            'name': 'Sabine Schröder',
                                                            'orcid': '0000-0002-0309-8010',
                                                            'phone': '+49-2461-61-6397'}}}]},
                         {'id': 2,
                          'label': 'CMA',
                          'order': 1,
                          'sampling_frequency': 'hourly',
                          'aggregation': 'mean',
                          'data_start_date': '2003-09-07T15:30:00+00:00',
                          'data_end_date': '2016-12-31T14:30:00+00:00',
                          'data_origin': 'instrument',
                          'data_origin_type': 'measurement',
                          'provider_version': 'N/A',
                          'sampling_height': 7.0,
                          'additional_metadata': {'original_units': {'since_19740101000000': 'nmol/mol'},
                                                  'measurement_method': 'uv_abs',
                                                  'absorption_cross_section': 'Hearn 1961',
                                                  'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, '
                                                                                                    'Lady, '
                                                                                                    'lady.unknown@unknown.com, '
                                                                                                    'some '
                                                                                                    'long '
                                                                                                    'division '
                                                                                                    'name, '
                                                                                                    'SHORT, '
                                                                                                    ', '
                                                                                                    '111 '
                                                                                                    'Streetname, '
                                                                                                    ', '
                                                                                                    'zipcode, '
                                                                                                    'Boulder, '
                                                                                                    'CO, '
                                                                                                    'USA',
                                                                                       'Data level': '2',
                                                                                       'Frameworks': 'GAW-WDCRG '
                                                                                                     'NOAA-ESRL',
                                                                                       'Station code': 'XXX',
                                                                                       'Station name': 'Secret'}},
                          'doi': '',
                          'coverage': -1.0,
                          'station': {'id': 3,
                                      'codes': ['China_test8'],
                                      'name': 'Test_China',
                                      'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                      'coordinate_validation_status': 'not checked',
                                      'country': 'China',
                                      'state': 'Shandong Sheng',
                                      'type': 'unknown',
                                      'type_of_area': 'unknown',
                                      'timezone': 'Asia/Shanghai',
                                      'additional_metadata': {},
                                      'aux_images': [],
                                      'aux_docs': [],
                                      'aux_urls': [],
                                      'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                     'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                     'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                     'climatic_zone_year2016': '6 (warm temperate dry)',
                                                     'htap_region_tier1_year2010': '10 (SAF Sub '
                                                                                   'Saharan/sub Sahel '
                                                                                   'Africa)',
                                                     'dominant_landcover_year2012': '10 (Cropland, '
                                                                                    'rainfed)',
                                                     'landcover_description_25km_year2012': '',
                                                     'dominant_ecoregion_year2017': '-1 (undefined)',
                                                     'ecoregion_description_25km_year2017': '',
                                                     'distance_to_major_road_year2020': -999.0,
                                                     'mean_stable_nightlights_1km_year2013': -999.0,
                                                     'mean_stable_nightlights_5km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year2013': -999.0,
                                                     'max_stable_nightlights_25km_year1992': -999.0,
                                                     'mean_population_density_250m_year2015': -1.0,
                                                     'mean_population_density_5km_year2015': -1.0,
                                                     'max_population_density_25km_year2015': -1.0,
                                                     'mean_population_density_250m_year1990': -1.0,
                                                     'mean_population_density_5km_year1990': -1.0,
                                                     'max_population_density_25km_year1990': -1.0,
                                                     'mean_nox_emissions_10km_year2015': -999.0,
                                                     'mean_nox_emissions_10km_year2000': -999.0,
                                                     'toar1_category': 'unclassified',
                                                     'toar2_category': 'suburban'},
                                      'changelog': []},
                          'variable': {'name': 'o3',
                                       'longname': 'ozone',
                                       'displayname': 'Ozone',
                                       'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                       'units': 'nmol mol-1',
                                       'chemical_formula': 'O3',
                                       'id': 5},
                          'programme': {'id': 0,
                                        'name': '',
                                        'longname': '',
                                        'homepage': '',
                                        'description': ''},
                          'roles': [{'id': 1,
                                     'role': 'resource provider',
                                     'status': 'active',
                                     'contact': {'id': 5,
                                                 'organisation': {'id': 2,
                                                                  'name': 'FZJ',
                                                                  'longname': 'Forschungszentrum '
                                                                              'Jülich',
                                                                  'kind': 'research',
                                                                  'city': 'Jülich',
                                                                  'postcode': '52425',
                                                                  'street_address': 'Wilhelm-Johnen-Straße',
                                                                  'country': 'Germany',
                                                                  'homepage': 'https://www.fz-juelich.de',
                                                                  'contact_url': 'mailto:toar-data@fz-juelich.de'}}}]}]
        assert response.json() == expected_resp


    def test_get_special(self, client, db):
        response = client.get("/timeseries/1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'id': 1, 'label': 'CMA', 'order': 1,
                         'sampling_frequency': 'hourly', 'aggregation': 'mean', 'data_origin_type': 'measurement',
                         'data_start_date': '2003-09-07T15:30:00+00:00', 'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0,
                         'data_origin': 'instrument', 'sampling_height': 7.0,
                         'provider_version': 'N/A', 'doi': '', 'additional_metadata': {'original_units': 'ppb'},
                         'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                                    'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                                          'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                                          'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}},
                                   {'id': 3, 'role': 'principal investigator', 'status': 'active',
                                    'contact': {'id': 3, 'person': {'email': 's.schroeder@fz-juelich.de', 'id': 3, 'isprivate': False,
                                                                    'name': 'Sabine Schröder', 'orcid': '0000-0002-0309-8010', 'phone': '+49-2461-61-6397'}}}],
                         'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                                      'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                                      'chemical_formula': 'C7H8', 'id': 7},
                         'station': {'id': 2, 'codes': ['SDZ54421'], 'name': 'Shangdianzi',
                                     'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                     'coordinate_validation_status': 'not checked',
                                     'country': 'China', 'state': 'Beijing Shi',
                                     'type': 'unknown', 'type_of_area': 'unknown',
                                     'timezone': 'Asia/Shanghai',
                                     'additional_metadata': {'add_type': 'nature reservation'},
                                     'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                                     'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                                    'distance_to_major_road_year2020': -999.0,
                                                    'dominant_ecoregion_year2017': '-1 (undefined)',
                                                    'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                                    'ecoregion_description_25km_year2017': '',
                                                    'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                                    'landcover_description_25km_year2012': '',
                                                    'max_stable_nightlights_25km_year1992': -999.0,
                                                    'max_stable_nightlights_25km_year2013': -999.0,
                                                    'max_population_density_25km_year1990': -1.0,
                                                    'max_population_density_25km_year2015': -1.0,
                                                    'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                    'mean_stable_nightlights_1km_year2013': -999.0,
                                                    'mean_stable_nightlights_5km_year2013': -999.0,
                                                    'mean_nox_emissions_10km_year2000': -999.0,
                                                    'mean_nox_emissions_10km_year2015': -999.0,
                                                    'mean_population_density_250m_year1990': -1.0,
                                                    'mean_population_density_250m_year2015': -1.0,
                                                    'mean_population_density_5km_year1990': -1.0,
                                                    'mean_population_density_5km_year2015': -1.0,
                                                    'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                    'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                    'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                    'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                    'toar1_category': 'unclassified',
                                                    'toar2_category': 'suburban'},
                                     'changelog': []},
                         'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}}
        assert response.json() == expected_resp


    def test_insert_new_wrong_credentials(self, client, db):
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA2", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 7,
                           "additional_metadata":"{}" 
                          }
                     },
                headers={"email": "j.doe@fz-juelich.de"}
                   )
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new_without_credentials(self, client, db):
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA2", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 7,
                           "additional_metadata":"{}" 
                          }
                     })
        expected_status_code=401
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Unauthorized.'}
        assert response.json() == expected_resp


    def test_insert_new_with_roles(self, client, db):
        add_meta_dict = {"absorption_cross_section": "Hearn1961", "measurement_method": "uv_abs"}
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA2", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 5,
                           "additional_metadata": json.dumps(add_meta_dict),
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"}] 
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}  
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'New timeseries created.', 'timeseries_id': 3}}
        assert response.json() == expected_resp

        new_response = client.get("/timeseries/3")
        new_expected_status_code = 200
        assert new_response.status_code == new_expected_status_code
        new_response_changelog = new_response.json()["changelog"][-1]
        expected_changelog = {'datetime': '2023-07-28T12:00:00+00:00', 'description': 'time series created', 'old_value': '', 'new_value': '',
                              'timeseries_id': 3, 'author_id': 1,
                              'type_of_change': 'created'}
        assert new_response_changelog == expected_changelog

                                                    
    def test_insert_new_without_existing_station(self, client, db):
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA2", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 97, "variable_id": 7,
                           "additional_metadata":"{}",
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"}] 
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 440
        assert response.status_code == expected_status_code
        expected_resp = 'Station (station_id: 97) not found in database.'
        assert response.json() == expected_resp


    def test_insert_duplicate_no_resource_provider(self, client, db):
        # one timeseries can be inserted without having a resource provider
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 7,
                           "additional_metadata":"{}",
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"}] 
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'New timeseries created.', 'timeseries_id': 3}}
        assert response.json() == expected_resp


    def test_insert_duplicate_wrong_resource_provider(self, client, db):
        # one timeseries can be inserted without having a resource provider
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 7,
                           "additional_metadata":"{}",
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"},
                                     {"role": "ResourceProvider", "contact_id": 1, "status": "Active"}] 
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 442
        assert response.status_code == expected_status_code
        expected_resp = 'Resource provider (contact_id: 1) not found in database.'
        assert response.json() == expected_resp


    def test_insert_duplicate(self, client, db):
        # one timeseries can be inserted without having a resource provider
        response = client.post("/timeseries/",
                json={"timeseries":
                          {"label": "CMA", "order": 1,
                           "sampling_frequency": "Hourly", "aggregation": "Mean", "data_origin_type": "Measurement",
                           "data_start_date": "2003-09-07T15:30:00+02:00",
                           "data_end_date": "2016-12-31T14:30:00+01:00",
                           'coverage': -1.0,
                           "data_origin": "Instrument", "sampling_height": 7.0,
                           "station_id": 2, "variable_id": 7,
                           "additional_metadata":"{}",
                           "roles": [{"role": "PointOfContact", "contact_id": 3, "status": "Active"},
                                     {"role": "Originator", "contact_id": 1, "status": "Active"},
                                     {"role": "ResourceProvider", "contact_id": 4, "status": "Active"}] 
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
                   )
        expected_status_code = 443
        assert response.status_code == expected_status_code
        expected_resp = {'detail': {'message': 'Timeseries already registered.', 'timeseries_id': 1}}
        assert response.json() == expected_resp

    def test_get_all_timeseries(self, client, db):
        response = client.get("/timeseries/?limit=1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [{
            'id': 1, 'label': 'CMA', 'order': 1, 'sampling_frequency': 'hourly',
            'aggregation': 'mean', 'data_start_date': '2003-09-07T15:30:00+00:00',
            'data_end_date': '2016-12-31T14:30:00+00:00', 'coverage': -1.0, 'data_origin': 'instrument', 'data_origin_type': 'measurement',
            'provider_version': 'N/A', 'doi': '', 'sampling_height': 7.0, 'additional_metadata': {'original_units': 'ppb'},
            'roles': [{'id': 2, 'role': 'resource provider', 'status': 'active',
                       'contact': {'id': 4, 'organisation': {'id': 1, 'name': 'UBA', 'longname': 'Umweltbundesamt',
                                                             'kind': 'government', 'city': 'Dessau-Roßlau', 'postcode': '06844', 'street_address': 'Wörlitzer Platz 1',
                                                             'country': 'Germany', 'homepage': 'https://www.umweltbundesamt.de', 'contact_url': 'mailto:immission@uba.de'}}},
                                  {'id': 3, 'role': 'principal investigator', 'status': 'active',
                                                           'contact': {'id': 3, 'person': {'email': 's.schroeder@fz-juelich.de','id': 3, 'isprivate': False,
                                                                                           'name': 'Sabine Schröder', 'orcid': '0000-0002-0309-8010', 'phone': '+49-2461-61-6397'}}}
                     ],
            'station': {'id': 2, 'codes': ['SDZ54421'], 'name': 'Shangdianzi',
                        'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                        'coordinate_validation_status': 'not checked', 'country': 'China',
                        'state': 'Beijing Shi', 'type': 'unknown', 'type_of_area': 'unknown',
                        'timezone': 'Asia/Shanghai',
                        'additional_metadata': {'add_type': 'nature reservation'},
                        'aux_images': [], 'aux_docs': [], 'aux_urls': [],
                        'globalmeta': {'climatic_zone_year2016': '6 (warm temperate dry)',
                                       'distance_to_major_road_year2020': -999.0,
                                       'dominant_ecoregion_year2017': '-1 (undefined)',
                                       'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                       'ecoregion_description_25km_year2017': '',
                                       'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                       'landcover_description_25km_year2012': '',
                                       'max_stable_nightlights_25km_year1992': -999.0,
                                       'max_stable_nightlights_25km_year2013': -999.0,
                                       'max_population_density_25km_year1990': -1.0,
                                       'max_population_density_25km_year2015': -1.0,
                                       'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                       'mean_stable_nightlights_1km_year2013': -999.0,
                                       'mean_stable_nightlights_5km_year2013': -999.0,
                                       'mean_nox_emissions_10km_year2000': -999.0,
                                       'mean_nox_emissions_10km_year2015': -999.0,
                                       'mean_population_density_250m_year1990': -1.0,
                                       'mean_population_density_250m_year2015': -1.0,
                                       'mean_population_density_5km_year1990': -1.0,
                                       'mean_population_density_5km_year2015': -1.0,
                                       'mean_topography_srtm_alt_1km_year1994': -999.0,
                                       'mean_topography_srtm_alt_90m_year1994': -999.0,
                                       'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                       'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                       'toar1_category': 'unclassified',
                                       'toar2_category': 'suburban'},
                        'changelog': []},
            'variable': {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                        'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                        'chemical_formula': 'C7H8', 'id': 7},
            'programme': {'id': 0, 'name': '', 'longname': '', 'homepage': '', 'description': ''}
            }]
        assert response.json() == expected_response


    def test_get_one_timeseries_with_fields(self, client, db):
        response = client.get("/timeseries/1?fields=additional_metadata")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'additional_metadata': {'original_units': 'ppb'} }
        assert response.json() == expected_response


    def test_get_all_timeseries_with_fields(self, client, db):
        response = client.get("/timeseries/?fields=additional_metadata")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [
                              {'additional_metadata': {'original_units': 'ppb'}},
                              {'additional_metadata':
                                  {'absorption_cross_section': 'Hearn 1961',
                                   'ebas_metadata_19740101000000_29y':
                                       {'Data level': '2',
                                        'Frameworks': 'GAW-WDCRG '
                                                      'NOAA-ESRL',
                                        'Station code': 'XXX',
                                        'Station name': 'Secret',
                                        'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA'
                                       },
                                   'measurement_method': 'uv_abs',
                                   'original_units': {'since_19740101000000': 'nmol/mol'}
                                  }
                              }
                            ]
        assert response.json() == expected_response


    def test_get_all_timeseries_filter_roles(self, client, db):
        response = client.get("/timeseries/?has_role=UBA")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [{'id': 1,
                              'label': 'CMA',
                              'order': 1,
                              'sampling_frequency': 'hourly',
                              'aggregation': 'mean',
                              'data_start_date': '2003-09-07T15:30:00+00:00',
                              'data_end_date': '2016-12-31T14:30:00+00:00',
                              'data_origin': 'instrument',
                              'data_origin_type': 'measurement',
                              'provider_version': 'N/A',
                              'sampling_height': 7.0,
                              'additional_metadata': {'original_units': 'ppb'},
                              'doi': '',
                              'coverage': -1.0,
                              'station': {'id': 2,
                                          'codes': ['SDZ54421'],
                                          'name': 'Shangdianzi',
                                          'coordinates': {'lat': 40.65, 'lng': 117.106, 'alt': 293.9},
                                          'coordinate_validation_status': 'not checked',
                                          'country': 'China',
                                          'state': 'Beijing Shi',
                                          'type': 'unknown',
                                          'type_of_area': 'unknown',
                                          'timezone': 'Asia/Shanghai',
                                          'additional_metadata': {'add_type': 'nature reservation'},
                                          'aux_images': [],
                                          'aux_docs': [],
                                          'aux_urls': [],
                                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                                         'htap_region_tier1_year2010': '11 (MDE Middle East: S. Arabia, Oman, etc, Iran, Iraq)',
                                                         'dominant_landcover_year2012': '11 (Cropland, rainfed, herbaceous cover)',
                                                         'landcover_description_25km_year2012': '',
                                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                                         'ecoregion_description_25km_year2017': '',
                                                         'distance_to_major_road_year2020': -999.0,
                                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year1992': -999.0,
                                                         'mean_population_density_250m_year2015': -1.0,
                                                         'mean_population_density_5km_year2015': -1.0,
                                                         'max_population_density_25km_year2015': -1.0,
                                                         'mean_population_density_250m_year1990': -1.0,
                                                         'mean_population_density_5km_year1990': -1.0,
                                                         'max_population_density_25km_year1990': -1.0,
                                                         'mean_nox_emissions_10km_year2015': -999.0,
                                                         'mean_nox_emissions_10km_year2000': -999.0,
                                                         'toar1_category': 'unclassified',
                                                         'toar2_category': 'suburban'
                                                        },
                                          'changelog': []
                                         },
                              'variable': {'name': 'toluene',
                                           'longname': 'toluene',
                                           'displayname': 'Toluene',
                                           'cf_standardname': 'mole_fraction_of_toluene_in_air',
                                           'units': 'nmol mol-1',
                                           'chemical_formula': 'C7H8',
                                           'id': 7
                                          },
                              'programme': {'id': 0,
                                            'name': '',
                                            'longname': '',
                                            'homepage': '',
                                            'description': ''
                                           },
                              'roles': [{'id': 2,
                                         'role': 'resource provider',
                                         'status': 'active',
                                         'contact': {'id': 4,
                                                     'organisation': {'id': 1,
                                                                      'name': 'UBA',
                                                                      'longname': 'Umweltbundesamt',
                                                                      'kind': 'government',
                                                                      'city': 'Dessau-Roßlau',
                                                                      'postcode': '06844',
                                                                      'street_address': 'Wörlitzer Platz 1',
                                                                      'country': 'Germany',
                                                                      'homepage': 'https://www.umweltbundesamt.de',
                                                                      'contact_url': 'mailto:immission@uba.de'
                                                                     }
                                                    }
                                        },
                                        {'id': 3,
                                         'role': 'principal investigator',
                                         'status': 'active',
                                         'contact': {'id': 3,
                                                     'person': {'email': 's.schroeder@fz-juelich.de',
                                                     'id': 3,
                                                     'isprivate': False,
                                                     'name': 'Sabine Schröder',
                                                     'orcid': '0000-0002-0309-8010',
                                                     'phone': '+49-2461-61-6397'}}}
                                       ]
                             }]
        assert response.json() == expected_response


    def test_get_all_timeseries_filter_not_roles(self, client, db):
        response = client.get("/timeseries/?has_role=~UBA")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [{'id': 2,
                              'label': 'CMA',
                              'order': 1,
                              'sampling_frequency': 'hourly',
                              'aggregation': 'mean',
                              'data_start_date': '2003-09-07T15:30:00+00:00',
                              'data_end_date': '2016-12-31T14:30:00+00:00',
                              'data_origin': 'instrument',
                              'data_origin_type': 'measurement',
                              'provider_version': 'N/A',
                              'sampling_height': 7.0,
                              'additional_metadata': {'original_units': {'since_19740101000000': 'nmol/mol'}, 'measurement_method': 'uv_abs', 'absorption_cross_section': 'Hearn 1961', 'ebas_metadata_19740101000000_29y': {'Submitter': 'Unknown, Lady, lady.unknown@unknown.com, some long division name, SHORT, , 111 Streetname, , zipcode, Boulder, CO, USA', 'Data level': '2', 'Frameworks': 'GAW-WDCRG NOAA-ESRL', 'Station code': 'XXX', 'Station name': 'Secret'}},
                              'doi': '',
                              'coverage': -1.0,
                              'station': {'id': 3,
                                          'codes': ['China_test8'],
                                          'name': 'Test_China',
                                          'coordinates': {'lat': 36.256, 'lng': 117.106, 'alt': 1534.0},
                                          'coordinate_validation_status': 'not checked',
                                          'country': 'China',
                                          'state': 'Shandong Sheng',
                                          'type': 'unknown',
                                          'type_of_area': 'unknown',
                                          'timezone': 'Asia/Shanghai',
                                          'additional_metadata': {},
                                          'aux_images': [],
                                          'aux_docs': [],
                                          'aux_urls': [],
                                          'globalmeta': {'mean_topography_srtm_alt_90m_year1994': -999.0,
                                                         'mean_topography_srtm_alt_1km_year1994': -999.0,
                                                         'max_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'min_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'stddev_topography_srtm_relative_alt_5km_year1994': -999.0,
                                                         'climatic_zone_year2016': '6 (warm temperate dry)',
                                                         'htap_region_tier1_year2010': '10 (SAF Sub Saharan/sub Sahel Africa)',
                                                         'dominant_landcover_year2012': '10 (Cropland, rainfed)',
                                                         'landcover_description_25km_year2012': '',
                                                         'dominant_ecoregion_year2017': '-1 (undefined)',
                                                         'ecoregion_description_25km_year2017': '',
                                                         'distance_to_major_road_year2020': -999.0,
                                                         'mean_stable_nightlights_1km_year2013': -999.0,
                                                         'mean_stable_nightlights_5km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year2013': -999.0,
                                                         'max_stable_nightlights_25km_year1992': -999.0,
                                                         'mean_population_density_250m_year2015': -1.0,
                                                         'mean_population_density_5km_year2015': -1.0,
                                                         'max_population_density_25km_year2015': -1.0,
                                                         'mean_population_density_250m_year1990': -1.0,
                                                         'mean_population_density_5km_year1990': -1.0,
                                                         'max_population_density_25km_year1990': -1.0,
                                                         'mean_nox_emissions_10km_year2015': -999.0,
                                                         'mean_nox_emissions_10km_year2000': -999.0,
                                                         'toar1_category': 'unclassified',
                                                         'toar2_category': 'suburban'},
                                          'changelog': []},
                              'variable': {'name': 'o3',
                                           'longname': 'ozone',
                                           'displayname': 'Ozone',
                                           'cf_standardname': 'mole_fraction_of_ozone_in_air',
                                           'units': 'nmol mol-1',
                                           'chemical_formula': 'O3',
                                           'id': 5},
                              'programme': {'id': 0,
                                            'name': '',
                                            'longname': '',
                                            'homepage': '',
                                            'description': ''},
                              'roles': [{'id': 1,
                                         'role': 'resource provider',
                                         'status': 'active',
                                         'contact': {'id': 5,
                                                     'organisation': {'id': 2,
                                                                      'name': 'FZJ',
                                                                      'longname': 'Forschungszentrum Jülich',
                                                                      'kind': 'research',
                                                                      'city': 'Jülich',
                                                                      'postcode': '52425',
                                                                      'street_address': 'Wilhelm-Johnen-Straße',
                                                                      'country': 'Germany',
                                                                      'homepage': 'https://www.fz-juelich.de',
                                                                      'contact_url': 'mailto:toar-data@fz-juelich.de'
                                                                     }
                                                    }
                                        }
                                       ]
                             }
                            ]
        assert response.json() == expected_response


    def test_get_timeseries_not_found(self, client, db):
        response = client.get("/timeseries/-1")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_response = {"detail": "Timeseries not found."}
        assert response.json() == expected_response


    def test_get_timeseries_invalid_type(self, client, db):
        response = client.get("/timeseries/A")
        expected_status_code = 422
        assert response.status_code == expected_status_code
        """expected_response = {"detail": "Timeseries not found."}
        assert response.json() == expected_response"""

    def test_get_timeseries2_not_found(self, client, db):
        response = client.get("/timeseries/id/-1")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_response = {"detail": "Timeseries not found."}
        assert response.json() == expected_response

    def test_get_timeseries2_invalid_type(self, client, db):
        response = client.get("/timeseries/id/A")
        expected_status_code = 422
        assert response.status_code == expected_status_code
        """expected_response = {"detail": "Timeseries not found."}
        assert response.json() == expected_response"""


#   def test_get_unique_timeseries(self, client, db):
#       response = client.get("/timeseries/unique/?station_id=2&variable_id=7")
#       expected_status_code = 200
#       assert response.status_code == expected_status_code

#       assert response.json() == client.get("/timeseries/1").json()


#   def test_get_unique_timeseries_not_found(self, client, db):
#       response = client.get("/timeseries/unique/?station_id=-1&variable_id=-1")
#       expected_status_code = 404
#       assert response.status_code == expected_status_code
#       expected_response = {"detail": "Timeseries not found."}
#       assert response.json() == expected_response


    def test_get_unique_timeseries_invalid_type(self, client, db):
        response = client.get("/timeseries/unique/?station_id=A&variable_id=A")
        expected_status_code = 422
        assert response.status_code == expected_status_code
        """expected_response = {"detail": "Timeseries not found."}
        assert response.json() == expected_response"""


    def test_get_citation(self, client, db):
        fixed_time = datetime(2023, 7, 28, 12, 0, 0)

        with patch('toardb.timeseries.crud.dt.datetime') as mock_datetime:
            mock_datetime.now.return_value = fixed_time
            response = client.get("/timeseries/citation/2")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = {'attribution': 'Test-Attributions to be announced',
                             'citation': 'Forschungszentrum Jülich: time series of o3 at Test_China, accessed from the TOAR database on 2023-07-28 12:00:00',
                             'license': 'This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/'
                            }
        assert response.json() == expected_response


    # test the endpoint needed for TOARgridding
    def test_get_contributors_json(self, client, db):
        response = client.post("/timeseries/request_contributors?format=json", files={"file": open("tests/fixtures/timeseries/timeseries_id.txt", "rb")})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [
                             {'contact': {'id': 5,
                                          'organisation': {'city': 'Jülich',
                                                           'contact_url': 'mailto:toar-data@fz-juelich.de',
                                                           'country': 'Germany',
                                                           'homepage': 'https://www.fz-juelich.de',
                                                           'id': 2,
                                                           'kind': 'research',
                                                           'longname': 'Forschungszentrum Jülich',
                                                           'name': 'FZJ',
                                                           'postcode': '52425',
                                                           'street_address': 'Wilhelm-Johnen-Straße'
                                                          },
                                          'person': None
                                         },
                              'id': 1,
                              'role': 'resource provider',
                              'status': 'active'
                             },
                             {'contact': {'id': 4,
                                          'organisation': {'city': 'Dessau-Roßlau',
                                                           'contact_url': 'mailto:immission@uba.de',
                                                           'country': 'Germany',
                                                           'homepage': 'https://www.umweltbundesamt.de',
                                                           'id': 1,
                                                           'kind': 'government',
                                                           'longname': 'Umweltbundesamt',
                                                           'name': 'UBA',
                                                           'postcode': '06844',
                                                           'street_address': 'Wörlitzer Platz 1'
                                                          },
                                          'person': None
                                         },
                              'id': 2,
                              'role': 'resource provider',
                              'status': 'active'
                             },
                             {'id': 3,
                              'role': 'principal investigator',
                              'status': 'active',
                              'contact': {'id': 3,
                                          'organisation': None,
                                          'person': {'email': 's.schroeder@fz-juelich.de',
                                                     'id': 3,
                                                     'isprivate': False,
                                                     'name': 'Sabine Schröder',
                                                     'orcid': '0000-0002-0309-8010',
                                                     'phone': '+49-2461-61-6397'}}}
                            ]
        assert response.json() == expected_response


    def test_get_contributors_text(self, client, db):
        response = client.post("/timeseries/request_contributors?format=text", files={"file": open("tests/fixtures/timeseries/timeseries_id.txt", "rb")})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = 'organisations: Forschungszentrum Jülich;Umweltbundesamt; persons:Sabine Schröder'
        assert response.json() == expected_response


    def test_get_contributors_default(self, client, db):
        response = client.post("/timeseries/request_contributors", files={"file": open("tests/fixtures/timeseries/timeseries_id.txt", "rb")})
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = 'organisations: Forschungszentrum Jülich;Umweltbundesamt; persons:Sabine Schröder'
        assert response.json() == expected_response


    def test_get_contributors_unknown_format(self, client, db):
        response = client.post("/timeseries/request_contributors?format=CMOR", files={"file": open("tests/fixtures/timeseries/timeseries_id.txt", "rb")})
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_response = 'not a valid format: CMOR'
        assert response.json() == expected_response 


    def test_register_contributors_list(self, client, db):
        response = client.post("/timeseries/register_timeseries_list_of_contributors/5f0df73a-bd0f-48b9-bb17-d5cd36f89598",
                               data='''[1,2]''',
                               headers={"email": "s.schroeder@fz-juelich.de"} )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = '5f0df73a-bd0f-48b9-bb17-d5cd36f89598 successfully registered.'
        assert response.json() == expected_response


    def test_register_duplicate_contributors_list(self, client, db):
        response = client.post("/timeseries/register_timeseries_list_of_contributors/7f0df73a-bd0f-48b9-bb17-d5cd36f89598",
                               data='''[1,2]''',
                               headers={"email": "s.schroeder@fz-juelich.de"} )
        expected_status_code = 443
        assert response.status_code == expected_status_code
        expected_response = '7f0df73a-bd0f-48b9-bb17-d5cd36f89598 already registered.'
        assert response.json() == expected_response


    def test_request_registered_contributors_list_json(self, client, db):
        response = client.get("/timeseries/request_timeseries_list_of_contributors/7f0df73a-bd0f-48b9-bb17-d5cd36f89598?format=json")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = [
                             {'contact': {'id': 5,
                                          'organisation': {'city': 'Jülich',
                                                           'contact_url': 'mailto:toar-data@fz-juelich.de',
                                                           'country': 'Germany',
                                                           'homepage': 'https://www.fz-juelich.de',
                                                           'id': 2,
                                                           'kind': 'research',
                                                           'longname': 'Forschungszentrum Jülich',
                                                           'name': 'FZJ',
                                                           'postcode': '52425',
                                                           'street_address': 'Wilhelm-Johnen-Straße'
                                                          },
                                          'person': None
                                         },
                              'id': 1,
                              'role': 'resource provider',
                              'status': 'active'
                             },
                             {'contact': {'id': 4,
                                          'organisation': {'city': 'Dessau-Roßlau',
                                                           'contact_url': 'mailto:immission@uba.de',
                                                           'country': 'Germany',
                                                           'homepage': 'https://www.umweltbundesamt.de',
                                                           'id': 1,
                                                           'kind': 'government',
                                                           'longname': 'Umweltbundesamt',
                                                           'name': 'UBA',
                                                           'postcode': '06844',
                                                           'street_address': 'Wörlitzer Platz 1'
                                                          },
                                          'person': None
                                         },
                              'id': 2,
                              'role': 'resource provider',
                              'status': 'active'
                             },
                             {'id': 3,
                              'role': 'principal investigator',
                              'status': 'active',
                              'contact': {'id': 3,
                                          'organisation': None,
                                          'person': {'email': 's.schroeder@fz-juelich.de',
                                                     'id': 3,
                                                     'isprivate': False,
                                                     'name': 'Sabine Schröder',
                                                     'orcid': '0000-0002-0309-8010',
                                                     'phone': '+49-2461-61-6397'}}}
                            ]
        assert response.json() == expected_response


    def test_request_registered_contributors_list_text(self, client, db):
        response = client.get("/timeseries/request_timeseries_list_of_contributors/7f0df73a-bd0f-48b9-bb17-d5cd36f89598?format=text")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_response = 'organisations: Forschungszentrum Jülich;Umweltbundesamt; persons:Sabine Schröder'
        assert response.json() == expected_response


    def test_request_registered_contributors_list_unknown_rid(self, client, db):
        response = client.get("/timeseries/request_timeseries_list_of_contributors/7f0df73a-bd0f-58b9-bb17-d5cd36f89598?format=text")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_response = 'not a registered request id: 7f0df73a-bd0f-58b9-bb17-d5cd36f89598'
        assert response.json() == expected_response


    # 3. tests updating timeseries metadata

    def test_patch_timeseries_no_description(self, client, db):
        response = client.patch("/timeseries/id/1",
                json={"timeseries":
                          {"sampling_frequency":"Daily"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "description text ist missing."}
        assert response.json() == expected_resp


    def test_patch_timeseries_not_found(self, client, db):
        response = client.patch("/timeseries/id/-1?description=changed sampling_frequency",
                json={"timeseries":
                          {"sampling_frequency":"Daily"}
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail": "Time series for patching not found."}
        assert response.json() == expected_resp


    def test_patch_timeseries_single_item(self, client, db):
        response = client.patch("/timeseries/id/1?description=changed sampling_frequency",
                json={"timeseries": 
                          {"sampling_frequency": "Daily"}
                          },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': { 'message': 'timeseries patched.',
                                     'timeseries_id': 1 }
                        } 
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/timeseries/id/{response_json['detail']['timeseries_id']}")
        response_json = response.json()
        # just check special changes
        assert response_json['sampling_frequency'] == "daily"
        assert response_json['changelog'][0]['old_value'] == "{'sampling_frequency': 'Hourly'}"
        assert response_json['changelog'][0]['new_value'] == "{'sampling_frequency': 'Daily'}"
        assert response_json['changelog'][0]['author_id'] == 1
        assert response_json['changelog'][0]['type_of_change'] == 'single value correction in metadata'


    def test_patch_timeseries_multiple_items(self, client, db):
        response = client.patch("/timeseries/id/1?description=changed some metadata",
                json={"timeseries": 
                          {"sampling_frequency": "Daily",
                           "aggregation": "MeanOf4Samples",
                           "data_origin": "COSMOREA6",
                           "data_origin_type": "Model"}
                          },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': { 'message': 'timeseries patched.',
                                     'timeseries_id': 1 }
                        } 
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/timeseries/id/{response_json['detail']['timeseries_id']}")
        response_json = response.json()
        # just check special changes
        assert response_json['sampling_frequency'] == "daily"
        assert response_json['changelog'][0]['old_value'] == "{'sampling_frequency': 'Hourly', 'aggregation': 'Mean', 'data_origin': 'Instrument', 'data_origin_type': 'Measurement'}"
        assert response_json['changelog'][0]['new_value'] == "{'sampling_frequency': 'Daily', 'aggregation': 'MeanOf4Samples', 'data_origin': 'COSMOREA6', 'data_origin_type': 'Model'}"
        assert response_json['changelog'][0]['author_id'] == 1
        assert response_json['changelog'][0]['type_of_change'] == 'comprehensive metadata revision'


    def test_patch_timeseries_roles(self, client, db):
        response = client.patch("/timeseries/id/1?description=changed roles",
                json={"timeseries": 
                          {"roles": [{"role": "ResourceProvider", "contact_id": 5, "status": "Active"}]
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': { 'message': 'timeseries patched.',
                                     'timeseries_id': 1 }
                        } 
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/timeseries/id/{response_json['detail']['timeseries_id']}")
        response_json = response.json()
        # just check special changes
        response_roles = [{'contact': {'id': 5,
                                       'organisation': {'city': 'Jülich',
                                                        'contact_url': 'mailto:toar-data@fz-juelich.de',
                                                        'country': 'Germany',
                                                        'homepage': 'https://www.fz-juelich.de',
                                                        'id': 2,
                                                        'kind': 'research',
                                                        'longname': 'Forschungszentrum Jülich',
                                                        'name': 'FZJ',
                                                        'postcode': '52425',
                                                        'street_address': 'Wilhelm-Johnen-Straße',
                                                    },
                                                },
                           'id': 1,
                           'role': 'resource provider',
                           'status': 'active'},
                          {'contact': {'id': 4,
                                       'organisation': {'city': 'Dessau-Roßlau',
                                                        'contact_url': 'mailto:immission@uba.de',
                                                        'country': 'Germany',
                                                        'homepage': 'https://www.umweltbundesamt.de',
                                                        'id': 1,
                                                        'kind': 'government',
                                                        'longname': 'Umweltbundesamt',
                                                        'name': 'UBA',
                                                        'postcode': '06844',
                                                        'street_address': 'Wörlitzer Platz 1',
                                                        'contact_url': 'mailto:immission@uba.de'}},
                           'id': 2,
                           'role': 'resource provider',
                           'status': 'active'},
                          {'id': 3,
                           'role': 'principal investigator',
                           'status': 'active',
                           'contact': {'id': 3,
                                       'person': {'email': 's.schroeder@fz-juelich.de',
                                                  'id': 3,
                                                  'isprivate': False,
                                                  'name': 'Sabine Schröder',
                                                  'orcid': '0000-0002-0309-8010',
                                                  'phone': '+49-2461-61-6397'}}}]
        set_expected_response_roles = {json.dumps(item, sort_keys=True) for item in response_roles}
        set_response_roles = {json.dumps(item, sort_keys=True) for item in response_json['roles']}
        assert set_response_roles == set_expected_response_roles
        assert response_json['changelog'][0]['old_value'] == "{'roles': [{'role': 'ResourceProvider', 'status': 'Active', 'contact_id': 4}, {'role': 'PrincipalInvestigator', 'status': 'Active', 'contact_id': 3}]}"
        assert response_json['changelog'][0]['new_value'] == "{'roles': [{'role': 'ResourceProvider', 'status': 'Active', 'contact_id': 4}, {'role': 'PrincipalInvestigator', 'status': 'Active', 'contact_id': 3}, {'role': 'ResourceProvider', 'contact_id': 5, 'status': 'Active'}]}"
        assert response_json['changelog'][0]['author_id'] == 1
        assert response_json['changelog'][0]['type_of_change'] == 'single value correction in metadata'


    def test_patch_timeseries_annotations(self, client, db):
        response = client.patch("/timeseries/id/1?description=changed annotations",
                json={"timeseries":
                          {"annotations": [{"kind": "User",
                                            "text": "some foo",
                                            "date_added": "2021-07-27 00:00",
                                            "approved": True,
                                            "contributor_id":1}]
                          }
                     },
                headers={"email": "s.schroeder@fz-juelich.de"}
        )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'detail': { 'message': 'timeseries patched.',
                                     'timeseries_id': 1 }
                        }
        response_json = response.json()
        assert response_json == expected_resp
        response = client.get(f"/timeseries/id/{response_json['detail']['timeseries_id']}")
        response_json = response.json()
        # just check special changes
        response_annotations = [{"kind": "user comment",
                                 "text": "some foo",
                                 "date_added": "2021-07-27 00:00",
                                 "approved": True,
                                 "contributor_id":1}
                               ]
        assert response_json['changelog'][0]['old_value'] == "{'annotations': []}"
        assert response_json['changelog'][0]['new_value'] == "{'annotations': [{'kind': 'User', 'text': 'some foo', 'date_added': '2021-07-27 00:00', 'approved': True, 'contributor_id': 1}]}"
        assert response_json['changelog'][0]['author_id'] == 1
        assert response_json['changelog'][0]['type_of_change'] == 'single value correction in metadata'


    """def test_get_timeseries_changelog(self, client, db):
        response = client.get("/timeseries_changelog/{id}")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        new_response = response.json()["datetime"] = ""
        expected_response = [{"datetime":"","description":"fake creation of timeseries","old_value":"","new_value":"","timeseries_id":8,"author_id":1,"type_of_change":0,"period_start":None,"period_end":None,"version":None}]
        assert response.json() == expected_response"""

