# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

import pytest
import json
from toardb.variables.models import Variable
# Required imports 'create_test_database'
from toardb.test_base import (
    client,
    get_test_db,
    create_test_database,
    url,
    get_test_engine,
    test_db_session as db,
)

class TestApps:
    def setup(self):
        self.application_url = "/variables/"

    """Set up all the data before each test
       If you want the setup only once (per test module),
       the scope argument is not working in the expected way, as discussed here:
       https://stackoverflow.com/questions/45817153/py-test-fixture-use-function-fixture-in-scope-fixture
    """
    @pytest.fixture(autouse=True)
    def setup_db_data(self, db):
        # id_seq will not be reset automatically between tests!
        _db_conn = get_test_engine()
        fake_conn = _db_conn.raw_connection()
        fake_cur = fake_conn.cursor()
        fake_cur.execute("ALTER SEQUENCE variables_id_seq RESTART WITH 1")
        fake_conn.commit()
        infilename = "tests/fixtures/variables/variables.json"
        with open(infilename, encoding="utf-8") as f:
            metajson=json.load(f)
            for entry in metajson:
                new_variable = Variable(**entry)
                db.add(new_variable)
                db.commit()
                db.refresh(new_variable)


    def test_get_variables(self, client, db):
        response = client.get("/variables/?limit=3")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'name': 'benzene', 'longname': 'benzene', 'displayname': 'Benzene',
                          'cf_standardname': 'mole_fraction_of_benzene_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'C6H6', 'id': 1},
                         {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                          'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'CO', 'id': 2},
                         {'name': 'no', 'longname': 'nitrogen monoxide', 'displayname': 'NO',
                          'cf_standardname': 'mole_fraction_of_nitrogen_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'NO', 'id': 3}]
        assert response.json() == expected_resp


    def test_get_all_variables_default_limit(self, client, db):
        response = client.get("/variables/")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'name': 'benzene', 'longname': 'benzene', 'displayname': 'Benzene',
                          'cf_standardname': 'mole_fraction_of_benzene_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'C6H6', 'id': 1},
                         {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                          'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'CO', 'id': 2},
                         {'name': 'no', 'longname': 'nitrogen monoxide', 'displayname': 'NO',
                          'cf_standardname': 'mole_fraction_of_nitrogen_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'NO', 'id': 3},
                         {'name': 'pm1', 'longname': 'particles up to 1 µm diameter', 'displayname': 'PM 1',
                          'cf_standardname': 'mass_concentration_of_pm1_ambient_aerosol_in_air', 'units': 'µg m-3', 'chemical_formula': '', 'id': 4},
                         {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                          'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'O3', 'id': 5},
                         {'name': 'no2', 'longname': 'nitrogen dioxide', 'displayname': 'NO2',
                          'cf_standardname': 'mole_fraction_of_nitrogen_dioxide_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'NO2', 'id': 6},
                         {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                          'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C7H8', 'id': 7},
                         {'name': 'so2', 'longname': 'Sulphur dioxide', 'displayname': 'SO2',
                          'cf_standardname': 'mole_fraction_of_sulfur_dioxide_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'SO2', 'id': 8},
                         {'name': 'ethane', 'longname': 'Ethane', 'displayname': 'Ethane',
                          'cf_standardname': 'mole_fraction_of_ethane_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C2H6', 'id': 9},
                         {'name': 'propane', 'longname': 'Propane', 'displayname': 'Propane',
                          'cf_standardname': 'mole_fraction_of_propane_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C3H8', 'id': 10}]
        assert response.json() == expected_resp

    def test_get_all_variables_unlimited(self, client, db):
        response = client.get("/variables/?limit=None")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'name': 'benzene', 'longname': 'benzene', 'displayname': 'Benzene',
                          'cf_standardname': 'mole_fraction_of_benzene_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'C6H6', 'id': 1},
                         {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                          'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'CO', 'id': 2},
                         {'name': 'no', 'longname': 'nitrogen monoxide', 'displayname': 'NO',
                          'cf_standardname': 'mole_fraction_of_nitrogen_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'NO', 'id': 3},
                         {'name': 'pm1', 'longname': 'particles up to 1 µm diameter', 'displayname': 'PM 1',
                          'cf_standardname': 'mass_concentration_of_pm1_ambient_aerosol_in_air', 'units': 'µg m-3', 'chemical_formula': '', 'id': 4},
                         {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                          'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'O3', 'id': 5},
                         {'name': 'no2', 'longname': 'nitrogen dioxide', 'displayname': 'NO2',
                          'cf_standardname': 'mole_fraction_of_nitrogen_dioxide_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'NO2', 'id': 6},
                         {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                          'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C7H8', 'id': 7},
                         {'name': 'so2', 'longname': 'Sulphur dioxide', 'displayname': 'SO2',
                          'cf_standardname': 'mole_fraction_of_sulfur_dioxide_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'SO2', 'id': 8},
                         {'name': 'ethane', 'longname': 'Ethane', 'displayname': 'Ethane',
                          'cf_standardname': 'mole_fraction_of_ethane_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C2H6', 'id': 9},
                         {'name': 'propane', 'longname': 'Propane', 'displayname': 'Propane',
                          'cf_standardname': 'mole_fraction_of_propane_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'C3H8', 'id': 10},
                         {'name': 'ox', 'longname': 'Ox', 'displayname': 'Ox',
                          'cf_standardname': '', 'units': 'nmol mol-1', 'chemical_formula': '', 'id': 11},
                         {'name': 'aswdir', 'longname': 'direct downward sw radiation', 'displayname': 'downward sw radiation',
                          'cf_standardname': 'surface_downwelling_shortwave_flux_in_air', 'units': 'W/m**2', 'chemical_formula': '', 'id': 12},
                         {'name': 'pm10', 'longname': 'particles up to 10 µm diameter', 'displayname': 'PM 10',
                          'cf_standardname': 'mass_concentration_of_pm10_ambient_aerosol_in_air', 'units': 'µg m-3', 'chemical_formula': '', 'id': 13},
                         {'name': 'rn', 'longname': 'radon', 'displayname': 'Radon',
                          'cf_standardname': '', 'units': 'mBq m-3', 'chemical_formula': 'Rn', 'id': 14},
                         {'name': 'mpxylene', 'longname': 'm,p-xylene', 'displayname': 'm,p-Xylene',
                          'cf_standardname': '', 'units': 'nmol mol-1', 'chemical_formula': 'C8H10', 'id': 15},
                         {'name': 'oxylene', 'longname': 'o-xylene', 'displayname': 'o-Xylene',
                          'cf_standardname': '', 'units': 'nmol mol-1', 'chemical_formula': 'C8H10', 'id': 16},
                         {'name': 'ch4', 'longname': 'Methane', 'displayname': 'CH4',
                          'cf_standardname': 'mole_fraction_of_methane_in_air', 'units': 'nmol mol-1', 'chemical_formula': 'CH4', 'id': 17},
                         {'name': 'wdir', 'longname': 'wind direction', 'displayname': 'Wind direction',
                          'cf_standardname': 'wind_from_direction', 'units': 'degrees', 'chemical_formula': '', 'id': 18},
                         {'name': 'pm2p5', 'longname': 'particles up to 2.5 µm diameter', 'displayname': 'PM 2.5',
                          'cf_standardname': 'mass_concentration_of_pm2p5_ambient_aerosol_in_air', 'units': 'µg m-3', 'chemical_formula': '', 'id': 19},
                         {'name': 'nox', 'longname': 'reactive nitrogen oxides (NO+NO2)', 'displayname': 'NOx',
                          'cf_standardname': 'mole_fraction_of_nox_expressed_as_nitrogen_in_air', 'units': 'nmol mol-1', 'chemical_formula': '', 'id': 20},
                         {'name': 'temp', 'longname': 'atmospheric temperature', 'displayname': 'Temperature',
                          'cf_standardname': 'air_temperature', 'units': 'degC', 'chemical_formula': '', 'id': 21},
                         {'name': 'wspeed', 'longname': 'wind speed', 'displayname': 'Wind speed',
                          'cf_standardname': 'wind_speed', 'units': 'm s-1', 'chemical_formula': '', 'id': 22},
                         {'name': 'press', 'longname': 'atmospheric pressure', 'displayname': 'Pressure',
                          'cf_standardname': 'air_pressure', 'units': 'hPa', 'chemical_formula': '', 'id': 23},
                         {'name': 'cloudcover', 'longname': 'total cloud cover', 'displayname': 'total cloud cover',
                          'cf_standardname': 'cloud_area_fraction', 'units': '%', 'chemical_formula': '', 'id': 24},
                         {'name': 'pblheight', 'longname': 'height of PBL', 'displayname': 'PBL height',
                          'cf_standardname': 'atmosphere_boundary_layer_thickness', 'units': 'm', 'chemical_formula': '', 'id': 25},
                         {'name': 'relhum', 'longname': 'relative humidity', 'displayname': 'relative humidity',
                          'cf_standardname': 'relative_humidity', 'units': '%', 'chemical_formula': '', 'id': 26},
                         {'name': 'totprecip', 'longname': 'total precipitation', 'displayname': 'total precipitation',
                          'cf_standardname': 'precipitation_amount', 'units': 'kg m-2', 'chemical_formula': '', 'id': 27},
                         {'name': 'u', 'longname': 'u-component (zonal) of wind', 'displayname': 'U',
                          'cf_standardname': 'x_wind', 'units': 'm s-1', 'chemical_formula': '', 'id': 28},
                         {'name': 'v', 'longname': 'v-component (meridional) of wind', 'displayname': 'V',
                          'cf_standardname': 'y_wind', 'units': 'm s-1', 'chemical_formula': '', 'id': 29},
                         {'name': 'albedo', 'longname': 'albedo', 'displayname': 'albedo',
                          'cf_standardname': 'surface_albedo', 'units': '%', 'chemical_formula': '', 'id': 30},
                         {'name': 'aswdifu', 'longname': 'diffuse upward sw radiation', 'displayname': 'upward sw radiation',
                          'cf_standardname': '', 'units': 'W/m**2', 'chemical_formula': '', 'id': 31},
                         {'name': 'humidity', 'longname': 'atmospheric humidity', 'displayname': 'Humidity',
                          'cf_standardname': 'specific_humidity', 'units': 'g kg-1', 'chemical_formula': '', 'id': 32},
                         {'name': 'irradiance', 'longname': 'global surface irradiance', 'displayname': 'solar irradiance',
                          'cf_standardname': '', 'units': 'W m-2', 'chemical_formula': '', 'id': 33}]
        assert response.json() == expected_resp


    def test_get_all_variables_using_fields(self, client, db):
        response = client.get("/variables/?limit=None&fields=name,units")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'name': 'benzene', 'units': 'nmol mol-1'},
                         {'name': 'co', 'units': 'nmol mol-1'},
                         {'name': 'no', 'units': 'nmol mol-1'},
                         {'name': 'pm1', 'units': 'µg m-3'},
                         {'name': 'o3', 'units': 'nmol mol-1'},
                         {'name': 'no2', 'units': 'nmol mol-1'},
                         {'name': 'toluene', 'units': 'nmol mol-1'},
                         {'name': 'so2', 'units': 'nmol mol-1'},
                         {'name': 'ethane', 'units': 'nmol mol-1'},
                         {'name': 'propane', 'units': 'nmol mol-1'},
                         {'name': 'ox', 'units': 'nmol mol-1'},
                         {'name': 'aswdir', 'units': 'W/m**2'},
                         {'name': 'pm10', 'units': 'µg m-3'},
                         {'name': 'rn', 'units': 'mBq m-3'},
                         {'name': 'mpxylene', 'units': 'nmol mol-1'},
                         {'name': 'oxylene', 'units': 'nmol mol-1'},
                         {'name': 'ch4', 'units': 'nmol mol-1'},
                         {'name': 'wdir', 'units': 'degrees'},
                         {'name': 'pm2p5', 'units': 'µg m-3'},
                         {'name': 'nox', 'units': 'nmol mol-1'},
                         {'name': 'temp', 'units': 'degC'},
                         {'name': 'wspeed', 'units': 'm s-1'},
                         {'name': 'press', 'units': 'hPa'},
                         {'name': 'cloudcover', 'units': '%'},
                         {'name': 'pblheight', 'units': 'm'},
                         {'name': 'relhum', 'units': '%'},
                         {'name': 'totprecip', 'units': 'kg m-2'},
                         {'name': 'u', 'units': 'm s-1'},
                         {'name': 'v', 'units': 'm s-1'},
                         {'name': 'albedo', 'units': '%'},
                         {'name': 'aswdifu', 'units': 'W/m**2'},
                         {'name': 'humidity', 'units': 'g kg-1'},
                         {'name': 'irradiance', 'units': 'W m-2'}]
        assert response.json() == expected_resp


    def test_get_all_variables_wrong_fieldname(self, client, db):
        response = client.get("/variables/?limit=None&fields=name,units,bla")
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = "Wrong field given: bla"
        assert response.json() == expected_resp


    def test_get_special(self, client, db):
        response = client.get("/variables/id/2")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                         'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air',
                         'units': 'nmol mol-1', 'chemical_formula': 'CO', 'id': 2}
        assert response.json() == expected_resp


    def test_get_special_out_of_index(self, client, db):
        response = client.get("/variables/id/99")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail":"Variable not found."}
        assert response.json() == expected_resp


    def test_get_special_by_name(self, client, db):
        response = client.get("/variables/CO")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                         'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air',
                         'units': 'nmol mol-1', 'chemical_formula': 'CO', 'id': 2}
        assert response.json() == expected_resp


    def test_get_special_by_units(self, client, db):
        response = client.get("/variables/?units=nmol mol-1")
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp = [{'name': 'benzene', 'longname': 'benzene', 'displayname': 'Benzene',
                          'cf_standardname': 'mole_fraction_of_benzene_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'C6H6', 'id': 1},
                         {'name': 'co', 'longname': 'carbon monoxide', 'displayname': 'CO',
                          'cf_standardname': 'mole_fraction_of_carbon_monoxide_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'CO', 'id': 2},
                         {'name': 'no', 'longname': 'nitrogen monoxide', 'displayname': 'NO',
                          'cf_standardname': 'mole_fraction_of_nitrogen_monoxide_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'NO', 'id': 3},
                         {'name': 'o3', 'longname': 'ozone', 'displayname': 'Ozone',
                          'cf_standardname': 'mole_fraction_of_ozone_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'O3', 'id': 5},
                         {'name': 'no2', 'longname': 'nitrogen dioxide', 'displayname': 'NO2',
                          'cf_standardname': 'mole_fraction_of_nitrogen_dioxide_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'NO2', 'id': 6},
                         {'name': 'toluene', 'longname': 'toluene', 'displayname': 'Toluene',
                          'cf_standardname': 'mole_fraction_of_toluene_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'C7H8', 'id': 7},
                         {'name': 'so2', 'longname': 'Sulphur dioxide', 'displayname': 'SO2',
                          'cf_standardname': 'mole_fraction_of_sulfur_dioxide_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'SO2', 'id': 8},
                         {'name': 'ethane', 'longname': 'Ethane', 'displayname': 'Ethane',
                          'cf_standardname': 'mole_fraction_of_ethane_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'C2H6', 'id': 9},
                         {'name': 'propane', 'longname': 'Propane', 'displayname': 'Propane',
                          'cf_standardname': 'mole_fraction_of_propane_in_air', 'units': 'nmol mol-1',
                          'chemical_formula': 'C3H8', 'id': 10},
                         {'name': 'ox', 'longname': 'Ox', 'displayname': 'Ox',
                          'cf_standardname': '', 'units': 'nmol mol-1',
                          'chemical_formula': '', 'id': 11}]
        assert response.json() == expected_resp


    def test_get_special_by_name_not_existing(self, client, db):
        response = client.get("/variables/sabinen")
        expected_status_code = 404
        assert response.status_code == expected_status_code
        expected_resp = {"detail":"Variable not found."}
        assert response.json() == expected_resp


#   def test_insert_new_without_credits(self):
#?      response = client.post("/variables/")
#       expected_status_code=401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp



#   def test_insert_new_wrong_credits(self):
#?      response = client.post("/variables/")
#       expected_status_code = 401
#       assert response.status_code == expected_status_code
#?      expected_resp = ...
#   assert response.json() == expected_resp


    def test_insert_new(self, client, db):
        response = client.post("/variables/",
                json= {"name": "sabinene",
                       "longname": "sabinene",
                       "displayname": "Sabinene",
                       "cf_standardname": "mole_fraction_of_sabinene_in_air",
                       "units": "nmol mol-1",
                       "chemical_formula": "C10H16"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp =  {'name': 'sabinene', 'longname': 'sabinene', 'displayname': 'Sabinene',
                          'cf_standardname': 'mole_fraction_of_sabinene_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'C10H16', 'id': 34}
        assert response.json() == expected_resp

                                                    
    def test_insert_new_again(self, client, db):
        response = client.post("/variables/",
                json= {"name": "sabinene",
                       "longname": "sabinene",
                       "displayname": "Sabinene",
                       "cf_standardname": "mole_fraction_of_sabinene_in_air",
                       "units": "nmol mol-1",
                       "chemical_formula": "C10H16"}
                   )
        expected_status_code = 200
        assert response.status_code == expected_status_code
        expected_resp =  {'name': 'sabinene', 'longname': 'sabinene', 'displayname': 'Sabinene',
                          'cf_standardname': 'mole_fraction_of_sabinene_in_air',
                          'units': 'nmol mol-1', 'chemical_formula': 'C10H16', 'id': 34}
        assert response.json() == expected_resp

                                                    
    def test_insert_duplicate(self, client, db):
        response = client.post("/variables/",
                json= {"name": "benzene",
                       "longname": "benzene",
                       "displayname": "Benzene",
                       "cf_standardname": "mole_fraction_of_benzene_in_air",
                       "units": "nmol mol-1",
                       "chemical_formula": "C6H6"}
                   )
        expected_status_code = 400
        assert response.status_code == expected_status_code
        expected_resp = {'detail': 'Variable already registered.'}
        assert response.json() == expected_resp
