from sqlalchemy.orm import Session
from fastapi import Request, Header, HTTPException
from starlette.datastructures import QueryParams

from typing import List
from datetime import datetime
import requests
import json

from . import models
from toardb.utils.settings import userinfo_endpoint, request_limitations, gridded_product_units
from toardb.utils.database import get_db


def get_user_role(userinfo, ltoken: bool = False):
    role = "anonymous"
    # has the user authorized themselves to the AAI?
    if ltoken:
        role = "logged_in"
    # is the user a power TOAR user?
    if "eduperson_entitlement" in userinfo and \
       f"urn:geant:helmholtz.de:res:toar-data:power-toar-user#login.helmholtz.de" \
       in userinfo["eduperson_entitlement"]:
        role = "power-toar-user"
    else:
        if "eduperson_entitlement" in userinfo and \
           f"urn:geant:helmholtz.de:res:toar-data#login.helmholtz.de" \
           in userinfo["eduperson_entitlement"]:
            role = "registered"
    return role


def get_eduperson_and_roles(request: Request, eduperson_unique_id: str = None, DoIncr: List[int]=[0,1]):
    # Do not use underscores; they are not valid in header attributes!
    role = 'unknown'
    status_code = 200
    userinfo = {}
    userinfo['name'] = 'fake name'
    userinfo['email'] = 'fakeEmail@fake.com'
    person_dict = {}
    access_token = request.headers.get('AccessToken')
    ltoken = access_token is not None
    if ltoken:
        us_info = requests.get(userinfo_endpoint, headers={'Authorization': f"Bearer {access_token}"})
        status_code = us_info.status_code
        if status_code != 401:
            userinfo = us_info.json()
            eduperson_unique_id = userinfo['eduperson_unique_id']
        else:
            person_dict['userinfo'] = userinfo
            person_dict['status_code'] = status_code
            return person_dict
    else:
        if eduperson_unique_id is None:
            role = 'anonymous'
            person_dict["num_timeseries"] = 0
            person_dict["num_gridded"] = 0
    if role != "anonymous":
        db = next(get_db())
        db_person = db.query(models.AuthUser).filter(models.AuthUser.eduperson_unique_id == eduperson_unique_id).first()
        if db_person:
            person_dict["num_timeseries"] = db_person.num_timeseries
            person_dict["num_gridded"] = db_person.num_gridded
        else:
            if ltoken:
                person_dict["eduperson_unique_id"] = userinfo['eduperson_unique_id']
                person_dict["email"] = userinfo['email']
                person_dict["username"] = userinfo['name']
                person_dict["num_timeseries"] = 0
                person_dict["num_gridded"] = 0
                db_person = models.AuthUser(**person_dict)
                db.add(db_person)
                db.commit()
                db.refresh(db_person)
            else:
                raise HTTPException(406, f"Person {eduperson_unique_id} not known")
    if eduperson_unique_id is not None:
        ltoken = True
    role = get_user_role(userinfo, ltoken)
    person_dict["role"] = role
    person_dict.update(request_limitations[role])
    match DoIncr[0]:
        case 1:
            person_dict["num_timeseries"] += DoIncr[1]
        case 2:
            person_dict["num_gridded"] += DoIncr[1]
    if role != "anonymous":
        match DoIncr[0]:
            case 1:
                db_person.num_timeseries += DoIncr[1]
            case 2:
                db_person.num_gridded += DoIncr[1]
        db.add(db_person)
        db.commit()
    person_dict['status_code'] = status_code
    person_dict['userinfo'] = userinfo
    return person_dict


def count_year_intervals(daterange):
    interval_length = gridded_product_units["years"]
    start_str, end_str = daterange.split(",")
    start = datetime.fromisoformat(start_str.strip())
    end = datetime.fromisoformat(end_str.strip())

    total_years = (end - start).days / 365.25
    intervals = int(total_years // interval_length)
    if total_years % interval_length > 0:
        intervals += 1
    return intervals



def modify_query_params(query_params, to_remove=["fields", "daterange", "format"], to_add=("limit", "None")):
    items = list(query_params.multi_items())
    filtered_items = [(k, v) for k, v in items if k not in to_remove]
    filtered_items.append(("fields", "id"))
    if not any(k == to_add[0] for k, _ in filtered_items):
        filtered_items.append(to_add)
    return QueryParams(filtered_items)


def pop_user_role(query_params, to_remove=["user_role"]):
    items = list(query_params.multi_items())
    filtered_items = [(k, v) for k, v in items if k not in to_remove]
    return QueryParams(filtered_items)


def count_timeseries_intervals(query_params, db):
    from toardb.timeseries.crud import search_all

    interval_length = gridded_product_units["timeseries"]
    filtered = modify_query_params(query_params)
    ret = search_all(db, path_params='', query_params=filtered)
    lret = [ x['id'] for x in ret ]
    if lret:
        min_id = min(lret)
        max_id = max(lret)
    else:
        min_id = 0
        max_id = 0
    total_timeseries = len(ret)
    intervals = int(total_timeseries// interval_length)
    if total_timeseries % interval_length > 0:
        intervals += 1
    return total_timeseries, intervals, min_id, max_id


def determine_increments(request: Request, db, user_role=None):
    query_params = request.query_params
    qps = pop_user_role(query_params)
    daterange = qps.get("daterange")
    if daterange is None:
        daterange = "1970-01-01,2025-12-31"
    year_faktor = count_year_intervals(daterange)
    num_timeseries, timeseries_faktor, min_id, max_id = count_timeseries_intervals(qps, db)
    in_allowed_tsrange = True
    if min_id < request_limitations["anonymous"]["min_tsid"] or max_id > request_limitations["anonymous"]["max_tsid"]:
        in_allowed_tsrange = False
    access = 200
    message = "Authorized access"
    if user_role:
        if user_role == "anonymous" and not in_allowed_tsrange:
            access = 401
            message = ("Request contains timeseries whose IDs are not in the range of "
                       f"{request_limitations['anonymous']['min_tsid']} to "
                       f"{request_limitations['anonymous']['max_tsid']}.")

    ret_dict = {"num_timeseries": num_timeseries,
                "year_faktor": year_faktor,
                "timeseries_faktor": timeseries_faktor,
                "num_gridded": year_faktor * timeseries_faktor,
                "in_allowed_tsrange": in_allowed_tsrange,
                "access": access,
                "message": message}
    return ret_dict
