# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Simple API for contacts management
"""

from typing import List, Union, Literal
from fastapi import APIRouter, Depends, HTTPException, Body, Request

from sqlalchemy.orm import Session
from . import crud, schemas
from toardb.utils.database import ToarDbSession, get_db

router = APIRouter()

# plain views to post and get contacts

# 1. persons

#get all entries of table persons
@router.get('/contacts/persons/', response_model=List[schemas.Person], response_model_exclude_none=True, response_model_exclude_unset=True)
def get_all_persons(request: Request, db: Session = Depends(get_db)):
    persons = crud.get_all_persons(db, path_params=request.path_params, query_params=request.query_params)
    return persons

#get all metadata of one person
@router.get('/contacts/persons/id/{person_id}', response_model=schemas.Person)
def get_person(person_id: int, db: Session = Depends(get_db)):
    db_person = crud.get_person(db, person_id=person_id)
    if db_person is None:
        raise HTTPException(status_code=404, detail="Person not found.")
    return db_person

@router.get('/contacts/persons/{name}', response_model=schemas.Person)
def get_person(name: str, db: Session = Depends(get_db)):
    db_person = crud.get_person_by_name(db, name=name)
    if db_person is None:
        raise HTTPException(status_code=404, detail="Person not found.")
    return db_person

@router.post('/contacts/persons/', response_model=schemas.Person)
def create_person(person: schemas.PersonCreate = Body(..., embed = True), db: Session = Depends(get_db)):
    db_person = crud.get_person_by_name(db, name=person.name)
    if db_person:
        raise HTTPException(status_code=400, detail="Person already registered.")
    return crud.create_person(db=db, person=person)


# 2. organisations

#get all entries of table organisations
@router.get('/contacts/organisations/', response_model=List[schemas.Organisation])
def get_all_organisations(offset: int = 0, limit: int | Literal["None"] = 10, db: Session = Depends(get_db)):
    if limit == "None":
        limit = None
    organisations = crud.get_all_organisations(db, offset=offset, limit=limit)
    return organisations

#get all metadata of one organisation
@router.get('/contacts/organisations/id/{organisation_id}', response_model=schemas.Organisation)
def get_organisation(organisation_id: int, db: Session = Depends(get_db)):
    db_organisation = crud.get_organisation (db, organisation_id=organisation_id)
    if db_organisation is None:
        raise HTTPException(status_code=404, detail="Organisation not found.")
    return db_organisation

@router.get('/contacts/organisations/{name}', response_model=schemas.Organisation)
def get_organisation(name: str, db: Session = Depends(get_db)):
    db_organisation = crud.get_organisation_by_name(db, name=name)
    if db_organisation is None:
        raise HTTPException(status_code=404, detail="Organisation not found.")
    return db_organisation

@router.post('/contacts/organisations/', response_model=schemas.Organisation)
def create_organisation(organisation: schemas.OrganisationCreate = Body(..., embed = True), db: Session = Depends(get_db)):
    db_organisation = crud.get_organisation_by_longname(db, longname=organisation.longname)
    if db_organisation:
        raise HTTPException(status_code=400, detail="Organisation already registered.")
    return crud.create_organisation(db=db, organisation=organisation)

# 3. combined (person + organisation)
#get all entries of table contacts
@router.get('/contacts/', response_model=List[schemas.Contact])
def get_all_contacts(offset: int = 0, limit: int | Literal["None"] = 10, db: Session = Depends(get_db)):
    if limit == "None":
        limit = None
    contacts = crud.get_all_contacts(db, offset=offset, limit=limit)
    return contacts

#get a special entry of table contacts (given by ID)
@router.get('/contacts/id/{contact_id}', response_model=Union[schemas.Organisation,schemas.Person])
def get_all_contacts(contact_id: int, db: Session = Depends(get_db)):
    contact = crud.get_contact(db, contact_id=contact_id)
    return contact

#get a special entry of table contacts (given by its organisation (long) name)
@router.get('/contacts/orga_name/{name}', response_model=schemas.Contact)
def get_all_contacts(name: str, db: Session = Depends(get_db)):
    contact = crud.get_contact_by_orga_name(db, name=name)
    if contact is None:
        raise HTTPException(status_code=404, detail="Contact not found.")
    return contact

#get a special entry of table contacts (given by its person name)
@router.get('/contacts/person_name/{name}', response_model=schemas.Contact)
def get_all_contacts(name: str, db: Session = Depends(get_db)):
    contact = crud.get_contact_by_person_name(db, name=name)
    return contact


#get a special entry of table contacts (given by its person email)
@router.get('/contacts/person_email/{email}', response_model=schemas.Contact)
def get_all_contacts(email: str, db: Session=Depends(get_db)):
    contact = crud.get_contact_by_person_email(db, email=email)
    return contact


#create new contact (either only person/organisation or both)
@router.post('/contacts/')
def create_contact(person_name: str = None, organisation_longname: str = None, db: Session = Depends(get_db)):
    if not person_name and not organisation_longname:
        raise HTTPException(status_code=400, detail="No contact data given.")
    contact = crud.create_contact(db,person_name=person_name,organisation_longname=organisation_longname)
    return contact
