# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Create, Read, Update, Delete functionality

"""

from sqlalchemy import text
from sqlalchemy.orm import Session
from fastapi.responses import JSONResponse
from . import models
from .schemas import PersonCreate, OrganisationCreate
from toardb.utils.utils import get_value_from_str, create_filter
import toardb


def get_organisation(db: Session, organisation_id: int):
    return db.query(models.Organisation).filter(models.Organisation.id == organisation_id).first()


def get_all_organisations(db: Session, limit: int, offset: int = 0):
    return db.query(models.Organisation).order_by(models.Organisation.id).offset(offset).limit(limit).all()


def get_organisation_by_name(db: Session, name: str):
    return db.query(models.Organisation).filter(models.Organisation.name == name.upper()).first()


def get_organisation_by_longname(db: Session, longname: str):
    return db.query(models.Organisation).filter(models.Organisation.longname == longname).first()
 

def create_organisation(db: Session, organisation: OrganisationCreate):
    db_organisation = models.Organisation(**organisation.dict())
    db_organisation.kind = get_value_from_str(toardb.toardb.OK_vocabulary,db_organisation.kind)
    db_organisation.country = get_value_from_str(toardb.toardb.CN_vocabulary,db_organisation.country)
    db.add(db_organisation)
    result = db.commit()
    db.refresh(db_organisation)
    db_contact = models.Contact(person_id=0,organisation_id=db_organisation.id)
    db.add(db_contact)
    result = db.commit()
    db.refresh(db_contact)
    return db_organisation


def get_person(db: Session, person_id: int):
    return db.query(models.Person).filter(models.Person.id == person_id).filter(models.Person.isprivate == False).first()


def get_all_persons(db: Session, path_params, query_params):
    try:
        limit, offset, fields, format, filters = create_filter(query_params, "persons")
        p_filter = filters["p_filter"]
    except (KeyError, ValueError) as e:
        status_code=400
        return JSONResponse(status_code=status_code, content=str(e))

    if fields:
        fields2 = fields.split(',')
        db_objects_l = db.query(*map(text,fields2)).select_from(models.Person). \
                           filter(models.Person.id != 0).filter(models.Person.isprivate == False). \
                           filter(text(p_filter)). \
                           order_by(models.Person.id). \
                           limit(limit).offset(offset)
        db_objects = []
        for db_object_immut in db_objects_l:
            db_object = dict(zip(fields.split(','),db_object_immut))
            db_objects.append(db_object)
    else:
        db_objects = db.query(models.Person).filter(models.Person.id != 0).filter(models.Person.isprivate == False). \
              filter(text(p_filter)). \
              order_by(models.Person.id).offset(offset).limit(limit).all()
    return db_objects


def get_person_by_name(db: Session, name: str):
    return db.query(models.Person).filter(models.Person.name == name).filter(models.Person.isprivate == False).first()


def get_person_by_email(db: Session, email: str):
    return db.query(models.Person).filter(models.Person.email == email).filter(models.Person.isprivate == False).first()


def _get_private_person_by_name(db: Session, name: str):
    return db.query(models.Person).filter(models.Person.name == name).first()


def create_person(db: Session, person: PersonCreate):
    db_person = models.Person(**person.dict())
    db.add(db_person)
    result = db.commit()
    db.refresh(db_person)
    db_contact = models.Contact(person_id=db_person.id,organisation_id=0)
    db.add(db_contact)
    result = db.commit()
    db.refresh(db_contact)
    return db_person


def get_all_contacts(db: Session, limit: int, offset: int = 0):
    return db.query(models.Contact).join(models.Person).filter(models.Contact.person_id == models.Person.id).filter(models.Person.isprivate == False).order_by(models.Contact.id).offset(offset).limit(limit).all()

def get_contact(db: Session, contact_id: int):
    db_object = db.query(models.Contact).join(models.Person).filter(models.Contact.person_id == models.Person.id).filter(models.Contact.id == contact_id).filter(models.Person.isprivate == False).first()
    if db_object:
        if db_object.person_id == 0:
            return db.query(models.Organisation).filter(models.Organisation.id == db_object.organisation_id).first()
        else:
            return db.query(models.Person).filter(models.Person.id == db_object.person_id).filter(models.Person.isprivate == False).first()
    else:
        return db_object


def get_contact_by_orga_name(db: Session, name: str):
    # issue with '/' in organisation longname ==> only possible with double encoding!
    name = name.replace('%2F', '/')
    db_object = db.query(models.Organisation).filter(models.Organisation.longname == name).first()
    if db_object:
        return db.query(models.Contact).filter(models.Contact.organisation_id == db_object.id).filter(models.Contact.person_id == 0).first()
    else:
        return None


def get_contact_by_person_name(db: Session, name: str):
    db_object = db.query(models.Person).filter(models.Person.name == name).first()
    return db.query(models.Contact).filter(models.Contact.person_id == db_object.id).first()


def get_contact_by_person_email(db: Session, email: str):
    db_object = db.query(models.Person).filter(models.Person.email == email).first()
    return db.query(models.Contact).filter(models.Contact.person_id == db_object.id).first()


def get_contact_by_unique_constraints(db: Session, person_id, organisation_id: int):
    return db.query(models.Contact).filter(models.Contact.organisation_id == organisation_id). \
                                    filter(models.Contact.person_id == person_id).first()


def create_contact(db: Session, person_name, organisation_longname: str):
    if person_name:
        db_person = _get_private_person_by_name(db,person_name)
    else:
        db_person = get_person(db,0)
    if organisation_longname:
        db_organisation = get_organisation_by_longname(db,organisation_longname)
    else:
        db_organisation = get_organisation(db,0)
    db_contact = get_contact_by_unique_constraints(db,db_person.id,db_organisation.id)
    if not db_contact:
        db_contact = models.Contact(person_id=db_person.id,organisation_id=db_organisation.id)
        db.add(db_contact)
        result = db.commit()
    db.refresh(db_contact)
    return db_contact
