# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Pydantic schemas for TOAR database

"""

from typing import List, Dict, Any
from pydantic import BaseModel, validator, Field
#from toardb.toardb import OK_vocabulary, CN_vocabulary
import toardb

class OrganisationBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    name: str = Field(..., description="Short name (abbreviation) of organisation")
    longname: str = Field(..., description="Long name of organisation")
    kind: str = Field(..., description="Kind of organisation (see controlled vocabulary: Kind Of Organization)")
    city: str = Field(..., description="City where organisation resides")
    postcode: str = Field(..., description="Postcode of organisation city")
    street_address: str = Field(..., description="Street address of organisation city")
    country: str = Field(..., description="Country to which organisation belongs (see controlled vocabulary: Country Code)")
    homepage: str = Field(..., description="Homepage of organisation")
    contact_url: str = Field(..., description="web form or general e-mail contact of organisation")

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.OK_vocabulary))[0].display_str

    @validator('country')
    def check_country(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CN_vocabulary))[0].display_str


class OrganisationCreate(OrganisationBase):
    kind: str
    country: str
    attribution: str = Field("", description="attribution when using data from this organisation")
    pass

    @validator('kind') 
    def check_kind(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.OK_vocabulary)):
            return v
        else:
            raise ValueError(f"kind of organisation not known: {v}")

    @validator('country')
    def check_country(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.CN_vocabulary)):
            return v
        else:
            raise ValueError(f"country not known: {v}")


class Organisation(OrganisationBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True


class PersonBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    name: str = Field(None, description="Name of person")
    email: str = Field(None, description="Email address of person")
    phone: str = Field(None, description="Phone number of person")
    orcid: str = Field(None, description="ORCID-iD of person")
    isprivate: bool = Field(None, description="Set this flag to true if the contact details shall not be exposed publicly")

    def __str__(self):
        return f"{self.name} <{self.email}>"

    class Meta:
        db_table = 'persons'
#       app_label = 'Person'

class PersonCreate(PersonBase):
    pass


class Person(PersonBase):
    id: int = Field(None, description="for internal use only")

    class Config:
        orm_mode = True

# ======== for nested view =========

class ContactBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    person: Person = Field(None, description="data for person of the contact information")
    organisation: Organisation = Field(None, description="data for organisation of the contact information")

    class Config:
        orm_mode = True

class Contact(ContactBase):
    id: int = Field(..., description="for internal use only")

    @validator('person')
    def check_contact_is_private(cls, v):
        if (v.id != 0):
            if (v.isprivate):
                return None
            else:
                return v
        else:
            return None

    @validator('organisation')
    def check_contact_is_organisation(cls, v):
        if (v.id != 0):
           return v

    class Config:
        orm_mode = True

