# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Create, Read, Update, Delete functionality

"""

import sys
from io import StringIO
from contextlib import closing
import requests
import json

from sqlalchemy import insert, delete, select, and_, text, func
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from geoalchemy2.elements import WKBElement, WKTElement
from fastapi import File, UploadFile
from typing import List
from fastapi.responses import JSONResponse, Response
import datetime as dt
import pandas as pd
import csv

from . import models, schemas
from toardb.variables import models as variables_models
from toardb.variables.crud import get_variable
from toardb.stationmeta import models as stationmeta_models
from toardb.stationmeta.crud import get_stationmeta_core
from toardb.stationmeta.schemas import get_coordinates_from_geom, get_geom_from_coordinates
from toardb.timeseries.models import TimeseriesChangelog
from toardb.timeseries.schemas import TimeseriesWithCitation
from toardb.timeseries.crud import get_timeseries_by_unique_constraints, get_timeseries, get_citation
from toardb.utils.utils import get_value_from_str, get_str_from_value, create_filter
import toardb
from toarqc import get_toarqc_config, run_toarqc
from toarqc.tests import RangeTest


def create_filter_from_aggreated_flags(i: int, filter_string: str, agg_flag_num: int):
    # definition of aggregated_flags:
    # agg. flag | composition of unique flags
    # ----------|----------------------------
    # 100       | 0-6
    # 101       | 0-2
    # 102       | 3-5
    # 103       | 0, 1, 3, 4
    # 104       | 2, 5, 6
    # 110       | 10-16
    # 111       | 10-12
    # 112       | 13-16
    # 120       | 20-28
    # 121       | 20-23
    # 122       | 24-28
    # 130       | 10-28
    # 131       | 10-12, 20-23
    # 132       | 13-16, 24-28
    # 140       | 7, 16, 28
    filter_dict = { 100: [ 0,  1,  2,  3,  4,  5,  6],
                    101: [ 0,  1,  2],
                    102: [ 3,  4,  5],
                    103: [ 0,  1,  3,  4],
                    104: [ 2,  5,  6],
                    110: [10, 11, 12, 13, 14, 15, 16],
                    111: [10, 11, 12],
                    112: [13, 14, 15, 16],
                    120: [20, 21, 22, 23, 24, 25, 26, 27, 28],
                    121: [20, 21, 22, 23],
                    122: [24, 25, 26, 27, 28],
                    130: [10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                          20, 21, 22, 23, 24, 25, 26, 27, 28],
                    131: [10, 11, 12,
                          20, 21, 22, 23],
                    132: [13, 16, 24, 25, 26, 27, 28],
                    140: [ 7, 16, 28] }
    for flag_num in filter_dict[agg_flag_num]:
        if i == 0:
            filter_string = f"(data.flags = {flag_num})"
        else:
            filter_string = filter_string + f" OR (data.flags = {flag_num})"
        i += 1
    return i, filter_string


def create_filter_from_flags(flags: str):
    i = 0
    filter_string = ''
    for flag in flags.split(','):
        flag_num = get_value_from_str(toardb.toardb.DF_vocabulary,flag)
        # check, whether an aggregated flag was given
        if (flag_num >= 100):
            i, filter_string = create_filter_from_aggreated_flags(i,filter_string,flag_num)
        else:
            if i == 0:
                filter_string = f"(data.flags = {flag_num})"
            else:
                filter_string = filter_string + f" OR (data.flags = {flag_num})"
            i += 1
    return f"({filter_string})"


def get_data(db: Session, timeseries_id: int, path_params, query_params):

    # BUT for some unknown reasons, get_timeseries does not return variable and programme!
    def get_timeseries_meta(timeseries_id):
        record = get_timeseries(db, timeseries_id=timeseries_id)
        # for some unknown reasons, variable, changelog, and programme have to be 'loaded' into the structure?!
        if record:
            dummy = record.variable
            dummy = record.programme
            dummy = record.changelog
            attribution, citation, license_txt = get_citation(db, timeseries_id=timeseries_id).values()
            record_dict = record.__dict__
            record_dict['license'] = license_txt
            record_dict['citation'] = citation
            if attribution != None:
                record_dict['attribution'] = attribution
            return TimeseriesWithCitation(**record_dict)
        else:
            return None

    try:
        # translation of flags should be done here!
        flags = ",".join([item.strip() for v in query_params.getlist("flags") for item in v.split(',')])
        if not flags:
            flags = None
        limit, offset, fields, format, filters = create_filter(query_params, "data")
        d_filter = filters["d_filter"]
        fields_list = []
        if fields:
            fields_list = fields.split(',')
        columns = ( [getattr(models.Data, field) for field in fields_list]
                    if fields_list
                    else list(models.Data.__table__.columns) )
    except KeyError as e:
        status_code=400
        return JSONResponse(status_code=status_code, content=str(e))
    if flags:
        filter_string = create_filter_from_flags(flags)
        data = db.query(*columns).filter(models.Data.timeseries_id == timeseries_id). \
                                     filter(text(filter_string)). \
                                     filter(text(d_filter)). \
                                     order_by(models.Data.datetime).all()
    else:
        data = db.query(*columns).filter(models.Data.timeseries_id == timeseries_id). \
                                     filter(text(d_filter)). \
                                     order_by(models.Data.datetime).all()
    # get advantages from pydantic, but without having another call of the REST API
    # (especially needed for testing with pytest!)
    metadata = get_timeseries_meta(timeseries_id)
# TOAR day 2022-08-09: We want the timezone information for the requested data
#   metadata['timezone'] = ...
    if not metadata:
        return None
    if format == 'json':
        composite = schemas.Composite(metadata=metadata, data=data)
        return composite
    elif format == 'csv':
        # start with metadata
        content = '#' +  metadata.json(indent=4, ensure_ascii=False).replace('\n', '\n#') + '\n'
        # add header
        content += ','.join(column.name for column in columns) + '\n'
        # now the data
        content += '\n'.join(','.join(f"{getattr(curr, column.name)}" for column in columns) for curr in data)
        return Response(content=content, media_type="text/csv")
    else:
        status_code=400
        message='Invalid format!'
        return JSONResponse(status_code=status_code, content=message)


def get_next_version(db: Session, timeseries_id: int, path_params, query_params):
    data_version = db.query(func.max(models.Data.version)). \
                      filter(models.Data.timeseries_id == timeseries_id). \
                      distinct().first()
    if is_preliminary(data_version[0]):
        status_code = 407
        message = 'Preliminary data has no next version!'
        return JSONResponse(status_code=status_code, content=message)
    else:
        splitted = data_version[0].split('.')
        lmajor = query_params.get("major", "False").lower() == "true"
        major = int(splitted[0])
        if lmajor:
            major += 1
            minor = 0
        else:
            minor = int(splitted[1]) + 1
        return f"{major:06d}.{minor:06d}.00000000000000"


def get_map_data(db:Session, variable_id, daterange):
    daterange = daterange.split(',')
    query_string = f"timeseries_id, value FROM timeseries t, data d WHERE d.timeseries_id=t.id AND t.variable_id={variable_id} AND d.datetime BETWEEN '{daterange[0]}' AND '{daterange[1]}'"
    data = db.execute(select(text(query_string))).all()
    return data


def get_data_with_staging(db: Session, timeseries_id: int, flags: str, format: str):

    # BUT for some unknown reasons, get_timeseries does not return variable and programme!
    def get_timeseries_meta(timeseries_id):
        record = get_timeseries(db, timeseries_id=timeseries_id)
        # for some unknown reasons, variable, changelog, and programme have to be 'loaded' into the structure?!
        if record:
            dummy = record.variable
            dummy = record.programme
            dummy = record.changelog
            attribution, citation, license_txt = get_citation(db, timeseries_id=timeseries_id).values()
            record_dict = record.__dict__
            record_dict['citation'] = citation
            record_dict['license'] =license_txt
            if attribution != None:
                record_dict['attribution'] = attribution
            return TimeseriesWithCitation(**record_dict)
        else:
            return None
    if flags:
        filter_string = create_filter_from_flags(flags)
        data = db.query(models.Data).filter(models.Data.timeseries_id == timeseries_id).filter(text(filter_string)).order_by(models.Data.datetime).all()
    else:
        query_txt = f"SELECT distinct ON (datetime) * \
                      FROM ( SELECT *,1 AS db FROM staging.data WHERE timeseries_id={timeseries_id} \
                             UNION \
                                 SELECT *,2 AS db FROM public.data WHERE timeseries_id={timeseries_id}) AS mix \
                      ORDER BY datetime, db"
        data = db.query(models.Data).from_statement(text(query_txt)).all()

    # get advantages from pydantic, but without having another call of the REST API
    # (especially needed for testing with pytest!)
    metadata = get_timeseries_meta(timeseries_id)
# TOAR day 2022-08-09: We want the timezone information for the requested data
#   metadata['timezone'] = ...
    if not metadata:
        return None
    if format == 'json':
        composite = schemas.Composite(metadata=metadata, data=data)
        return composite
    elif format == 'csv':
        # start with metadata
        content = '#' +  metadata.json(indent=4, ensure_ascii=False).replace('\n', '\n#') + '\n'
        # add header
        content += ','.join(column.name for column in models.Data.__mapper__.columns) + '\n'
        # now the data
        content += '\n'.join(','.join(f"{getattr(curr, column.name)}" for column in models.Data.__mapper__.columns) for curr in data)
        return Response(content=content, media_type="text/csv")
    else:
        status_code=400
        message='Invalid format!'
        return JSONResponse(status_code=status_code, content=message)


def get_data_by_datetime_and_timeseriesid(db: Session, datetime: dt.datetime, timeseries_id: int):
    return db.query(models.Data).filter([models.Data.datetime== datetime, models.Data.timeseries_id == timeseries_id]).first()


def get_all_data(db: Session, limit: int, offset: int = 0):
    return db.query(models.Data).order_by(models.Data.datetime).limit(limit).all()

def is_preliminary(version: str):
    return (version.split('.')[0] == '000000')

def create_data_record(db: Session, engine: Engine,
        series_id: int, datetime: dt.datetime,
        value: float, flag: str, version: str,
        author_id: int):
    toarqc_config_type: str = 'standard'
    timeseries = get_timeseries(db=db,timeseries_id=series_id)
    variable = get_variable(db=db, variable_id=timeseries.variable_id)
    parameter = variable.name
    data_dict = {"datetime": datetime,
                 "value": value,
                 "flags": flag,
                 "version": version,
                 "timeseries_id": series_id}
    df = pd.DataFrame([data_dict]).set_index("datetime")
    try:
        test_config = get_toarqc_config('static/../toardb/data/toarqc_config',parameter, toarqc_config_type)
    except FileNotFoundError:
        message = f'no toarqc configuration found for {parameter}, {toarqc_config_type}'
        status_code = 400
        return JSONResponse(status_code=status_code, content=message)
    result = run_toarqc(test_config, df, ok_limit=0.85, questionable_limit=0.6)
    combined_flag_matrix = {( 'OK',           'OK')        : 'OKPreliminaryNotChecked',
                            ( 'Questionable', 'OK')        : 'QuestionablePreliminaryNotChecked',
                            ( 'Erroneous',    'OK')        : 'ErroneousPreliminaryNotChecked',
                            ( 'OK',           'Erroneous') : 'ErroneousPreliminaryFlagged1',
                            ( 'Questionable', 'Erroneous') : 'ErroneousPreliminaryFlagged2',
                            ( 'Erroneous',    'Erroneous') : 'Erroneous_Preliminary_Confirmed'}
    combined_flag = combined_flag_matrix[(flag,result['flags'].iloc[0])]
    data_dict["flags"] = get_value_from_str(toardb.toardb.DF_vocabulary,combined_flag.strip())
    data = models.Data(**data_dict)
    db.rollback()
    db.add(data)
    # adjust data_start_date, data_end_date
    datetime = datetime.replace(tzinfo=timeseries.data_end_date.tzinfo)
    if datetime < timeseries.data_start_date:
        timeseries.data_start_date = datetime
    if datetime > timeseries.data_end_date:
        timeseries.data_end_date = datetime
    db.add(timeseries)
    result = db.commit()
    db.refresh(data)
    # if not preliminary data: create changelog entry
    if not is_preliminary(version):
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Created")
        description="data record created"
        db_changelog = TimeseriesChangelog(description=description, timeseries_id=series_id, author_id=author_id, type_of_change=type_of_change,
                                           old_value='', new_value='', period_start=datetime, period_end=datetime, version=version)
        db.add(db_changelog)
        db.commit()
    status_code=200
    message='Data successfully inserted.'
    return JSONResponse(status_code=status_code, content=message)


def insert_dataframe (db: Session, engine: Engine, df: pd.DataFrame, toarqc_config_type: str = 'standard', dry_run: bool = False, parameter: str = 'o3', preliminary: bool = False, force: bool = False):
    # df: pandas.DataFrame
    #     index: datetime
    #     1st column: value
    #     2nd column: flags (0: 'OK', 1: 'Questionable', 2: 'Erroneous')
    #     3rd column: timeseries_id
    #     4th column: version
    df['flags'] = df['flags'].replace([0],'OK')
    df['flags'] = df['flags'].replace([1],'Questionable')
    df['flags'] = df['flags'].replace([2],'Erroneous')
    try:
        test_config = get_toarqc_config('static/../toardb/data/toarqc_config',parameter, toarqc_config_type)
    except FileNotFoundError:
        message = f'no toarqc configuration found for {parameter}, {toarqc_config_type}'
        status_code = 400
        return JSONResponse(status_code=status_code, content=message)
    # get flags and qc-statistics from toarqc 
    # available keys in result (at the moment):
    # 'time_series'
    # 'test_config'
    # 'ok_limit'
    # 'questionable_limit'
    # 'metadata'
    # 'plot'
    # 'outfile'
    # 'plot_type'
    # 'results'
    # 'qc-score'
    # 'flags'
    # 'number_OK'
    # 'number_Questionable'
    # 'number_Erroneous'
    # 'percentage_OK'
    # 'percentage_Questionable'
    # 'percentage_Erroneous'
    # 'qc-score_g0'
    # 'flags_g0'
    # 'qc-score_g1'
    # 'flags_g1'
    # 'qc-score_g2'
    # 'flags_g2'
    # 'qc-score_g3'
    # 'flags_g3'
    # 'qc-score_g0_range_test'
    # 'flags_g0_range_test'
    # 'qc-score_g1_sigma_test'
    # 'flags_g1_sigma_test'
    # 'qc-score_g2_constant_value_test'
    # 'flags_g2_constant_value_test'
    # 'qc-score_g2_positive_spike_test'
    # 'flags_g2_positive_spike_test'
    # 'qc-score_g2_negative_spike_test'
    # 'flags_g2_negative_spike_test'
    # 'qc-score_g3_before_nan_test'
    # 'flags_g3_before_nan_test'
    # 'qc-score_g3_after_nan_test'
    # 'flags_g3_after_nan_test'
    # toarqc expects pandas.Series
    # --> if getting pandas.DataFrame, it will take index and first column
    # (other columns will be ignored)
    if toarqc_config_type != 'realtime':
        try:
            result = run_toarqc(test_config, df, ok_limit=0.85, questionable_limit=0.6)
        except ValueError as ve:
            message = {"detail":{"message":f'Aborted by automatic quality control: {ve}'}}
            status_code = 445
            return JSONResponse(status_code=status_code, content=message)
        percentage_OK = result['percentage_OK']
        if percentage_OK >= 0.9 or force:
            target_schema='public'
            message = {"detail":{"message":"Data successfully inserted."}}
            status_code = 200
        else:
            target_schema='staging'
            message = f'Aborted by automatic quality control: percentage_OK = {percentage_OK}'
            status_code = 446
        toarqc_flags = result['flags']
        # combine toarqc flags with flags given by provider (see matrix: TOAR_UG_Vol03_Database_2021-05.docx, chapter 5.2)
        #                      preliminary  provider           toarqc            combined
        combined_flag_matrix = {( False,       'OK',           'OK')           : 'OKValidatedQCPassed',
                                ( False,       'OK',           'Questionable') : 'QuestionableValidatedFlagged',
                                ( False,       'OK',           'Erroneous')    : 'ErroneousValidatedFlagged1',
                                ( False,       'Questionable', 'OK')           : 'QuestionableValidatedUnconfirmed',
                                ( False,       'Questionable', 'Questionable') : 'QuestionableValidatedConfirmed',
                                ( False,       'Questionable', 'Erroneous')    : 'ErroneousValidatedFlagged2',
                                ( False,       'Erroneous',    'OK')           : 'ErroneousValidatedUnconfirmed',
                                ( False,       'Erroneous',    'Questionable') : 'ErroneousValidatedConfirmed',
                                ( False,       'Erroneous',    'Erroneous')    : 'ErroneousValidatedConfirmed',
                                ( True,        'OK',           'OK')           : 'OKPreliminaryQCPassed',
                                ( True,        'OK',           'Questionable') : 'QuestionablePreliminaryFlagged',
                                ( True,        'OK',           'Erroneous')    : 'ErroneousPreliminaryFlagged1',
                                ( True,        'Questionable', 'OK')           : 'QuestionablePreliminaryUnconfirmed',
                                ( True,        'Questionable', 'Questionable') : 'QuestionablePreliminaryConfirmed',
                                ( True,        'Questionable', 'Erroneous')    : 'ErroneousPreliminaryFlagged2',
                                ( True,        'Erroneous',    'OK')           : 'ErroneousPreliminaryUnconfirmed',
                                ( True,        'Erroneous',    'Questionable') : 'ErroneousPreliminaryConfirmed',
                                ( True,        'Erroneous',    'Erroneous')    : 'ErroneousPreliminaryConfirmed'}
        combined_flags = [ combined_flag_matrix[(preliminary,provider_flag,toarqc_flag)] for provider_flag, toarqc_flag in zip(df['flags'],toarqc_flags) ]
    else:
        range_test = RangeTest(**test_config[0]["range_test"])
        toarqc_flags = range_test.run(df['value'])
        toarqc_flags = toarqc_flags.replace([1],'OK')
        toarqc_flags = toarqc_flags.replace([0],'Erroneous')
        target_schema='public'
        status_code = 200
        message = {"detail":{"message":"Data successfully inserted."}}

        # combine toarqc flags with flags given by provider (see matrix: TOAR_UG_Vol03_Database_2021-05.docx, chapter 5.2)
        # the range test is not a full QC test ==> the range test only returns 'OK' and 'Erroneous'
        #                         provider        toarqc         combined
        combined_flag_matrix = {( 'OK',           'OK')        : 'OKPreliminaryNotChecked',
                                ( 'Questionable', 'OK')        : 'QuestionablePreliminaryNotChecked',
                                ( 'Erroneous',    'OK')        : 'ErroneousPreliminaryNotChecked',
                                ( 'OK',           'Erroneous') : 'ErroneousPreliminaryFlagged1',
                                ( 'Questionable', 'Erroneous') : 'ErroneousPreliminaryFlagged2',
                                ( 'Erroneous',    'Erroneous') : 'Erroneous_Preliminary_Confirmed'}
        combined_flags = [ combined_flag_matrix[(provider_flag, toarqc_flag)] for provider_flag, toarqc_flag in zip(df['flags'],toarqc_flags) ]
    # exchange combined flags in dataframe
    flag_num = [get_value_from_str(toardb.toardb.DF_vocabulary,flag.strip()) for flag in combined_flags]
    del df['flags']
    df.insert(1, 'flags', flag_num)
    if dry_run:
        return JSONResponse(status_code=status_code, content=df.to_json(orient='table', date_format='iso'))
    else:
        buf = StringIO()
        df.to_csv(buf, header=False)
        buf.pos = 0
        buf.seek(0)
        with closing(engine.raw_connection()) as fake_conn:
            fake_cur = fake_conn.cursor()
            try:
                fake_cur.copy_expert(f"COPY  {target_schema}.data (datetime, value, flags, timeseries_id, version) FROM STDIN WITH CSV DELIMITER ',';", buf)
                fake_conn.commit()
            except:
                e = sys.exc_info()[0]
                message = f"An error occurred in {target_schema}.data insertion: %s" % (e,)
                status_code = 400
            fake_cur.close()
        return JSONResponse(status_code=status_code, content=message)


def create_data(db: Session, engine: Engine, author_id: int, input_handle: UploadFile = File(...), toarqc_config_type: str = 'standard', dry_run: bool = False, 
        force: bool = False):
    # a timeseries is defined by the unique_constraint of (station_id, variable_id, ...)
    # station_id: from header
    # variable_id: from database (with variable_name -- from filename)
    # get variable_name from filename
    variable_name = input_handle.filename.split('_')[0]
    variable = db.query(variables_models.Variable).filter(variables_models.Variable.name == variable_name).first()
    variable_id = variable.id
    # get header information (station_id, contributor_shortname, timeshift_from_utc)
    line = '#bla'
    f = input_handle.file
    prev = pos = 0
    while line[0] == '#':
        line = f.readline().decode('utf-8')
        key = line.split(':')[0].lower().strip()
        if key == "#station_id":
            station_id = line.split(':')[1]
        if key == "#timeshift_from_utc":
            timeoffset = dt.timedelta(hours=float(line.split(':')[1]))
        if key == "#dataset_contributor_organisation_longname":
            contributor = line.split(':')[1].strip()
        if key == "#dataset_pi_organisation_longname":
            contributor = line.split(':')[1].strip()
        if key == "#dataset_resourceprovider_organisation_longname":
            contributor = line.split(':')[1].strip()
        prev, pos = pos, f.tell()
    f.seek(prev)
    station_code = station_id.strip()
    stationmeta_core = get_stationmeta_core(db=db,station_code=station_code)
    station_id = stationmeta_core.id
    timeseries = get_timeseries_by_unique_constraints(db=db,station_id=station_id,variable_id=variable_id,resource_provider=contributor)
    # again problems with converted coordinates!
    db.rollback()
    version = '000001.000000.00000000000000'
    if timeseries:
        timeseries_id = timeseries.id
        # open SpooledTemporaryFile, skip header (and also try to insert timeseries_id!)
        # python3.8: bug in SpooledTemporaryFile --> https://bugs.python.org/issue26175
        #                                        --> https://github.com/fedspendingtransparency/usaspending-api/pull/2963
        # ==> I have to check for a workaround!
        df = pd.read_csv(input_handle.file, comment='#', header=None, sep=';',names=["time","value","flags"],parse_dates=["time"],index_col="time")
        # substract timeshift to convert data to UTC
        df.index = df.index - timeoffset
        # now insert the timeseries_id to the end of the data frame
        df.insert(2, 'timeseries_id', timeseries_id)
        # also insert version
        df.insert(3, 'version', version)
        # datetime needs timezone information
        df = df.tz_localize('UTC')
        result = insert_dataframe (db, engine, df, toarqc_config_type=toarqc_config_type, parameter=variable_name, dry_run=dry_run, force=force)
        # adjust data_start_date, data_end_date
        if result.status_code == 200:
            data_start_date = min(df.index)
            data_end_date = max(df.index)
            if data_start_date < timeseries.data_start_date:
                timeseries.data_start_date = data_start_date
            if data_end_date > timeseries.data_end_date:
                timeseries.data_end_date = data_end_date
            db.add(timeseries)
            db.commit()
        return result
    else:
        message = f'Timeseries not found for station {station_code.strip()}, variable {variable_name}'
        status_code = 400
    return JSONResponse(status_code=status_code, content=message)


def create_bulk_data(db: Session, engine: Engine, bulk: List[schemas.DataCreate], author_id: int,
        toarqc_config_type: str = 'standard', dry_run: bool = False, force: bool = False):
    df = pd.DataFrame([x.dict() for x in bulk]).set_index("datetime")
    # bulk data: to be able to do at least a range test, all data needs to be from the same parameter
    #            variable_name is therefore determined for the first entry of the dataframe
    timeseries = get_timeseries(db=db,timeseries_id=int(df['timeseries_id'].iloc[0]))
    # again problems with converted coordinates!
    db.rollback()
    variable = get_variable(db=db, variable_id=timeseries.variable_id)
    variable_name = variable.name
    insert_result = insert_dataframe (db, engine, df, toarqc_config_type=toarqc_config_type, parameter=variable_name, dry_run=dry_run, force=force)
    # adjust data_start_date, data_end_date
    if insert_result.status_code == 200:
        data_start_date = min(df.index)
        data_end_date = max(df.index)
        if data_start_date < timeseries.data_start_date:
            timeseries.data_start_date = data_start_date
        if data_end_date > timeseries.data_end_date:
            timeseries.data_end_date = data_end_date
        db.add(timeseries)
        db.commit()
    return insert_result


def patch_data(db: Session, engine: Engine, description: str, version: str, toarqc_config_type: str,
        force: bool, author_id: int, input_handle: UploadFile = File(...)):
    # a timeseries is defined by the unique_constraint of (station_id, variable_id, ...)
    # station_id: from header
    # variable_id: from database (with variable_name -- from filename)
    # get variable_name from filename

    # versionlabel has to be unique for this timeseries ==> to be checked!

    variable_name = input_handle.filename.split('_')[0]
    variable = db.query(variables_models.Variable).filter(variables_models.Variable.name == variable_name).first()
    variable_id = variable.id
    # get header information (station_id, contributor_shortname, timeshift_from_utc)
    line = '#bla'
    f = input_handle.file
    prev = pos = 0
    while line[0] == '#':
        line = f.readline().decode('utf-8')
        key = line.split(':')[0].lower().strip()
        if key == "#station_id":
            station_id = line.split(':')[1]
        if key == "#timeshift_from_utc":
            timeoffset = dt.timedelta(hours=float(line.split(':')[1]))
        if key == "#dataset_contributor_organisation_longname":
            contributor = line.split(':')[1].strip()
        if key == "#dataset_pi_organisation_longname":
            contributor = line.split(':')[1].strip()
        if key == "#dataset_resourceprovider_organisation_longname":
            contributor = line.split(':')[1].strip()
        prev, pos = pos, f.tell()
    f.seek(prev)
    station_code = station_id
    stationmeta_core = get_stationmeta_core(db=db,station_code=station_code)
    # there is a mismatch with coordinates
    stationmeta_core.coordinates = get_geom_from_coordinates(stationmeta_core.coordinates)
    station_id = stationmeta_core.id
    timeseries = get_timeseries_by_unique_constraints(db=db,station_id=station_id,variable_id=variable_id, resource_provider=contributor)
    # again problems with converted coordinates!
    db.rollback()
    if timeseries:
        timeseries_id = timeseries.id
        # open SpooledTemporaryFile, skip header (and also try to insert timeseries_id!)
        df = pd.read_csv(input_handle.file, comment='#', header=None, sep=';',names=["time","value","flags"],parse_dates=["time"],index_col="time")
        # substract timeshift to convert data to UTC
        df.index = df.index - timeoffset
        # now insert the timeseries_id to the end of the data frame
        df.insert(2, 'timeseries_id', timeseries_id)
        # also insert version
        df.insert(3, 'version', version)
        # datetime needs timezone information
        df = df.tz_localize('UTC')
        # determine period_start and period_end of data
        period_start = min(df.index)
        period_end = max(df.index)
        # mv data from this period to data_archive
        db.execute(insert(models.DataArchive).from_select((models.Data.datetime,models.Data.value,models.Data.flags,models.Data.version,models.Data.timeseries_id),
            select([models.Data]).where(
            and_(and_(models.Data.timeseries_id == timeseries_id,
                      models.Data.datetime >= period_start),
                      models.Data.datetime <= period_end))))
        db.execute(delete(models.Data).where(
            and_(and_(models.Data.timeseries_id == timeseries_id,
                      models.Data.datetime >= period_start),
                      models.Data.datetime <= period_end)))
        db.commit()
        # now insert new data for this period from file
        buf = StringIO()
        df.to_csv(buf, header=False)
        buf.pos = 0
        buf.seek(0)
        with closing(engine.raw_connection()) as fake_conn:
            fake_cur = fake_conn.cursor()
            try:
                fake_cur.copy_from(buf, 'data', sep=',', columns=('datetime','value','flags','timeseries_id', 'version'))
                fake_conn.commit()
                # adjust data_start_date, data_end_date
                if period_start < timeseries.data_start_date:
                    timeseries.data_start_date = period_start
                if period_end > timeseries.data_end_date:
                    timeseries.data_end_date = period_end
                db.add(timeseries)
                db.commit()
                message = {"detail":{"message":"Data successfully inserted."}}
                status_code = 200
            except:
                e = sys.exc_info()[0]
                message = {"detail":{"message":"An error occurred in data insertion: %s" % (e,)}}
                status_code = 400
                return JSONResponse(status_code=status_code, content=message)
            fake_cur.close()
        # create changelog entry
        # how to determine type_of_change?
        # 4 – unspecified data value corrections (this holds also, if there is only one single value to be corrected; the addition "unspecified" keeps all possibilities open to add "specified" corrections later (e. g. from QC)
        # 5 – replaced data with a new version
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"UnspecifiedData")
        db_changelog = TimeseriesChangelog(description=description, timeseries_id=timeseries_id, author_id=author_id, type_of_change=type_of_change,
                                            old_value="", new_value="", period_start=period_start, period_end=period_end, version=version)
        db.add(db_changelog)
        db.commit()
    else:
        message = f'Timeseries not found for station {station_code.strip()}, variable {variable_name}'
        status_code = 400
    return JSONResponse(status_code=status_code, content=message)

def patch_bulk_data(db: Session, engine: Engine, description: str, version: str, bulk: List[schemas.DataPatch], author_id: int,
        toarqc_config_type: str = 'standard', force: bool = False, no_archive: bool = False):
    df = pd.DataFrame([x.dict() for x in bulk]).set_index("datetime")
    # bulk data: to be able to do at least a range test, all data needs to be from the same parameter
    #            variable_name is therefore determined for the first entry of the dataframe
    timeseries_id = int(df['timeseries_id'].iloc[0])
    timeseries = get_timeseries(db=db,timeseries_id=timeseries_id)
    variable = get_variable(db=db, variable_id=timeseries.variable_id)
    variable_name = variable.name
    timeseries_list = pd.unique(df['timeseries_id'])
    # next command just to avoid problems with converted coordinates
    db.rollback()
    for t_id in timeseries_list:
        df2 = df[df['timeseries_id'] == t_id]
        timeseries_id = int(t_id)
        # determine period_start and period_end of data
        period_start = min(df2.index)
        period_end = max(df2.index)
        # mv data from this period to data_archive
        if not no_archive:
            db.execute(insert(models.DataArchive).from_select((models.Data.datetime,models.Data.value,models.Data.flags,models.Data.version,models.Data.timeseries_id),
                select([models.Data]).where(
                and_(and_(models.Data.timeseries_id == timeseries_id,
                          models.Data.datetime >= period_start),
                          models.Data.datetime <= period_end))))
        # delete all data in period that is patched
        db.execute(delete(models.Data).where(
            and_(and_(models.Data.timeseries_id == timeseries_id,
                      models.Data.datetime >= period_start),
                      models.Data.datetime <= period_end)))
        db.commit()
        # adjust data_start_date, data_end_date
        timeseries = get_timeseries(db=db,timeseries_id=timeseries_id)
        # next command just to avoid problems with converted coordinates
        db.rollback()
        if period_start < timeseries.data_start_date:
            timeseries.data_start_date = period_start
        if period_end > timeseries.data_end_date:
            timeseries.data_end_date = period_end
        db.add(timeseries)
        db.commit()
    result = insert_dataframe (db, engine, df, toarqc_config_type=toarqc_config_type, parameter=variable_name, force=force)
    if not no_archive:
        # create changelog entry
        # how to determine type_of_change?
        # 4 – unspecified data value corrections (this holds also, if there is only one single value to be corrected; the addition "unspecified" keeps all possibilities open to add "specified" corrections later (e. g. from QC)
        # 5 – replaced data with a new version
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Replaced")
        db_changelog = TimeseriesChangelog(description=description, timeseries_id=timeseries_id, author_id=author_id, type_of_change=type_of_change,
                                            old_value="", new_value="", period_start=period_start, period_end=period_end, version=version)
        db.add(db_changelog)
        db.commit()
    return result

