# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Simple test API for data management
"""

import json
from typing import List
from fastapi import APIRouter, Depends, HTTPException, \
                    File, UploadFile, Request
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from . import crud, schemas
from toardb.utils.database import ToarDbSession, engine, get_engine, get_db
from toardb.utils.utils import (
        get_data_download_access_rights,
        get_map_data_download_access_rights,
        get_admin_access_rights,
        get_data_change_access_rights
)
from toardb.utils.settings import request_limitations
import datetime as dt

router = APIRouter()

# CRUD: create, retrieve, update, delete

# 1. retrieve

#get all data of one timeseries
@router.get('/data/timeseries/{timeseries_id}', response_model=schemas.Composite, response_model_exclude_unset=True, response_model_exclude_none=True)
def get_data(timeseries_id: int, request: Request, 
             access: dict = Depends(get_data_download_access_rights),
             db: Session = Depends(get_db)):
    # check whether the user is sending the request via the dashboard
    if not access['lfromdashboard']:
        if access['status_code'] == 200:
            if access['role'] == 'anonymous' and not (request_limitations["anonymous"]["min_tsid"] <= timeseries_id <= request_limitations["anonymous"]["max_tsid"]):
                raise HTTPException(status_code=401, detail="Anonymous access only to timeseries within the ID range of 1000 to 2000.")
        else:
            raise HTTPException(status_code=401, detail="Unauthorized.")

    db_data = crud.get_data(db, timeseries_id=timeseries_id, path_params=request.path_params, query_params=request.query_params)
    if db_data is None:
        raise HTTPException(status_code=404, detail="Data not found.")
    return db_data


#get all data of one timeseries
@router.get('/data/timeseries/id/{timeseries_id}', response_model=schemas.Composite, response_model_exclude_unset=True, response_model_exclude_none=True)
def get_data2(timeseries_id: int, request: Request, 
              access: dict = Depends(get_data_download_access_rights),
              db: Session = Depends(get_db)):
    if access['status_code'] == 200:
        if access['role'] == 'anonymous' and not 1000 <= timeseries_id <= 2000:
            raise HTTPException(status_code=401, detail="Anonymous access only to timeseries within the ID range of 1000 to 2000.")

        db_data = crud.get_data(db, timeseries_id=timeseries_id, path_params=request.path_params, query_params=request.query_params)
        if db_data is None:
            raise HTTPException(status_code=404, detail="Data not found.")
        return db_data
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


#get all data of one timeseries (including staging data)
@router.get('/data/timeseries_with_staging/id/{timeseries_id}', response_model=schemas.Composite, response_model_exclude_unset=True, response_model_exclude_none=True)
def get_data_with_staging(timeseries_id: int, flags: str = None, format: str = 'json',
                          access: dict = Depends(get_data_download_access_rights),
                          db: Session = Depends(get_db)):
    if access['status_code'] == 200:
        if access['role'] == 'anonymous' and not 1000 <= timeseries_id <= 2000:
            raise HTTPException(status_code=401, detail="Anonymous access only to timeseries within the ID range of 1000 to 2000.")

        db_data = crud.get_data_with_staging(db, timeseries_id=timeseries_id, flags=flags, format=format)
        if db_data is None:
            raise HTTPException(status_code=404, detail="Data not found.")
        return db_data
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


#get map data (for a special variable and timestamp)
@router.get('/data/map/')
def get_map_data(variable_id: int = 5, daterange: str = '2023-02-22 12:00,2023-02-22 12:00',
                 access: dict = Depends(get_map_data_download_access_rights),
                 db: Session = Depends(get_db)):
    if access['status_code'] == 200:
        db_data = crud.get_map_data(db, variable_id=variable_id, daterange=daterange)
        if db_data is None:
            raise HTTPException(status_code=404, detail="Data not found.")
        return db_data
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


#get the next available version for one timeseries
@router.get('/data/timeseries/next_version/{timeseries_id}')
def get_version(timeseries_id: int, request: Request, db: Session = Depends(get_db)):
    version = crud.get_next_version(db, timeseries_id=timeseries_id, path_params=request.path_params, query_params=request.query_params)
    return version


# post and patch only via authorization by Helmholtz AAI

# 2. create

@router.post('/data/timeseries/')
def create_data(file: UploadFile = File(...),
                toarqc_config_type: str = 'standard',
                dry_run: bool = False,
                force: bool = False,
                access: dict = Depends(get_admin_access_rights),
                db: Session = Depends(get_db),
                engine: Engine = Depends(get_engine)):
#   # the next three lines are automatically done by database management,
#   # but we do want helpful error messages!
#   db_data = crud.get_data_by_datetime_and_timeseriesid(db, datetime=data.datetime, timeseries_id=data.timeseries_id)
#   if db_data:
#       raise HTTPException(status_code=400, detail="Data already registered.")

#   BUT:
#   we want to upload a whole file!
#
    # check whether the post is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        response = crud.create_data(db, engine, author_id=access['auth_user_id'], input_handle=file,
                toarqc_config_type=toarqc_config_type, dry_run=dry_run, force=force)
        if response.status_code != 200 and response.status_code != 446:
            msg = json.loads(response.body.decode('utf-8'))
            # try to parse error messages from DBS (to be more understandable)
            msg2 = "An error occurred in public.data insertion: <class 'psycopg2.errors.UniqueViolation'>"
            if (msg == msg2):
                msg = 'Data for timeseries already registered.'
            raise HTTPException(status_code=400, detail=msg)
        return response
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


@router.post('/data/timeseries/bulk/')
def create_bulk_data(bulk: List[schemas.DataCreate],
                     toarqc_config_type: str = 'standard',
                     dry_run: bool = False,
                     force: bool = False,
                     access: dict = Depends(get_admin_access_rights),
                     db: Session = Depends(get_db),
                     engine: Engine = Depends(get_engine)):
    # check whether the post is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        response = crud.create_bulk_data(db, engine, bulk=bulk, 
                author_id=access['auth_user_id'], toarqc_config_type=toarqc_config_type, dry_run=dry_run,
                force=force)
        if response.status_code != 200:
            msg = response.body.decode('utf-8')
            # try to parse error messages from DBS (to be more understandable)
            msg2 = '"An error occurred in public.data insertion: <class \'psycopg2.errors.UniqueViolation\'>"'
            if (msg == msg2):
                msg = 'Data for timeseries already registered.'
            raise HTTPException(status_code=400, detail=msg)
        return response
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


@router.post('/data/timeseries/record/')
def create_data_record(series_id: int,
                       datetime: dt.datetime,
                       value: float,
                       flag: str,
                       version: str = None,
                       suppress_unique_violation: bool = False,
                       access: dict = Depends(get_admin_access_rights),
                       db: Session = Depends(get_db),
                       engine: Engine = Depends(get_engine)):
    # check whether the post is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        response = crud.create_data_record(db, engine, series_id=series_id, datetime=datetime,
                                           value=value, flag=flag, version=version,
                                           author_id=access['auth_user_id'])
        if response.status_code != 200:
            msg = response.body.decode('utf-8')
            # try to parse error messages from DBS (to be more understandable)
            msg2 = '"An error occurred in data insertion: <class \'psycopg2.errors.UniqueViolation\'>"'
            if (msg == msg2):
                if not suppress_unique_violation:
                    msg = 'Data for timeseries already registered.'
                    raise HTTPException(status_code=400, detail=msg)
            else:
                raise HTTPException(status_code=400, detail=msg)
        return response
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")

# 3. update

@router.patch('/data/timeseries/')
def patch_data(description: str,
               version: str,
               file: UploadFile = File(...),
               toarqc_config_type: str = 'standard',
               force: bool = False,
               access: dict = Depends(get_data_change_access_rights),
               db: Session = Depends(get_db),
               engine: Engine = Depends(get_engine)):
    # check whether the patch is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        return crud.patch_data(db, engine, description=description, version=version, toarqc_config_type=toarqc_config_type,
                force=force, author_id=access['auth_user_id'], input_handle=file)
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")

@router.patch('/data/timeseries/bulk/')
def patch_bulk_data(description: str,
                    version: str,
                    bulk: List[schemas.DataPatch],
                    toarqc_config_type: str = 'standard',
                    force: bool = False,
                    no_archive: bool = False,
                    access: dict = Depends(get_data_change_access_rights),
                    db: Session = Depends(get_db),
                    engine: Engine = Depends(get_engine)):
    # check whether the patch is authorized (401: Unauthorized)
    if access['status_code'] == 200:

        # for Max's bachelor thesis: extend description string to be more explicit
        # still to be done: GUI and REST should use the same user database!
        description = ": ".join([f"{access['user_name']} ({access['user_email']})", description])

        return crud.patch_bulk_data(db, engine, description=description, version=version, bulk=bulk, author_id=access['auth_user_id'],
                toarqc_config_type=toarqc_config_type, force=force, no_archive=no_archive)
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")

