# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Create, Read, Update, Delete functionality

"""

import json
import requests

from typing import List
from pydantic import ValidationError
from geoalchemy2.shape import to_shape
from geoalchemy2.types import Geometry
from geoalchemy2.elements import WKBElement, WKTElement
from sqlalchemy import cast, Text, insert, update, delete, select, and_, String, text
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from sqlalchemy.dialects.postgresql import JSONB, ARRAY
from sqlalchemy.inspection import inspect
from fastapi import File, UploadFile
from fastapi.responses import JSONResponse

from . import models
from .models import StationmetaCore, StationmetaGlobal, StationmetaChangelog, \
                    stationmeta_core_stationmeta_roles_table, stationmeta_core_stationmeta_annotations_table
from .schemas import get_coordinates_from_geom, get_geom_from_coordinates, get_abbreviation_from_code_description, get_code_description_from_abbreviation, \
                     Stationmeta, StationmetaCreate, StationmetaPatch, StationmetaGlobalFields, Coordinates, get_coordinates_from_string
from toardb.utils.utils import get_value_from_str, get_str_from_value, get_hr_value, create_filter
from toardb.utils.settings import base_geopeas_url
import toardb


def get_stationmeta_core(db: Session, station_code: str):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.codes.any(station_code.strip())).first()
    # there is a mismatch with coordinates and additional_metadata
    if db_object:
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_object


def get_stationmeta(db: Session, station_code: str, fields: str = None):
    if fields:
        global_params = {column.name for column in inspect(StationmetaGlobal).c if column.name not in ['id','station_id']}
        lconstr_glob = any(field in global_params for field in fields.split(','))
        if fields.find("globalmeta") >= 0:
            fields = fields.replace("globalmeta","")
            fields += ','.join(global_params)
            lconstr_glob = True
        fields2 = []
        for field in fields.split(','):
            if field == "id":
                fields2.append("stationmeta_core.id")
            elif field == "coordinates":
                fields2.append("ST_AsText(coordinates)")
            else:
                fields2.append(field)
        tmp_object = db.query(*map(text,fields2)). \
                        select_from(StationmetaCore,StationmetaGlobal). \
                        filter(and_(models.StationmetaCore.codes.any(station_code.strip()), \
                                   (models.StationmetaCore.id == models.StationmetaGlobal.station_id))).first()
        if tmp_object:
            db_object = dict(zip(fields.split(','),tmp_object))
        else:
            return tmp_object
        # there is a mismatch with coordinates and additional_metadata
        if lconstr_glob:
            db_object_cp = db_object.copy()
            global_atts = {}
            for key in db_object:
                if key in global_params:
                    global_atts[key] = db_object[key]
                    del db_object_cp[key]
            db_object = db_object_cp
            db_object["globalmeta"] = StationmetaGlobalFields(**global_atts)
        try:
            db_object['coordinates'] = get_coordinates_from_string(db_object['coordinates'])
        except:
            pass
        try:
            db_object['additional_metadata'] = str(db_object['additional_metadata']).replace("'",'"')
        except:
            pass
    else:
        db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.codes.any(station_code.strip())).first()
        # there is a mismatch with coordinates and additional_metadata
        if db_object:
            try:
                if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
                    db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
            except:
                pass
            try:
                db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
            except:
                pass
    return db_object


def get_stationmeta_by_id(db: Session, station_id: int):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.id == station_id).first()
    # there is a mismatch with coordinates and additional_metadata
    if db_object:
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_object


def station_id_exists(db: Session, station_id: int):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.id == station_id).first()
    return (db_object != None)


def get_all_stationmeta(db: Session, path_params, query_params):
    try:
        limit, offset, fields, format, filters = create_filter(query_params, "stationmeta")
        s_c_filter = filters["s_c_filter"]
        s_g_filter = filters["s_g_filter"]
    except (KeyError, ValueError) as e:
        status_code=400
        return JSONResponse(status_code=status_code, content=str(e))
    global_params = {column.name for column in inspect(StationmetaGlobal).c if column.name not in ['id','station_id']}

    if fields:
        fields2 = []
        for field in fields.split(','):
            if field == "id":
                fields2.append("stationmeta_core.id")
            elif field == "coordinates":
                fields2.append("ST_AsText(coordinates)")
            else:
                fields2.append(field)
        db_objects_l = db.query(*map(text,fields2)).select_from(StationmetaCore,StationmetaGlobal).filter(text(s_c_filter)). \
                           join(StationmetaGlobal).filter(and_(StationmetaCore.id == StationmetaGlobal.station_id, text(s_g_filter))). \
                           order_by(models.StationmetaCore.id). \
                           limit(limit).offset(offset)
        db_objects = []
        for db_object_immut in db_objects_l:
            db_object = dict(zip(fields.split(','),db_object_immut))
            # there is a mismatch with coordinates and additional_metadata
            try:
                db_object['coordinates'] = get_coordinates_from_string(db_object['coordinates'])
            except:
                pass
            try:
                db_object['additional_metadata'] = str(db_object['additional_metadata']).replace("'",'"')
            except:
                pass
            db_object_cp = db_object.copy()
            global_atts = {}
            for key in db_object:
                if key in global_params:
                    global_atts[key] = db_object[key]
                    del db_object_cp[key]
            db_object = db_object_cp
            if global_atts:
                db_object["globalmeta"] = StationmetaGlobalFields(**global_atts)
            db_objects.append(db_object)
    else:
        db_objects = db.query(models.StationmetaCore).filter(text(s_c_filter)). \
                           join(StationmetaGlobal).filter(and_(StationmetaCore.id == StationmetaGlobal.station_id, text(s_g_filter))). \
                           order_by(models.StationmetaCore.id). \
                           limit(limit).offset(offset).all()

        for db_object in db_objects:
            # there is a mismatch with coordinates and additional_metadata
            if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
                   db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
            db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')

    if limit:
        return db_objects[:limit]
    else:
        return db_objects


# is this internal, or should this also go to public REST api?
# - get_stationmeta_roles
# - get_stationmeta_changelog
# - get_unique_stationmeta_role
# - get_stationmeta_role_by_id
# - get_unique_stationmeta_annotation
def get_stationmeta_roles(db: Session, station_id: int):
    return db.execute(select([stationmeta_core_stationmeta_roles_table]).where(stationmeta_core_stationmeta_roles_table.c.station_id == station_id))


def get_stationmeta_changelog(db: Session, station_id: int):
    return db.query(models.StationmetaChangelog).filter(models.StationmetaChangelog.station_id == station_id).order_by(models.StationmetaChangelog.datetime).all()


def get_unique_stationmeta_role(db: Session, role: int, contact_id: int, status: int):
    db_object = db.query(models.StationmetaRole).filter(models.StationmetaRole.role == role) \
                                                .filter(models.StationmetaRole.contact_id == contact_id) \
                                                .filter(models.StationmetaRole.status == status) \
                                                .first()
    return db_object

def get_stationmeta_role_by_id(db: Session, role_id):
    return db.query(models.StationmetaRole).filter(models.StationmetaRole.id == role_id).first()


def get_unique_stationmeta_annotation(db: Session, text: str, contributor_id: int):
    db_object = db.query(models.StationmetaAnnotation).filter(models.StationmetaAnnotation.text == text) \
                                                      .filter(models.StationmetaAnnotation.contributor_id == contributor_id) \
                                                      .first()
    return db_object


# is this internal, or should this also go to public REST api?
def get_stationmeta_annotations(db: Session, station_id: int):
    return db.execute(select([stationmeta_core_stationmeta_annotations_table]).where(stationmeta_core_stationmeta_annotations_table.c.station_id == station_id))


def get_stationmeta_global(db: Session, station_id: int):
    db_object = db.query(models.StationmetaGlobal).filter(models.StationmetaGlobal.station_id == station_id) \
                                                      .first()
    return db_object


def determine_stationmeta_global(db, tmp_coordinates, country):
    # GEO-PEAS expects country to be a two-letter-string
    country = get_str_from_value(toardb.toardb.CN_vocabulary,country)
    db_globalservices = db.query(models.StationmetaGlobalService).all()
    globalmeta_dict = {}
    for db_service in db_globalservices:
        request=db_service.service_url.format(base_url=f'{base_geopeas_url}',lat=tmp_coordinates["lat"],lon=tmp_coordinates["lng"],
                                                                             country=country)
        try:
            result = requests.get(request).json()
            # htap_region_tier1 (also major_road) does not return "agg_function"!
            try:
                key = result["properties"]["agg_function"]
            except:
                if db_service.variable_name != 'distance_to_major_road_year2020':
                    key = "value"
                else:
                    key = 'distance'
            try:
                value = result["properties"][key]
            except:
                value = result[key]
        except:
            continue
        globalmeta_dict[db_service.variable_name] = value
    return globalmeta_dict


def create_stationmeta(db: Session, engine: Engine, stationmeta: StationmetaCreate,
                       author_id: int, force: bool):
    stationmeta_dict = stationmeta.dict()
    roles_data         = stationmeta_dict.pop('roles', None)
    annotations_data   = stationmeta_dict.pop('annotations', None)
    aux_images_data    = stationmeta_dict.pop('aux_images', None)
    aux_docs_data      = stationmeta_dict.pop('aux_docs', None)
    aux_urls_data      = stationmeta_dict.pop('aux_urls', None)
    db_stationmeta = models.StationmetaCore(**stationmeta_dict)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    tmp_coordinates = db_stationmeta.coordinates
    db_stationmeta.coordinates = get_geom_from_coordinates(Coordinates(**db_stationmeta.coordinates))
    # there's also a mismatch with additional_metadata --> BUT: this should not be switched back!
    # in upload command, we have now: "additional_metadata": "{}"
    # but return from this method gives: "additional_metadata": {}
    # ==> there is a mismatch between model(JSONB) and schema(JSON)
    db_stationmeta.additional_metadata = json.loads(str(db_stationmeta.additional_metadata).replace("'",'"'))
    num_records = 0
    if not force:
        with engine.begin() as connection:
            radius = 10
            db_cmd = f"select id,codes,name,coordinates,state,country from stationmeta_core where ST_DistanceSphere(stationmeta_core.coordinates, '{db_stationmeta.coordinates}') < {radius}"
            records = connection.execute(db_cmd)
            num_records = records.rowcount
            if num_records > 0:
                status_code=444
                if num_records == 1:
                    message={"message": f'already in database --> patch record (add station code)'}
                else:
                    message = {"message": 'more than one station falls within the given radius!\n' + \
                              'choose which station record to patch (add station code)'}
                station_ids = []
                station_records = []
                for record in records: 
                    station_ids.append(record[0])
                    station_records.append((*record[1:2],str(get_coordinates_from_geom(WKBElement(record[3]))),*record[4:]))
                message["station_ids"] = station_ids
                message["station_records"] = station_records
        engine.dispose()
    if num_records == 0:
        status_code=200
        db_stationmeta.country = get_value_from_str(toardb.toardb.CN_vocabulary,db_stationmeta.country)
        db_stationmeta.timezone = get_value_from_str(toardb.toardb.TZ_vocabulary,db_stationmeta.timezone)
        db_stationmeta.type = get_value_from_str(toardb.toardb.ST_vocabulary,db_stationmeta.type)
        db_stationmeta.type_of_area = get_value_from_str(toardb.toardb.TA_vocabulary,db_stationmeta.type_of_area)
        db_stationmeta.coordinate_validation_status = get_value_from_str(toardb.toardb.CV_vocabulary,db_stationmeta.coordinate_validation_status)
        db.add(db_stationmeta)
        result = db.commit()
        db.refresh(db_stationmeta)
        # get stationmeta_core_id
        stationmeta_core_id = db_stationmeta.id
        message={"message":f'new station: {db_stationmeta.codes},{db_stationmeta.name},{tmp_coordinates}',
                 "station_id": stationmeta_core_id}
        # store roles and update association table
        if roles_data:
            for r in roles_data:
                db_role = models.StationmetaRole(**r)
                db_role.role = get_value_from_str(toardb.toardb.RC_vocabulary,db_role.role)
                db_role.status = get_value_from_str(toardb.toardb.RS_vocabulary,db_role.status)
                # check whether role is already present in database
                db_object = get_unique_stationmeta_role(db, db_role.role, db_role.contact_id, db_role.status)
                if db_object:
                    role_id = db_object.id
                else:
                    # Something is going wrong here!
                    # Is the model StationmetaRole correctly defined?!
                    del db_role.contact
                    db.add(db_role)
                    db.commit()
                    db.refresh(db_role)
                    role_id = db_role.id
                db.execute(insert(stationmeta_core_stationmeta_roles_table).values(station_id=stationmeta_core_id, role_id=role_id))
                db.commit()
        # store annotations and update association table
        if annotations_data:
            for a in annotations_data:
                db_annotation = models.StationmetaAnnotation(**a)
                db_annotation.kind = get_value_from_str(toardb.toardb.AK_vocabulary,db_annotation.kind)
                # check whether annotation is already present in database
                db_object = get_unique_stationmeta_annotation(db, db_annotation.text, db_annotation.contributor_id)
                if db_object:
                    annotation_id = db_object.id
                else:
                    db.add(db_annotation)
                    db.commit()
                    db.refresh(db_annotation)
                    annotation_id = db_annotation.id
                db.execute(insert(stationmeta_core_stationmeta_annotations_table).values(station_id=stationmeta_core_id, annotation_id=annotation_id))
                db.commit()
        # store aux_images
        if aux_images_data:
            for i in aux_images_data:
                db_aux_image = models.StationmetaAuxImage(**i)
                db_aux_image.station_id = stationmeta_core_id
                db.add(db_aux_image)
                db.commit()
                db.refresh(db_aux_image)
        # store aux_docs
        if aux_docs_data:
            for d in aux_docs_data:
                db_aux_doc = models.StationmetaAuxDoc(**d)
                db_aux_doc.station_id = stationmeta_core_id
                db.add(db_aux_doc)
                db.commit()
                db.refresh(db_aux_doc)
        # store aux_urls
        if aux_urls_data:
            for u in aux_urls_data:
                db_aux_url = models.StationmetaAuxUrl(**u)
                db_aux_url.station_id = stationmeta_core_id
                db.add(db_aux_url)
                db.commit()
                db.refresh(db_aux_url)
        # determine stationmeta_global
        globalmeta_dict = determine_stationmeta_global(db, tmp_coordinates,db_stationmeta.country)
        db_globalmeta = models.StationmetaGlobal(**globalmeta_dict)
        db_globalmeta.station_id = stationmeta_core_id
        db.add(db_globalmeta)
        db.commit()
        db.refresh(db_globalmeta)
        # create changelog entry
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Created")
        description="station created"
        db_changelog = StationmetaChangelog(description=description, station_id=stationmeta_core_id, author_id=author_id, type_of_change=type_of_change,
                                            old_value='', new_value='')
        db.add(db_changelog)
        db.commit()
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    res ={}
    res['detail'] = message
    return JSONResponse(status_code=status_code, content=res)


def get_field_from_record(db, station_id, field, db_stationmeta):
    value = str(getattr(db_stationmeta,field)) 
    # if needed, convert this to human readable format (controlled vocabulary)
    value = get_hr_value('stationmeta',field,value)
    if isinstance(value, str):
        value = f"'{value}'"
    # I forgot to add the code to the following if/else-construct?!
    if field in ['roles', 'annotations', 'aux_images', 'aux_docs', 'aux_urls', 'globalmeta']:
        field =  f"'{field}': " + value
    else:
        field =  f"'{field}': " + value
    return field


def patch_stationmeta(db: Session, description: str,
                      station_id: int, stationmeta: StationmetaPatch,
                      author_id: int):

    # still to be done:
    # collect all old/new_values from different sections!

    no_log = (description == 'NOLOG')
    stationmeta_dict       = stationmeta.dict()
    # delete empty fields from stationmeta_dict already at this place, to be able to
    # distinguish between "single value correction in metadata" and "comprehensive metadata revision"
    # (see controlled vocabulary "CL_vocabulary")
    stationmeta_dict2 = {k: v for k, v in stationmeta_dict.items() if v is not None}
    number_of_elements = len(stationmeta_dict2)
    if not no_log:
        if (number_of_elements == 1):
            type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"SingleValue")
        else:
            type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Comprehensive")
    roles_data         = stationmeta_dict.pop('roles', None)
    annotations_data   = stationmeta_dict.pop('annotations', None)
    aux_images_data    = stationmeta_dict.pop('aux_images', None)
    aux_docs_data      = stationmeta_dict.pop('aux_docs', None)
    aux_urls_data      = stationmeta_dict.pop('aux_urls', None)
    globalmeta_data    = stationmeta_dict.pop('globalmeta', None)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    # ==> the following two commands are not working
    # ==> workaround
    # db.query(models.StationmetaCore).filter(models.StationmetaCore.id == stationm_id).update(stationmeta_dict2)
    db_obj = models.StationmetaCore(**stationmeta_dict)
    tmp_coordinates = db_obj.coordinates
    db_stationmeta = db.query(models.StationmetaCore).get(station_id)
    db_stationmeta.coordinates = get_geom_from_coordinates(db_stationmeta.coordinates)
    if not no_log:
        old_value="{"
        number_of_commas = number_of_elements - 1
    for k, v in stationmeta_dict.items():
        if v is not None:
            # prepare changelog entry
            if not no_log:
                field = get_field_from_record(db, station_id, k, db_stationmeta)
                if (number_of_commas == 0):
                    old_value=old_value + field + '}'
                else:
                    old_value=old_value + field + ','
                number_of_commas = number_of_commas - 1
            setattr(db_stationmeta,k,stationmeta_dict[k])
    # do the following for every entry that uses the controlled vocabulary!
    # this should be improved!
    if db_obj.country:
        db_stationmeta.country = get_value_from_str(toardb.toardb.CN_vocabulary, db_obj.country)
    if db_obj.timezone:
        db_stationmeta.timezone = get_value_from_str(toardb.toardb.TZ_vocabulary, db_obj.timezone)
    if db_obj.type:
        db_stationmeta.type = get_value_from_str(toardb.toardb.ST_vocabulary, db_obj.type)
    if db_obj.type_of_area:
        db_stationmeta.type_of_area = get_value_from_str(toardb.toardb.TA_vocabulary, db_obj.type_of_area)
    if db_obj.coordinate_validation_status:
        db_stationmeta.coordinate_validation_status = get_value_from_str(toardb.toardb.CV_vocabulary, db_obj.coordinate_validation_status)
    result = db.commit()
    # store roles and update association table
    if roles_data:
        # prepare changelog entry/entries
        if not no_log:
            description = description + f"; add role"
            db_old_roles = get_stationmeta_roles(db, station_id)
            old_value = old_value + "'roles': {"
            for oldr in db_old_roles:
                old_role = get_stationmeta_role_by_id(db, oldr.role_id)
                old_value=old_value + "{'role': '" + get_str_from_value(toardb.toardb.RC_vocabulary,old_role.role) + \
                                    "', 'status': '" + get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status) + \
                                    "', 'contact_id': " + str(old_role.contact_id) + '}'
            if (number_of_commas == 0):
                old_value=old_value + '}'
            else:
                old_value=old_value + ','
            number_of_commas = number_of_commas - 1
        for r in roles_data:
            db_role = models.StationmetaRole(**r)
            db_role.role = get_value_from_str(toardb.toardb.RC_vocabulary,db_role.role)
            db_role.status = get_value_from_str(toardb.toardb.RS_vocabulary,db_role.status)
            # check whether role is already present in database
            db_object = get_unique_stationmeta_role(db, db_role.role, db_role.contact_id, db_role.status)
            if db_object:
                role_id = db_object.id
            else:
                db.add(db_role)
                db.commit()
                db.refresh(db_role)
                role_id = db_role.id
            db.execute(insert(stationmeta_core_stationmeta_roles_table).values(station_id=station_id, role_id=role_id))
            db.commit()
    # store annotations and update association table
    if annotations_data:
        # prepare changelog entry/entries
        if not no_log:
            description = description + f"; add annotation"
            db_old_annotations = get_stationmeta_annotations(db, station_id)
            for olda in db_old_annotations:
                old_annotation = get_stationmeta_annotation_by_id(db, olda.annotation_id)
                old_value=old_value + "{'kind': '" + get_str_from_value(toardb.toardb.RC_vocabulary,old_role.kind) + \
                                    "', 'text': '" + get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status) + \
                                    "', 'date_added': '" + get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status) + \
                                    "', 'approved': '" + get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status) + \
                                    "', 'contributor_id': " + str(old_role.contact_id) + '}'
            if (number_of_commas == 0):
                old_value=old_value + '}'
            else:
                old_value=old_value + ','
            number_of_commas = number_of_commas - 1
        for a in annotations_data:
            db_annotation = models.StationmetaAnnotation(**a)
            db_annotation.kind = get_value_from_str(toardb.toardb.AK_vocabulary,db_annotation.kind)
            # check whether annotation is already present in database
            db_object = get_unique_stationmeta_annotation(db, db_annotation.text, db_annotation.contributor_id)
            if db_object:
                annotation_id = db_object.id
            else:
                db.add(db_annotation)
                db.commit()
                db.refresh(db_annotation)
                annotation_id = db_annotation.id
            db.execute(insert(stationmeta_core_stationmeta_annotations_table).values(station_id=station_id, annotation_id=annotation_id))
            db.commit()
    # patch aux_images
    if aux_images_data:
        for i in aux_images_data:
            db_aux_image = models.StationmetaAuxImage(**i)
            db_aux_image.station_id = station_id
            db.add(db_aux_image)
            db.commit()
            db.refresh(db_aux_image)
    # patch aux_docs
    if aux_docs_data:
        for d in aux_docs_data:
            db_aux_doc = models.StationmetaAuxDoc(**d)
            db_aux_doc.station_id = station_id
            db.add(db_aux_doc)
            db.commit()
            db.refresh(db_aux_doc)
    # patch aux_urls
    if aux_urls_data:
        for u in aux_urls_data:
            db_aux_url = models.StationmetaAuxUrl(**u)
            db_aux_url.station_id = station_id
            db.add(db_aux_url)
            db.commit()
            db.refresh(db_aux_url)
    # patch globalmeta
    if globalmeta_data:
        db_globalmeta = get_stationmeta_global(db,station_id)
        old_global = {}
        if db_globalmeta:
            # combine old and new data:
            for key, value in globalmeta_data.items():
                # "if value" would also skip a value of 0.0!!!
                if value != None:
                    old_value = getattr(db_globalmeta, key)
                    # do the following for every entry that uses the controlled vocabulary!
                    # this should be improved!
                    if key == "climatic_zone_year2016":
                        value = get_value_from_str(toardb.toardb.CZ_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.CZ_vocabulary, old_value)
                    elif key == "toar1_category":
                        value = get_value_from_str(toardb.toardb.TC_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.TC_vocabulary, old_value)
                    elif key == "toar2_category":
                        value = get_value_from_str(toardb.toardb.TA_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.TA_vocabulary, old_value)
                    elif key == "htap_region_tier1_year2010":
                        value = get_value_from_str(toardb.toardb.TR_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.TR_vocabulary, old_value)
                    elif key == "dominant_landcover_year2012":
                        value = get_value_from_str(toardb.toardb.LC_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.LC_vocabulary, old_value)
                    elif key == "landcover_description_25km_year2012":
                        value = get_abbreviation_from_code_description(toardb.toardb.LC_vocabulary, value)
                        old_value = get_code_description_from_abbreviation(toardb.toardb.LC_vocabulary, old_value)
                    elif key == "dominant_ecoregion_year2017":
                        value = get_value_from_str(toardb.toardb.ER_vocabulary, value)
                        old_value = get_str_from_value(toardb.toardb.ER_vocabulary, old_value)
                    elif key == "ecoregion_description_25km_year2017":
                        value = get_abbreviation_from_code_description(toardb.toardb.ER_vocabulary, value)
                        old_value = get_code_description_from_abbreviation(toardb.toardb.ER_vocabulary, old_value)
                    else:
                        pass
                    old_global[key] = old_value
                    setattr(db_globalmeta, key, value)
        else:
            db_globalmeta = models.StationmetaGlobal(**globalmeta_data)
        db.add(db_globalmeta)
        db.commit()
        stationmeta_dict2 = {k: v for k, v in stationmeta_dict2['globalmeta'].items() if v is not None}
        old_value = str(old_global)
    if not no_log:
        db_changelog = StationmetaChangelog(description=description, station_id=station_id, author_id=author_id, type_of_change=type_of_change,
                                            old_value=old_value, new_value=str(stationmeta_dict2))
        db.add(db_changelog)
        db.commit()
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    status_code = 200
    message={"message": f'patched stationmeta record for station_id {station_id}'}
    message["station_id"] = station_id
    return JSONResponse(status_code=status_code, content=message)


def delete_stationmeta_field(db: Session, station_id: int, field: str, author_id: int):
    # id can never be deleted (and of course also not changed)!!!
    # there are mandatory fields (from stationmeta_core), that cannot be deleted!
    # --> set these to their default value
    new_value = ""
    field_table = {'roles': stationmeta_core_stationmeta_roles_table,
                   'annotations': stationmeta_core_stationmeta_annotations_table}
#                  'aux_images': stationmeta_core_stationmeta_aux_images_table,
#                  'aux_docs': stationmeta_core_stationmeta_aux_docs_table,
#                  'aux_urls': stationmeta_core_stationmeta_aux_urls_table,
    if (field in field_table):
        db.execute(delete(field_table[field]).where(field_table[field].c.station_id==station_id))
        new_value = f"'{field}': []"
    # problem with automatic conversion of coordinates (although not explicitly fetched from database)
    # ==> next two lines are a workaround
    db_stationmeta = db.query(models.StationmetaCore).get(station_id)
    tmp_coordinates = db_stationmeta.coordinates
    db_stationmeta.coordinates = get_geom_from_coordinates(db_stationmeta.coordinates)
    # also write changelog entry
    type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"SingleValue")
    description=f"delete field {field}"
    old_value = get_field_from_record(db, station_id, field, db_stationmeta)
    db_changelog = StationmetaChangelog(description=description, station_id=station_id, author_id=author_id, type_of_change=type_of_change,
                                        old_value=old_value, new_value=new_value)
    db.add(db_changelog)
    db.commit()
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    # hotfix: db_stationmeta.global needs to be retranslated to the dict that is understood by StationmetaGlobal
    db_stationmeta.globalmeta = None
    return db_stationmeta
