# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
class StationmetaGlobal (Base)
==============================
"""
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, String, \
                       Text, text, CheckConstraint, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import JSONB, ARRAY
from .models_core import StationmetaCore
from toardb.base import Base

STATIONMETA_GLOBAL_ID_SEQ = Sequence('stationmeta_global_id_seq')  # define sequence explicitly
class StationmetaGlobal(Base):
    """
    Table "public.stationmeta_global"

    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    |                Column                            |       Type             | Collation | Nullable |                    Default                     |
    +=======================================+========================+===========+==========+================================================+
    | id                                               | integer                |           | not null | nextval('stationmeta_global_id_seq'::regclass) |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_topography_srtm_alt_90m_year1994            | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_topography_srtm_alt_1km_year1994            | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | max_topography_srtm_relative_alt_5km_year1994    | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | min_topography_srtm_relative_alt_5km_year1994    | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | stddev_topography_srtm_relative_alt_5km_year1994 | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | climatic_zone_year2016                           | integer                |           | not null | '-1'::integer                                  |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | htap_region_tier1_year2010                       | integer                |           | not null | '-1'::integer                                  |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | dominant_landcover_year2012                      | integer                |           | not null | '-1'::integer                                  |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | landcover_description_25km_year2012              | character varying(400) |           | not null | ''::character varying                          |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | dominant_ecoregion_year2017                      | integer                |           | not null | '-1'::integer                                  |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | ecoregion_description_25km_year2017              | character varying(400) |           | not null | ''::character varying                          |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | distance_to_major_road_year2020                  | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_stable_nightlights_1km_year2013             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_stable_nightlights_5km_year2013             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | max_stable_nightlights_25km_year2013             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | max_stable_nightlights_25km_year1992             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_population_density_250m_year2015            | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_population_density_5km_year2015             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | max_population_density_25km_year2015             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_population_density_250m_year1990            | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_population_density_5km_year1990             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | max_population_density_25km_year1990             | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_nox_emissions_10km_year2015                 | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | mean_nox_emissions_10km_year2000                 | double precision       |           | not null | '-999.0'::numeric                              |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | toar1_category                                   | integer                |           | not null | '-1'::integer                                  |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | toar2_category                                   | integer                |           | not null | '0'::integer                                   |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+
    | station_id                                       | integer                |           | not null |                                                |
    +--------------------------------------------------+------------------------+-----------+----------+------------------------------------------------+

    Indexes:
     "stationmeta_global_pkey" PRIMARY KEY, btree (id)
     "stationmeta_global_station_id_key" UNIQUE CONSTRAINT, btree (station_id)
    Check constraints:
     "stationmeta_global_climatic_zone_check" CHECK (climatic_zone_year2016 >= 0)
     "stationmeta_global_dominant_landcover_year2012_check" CHECK (dominant_landcover_year2012 >= 0)
     "stationmeta_global_htap_region_tier1_check" CHECK (htap_region_tier1_year2010 >= 0)
    Foreign-key constraints:
     "stationmeta_global_station_id_29ff53dd_fk_stationmeta_core_id" FOREIGN KEY (station_id) REFERENCES stationmeta_core(id) DEFERRABLE INITIALLY DEFERRED
     "stationmeta_glob_dominant_ecoregion_year2017_fk_er_voc_enum_val" FOREIGN KEY (dominant_ecoregion_year2017) REFERENCES er_vocabulary(enum_val)
     "stationmeta_glob_dominant_landcover_year2012_fk_lc_voc_enum_val" FOREIGN KEY (dominant_landcover_year2012) REFERENCES lc_vocabulary(enum_val)
     "stationmeta_global_climatic_zone_fk_cz_at_vocabulary_enum_val" FOREIGN KEY (climatic_zone_year2016) REFERENCES cz_vocabulary(enum_val)
     "stationmeta_global_htap_region_tier1_fk_tr_vocabulary_enum_val" FOREIGN KEY (htap_region_tier1_year2010) REFERENCES tr_vocabulary(enum_val)
     "stationmeta_global_toar1_category_fk_tc_vocabulary_enum_val" FOREIGN KEY (toar1_category) REFERENCES tc_vocabulary(enum_val)
     "stationmeta_global_toar2_category_fk_tc_vocabulary_enum_val" FOREIGN KEY (toar2_category) REFERENCES ta_vocabulary(enum_val)
    """
    __tablename__ = 'stationmeta_global'
#   Default values do not fit CheckConstraints  == >  Check, how both can be done!!
#   __table_args__ = (
#                        CheckConstraint('climatic_zone_year2016 >= 0'),
#                        CheckConstraint('dominant_landcover_year2012 >= 0'),
#                        CheckConstraint('htap_region_tier1_year2010 >= 0'),
#                    )

    id = Column(Integer, STATIONMETA_GLOBAL_ID_SEQ, primary_key=True, server_default=STATIONMETA_GLOBAL_ID_SEQ.next_value())
    mean_topography_srtm_alt_90m_year1994 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    mean_topography_srtm_alt_1km_year1994 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    max_topography_srtm_relative_alt_5km_year1994 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    min_topography_srtm_relative_alt_5km_year1994 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    stddev_topography_srtm_relative_alt_5km_year1994 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    climatic_zone_year2016 = Column(ForeignKey('cz_vocabulary.enum_val'), nullable=False, server_default=text("'-1'::integer"))
    htap_region_tier1_year2010 = Column(ForeignKey('tr_vocabulary.enum_val'), nullable=False, server_default=text("'-1'::integer"))
    dominant_landcover_year2012 = Column(ForeignKey('lc_vocabulary.enum_val'), nullable=False, server_default=text("'-1'::integer"))
    landcover_description_25km_year2012 = Column(String(400), nullable=False, server_default=text("''::character varying"))
    dominant_ecoregion_year2017 = Column(ForeignKey('er_vocabulary.enum_val'), nullable=False, server_default=text("'-1'::integer"))
    ecoregion_description_25km_year2017 = Column(String(400), nullable=False, server_default=text("''::character varying"))
    distance_to_major_road_year2020 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    mean_stable_nightlights_1km_year2013 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    mean_stable_nightlights_5km_year2013 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    max_stable_nightlights_25km_year2013 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    max_stable_nightlights_25km_year1992 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    mean_population_density_250m_year2015 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    mean_population_density_5km_year2015 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    max_population_density_25km_year2015 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    mean_population_density_250m_year1990 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    mean_population_density_5km_year1990 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    max_population_density_25km_year1990 = Column(Integer, nullable=False, server_default=text("'-999.0'::numeric"))
    mean_nox_emissions_10km_year2015 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    mean_nox_emissions_10km_year2000 = Column(Float(53), nullable=False, server_default=text("'-999.0'::numeric"))
    toar1_category = Column(ForeignKey('tc_vocabulary.enum_val'), nullable=False, server_default=text("'-1'::integer"))
    toar2_category = Column(ForeignKey('ta_vocabulary.enum_val'), nullable=False, server_default=text("'0'::integer"))
# do not use string declaration here (not working for pytest)
# use the explicit class name here,
# see: https://groups.google.com/forum/#!topic/sqlalchemy/YjGhE4d6K4U
#   station_id = Column(ForeignKey('stationmeta_core.id', deferrable=True, initially='DEFERRED'), nullable=False, unique=True)
    station_id = Column(ForeignKey(StationmetaCore.id))

    #uselist=False ==> 1:1 relationship
#   station = relationship('StationmetaCore', uselist=False)
    station = relationship('StationmetaCore', back_populates='globalmeta')

#   cz_vocabulary = relationship('CzVocabulary')
#   lc_vocabulary = relationship('LcVocabulary')
#   er_vocabulary = relationship('ErVocabulary')
#   tr_vocabulary = relationship('TrVocabulary')
#   tc_vocabulary = relationship('TcVocabulary')

