# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Pydantic schemas for TOAR database

"""

from typing import List, Dict
from pydantic import BaseModel, Field, BaseConfig, Json, validator
from geoalchemy2 import WKTElement
from geoalchemy2.shape import to_shape
import datetime as dt
from toardb.contacts.schemas import Contact
import toardb
from toardb.utils.utils import provenance


# the following class was taken from:
# https://github.com/tiangolo/fastapi/issues/312
class Coordinates(BaseModel):
    lat: float = Field(..., ge=-90, le=90, description="latitude coordinate of station (decimal degrees_north). This is our best estimate of the station location which is not always identical to the official station coordinates (see potential changelog entry).")
    lng: float = Field(..., ge=-180, le=180, description="longitude coordinate of station (decimal degrees_east). This is our best estimate of the station location which is not always identical to the official station coordinates (see potential changelog entry).")
    alt: float = Field(..., description="altitude of station (in m above sea level). This is our best estimate of the station altitude, which is not always identical to the reported station altitude, but frequently uses the elevation from google earth instead (see potential changelog entry).")

# ======== StationmetaCore =========

class StationmetaCoreBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    codes: List[str] = Field(None, description="list of station's codes")
    name: str = Field(None, description="Station name for use in TOAR; normally equal to data provider’s name")
    coordinates: Coordinates = Field(None, description="Best estimate of station latitude, longitude and altitude")
    coordinate_validation_status: str = Field(None, description="Flag indicating whether the location of a station has been verified (see controlled vocabulary: Coordinate Validity)")
    country: str = Field(None, description="The country, where the station resides, or which operates the station (e.g. in Antarctica) (see controlled vocabulary: Country Code)")
    state: str = Field(None, description="The state or province, where the station resides")
    type: str = Field(None, description="Type of station (see controlled vocabulary: Station Type)")
    type_of_area: str = Field(None, description="Type of station area (see controlled vocabulary: Station Type Of Area)")
    timezone: str = Field(None, description="Station timezone (see controlled vocabulary: Timezone)")
    additional_metadata: Json = Field(None, description="Any additional metadata about the station as a JSON data structure")

    class Config(BaseConfig):
        orm_mode = True

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CV_vocabulary))[0].display_str

    @validator('country')
    def check_country(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CN_vocabulary))[0].display_str

    @validator('type')
    def check_type(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.ST_vocabulary))[0].display_str

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TA_vocabulary))[0].display_str

    @validator('timezone')
    def check_timezone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TZ_vocabulary))[0].display_str


class StationmetaCorePatch(BaseModel):
    codes: List[str] = None
    name: str = None
    coordinates: Coordinates = None
    coordinate_validation_status: str = None
    country: str = None
    state: str = None
    type: str = None
    type_of_area: str = None
    timezone: str = None
    additional_metadata: Json = None

    class Config(BaseConfig):
        orm_mode = True


class StationmetaCoreCreate(StationmetaCoreBase):
    pass

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.CV_vocabulary)):
            return v
        else:
            raise ValueError(f"coordinate validation status not known: {v}")

    @validator('country')
    def check_country(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.CN_vocabulary)):
            return v
        else:
            raise ValueError(f"country not known: {v}")

    @validator('type')
    def check_type(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.ST_vocabulary)):
            return v
        else:
            raise ValueError(f"type of environment not known: {v}")

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TA_vocabulary)):
            return v
        else:
            raise ValueError(f"type of area not known: {v}")

    @validator('timezone')
    def check_timezone(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TZ_vocabulary)):
            return v
        else:
            raise ValueError(f"timezone not known: {v}")


class StationmetaCore(StationmetaCoreBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True


def get_geom_from_coordinates(coordinates: Coordinates) -> str:
    geom_wkte = f'SRID=4326;POINTZ({coordinates.lng} {coordinates.lat} {coordinates.alt})'
    return geom_wkte


def get_coordinates_from_geom(geom: WKTElement) -> Coordinates:
    shply_geom = to_shape(geom)
    coordinates = Coordinates(lng=shply_geom.x, lat=shply_geom.y, alt=shply_geom.z)
    return coordinates

def get_coordinates_from_string(point: str) -> Coordinates:
    coordinates = Coordinates(**dict(zip(['lng','lat','alt'],point[9:-1].split())))
    return coordinates

# ======== StationmetaAnnotation =========

class StationmetaAnnotationBase(BaseModel):
    id: int = None
    kind: str = Field(..., description="kind of annotation (see controlled vocabulary: Kind Of Annotation)")
    text: str = Field(..., description="text of annotation")
    date_added: dt.datetime = Field(..., description="timestamp when annotation was added")
    approved: bool = Field(..., description="Flag indicating whether the annotation of a station has been verified")
    contributor_id: int = Field(..., description="ID of contributor who added the annotation")

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AK_vocabulary))[0].display_str


class StationmetaAnnotationPatch(BaseModel):
    kind: int = None
    text: str = None
    date_added: dt.datetime = None
    approved: bool = None
    contributor_id: int = None

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AK_vocabulary))[0].display_str


class StationmetaAnnotationCreate(StationmetaAnnotationBase):

    @validator('kind')
    def check_kind(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.AK_vocabulary)):
            return v
        else:
            raise ValueError(f"kind of annotation code not known: {v}")



class StationmetaAnnotation(StationmetaAnnotationBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# ======== StationmetaAux =========

# -------- StationmetaAuxDoc ---------

class StationmetaAuxDocBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="auxillary document file")
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxDocPatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    station_id: int = None


class StationmetaAuxDocCreate(StationmetaAuxDocBase):
    pass


class StationmetaAuxDoc(StationmetaAuxDocBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# -------- StationmetaAuxImage ---------

class StationmetaAuxImageBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="auxillary image file")
    image_height: int = Field(..., description="Number of image pixels in direction of height")
    image_width: int = Field(..., description="Number of image pixels in direction of width")
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxImagePatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    image_height: int = None
    image_width: int = None
    station_id: int = None


class StationmetaAuxImageCreate(StationmetaAuxImageBase):
    pass


class StationmetaAuxImage(StationmetaAuxImageBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# -------- StationmetaAuxUrl ---------

class StationmetaAuxUrlBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="URL")
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxUrlPatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    station_id: int = None


class StationmetaAuxUrlCreate(StationmetaAuxUrlBase):
    pass


class StationmetaAuxUrl(StationmetaAuxUrlBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# ======== StationmetaGlobal =========

class StationmetaGlobalBase(BaseModel):
    mean_topography_srtm_alt_90m_year1994: float = Field(..., description="mean value within a radius of 90 m around station location of the following data of the year 1994: " + str(provenance['topography_srtm']))
    mean_topography_srtm_alt_1km_year1994: float = Field(..., description="mean value within a radius of 1 km around station location of the following data of the year 1994: " + str(provenance['topography_srtm']))
    max_topography_srtm_relative_alt_5km_year1994: float = Field(..., description="maximum value within a radius of 5 km around station location with relative altitude of the following data of the year 1994: " + str(provenance['topography_srtm']))
    min_topography_srtm_relative_alt_5km_year1994: float = Field(..., description="minimum value within a radius of 5 km around station location with relative altitude of the following data of the year 1994: " + str(provenance['topography_srtm']))
    stddev_topography_srtm_relative_alt_5km_year1994: float = Field(..., description="standard deviation within a radius of 5 km around station location with relative altitude of the following data of the year 1994: " + str(provenance['topography_srtm']))
    climatic_zone_year2016: str = Field(..., description= "value for the year 2016 of the following data: " + str(provenance['climatic_zone']) + " (see controlled vocabulary: Climatic Zone 2019)")
    htap_region_tier1_year2010: str = Field(..., description= "value for the year 2010 of the following data: " + str(provenance['htap_region_tier1']) + " (see controlled vocabulary: Station HTAP Region)")
    dominant_landcover_year2012: str = Field(..., description="value for the year 2012 of the following data: " + str(provenance['landcover']) + " (see controlled vocabulary: Station Landcover Type)")
    landcover_description_25km_year2012: str = Field(..., description="description of the values for the year 2012 within a radius of 25 km around station location of the following data: " + str(provenance['landcover']) + " (see controlled vocabulary: Station Landcover Type)")
    dominant_ecoregion_year2017: str = Field(..., description="value for the year 2017 of the following data: " + str(provenance['ecoregion']) + " (see controlled vocabulary: Station ECO Region Type)")
    ecoregion_description_25km_year2017: str = Field(..., description="description of the values for the year 2017 within a radius of 25 km around station location of the following data: " + str(provenance['ecoregion']) + " (see controlled vocabulary: Station ECO Region Type)")
    distance_to_major_road_year2020: float = Field(..., description="value for the year 2020 of the following data: " + str(provenance['major_road']))
    mean_stable_nightlights_1km_year2013: float = Field(..., description="mean value within a radius of 1 km around station location of the following data of the year 2013: " + str(provenance['stable_nightlights']))
    mean_stable_nightlights_5km_year2013: float = Field(..., description="mean value within a radius of 5 km around station location of the following data of the year 2013: " + str(provenance['stable_nightlights']))
    max_stable_nightlights_25km_year2013: float = Field(..., description="maximum value within a radius of 5 km around station location of the following data of the year 2013: " + str(provenance['stable_nightlights']))
    max_stable_nightlights_25km_year1992: float = Field(..., description="maximum value within a radius of 25 km around station location of the following data of the year 2013: " + str(provenance['stable_nightlights']))
    mean_population_density_250m_year2015: float = Field(..., description="mean value within a radius of 250 m around station location of the following data of the year 2015: " + str(provenance['population_density']))
    mean_population_density_5km_year2015: float = Field(..., description="mean value within a radius of 5 km around station location of the following data of the year 2015: " + str(provenance['population_density']))
    max_population_density_25km_year2015: float = Field(..., description="maximum value within a radius of 25 km around station location of the following data of the year 2015: " + str(provenance['population_density']))
    mean_population_density_250m_year1990: float = Field(..., description="human population on a square of 250 m for the year 1990 (residents km-2)")
    mean_population_density_5km_year1990: float = Field(..., description="mean value within a radius of 250 m around station location of the following data of the year 1990: " + str(provenance['population_density']))
    max_population_density_25km_year1990: float = Field(..., description="maximum value within a radius of 25 km around station location of the following data of the year 1990: " + str(provenance['population_density']))
    mean_nox_emissions_10km_year2015: float = Field(..., description="mean value within a radius of 10 km around station location of the following data of the year 2015: " + str(provenance['nox_emissions']))
    mean_nox_emissions_10km_year2000: float = Field(..., description="mean value within a radius of 10 km around station location of the following data of the year 2000: " + str(provenance['nox_emissions']))
    toar1_category: str = Field(..., description="The station classification for the Tropsopheric Ozone Assessment Report based on the station proxy data that are stored in the TOAR database (see controlled vocabulary: Station TOAR Category)")
    toar2_category: str = Field(..., description="The station classification for the TOAR-II based on the station proxy data that are stored in the TOAR database and obtained from an ML approach (see controlled vocabulary: Station TOAR Category)")
#   station_id: int = Field(..., description="internal station_id to which these global data belong")


    @validator('climatic_zone_year2016')
    def check_climatic_zone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CZ_vocabulary))[0].display_str

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TC_vocabulary))[0].display_str

    @validator('toar2_category')
    def check_toar2_category(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TA_vocabulary))[0].display_str

    @validator('htap_region_tier1_year2010')
    def check_htap_region_tier1(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TR_vocabulary))[0].display_str

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.LC_vocabulary))[0].display_str

    @validator('landcover_description_25km_year2012')
    def check_landcover_description_25km_year2012(cls, v):
        return get_full_description_from_abbreviation(toardb.toardb.LC_vocabulary, v)

    @validator('dominant_ecoregion_year2017')
    def check_dominant_ecoregion_year2017(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.ER_vocabulary))[0].display_str

    @validator('ecoregion_description_25km_year2017')
    def check_ecoregion_description_25km_year2017(cls, v):
        return get_full_description_from_abbreviation(toardb.toardb.ER_vocabulary, v)


class StationmetaGlobalPatch(BaseModel):
    mean_topography_srtm_alt_90m_year1994: float = None
    mean_topography_srtm_alt_1km_year1994: float = None
    max_topography_srtm_relative_alt_5km_year1994: float = None
    min_topography_srtm_relative_alt_5km_year1994: float = None
    stddev_topography_srtm_relative_alt_5km_year1994: float = None
    climatic_zone_year2016: str = None
    htap_region_tier1_year2010: str = None
    dominant_landcover_year2012: str = None
    landcover_description_25km_year2012: str = None
    dominant_ecoregion_year2017: str = None
    ecoregion_description_25km_year2017: str = None
    distance_to_major_road_year2020: float = None
    mean_stable_nightlights_1km_year2013: float = None
    mean_stable_nightlights_5km_year2013: float = None
    max_stable_nightlights_25km_year2013: float = None
    max_stable_nightlights_25km_year1992: float = None
    mean_population_density_250m_year2015: float = None
    mean_population_density_5km_year2015: float = None
    max_population_density_25km_year2015: float = None
    mean_population_density_250m_year1990: float = None
    mean_population_density_5km_year1990: float = None
    max_population_density_25km_year1990: float = None
    mean_nox_emissions_10km_year2015: float = None
    mean_nox_emissions_10km_year2000: float = None
    toar1_category: str = None
    toar2_category: str = None
    station_id: int = None

    @validator('climatic_zone_year2016')
    def check_climatic_zone(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.CZ_vocabulary)):
            return v
        else:
            raise ValueError(f"climatic zone not known: {v}")

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TC_vocabulary)):
            return v
        else:
            raise ValueError(f"TOAR1 category not known: {v}")

    @validator('toar2_category')
    def check_toar2_category(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TA_vocabulary)):
            return v
        else:
            raise ValueError(f"TOAR2 category not known: {v}")

    @validator('htap_region_tier1_year2010')
    def check_htap_region_tier1(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TR_vocabulary)):
            return v
        else:
            raise ValueError(f"HTAP region TIER1 not known: {v}")

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.LC_vocabulary)):
            return v
        else:
            raise ValueError(f"dominant landcover (year2012) not known: {v}")

    @validator('dominant_ecoregion_year2017')
    def check_dominant_ecoregion_year2017(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.ER_vocabulary)):
            return v
        else:
            raise ValueError(f"dominant ecoregion (year2017) not known: {v}")

    @validator('landcover_description_25km_year2012')
    def check_landcover_description_25km_year2012(cls, v):
        try:
            desc = get_abbreviation_from_code_description(toardb.toardb.LC_vocabulary, v)
            return v
        except:
            raise ValueError(f"landcover_description_25km_year2012 not known: {v}")

    @validator('ecoregion_description_25km_year2017')
    def check_ecoregion_description_25km_year2017(cls, v):
        try:
            desc = get_abbreviation_from_code_description(toardb.toardb.ER_vocabulary, v)
            return v
        except:
            raise ValueError(f"ecoregion_description_25km_year2017 not known: {v}")


def get_full_description_from_abbreviation (vocabulary, abbr_desc: str):
    full_description = ''
    if abbr_desc != '':
        for desc in abbr_desc.split(','):
            code, perc = desc.split(':')
            ftype = tuple(filter(lambda x: x.value == int(code), vocabulary))[0].display_str
            if full_description:
                full_description += ', ' + ftype + ': ' + perc
            else:
                full_description = ftype + ': ' + perc
    return full_description


def get_code_description_from_abbreviation (vocabulary, abbr_desc: str):
    code_description = ''
    if abbr_desc != '':
        for desc in abbr_desc.split(','):
            code, perc = desc.split(':')
            ftype = tuple(filter(lambda x: x.value == int(code.strip()), vocabulary))[0].string
            if code_description:
                code_description += ',' + ftype + ':' + perc.strip()
            else:
                code_description = ftype + ':' + perc.strip()
    return code_description


def get_abbreviation_from_code_description(vocabulary, full_desc: str):
    full_description = ''
    if full_desc != '':
        for desc in full_desc.split(','):
            code, perc = desc.split(':')
            abbr = tuple(filter(lambda x: x.string == code.strip(), vocabulary))[0].value
            if full_description:
                full_description += ',' + str(abbr) + ':' + perc.strip()
            else:
                full_description = str(abbr) + ':' + perc.strip()
    return full_description


class StationmetaGlobalFields(StationmetaGlobalPatch):
    pass

    @validator('climatic_zone_year2016')
    def check_climatic_zone(cls, v):
        return v

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        return v

    @validator('toar2_category')
    def check_toar2_category(cls, v):
        return v

    @validator('htap_region_tier1_year2010')
    def check_htap_region_tier1(cls, v):
        return v

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        return v

    @validator('dominant_ecoregion_year2017')
    def check_dominant_ecoregion_year2017(cls, v):
        return v

    @validator('landcover_description_25km_year2012')
    def check_landcover_description_25km_year2012(cls, v):
        return v

    @validator('ecoregion_description_25km_year2017')
    def check_ecoregion_description_25km_year2017(cls, v):
        return v


class StationmetaGlobalCreate(StationmetaGlobalBase):
    pass

    @validator('climatic_zone_year2016')
    def check_climatic_zone(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.CZ_vocabulary)):
            return v
        else:
            raise ValueError(f"climatic zone not known: {v}")

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TC_vocabulary)):
            return v
        else:
            raise ValueError(f"TOAR1 category not known: {v}")

    @validator('toar2_category')
    def check_toar2_category(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TA_vocabulary)):
            return v
        else:
            raise ValueError(f"TOAR2 category not known: {v}")

    @validator('htap_region_tier1_year2010')
    def check_htap_region_tier1(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.TR_vocabulary)):
            return v
        else:
            raise ValueError(f"HTAP region TIER1 not known: {v}")

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.LC_vocabulary)):
            return v
        else:
            raise ValueError(f"dominant landcover (year2012) not known: {v}")

    @validator('dominant_ecoregion_year2017')
    def check_dominant_ecoregion_year2017(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.ER_vocabulary)):
            return v
        else:
            raise ValueError(f"dominant ecoregion (year2017) not known: {v}")


class StationmetaGlobal(StationmetaGlobalBase):
    class Config:
        orm_mode = True


class StationmetaGlobalBaseNested(StationmetaGlobalBase):
    mean_topography_srtm_alt_90m_year1994: float = None
    mean_topography_srtm_alt_1km_year1994: float = None
    max_topography_srtm_relative_alt_5km_year1994: float = None
    min_topography_srtm_relative_alt_5km_year1994: float = None
    stddev_topography_srtm_relative_alt_5km_year1994: float = None
    climatic_zone_year2016: str = None
    htap_region_tier1_year2010: str = None
    dominant_landcover_year2012: str = None
    landcover_description_25km_year2012: str = None
    full_description_year2012: str = None
    dominant_ecoregion_year2017: str = None
    ecoregion_description_25km_year2017: str = None
    distance_to_major_road_year2020: float = None
    mean_stable_nightlights_1km_year2013: float = None
    mean_stable_nightlights_5km_year2013: float = None
    max_stable_nightlights_25km_year2013: float = None
    max_stable_nightlights_25km_year1992: float = None
    mean_population_density_250m_year2015: float = None
    mean_population_density_5km_year2015: float = None
    max_population_density_25km_year2015: float = None
    mean_population_density_250m_year1990: float = None
    mean_population_density_5km_year1990: float = None
    max_population_density_25km_year1990: float = None
    mean_nox_emissions_10km_year2015: float = None
    mean_nox_emissions_10km_year2000: float = None
    toar1_category: str = None
    toar2_category: str = None


class StationmetaGlobalNestedCreate(StationmetaGlobalBaseNested):
    pass


class StationmetaGlobalNested(StationmetaCoreBase, StationmetaGlobalBaseNested):
    class Config:
        orm_mode = True


# ======== StationmetaGlobalService =========

class StationmetaGlobalServiceBase(BaseModel):
    variable_name: str = Field(..., description="Variable name which will be returned by the service URL")
    service_url: str = Field(..., description="Service URL with %%LAT%% and %%LON%% placeholders to enter station or point location in floating point degrees")


class StationmetaGlobalService(StationmetaGlobalServiceBase):

    class Config:
        orm_mode = True

# ======== StationmetaRole =========

class StationmetaRoleBase(BaseModel):
    id: int = None
    role: str = Field(..., description="Role of contact (see controlled vocabulary: Role Code)")
    status: str = Field(..., description="Status of contact (see controlled vocabulary: Role Status)")
    contact: Contact

    @validator('role')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.RC_vocabulary))[0].display_str

    @validator('status')
    def check_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.RS_vocabulary))[0].display_str

    @validator('contact')
    def check_contact_is_private(cls, v):
        if v.person:
            if (v.person.id != 0):
                if (v.person.isprivate):
                    return None
                else:
                    return v.person
        else:
            return v.organisation

    class Config:
        orm_mode = True


class StationmetaRolePatch(BaseModel):
    role: str
    status: str
#   at the moment contact_id has to be given...
#   ==> in the future: give unique contact_email
#   patching stationmeta should not result in creating new contacts!
#   ==> still to do: check, whether contact already exists (otherwise patching cannot be done)
    contact_id: int

    class Config:
        orm_mode = True


class StationmetaRoleCreate(StationmetaRoleBase):
    contact_id: int
    contact: Contact = None

    @validator('role')
    def check_role(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.RC_vocabulary)):
            return v
        else:
            raise ValueError(f"role code not known: {v}")

    @validator('status')
    def check_status(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.RS_vocabulary)):
            return v
        else:
            raise ValueError(f"role status not known: {v}")


class StationmetaRole(StationmetaRoleBase):
    id: int = Field(..., description="for internal use only")
    contact: Contact = Field(..., description="Contact for this role")

    class Config:
        orm_mode = True


# ======== StationmetaChangelog =========

class StationmetaChangelogBase(BaseModel):
    datetime: dt.datetime = Field(..., description="Date of change to the TOAR database")
    description: str = Field(..., description="Description of change")
    old_value: str = Field(..., description="value that has been changed in the TOAR database")
    new_value: str = Field(..., description="new, changed value")
    station_id: int = Field(..., description="internal ID of station to which this change belongs")
    author_id: int = Field(..., description="internal ID of person who submitted the change to the TOAR database")
    type_of_change: int = Field(..., description="Type of change (see controlled vocabulary: Type Of Change)")


    @validator('type_of_change')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CL_vocabulary))[0].display_str

    class Config:
        orm_mode = True

class StationmetaChangelog(StationmetaChangelogBase):

    class Config:
        orm_mode = True


# ======== for nested view/upload =========

class StationmetaBase(StationmetaCoreBase):
    roles: List[StationmetaRole] = None
    annotations: List[StationmetaAnnotation] = None
    aux_images: List[StationmetaAuxImage] = None
    aux_docs: List[StationmetaAuxDoc] = None
    aux_urls: List[StationmetaAuxUrl] = None
    globalmeta: StationmetaGlobal = None
    changelog: List[StationmetaChangelog] = None

    class Config:
       orm_mode = True

    @validator('changelog')
    def order_changelog(cls, v):
        return sorted(v, key=lambda x: x.datetime)

    @validator('roles')
    def check_roles(cls, v):
        if v == []:
            return None
        else:
            return v

    @validator('annotations')
    def check_annotations(cls, v):
        if v == []:
            return None
        else:
            return v


class StationmetaPatch(StationmetaCoreCreate):
    #roles: List[StationmetaRolePatch] = None
    #annotations: List[StationmetaAnnotationPatch] = None
    # just to get things working
    roles: list = None
    annotations: list = None
    aux_images: List[StationmetaAuxImagePatch] = None
    aux_docs: List[StationmetaAuxDocPatch] = None
    aux_urls: List[StationmetaAuxUrlPatch] = None
    globalmeta: StationmetaGlobalPatch = None

    class Config:
       orm_mode = True


class StationmetaCreate(StationmetaCoreCreate):
    roles: List[StationmetaRoleCreate] = None
    annotations: List[StationmetaAnnotationCreate] = None
    aux_images: List[StationmetaAuxImage] = None
    aux_docs: List[StationmetaAuxDoc] = None
    aux_urls: List[StationmetaAuxUrl] = None
    globalmeta: StationmetaGlobalNestedCreate = None

    class Config:
        orm_mode = True


class StationmetaFields(StationmetaCorePatch):
    roles: List[StationmetaRolePatch] = None
    annotations: List[StationmetaAnnotationPatch] = None
    aux_images: List[StationmetaAuxImagePatch] = None
    aux_docs: List[StationmetaAuxDocPatch] = None
    aux_urls: List[StationmetaAuxUrlPatch] = None
    globalmeta: StationmetaGlobalFields = None

    class Config:
       orm_mode = True


class Stationmeta(StationmetaBase):
    id: int = Field(None, description="for internal use only")

    class Config:
      orm_mode = True

