# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Simple test API for stationmeta management
"""
from typing import List
from fastapi import APIRouter, Depends, HTTPException, Body, Request
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.engine import Engine
from . import crud, schemas
from toardb.utils.database import get_db, get_engine
from toardb.utils.utils import (
        get_admin_access_rights,
        get_station_md_change_access_rights
)


router = APIRouter()

# CRUD: create, retrieve, update, delete

# 1. retrieve

#get all entries of table stationmeta
@router.get('/stationmeta/', response_model=List[schemas.Stationmeta  | schemas.StationmetaFields], response_model_exclude_none=True, response_model_exclude_unset=True)
def get_all_stationmeta(request: Request, db: Session = Depends(get_db)):
    return crud.get_all_stationmeta(db, path_params=request.path_params, query_params=request.query_params)


#get all core metadata of one station
@router.get('/stationmeta/{station_code}', response_model=schemas.Stationmeta | schemas.StationmetaFields, response_model_exclude_none=True, response_model_exclude_unset=True)
def get_stationmeta(station_code: str, fields: str = None, db: Session = Depends(get_db)):
    db_stationmeta = crud.get_stationmeta(db, station_code=station_code, fields=fields)
    if db_stationmeta is None:
        raise HTTPException(status_code=404, detail=f"Metadata for station '{station_code}' not found.")
    return db_stationmeta


#get all core metadata of one station (given its ID)
@router.get('/stationmeta/id/{station_id}', response_model=schemas.Stationmeta, response_model_exclude_none=True, response_model_exclude_unset=True)
def get_stationmeta_by_id(station_id: int, db: Session = Depends(get_db)):
    db_stationmeta = crud.get_stationmeta_by_id(db, station_id=station_id)
    if db_stationmeta is None:
        raise HTTPException(status_code=404, detail=f"Metadata for station with ID '{station_id}' not found.")
    return db_stationmeta


#to be deleted?
@router.get('/stationmeta_changelog/{station_id}', response_model=List[schemas.StationmetaChangelog])
def get_stationmeta_changelog(station_id: int, db: Session = Depends(get_db)):
    db_changelog = crud.get_stationmeta_changelog(db, station_id=station_id)
    return db_changelog


# post and patch only via authorization by Helmholtz AAI

# 2. create

@router.post('/stationmeta/')
async def create_stationmeta_core(request: Request,
                            stationmeta: schemas.StationmetaCreate = Body(..., embed = True),
                            force: bool = False,
                            access: dict = Depends(get_admin_access_rights),
                            db: Session = Depends(get_db),
                            engine: Engine = Depends(get_engine)):
    # check whether the post is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        for station_code in stationmeta.codes:
            db_stationmeta_core= crud.get_stationmeta_core(db, station_code=station_code)
            if db_stationmeta_core:
                raise HTTPException(status_code=443, detail={"message":"Station already registered!", "station_id": db_stationmeta_core.id})
        return crud.create_stationmeta(db=db, engine=engine, stationmeta=stationmeta,
                                       author_id=access['auth_user_id'], force=force)
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


# 3. update

#@router.patch('/stationmeta/{station_code}', response_model=schemas.StationmetaPatch)
@router.patch('/stationmeta/{station_code}')
def patch_stationmeta_core(request: Request,
                           station_code: str,
                           stationmeta: schemas.StationmetaPatch = Body(..., embed = True),
                           access: dict = Depends(get_station_md_change_access_rights),
                           db: Session = Depends(get_db)):
    # check whether the patch is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        query_params=request.query_params
        # check, if description has been given in query_params!
        if not "description" in query_params:
            raise HTTPException(status_code=404, detail="description text ist missing.")
        else:
            description = query_params.get("description")
            db_stationmeta = crud.get_stationmeta(db, station_code=station_code)
            if db_stationmeta is None:
                raise HTTPException(status_code=404, detail="Station for patching not found.")

            # for Max's bachelor thesis: extend description string to be more explicit
            # still to be done: GUI and REST should use the same user database!
            description = ": ".join([f"{access['user_name']} ({access['user_email']})", description])

            return crud.patch_stationmeta(db=db, description=description,
                                          station_id=db_stationmeta.id, stationmeta=stationmeta,
                                          author_id=access['auth_user_id'])
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


@router.patch('/stationmeta/delete_field/{station_code}', response_model=schemas.StationmetaBase, response_model_exclude_none=True, response_model_exclude_unset=True)
def delete_field_from_stationmeta_core(request: Request,
                           station_code: str,
                           field: str,
                           access: dict = Depends(get_station_md_change_access_rights),
                           db: Session = Depends(get_db)):
    # check whether the patch is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        db_stationmeta = crud.get_stationmeta(db, station_code=station_code)
        if db_stationmeta is None:
            raise HTTPException(status_code=404, detail="Station for deleting field not found.")
        return crud.delete_stationmeta_field(db=db, station_id=db_stationmeta.id, field=field,
                author_id=access['auth_user_id'])
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")

