# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

# taken from: https://github.com/tiangolo/fastapi/issues/831
import pytest
from starlette.testclient import TestClient
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import database_exists, create_database, drop_database
from fastapi import Request

from toardb.base import Base
from toardb.toardb import app
from toardb.auth_user.models import AuthUser
from toardb.utils.database import DATABASE_URL, get_db, get_engine
from toardb.utils.utils import (
    get_admin_access_rights,
    get_station_md_change_access_rights,
    get_timeseries_md_change_access_rights,
    get_data_change_access_rights
)

url = "postgresql://postgres:postgres@postgres:5432/postgres"
_db_conn = create_engine(url, pool_pre_ping=True, pool_size=32, max_overflow=128)
sess = sessionmaker(bind=_db_conn, autoflush=False, autocommit=False)


def get_test_engine() -> Engine:
    assert _db_conn is not None
    return _db_conn


def get_test_db():
    test_db = sess()
    try:
        yield test_db
    finally:
        test_db.close()


async def override_dependency(request: Request):
    db = next(get_test_db())
    email = request.headers.get("email")
    db_user = db.query(AuthUser).filter(AuthUser.email == email).first()
    # status_code will be taken from the AAI (here: faked)
    status_code = 401
    if db_user:
        # status_code will be taken from the AAI (here: faked)
        status_code = 200
        access_dict = {
            "status_code": status_code,
            "user_name": "Sabine Schröder",
            "user_email": email,
            "auth_user_id": db_user.id
        }
    else:
        # the user needs to be added to the database!
        # (maybe users already have the credentials (in the AAI),
        # but they also need a permanent auth_user_id related to the TOAR database)
        access_dict = {
            "status_code": status_code,
            "user_name": "Something from AAI",
            "user_email": email,
            "auth_user_id": -1
        }
    return access_dict


app.dependency_overrides[get_admin_access_rights] = override_dependency
app.dependency_overrides[get_station_md_change_access_rights] = override_dependency
app.dependency_overrides[get_timeseries_md_change_access_rights] = override_dependency
app.dependency_overrides[get_data_change_access_rights] = override_dependency


@pytest.fixture(scope="session", autouse=True)
def create_test_database():
    """
    Create a clean database on every test case.
    We use the `sqlalchemy_utils` package here for a few helpers in consistently
    creating and dropping the database.
    """
#   if database_exists(url):
#       drop_database(url)
#   create_database(url)  # Create the test database.
    # 'create_all' does not work (because of undefined 'Geometry')!
    # declare PostGIS extension! (and toar_controlled_vocabulary)
#   fake_conn = _db_conn.raw_connection()
#   fake_cur = fake_conn.cursor()
#   fake_cur.execute("CREATE EXTENSION IF NOT EXISTS postgis")
#   fake_conn.commit()
#   fake_cur.execute("CREATE EXTENSION IF NOT EXISTS toar_controlled_vocabulary")
#   fake_conn.commit()
#   fake_cur.execute("SET TIMEZONE='UTC'")
#   fake_conn.commit()
    Base.metadata.create_all(_db_conn)  # Create the tables.
    # try with the basics
    app.dependency_overrides[get_db] = get_test_db  # Mock the Database Dependency
    app.dependency_overrides[get_engine] = get_test_engine  # Mock the Database Dependency
    yield  # Run the tests.
#   drop_database(url)  # Drop the test database.


@pytest.fixture
def test_db_session():
    """Returns an sqlalchemy session, and after the test tears down everything properly."""

    session = sessionmaker(bind=_db_conn, autoflush=False, autocommit=False)()
    yield session
    # Drop all data after each test
    for tbl in reversed(Base.metadata.sorted_tables):
    # otherwiese all tables from "toar_controlled_vocabulary" will get lost!
        if not tbl.name.endswith("_vocabulary"):
            _db_conn.execute(tbl.delete())
    _db_conn.execute("DELETE FROM staging.data;")
    fake_conn = _db_conn.raw_connection()
    fake_cur = fake_conn.cursor()
    fake_cur.execute("ALTER TABLE timeseries_changelog ALTER COLUMN datetime SET DEFAULT now();")
    fake_conn.commit()
    # put back the connection to the connection pool
    session.close()


@pytest.fixture()
def client():
    """
    When using the 'client' fixture in test cases, we'll get full database
    rollbacks between test cases:
    """
    with TestClient(app) as client:
        yield client
