# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Create, Read, Update, Delete functionality

"""

from sqlalchemy import insert, select, and_, func, text
from sqlalchemy.orm import Session, load_only
from sqlalchemy.util import _collections
from sqlalchemy.exc import IntegrityError
from geoalchemy2.elements import WKBElement, WKTElement
from fastapi import File, UploadFile
from fastapi.responses import JSONResponse
import datetime as dt
import json
from . import models
from .models import TimeseriesChangelog, timeseries_timeseries_roles_table, \
                    timeseries_timeseries_annotations_table, s1_contributors_table
from toardb.stationmeta.models import StationmetaCore, StationmetaGlobal
from toardb.stationmeta.schemas import get_coordinates_from_geom, get_geom_from_coordinates, get_coordinates_from_string
from toardb.stationmeta.crud import get_stationmeta_by_id, get_stationmeta_core, station_id_exists, get_stationmeta_changelog
from toardb.contacts.crud import get_organisation_by_name, get_contact
from toardb.contacts.models import Organisation, Person, Contact
from toardb.contacts.schemas import ContactBase
from toardb.variables.models import Variable
from toardb.variables.crud import get_variable
from .schemas import TimeseriesCreate, TimeseriesPatch, TimeseriesRoleNoCreate, TimeseriesRoleFields
from toardb.utils.utils import get_value_from_str, get_str_from_value, create_filter, roles_params
import toardb


def clean_additional_metadata(ad_met_dict):
    # all changes are permanent!
    if not isinstance(ad_met_dict,dict):
        tmp = ad_met_dict.replace('"','\\"')
        return tmp.replace("'",'"')
    # there is a mismatch with additional_metadata
    additional_metadata = ad_met_dict
    for key, value in additional_metadata.items():
        if isinstance(value,dict):
            for key2, value2 in value.items():
                if isinstance(value2,str):
                    additional_metadata[key][key2] = value2.replace("'","$apostroph$")
        elif isinstance(value,str):
            additional_metadata[key] = value.replace("'","$apostroph$")
    additional_metadata = str(additional_metadata).replace('"','\\"')
    additional_metadata = str(additional_metadata).replace("'",'"')
    additional_metadata = str(additional_metadata).replace("$apostroph$","'")
    return additional_metadata


def get_timeseries(db: Session, timeseries_id: int, fields: str = None):
    if fields:
        fields = ','.join('"{}"'.format(word) for word in fields.split(','))
        # this command returns a tuple (which is unmutable --> problems with additional_metadata and coordinates!)
        # also this command could not deal with reserved words used as column names (like "order")
        resultproxy = db.execute(f'SELECT {fields} FROM timeseries WHERE id={timeseries_id} LIMIT 1')
        db_object_dict = [ rowproxy._asdict() for rowproxy in resultproxy ][0]
        if fields.find('additional_metadata') != -1:
            db_object_dict['additional_metadata'] = clean_additional_metadata(db_object_dict['additional_metadata'])
        db_object = models.Timeseries(**db_object_dict)
    else:
        db_object = db.query(models.Timeseries).filter(models.Timeseries.id == timeseries_id).first()
        if db_object:
            # only for internal use!
            if db_object.data_license_accepted:
                del db_object.data_license_accepted
            if db_object.dataset_approved_by_provider:
                del db_object.dataset_approved_by_provider
        if db_object:
            try:
                # there is a mismatch with additional_metadata
                db_object.additional_metadata = clean_additional_metadata(db_object.additional_metadata)
            except:
                pass
            try:
                # there is also a mismatch with coordinates and additional_metadata from station object
                if isinstance(db_object.station.coordinates, (WKBElement, WKTElement)):
                    db_object.station.coordinates = get_coordinates_from_geom(db_object.station.coordinates)
                # there is a mismatch with additional_metadata
                if isinstance(db_object.station.additional_metadata, dict):
                    db_object.station.additional_metadata = json.dumps(db_object.station.additional_metadata)
            except:
                pass
    return db_object

# https://www.reddit.com/r/flask/comments/ypqk40/default_datetimenow_not_using_current_time/
# dt.datetime.now(dt.timezone.utc) cannot be used as default value, since it will already be
# evaluated at the start of the worker process and **not** when actually accessing the data
# def get_citation(db: Session, timeseries_id: int, datetime: dt.datetime = dt.datetime.now(dt.timezone.utc)):
def get_citation(db: Session, timeseries_id: int, datetime: dt.datetime = None):
    if not datetime:
        datetime = dt.datetime.now(dt.timezone.utc)
    db_object = db.query(models.Timeseries).filter(models.Timeseries.id == timeseries_id).first()
    # there is a mismatch with additional_metadata
    PI = "unknown"
    originators = False
    attribution = None
    if db_object:
        pi_role = get_value_from_str(toardb.toardb.RC_vocabulary,'PrincipalInvestigator')
        originator_role = get_value_from_str(toardb.toardb.RC_vocabulary,'Originator')
        contributor_role = get_value_from_str(toardb.toardb.RC_vocabulary,'Contributor')
        list_of_originators = []
        for db_role in db_object.roles:
            if (db_role.role == pi_role):
                db_contact = get_contact(db, contact_id = db_role.contact_id)
                PI = db_contact.name
            elif (db_role.role == originator_role):
                originators = True
                db_contact = get_contact(db, contact_id = db_role.contact_id)
                list_of_originators.append(db_contact.name)
            elif (db_role.role == contributor_role):
                db_contact = get_contact(db, contact_id = db_role.contact_id)
                list_of_originators.append(db_contact.name)
        list_of_data_originators = ", ".join(list_of_originators)
        # if no PI is given, the resource provider's longname should be named
        if PI == "unknown":
            role = get_value_from_str(toardb.toardb.RC_vocabulary,'ResourceProvider')
            for db_role in db_object.roles:
                if (db_role.role == role):
                    db_contact = get_contact(db, contact_id = db_role.contact_id)
                    PI = db_contact.longname
                    if db_contact.attribution != '':
                        attribution = db_contact.attribution
        var = get_variable(db, variable_id=db_object.variable_id).name
        station = get_stationmeta_by_id(db, station_id=db_object.station_id).name
        dataset_version = db_object.provider_version.strip()
        dataset_doi = db_object.doi.strip()
    citation = f"{PI}: time series of {var} at {station}, accessed from the TOAR database on {datetime}"
    if attribution and list_of_data_originators:
        attribution = attribution.format(orga_or_origs="data originators" if originators else "contributing organisations", list_of_data_originators=list_of_data_originators)
    if dataset_version != 'N/A':
        citation += f", original dataset version {dataset_version}"
    if dataset_doi!= '':
        citation += f", original dataset doi: {dataset_doi}"
    license_txt = "This data is published under a Creative Commons Attribution 4.0 International (CC BY 4.0). https://creativecommons.org/licenses/by/4.0/"
    return {"attribution": attribution, "citation": citation, "license": license_txt}

def adapt_db_object(db_object, db, fields=False, lconstr_roles=False):
    if fields:
        db_object = dict(zip((field for field in fields if field not in {"station_changelog", "changelog"}), db_object))

        # there is a mismatch with coordinates and additional_metadata
        if "coordinates" in db_object:
            db_object["coordinates"] = get_coordinates_from_string(db_object["coordinates"])

        if "additional_metadata" in db_object:
            db_object["additional_metadata"] = clean_additional_metadata(db_object["additional_metadata"])

        if "station_id" in db_object:
            station_id = {"id": db_object["station_id"]}
            db_object["station"] = station_id
            del db_object["station_id"]

        if "variable_id" in db_object:
            variable_id = {"id": db_object["variable_id"]}
            db_object["variable"] = variable_id
            del db_object["variable_id"]

        if "changelog" in db_object:
            db_object["changelog"] = get_timeseries_changelog(db, db_object["id"])

        if "station_changelog" in db_object:
            try:
                db_object["station_changelog"] = get_stationmeta_changelog(db, db_object["station_id"])
            except Exception:
                pass

        if lconstr_roles:
            # example, how to put the roles explicitly (not needed at the moment)
            # organisation = get_contact(db, contact_id=39)
            # roles_atts["contact"] = {"id": 39, "organisation": organisation.__dict__}
            roles_atts = {key: value for key, value in db_object.items() if key in roles_params}
            db_object = {key: value for key, value in db_object.items() if key not in roles_params}
            db_object["roles"] = TimeseriesRoleFields(**roles_atts)
    else:
        if isinstance(db_object.station.coordinates, (WKBElement, WKTElement)):
            db_object.station.coordinates = get_coordinates_from_geom(db_object.station.coordinates)
        # there is a mismatch with additional_metadata
        if isinstance(db_object.station.additional_metadata, dict):
            db_object.station.additional_metadata = json.dumps(db_object.station.additional_metadata)
        db_object.additional_metadata = clean_additional_metadata(db_object.additional_metadata)


        #Internall use
        try:
            del db_object.data_license_accepted
        except AttributeError:
            pass

        try:
            del db_object.dataset_approved_by_provider
        except AttributeError:
            pass

    return db_object

class TimeseriesQuery:
    def __init__(self, sign, query, fields, lconstr_roles):
        self.sign = sign
        self.query = query
        self.fields = fields
        self.lconstr_roles = lconstr_roles
    
    @staticmethod
    def aggregate(querys):
        aggregated_query = next(querys)
        for query in querys:
            if aggregated_query.fields != query.fields:
                raise ValueError("Fields of subquerys are diffrent")

            aggregated_query = TimeseriesQuery(
                True,
                aggregated_query.query.union(query.query)
                if query.sign
                else aggregated_query.query.except_(query.query),
                aggregated_query.fields,
                aggregated_query.lconstr_roles or query.lconstr_roles,
            )
        return aggregated_query

    @staticmethod
    def from_query_params(query_params, db, endpoint="timeseries", sign=True):
        limit, offset, fields, format, filters = create_filter(query_params, endpoint)
        t_filter = filters["t_filter"]
        t_r_filter = filters["t_r_filter"]
        s_c_filter = filters["s_c_filter"]
        s_g_filter = filters["s_g_filter"]

        if fields:
            # If only certain fields are selected the return type is not a orm object anymore but a dict
            # sort input fields to be sure to replace station_changelog before changelog
            fields = sorted(fields.split(","), reverse=True)
            if "role" in fields:
                fields.remove("role")
                fields.extend(roles_params)

            field_map = {
                "id": models.Timeseries.id,
                "order": models.Timeseries.order,
                "additional_metadata": models.Timeseries.additional_metadata,
                "station_id": StationmetaCore.id,
                "variable_id": Variable.id,
                "name": StationmetaCore.name,
                "coordinates": func.ST_AsText(StationmetaCore.coordinates),
                "station_country": StationmetaCore.country,
            }

            query_select = [field_map.get(field, text(field)) for field in fields]

        else:
            query_select = [models.Timeseries]

        query = (
            db.query(*query_select)
            .select_from(models.Timeseries)
            .distinct()
            .filter(text(t_filter), text(s_c_filter), text(s_g_filter), text(t_r_filter))
            .join(StationmetaCore)
            .join(StationmetaGlobal)
            .join(timeseries_timeseries_roles_table)
            .join(models.TimeseriesRole)
            .join(Contact)
            .join(Organisation)
            .join(Person)
            .join(Variable)
            .execution_options(stream_results=True)
            .order_by(models.Timeseries.id)
        )

        # Apply NOT filter with role logic
        if "NOT" in t_r_filter:
            role_ids = get_role_id_from_string(db, query_params.get("has_role")[1:])
            query = query.filter(
                ~models.Timeseries.id.in_(
                    select(timeseries_timeseries_roles_table.c.timeseries_id).where(
                        timeseries_timeseries_roles_table.c.role_id.in_(role_ids)
                    )
                )
            )

        if limit:
            query = query.limit(limit).offset(offset)

        return TimeseriesQuery(sign, query, fields, lconstr_roles=any(field in roles_params for field in fields))

    def adapt_objects(self, db):
        return [adapt_db_object(db_object, db, self.fields, self.lconstr_roles) for db_object in self.query]


def search_all(db, path_params, query_params, lts=False):
    endpoint = "timeseries" if lts else "search"

    try:
        ts_list = TimeseriesQuery.from_query_params(query_params, db, endpoint).adapt_objects(db)
        # remove duplicates
        if ts_list and isinstance(ts_list[0], dict):
            try:
                ts_set = set(json.dumps(ts, sort_keys=True) for ts in ts_list)
                return [json.loads(ts) for ts in ts_set]
            except: # not correct, because duplicates are NOT removed!!!
                return ts_list
        else:
            return ts_list
    except (KeyError, ValueError) as e:
        status_code = 400
        return JSONResponse(status_code=status_code, content=str(e))


def search_all_aggregation(db, path_params, signs, query_params_list, lts=False):
    endpoint = "timeseries" if lts else "search"

    try:
        ts_list = TimeseriesQuery.aggregate(
            TimeseriesQuery.from_query_params(query_params, db, endpoint, sign)
            for sign, query_params in zip(signs, query_params_list)
        ).adapt_objects(db)
        # remove duplicates
        if ts_list and isinstance(ts_list[0], dict):
            try:
                ts_set = set(json.dumps(ts, sort_keys=True) for ts in ts_list)
                return [json.loads(ts) for ts in ts_set]
            except: # not correct, because duplicates are NOT removed!!!
                return ts_list
        else:
            return ts_list
    except (KeyError, ValueError) as e:
        status_code = 400
        return JSONResponse(status_code=status_code, content=str(e))



def get_timeseries_by_unique_constraints(db: Session, station_id: int, variable_id: int, resource_provider: str = None,
                                                      sampling_frequency: str = None, provider_version: str = None, data_origin_type: str = None,
                                                      data_origin: str = None, sampling_height: float = None, label: str = None):
    """
    Criteria taken from TOAR_TG_Vol02_Data_Processing, 'Step 14: Identify Time Series'
    Criterion 14.1: id of the corresponding station
    Criterion 14.2: variable id
    Criterion 14.3: role: resource_provider (organisation)
    Criterion 14.4: sampling_frequency
    Criterion 14.5: version number
    Criterion 14.6: data_origin_type (measurement or model)
    Criterion 14.7: data origin
    Criterion 14.8: sampling height
    Criterion 14.9: data filtering procedures or other special dataset identifiers (use database field 'label')
    """

#   print("in get_timeseries_by_unique_constraints")
#   print(f"station_id: {station_id}, variable_id: {variable_id}, resource_provider: {resource_provider}, ", \
#         f"sampling_frequency: {sampling_frequency}, provider_version: {provider_version}, data_origin_type: {data_origin_type}, ", \
#         f"data_origin: {sampling_frequency}, sampling_height: {sampling_height}, label: {label}")

    # filter for criterion 14.1 and 14.2
    ret_db_object = db.query(models.Timeseries).filter(models.Timeseries.station_id == station_id) \
                                               .filter(models.Timeseries.variable_id == variable_id).all()
    # if already not found: return None
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None


    # filter for criterion 14.3
    if resource_provider:
        # issue with '/' in organisation longname ==> only possible with double encoding!
        resource_provider = resource_provider.replace('%2F', '/')
        role_num = get_value_from_str(toardb.toardb.RC_vocabulary,'ResourceProvider')
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            found = False
            for role in db_object.roles:
                # resource provider is always an organisation!
                organisation = get_contact(db, contact_id=role.contact_id)
                if ((role_num == role.role) and (organisation.longname == resource_provider)):
                    found = True
            if not found:
                ret_db_object.pop(counter)
            else:
                counter += 1
    else:
        # time series that do not have a resource_provider are not identical to those who do not!
        role_num = get_value_from_str(toardb.toardb.RC_vocabulary,'ResourceProvider')
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            found = False
            for role in db_object.roles:
                if (role_num == role.role):
                    found = True
            if found:
                ret_db_object.pop(counter)
            else:
                counter += 1


    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.4
    if sampling_frequency:
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            if not (db_object.sampling_frequency == sampling_frequency):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.5
    if provider_version:
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_objects:
            if not (db_object.provider_version == provider_version):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.6
    if data_origin_type:
        data_origin_type_num = get_value_from_str(toardb.toardb.OT_vocabulary,data_origin_type)
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            if not (db_object.data_origin_type == data_origin_type_num):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.7
    if data_origin:
        data_origin_num = get_value_from_str(toardb.toardb.DO_vocabulary,data_origin)
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            if not (db_object.data_origin == data_origin_num):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.8
    if sampling_height:
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            if not (db_object.sampling_height == sampling_height):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # if already only none object --> return
    # if only one single object is found, it has to be checked whether all criterions are fullfilled
    if len(ret_db_object) == 0:
        return None

    # filter for criterion 14.9
    if label:
        iter_obj = ret_db_object.copy()
        counter=0
        for db_object in iter_obj:
            if not (db_object.label == label):
                ret_db_object.pop(counter)
            else:
                counter += 1

    # check that only one object is left!!!
    # adapt mismatches for return value
    if len(ret_db_object) == 0:
        ret_db_object = None
    else:
        if len(ret_db_object) == 1:
            ret_db_object = ret_db_object[0]
            # there is a mismatch with additional_metadata
            ret_db_object.additional_metadata = clean_additional_metadata(ret_db_object.additional_metadata)
            # there is also a mismatch with coordinates and additional_metadata from station object
            if isinstance(ret_db_object.station.coordinates, (WKBElement, WKTElement)):
                ret_db_object.station.coordinates = get_coordinates_from_geom(ret_db_object.station.coordinates)
            # there is a mismatch with additional_metadata
            if isinstance(ret_db_object.station.additional_metadata, dict):
                ret_db_object.station.additional_metadata = json.dumps(ret_db_object.station.additional_metadata)
        else:
            status_code=405
            message=f"Timeseries not unique, more criteria need to be defined."
            return JSONResponse(status_code=status_code, content=message)

    return ret_db_object


def get_timeseries_changelog(db: Session, timeseries_id: int):
    return db.query(models.TimeseriesChangelog).filter(models.TimeseriesChangelog.timeseries_id == timeseries_id).all()


def get_timeseries_programme(db: Session, name: str):
    return db.query(models.TimeseriesProgramme).filter(models.TimeseriesProgramme.name == name).all()


# is this internal, or should this also go to public REST api?
# do we need this at all?
def get_role_ids_of_timeseries(db: Session, timeseries_id: int):
    db_objects = db.query(models.TimeseriesTimeseriesRoles) \
                                      .filter(models.TimeseriesTimeseriesRoles.timeseries_id == timeseries_id) \
                                      .all()
    return db_objects


# is this internal, or should this also go to public REST api?
def get_unique_timeseries_role(db: Session, role: int, contact_id: int, status: int):
    db_object = db.query(models.TimeseriesRole).filter(models.TimeseriesRole.role == role) \
                                      .filter(models.TimeseriesRole.contact_id == contact_id) \
                                      .filter(models.TimeseriesRole.status == status) \
                                      .first()
    return db_object


def get_role_id_from_string(db: Session, role_string: str):
    sql_command = f"""
                   SELECT distinct(r.id) FROM timeseries_timeseries_roles tr,
                                    timeseries_roles r,
                                    contacts c,
                                    organisations o,
                                    persons p
                   WHERE (o.longname IN ('{role_string}') OR
                          o.name IN ('{role_string}') OR
                          o.city IN ('{role_string}') OR
                          o.homepage IN ('{role_string}') OR
                          o.contact_url IN ('{role_string}') OR
                          p.email IN ('{role_string}') OR
                          p.name IN ('{role_string}') OR
                          p.orcid IN ('{role_string}')) AND
                          tr.role_id=r.id AND
                          r.contact_id=c.id AND
                          c.person_id=p.id AND
                          c.organisation_id=o.id
                   """
    resultproxy = db.execute(sql_command)
    id_list = [ rowproxy._asdict()['id'] for rowproxy in resultproxy ]
    return id_list

# is this internal, or should this also go to public REST api?
def get_unique_timeseries_programme(db: Session, name: str, homepage: str):
    db_object = db.query(models.TimeseriesProgramme).filter(models.TimeseriesProgramme.name == name) \
                                      .filter(models.TimeseriesProgramme.homepage == homepage) \
                                      .first()
    return db_object


# is this internal, or should this also go to public REST api?
def get_unique_timeseries_annotation(db: Session, text: str, contributor_id: int):
    db_object = db.query(models.TimeseriesAnnotation).filter(models.TimeseriesAnnotation.text == text) \
                                      .filter(models.TimeseriesAnnotation.contributor_id == contributor_id) \
                                      .first()
    return db_object


def get_contributors_string(programmes, roles):
    # sort every section alphabetically and have sub-section titles
    # programmes are already unique, but they still need to be sorted
    resultp = "programmes: " + ";".join(sorted([programme.longname for programme in programmes])) if len(programmes) > 0 else ''
    # organisations might contain duplicates
    organisations = set()
    [ organisations.add(role.contact.organisation.longname) for role in roles ]
    # eliminate dummy organisation
    organisations.discard('')
    resulto = "organisations: " + ";".join(sorted(organisations)) if len(organisations) > 0 else ''
    # persons might contain duplicates
    persons = set()
    [ persons.add(role.contact.person.name) for role in roles ]
    # eliminate dummy person
    persons.discard('')
    resultc = "persons:" + ";".join(sorted(persons)) if len(persons) > 0 else ''
    result = resultp if resultp != '' else ''
    if resulto != '':
        result = resulto if result == '' else result + '; ' + resulto
    if resultc != '':
        result = resultc if result == '' else result + '; ' + resultc
    return result


def get_contributors_list(db: Session, timeseries_ids, format: str = 'text'):
    # get time series' programmes
    # join(models.Timeseries, Timeseries.programme_id == TimeseriesProgramme.id) is implicit given due to the foreign key
    programmes = db.query(models.TimeseriesProgramme) \
                   .join(models.Timeseries) \
                   .filter(models.Timeseries.id.in_(timeseries_ids), models.TimeseriesProgramme.id != 0).all()
    # get all time series roles (with duplicates)
    roles = db.execute(select([timeseries_timeseries_roles_table]).where(timeseries_timeseries_roles_table.c.timeseries_id.in_(timeseries_ids)))
    # eliminate duplicates
    role_ids = set()
    [ role_ids.add(role.role_id) for role in roles ]
    roles = db.query(models.TimeseriesRole).filter(models.TimeseriesRole.id.in_(role_ids)).all()
    # return both programmes and roles
    if format == 'text':
        result = get_contributors_string(programmes, roles)
    elif format == 'json':
        result = programmes + roles
    else:
        status_code=400
        message=f"not a valid format: {format}"
        result = JSONResponse(status_code=status_code, content=message)
    return result


def get_request_contributors(db: Session, format: str = 'text', input_handle: UploadFile = File(...)):
    f = input_handle.file
    timeseries_ids = [int(line.strip()) for line in f.readlines()]
    return get_contributors_list(db, timeseries_ids, format)


def get_registered_request_contributors(db: Session, rid, format: str = 'text'):
    try:
        timeseries_ids = db.execute(select([s1_contributors_table]).\
                                    where(s1_contributors_table.c.request_id == rid)).mappings().first()['timeseries_ids']
        return get_contributors_list(db, timeseries_ids, format)
    except:
        status_code=400
        message=f"not a registered request id: {rid}"
        return JSONResponse(status_code=status_code, content=message)


def register_request_contributors(db: Session, rid, ids):
    try:
        db.execute(insert(s1_contributors_table).values(request_id=rid, timeseries_ids=ids))
        db.commit()
        status_code = 200
        message=f'{rid} successfully registered.'
    except IntegrityError as e:
        error_code = e.orig.pgcode
        if error_code == "23505":
            status_code = 443
            message=f'{rid} already registered.'
        else:
            status_code = 442
            message=f'database error: error_code'
    result = JSONResponse(status_code=status_code, content=message)
    return result


# is this internal, or should this also go to public REST api?
def get_timeseries_roles(db: Session, timeseries_id: int):
    return db.execute(select([timeseries_timeseries_roles_table]).where(timeseries_timeseries_roles_table.c.timeseries_id == timeseries_id))


# is this internal, or should this also go to public REST api?
def get_timeseries_role_by_id(db: Session, role_id):
    return db.query(models.TimeseriesRole).filter(models.TimeseriesRole.id == role_id).first()


# is this internal, or should this also go to public REST api?
def get_timeseries_annotations(db: Session, timeseries_id: int):
    return db.execute(select([timeseries_timeseries_annotations_table]).where(timeseries_timeseries_annotations_table.c.timeseries_id == timeseries_id))


# is this internal, or should this also go to public REST api?
def get_timeseries_annotation_by_id(db: Session, annotation_id):
    return db.query(models.TimeseriesAnnotation).filter(models.TimeseriesAnnotation.id == annotation_id).first()


def create_timeseries(db: Session, timeseries: TimeseriesCreate, author_id: int):
    timeseries_dict = timeseries.dict()
    # no timeseries can be created, if station_id or variable_id are not found in the database
    if not station_id_exists(db,timeseries.station_id):
        status_code=440
        message=f"Station (station_id: {timeseries.station_id}) not found in database."
        return JSONResponse(status_code=status_code, content=message)
    if timeseries_dict['additional_metadata']:
        for key, value in timeseries_dict['additional_metadata'].items():
            if isinstance(value,dict):
                for key2, value2 in value.items():
                    timeseries_dict['additional_metadata'][key][key2] = value2.replace("''","'")
            else:
                timeseries_dict['additional_metadata'][key] = value.replace("''","'")
        if 'absorption_cross_section' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.CS_vocabulary,timeseries_dict['additional_metadata']['absorption_cross_section'])
            timeseries_dict['additional_metadata']['absorption_cross_section'] = value
        if 'calibration_type' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.CT_vocabulary,timeseries_dict['additional_metadata']['calibration_type'])
            timeseries_dict['additional_metadata']['calibration_type'] = value
        if 'sampling_type' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.KS_vocabulary,timeseries_dict['additional_metadata']['sampling_type'])
            timeseries_dict['additional_metadata']['sampling_type'] = value
    db_variable = get_variable(db,timeseries.variable_id)
    if not db_variable:
        status_code=441
        message=f"Variable (variable_id: {timeseries.variable_id}) not found in database."
        return JSONResponse(status_code=status_code, content=message)
    # for networks: we do not want data without a resource_provider (organisation)
    resource_provider = None
    if timeseries.roles:
        for role in timeseries.roles:
            if role.role == 'ResourceProvider':
                # resource provider is always an organisation!
                organisation = get_contact(db, contact_id=role.contact_id)
                if organisation:
                    resource_provider=organisation.longname
                else:
                    status_code=442
                    message=f"Resource provider (contact_id: {role.contact_id}) not found in database."
                    return JSONResponse(status_code=status_code, content=message)
    db_timeseries = get_timeseries_by_unique_constraints(db, station_id=timeseries.station_id,
                         variable_id=timeseries.variable_id, label=timeseries.label, resource_provider=resource_provider)
    if db_timeseries:
        if isinstance(db_timeseries, list):
            status_code=444
            message = {"detail":{"message":"Given constraints match more than one timeseries.",
                                 "timeseries_ids": [ db_timeseries[i].id for i in range(len(db_timeseries)) ]}}
            return JSONResponse(status_code=status_code, content=message)
        else:
            status_code=443
            message = {"detail":{"message":"Timeseries already registered.","timeseries_id":db_timeseries.id}}
            return JSONResponse(status_code=status_code, content=message)
    roles_data = timeseries_dict.pop('roles', None)
    annotations_data = timeseries_dict.pop('annotations', None)
    db_timeseries = models.Timeseries(**timeseries_dict)
    # there is a mismatch with additional_metadata
    # in upload command, we have now: "additional_metadata": "{}"
    # but return from this method gives (=database): "additional_metadata": {}
#   print(db_timeseries.additional_metadata)
#   db_timeseries.additional_metadata = json.loads(str(db_timeseries.additional_metadata).replace("'",'"'))
    db_timeseries.sampling_frequency = get_value_from_str(toardb.toardb.SF_vocabulary,db_timeseries.sampling_frequency)
    db_timeseries.aggregation = get_value_from_str(toardb.toardb.AT_vocabulary,db_timeseries.aggregation)
    db_timeseries.data_origin_type = get_value_from_str(toardb.toardb.OT_vocabulary,db_timeseries.data_origin_type)
    db_timeseries.data_origin = get_value_from_str(toardb.toardb.DO_vocabulary,db_timeseries.data_origin)
    db.add(db_timeseries)
    result = db.commit()
    db.refresh(db_timeseries)
    # get timeseries_id
    timeseries_id = db_timeseries.id
    # store roles and update association table
    if roles_data:
        for r in roles_data:
            db_role = models.TimeseriesRole(**r)
            db_role.role = get_value_from_str(toardb.toardb.RC_vocabulary,db_role.role)
            db_role.status = get_value_from_str(toardb.toardb.RS_vocabulary,db_role.status)
            db_object = get_unique_timeseries_role(db, db_role.role, db_role.contact_id, db_role.status)
            if db_object:
                role_id = db_object.id
            else:
                # Something is going wrong here!
                # Is the model TimeseriesRole correctly defined?!
                del db_role.contact
                db.add(db_role)
                db.commit()
                db.refresh(db_role)
                role_id = db_role.id
            db.execute(insert(timeseries_timeseries_roles_table).values(timeseries_id=timeseries_id, role_id=role_id))
            db.commit()
    # store annotations and update association table
    if annotations_data:
        for a in annotations_data:
            db_annotation = models.TimeseriesAnnotation(**a)
            # check whether annotation is already present in database
            db_object = get_unique_timeseries_annotation(db, db_annotation.text, db_annotation.contributor_id)
            if db_object:
                annotation_id = db_object.id
            else:
                db.add(db_annotation)
                db.commit()
                db.refresh(db_annotation)
                annotation_id = db_annotation.id
            db.execute(insert(timeseries_timeseries_annotations_table).values(timeseries_id=timeseries_id, annotation_id=annotation_id))
            db.commit()
    # create changelog entry
    type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Created")
    description="time series created"
    db_changelog = TimeseriesChangelog(description=description, timeseries_id=timeseries_id, author_id=author_id, type_of_change=type_of_change,
                                        old_value='', new_value='')
    db.add(db_changelog)
    db.commit()
    status_code=200
    message = {"detail":{"message":"New timeseries created.","timeseries_id":db_timeseries.id}}
    return JSONResponse(status_code=status_code, content=message)
#   return db_timeseries


def patch_timeseries(db: Session, description: str, timeseries_id: int, timeseries: TimeseriesPatch,
        author_id: int):
    timeseries_dict = timeseries.dict()
    # check for controlled vocabulary in additional metadata
    # do this in dictionary that always contains additional_metadata entry
    if timeseries_dict['additional_metadata']:
        for key, value in timeseries_dict['additional_metadata'].items():
            if isinstance(value,dict):
                for key2, value2 in value.items():
                    timeseries_dict['additional_metadata'][key][key2] = value2.replace("''","'")
            else:
                timeseries_dict['additional_metadata'][key] = value.replace("''","'")
        if 'absorption_cross_section' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.CS_vocabulary,timeseries_dict['additional_metadata']['absorption_cross_section'])
            timeseries_dict['additional_metadata']['absorption_cross_section'] = value
        if 'calibration_type' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.CT_vocabulary,timeseries_dict['additional_metadata']['calibration_type'])
            timeseries_dict['additional_metadata']['calibration_type'] = value
        if 'sampling_type' in timeseries_dict['additional_metadata']:
            value = get_value_from_str(toardb.toardb.KS_vocabulary,timeseries_dict['additional_metadata']['sampling_type'])
            timeseries_dict['additional_metadata']['sampling_type'] = value
    # delete empty fields from timeseries_dict already at this place, to be able to
    # distinguish between "single value correction in metadata" and "comprehensive metadata revision"
    # (see controlled vocabulary "CL_vocabulary")
    roles_data = timeseries_dict.pop('roles', None)
    annotations_data = timeseries_dict.pop('annotations', None)
    timeseries_dict2 = {k: v for k, v in timeseries_dict.items() if v is not None}
    number_of_elements = len(timeseries_dict2)
    if roles_data:
        number_of_elements +=1
    if annotations_data:
        number_of_elements +=1
    if (number_of_elements == 1):
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"SingleValue")
    else:
        type_of_change = get_value_from_str(toardb.toardb.CL_vocabulary,"Comprehensive")
    db_obj = models.Timeseries(**timeseries_dict2)
#   also the sqlalchemy get will call get_timeseries here!!!
#   --> therefore call it right away
    db_timeseries = get_timeseries(db, timeseries_id, fields="dataset_approved_by_provider,data_license_accepted")
    dataset_approved_by_provider = db_timeseries.dataset_approved_by_provider
    data_license_accepted = db_timeseries.data_license_accepted
    db_timeseries = get_timeseries(db, timeseries_id)
    # for some unknown reasons, format of coordinates and additional_metadata (for timeseries, but also for its related station!) have changed!
    db.rollback()
    db_timeseries.dataset_approved_by_provider = dataset_approved_by_provider
    db_timeseries.data_license_accepted = data_license_accepted
    # still problems with coordinates and additional metadata (from STATION!!!)...
    try:
        db_timeseries.station.coordinates = get_geom_from_coordinates(db_timeseries.station.coordinates)
    except:
        pass
#   try:
#       db_timeseries.station.additional_metadata = json.loads(str(db_timeseries.station.additional_metadata).replace("'",'"'))
#   except:
#       pass
    # prepare changelog entry/entries
    no_log = (description == 'NOLOG')
    if not no_log:
        old_values={}
        new_values={}
        for k, v in timeseries_dict2.items():
            field=str(getattr(db_timeseries,k))
            if k == 'additional_metadata':
                old_values[k] = db_timeseries.additional_metadata
            else:
                if k == "sampling_frequency":
                    old_values[k] = get_str_from_value(toardb.toardb.SF_vocabulary, int(field))
                elif k == "aggregation":
                    old_values[k] = get_str_from_value(toardb.toardb.AT_vocabulary, int(field))
                elif k == "data_origin":
                    old_values[k] = get_str_from_value(toardb.toardb.DO_vocabulary, int(field))
                elif k == "data_origin_type":
                    old_values[k] = get_str_from_value(toardb.toardb.OT_vocabulary, int(field))
                else:
                    old_values[k] = field
    for k, v in timeseries_dict2.items():
        setattr(db_timeseries,k,timeseries_dict[k])
    # there is a mismatch with additional_metadata
    # in upload command, we have now: "additional_metadata": "{}"
    # but return from this method gives (=database): "additional_metadata": {}
#   if timeseries_dict['additional_metadata']:
#       db_timeseries.additional_metadata = clean_additional_metadata(db_timeseries.additional_metadata)
    # do the following for every entry that uses the controlled vocabulary!
    # this should be improved!
    if db_obj.sampling_frequency:
        db_timeseries.sampling_frequency = get_value_from_str(toardb.toardb.SF_vocabulary, db_obj.sampling_frequency)
    if db_obj.aggregation:
        db_timeseries.aggregation = get_value_from_str(toardb.toardb.AT_vocabulary, db_obj.aggregation)
    if db_obj.data_origin:
        db_timeseries.data_origin = get_value_from_str(toardb.toardb.DO_vocabulary, db_obj.data_origin)
    if db_obj.data_origin_type:
        db_timeseries.data_origin_type = get_value_from_str(toardb.toardb.OT_vocabulary, db_obj.data_origin_type)
    db.add(db_timeseries)
    result = db.commit()
    # store roles and update association table
    if roles_data:
        if not no_log:
            # prepare changelog entry/entries
            description = description + f"; add role"
            db_old_roles = get_timeseries_roles(db, timeseries_id)
            old_roles = []
            for oldr in db_old_roles:
                old_role = get_timeseries_role_by_id(db, oldr.role_id)
                old_value = {}
                old_value['role']  =  get_str_from_value(toardb.toardb.RC_vocabulary,old_role.role)
                old_value['status'] =  get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status)
                old_value['contact_id'] =  old_role.contact_id
                old_roles.append(old_value)
            old_values['roles'] = old_roles
            new_roles = old_roles.copy()
        for r in roles_data:
            db_role = models.TimeseriesRole(**r)
            if not no_log:
                new_roles.append(r)
            db_role.role = get_value_from_str(toardb.toardb.RC_vocabulary,db_role.role)
            db_role.status = get_value_from_str(toardb.toardb.RS_vocabulary,db_role.status)
            # check whether role is already present in database
            db_object = get_unique_timeseries_role(db, db_role.role, db_role.contact_id, db_role.status)
            if db_object:
                role_id = db_object.id
            else:
                db.add(db_role)
                db.commit()
                db.refresh(db_role)
                role_id = db_role.id
            db.execute(insert(timeseries_timeseries_roles_table).values(timeseries_id=timeseries_id, role_id=role_id))
            db.commit()
        if not no_log:
            new_values['roles'] = new_roles
    # store annotations and update association table
    if annotations_data:
        if not no_log:
            # prepare changelog entry/entries
            description = description + f"; add annotation"
            db_old_annotations = get_timeseries_annotations(db, timeseries_id)
            old_annotations = []
            for olda in db_old_annotations:
                old_annotation = get_timeseries_annotation_by_id(db, olda.annotation_id)
                old_value = {}
                old_value['kind'] = get_str_from_value(toardb.toardb.RC_vocabulary,old_role.kind)
                old_value['text'] = get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status)
                old_value['date_added'] = get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status)
                old_value['approved'] = get_str_from_value(toardb.toardb.RS_vocabulary,old_role.status)
                old_value['contributor_id'] = str(old_role.contact_id)
                old_annotations.append(old_value)
            old_values['annotations'] = old_annotations
            new_annotations = []
        for a in annotations_data:
            db_annotation = models.TimeseriesAnnotation(**a)
            # check whether annotation is already present in database
            if not no_log:
                new_annotations.append(a)
            db_annotation.kind = get_value_from_str(toardb.toardb.AK_vocabulary,db_annotation.kind)
            db_object = get_unique_timeseries_annotation(db, db_annotation.text, db_annotation.contributor_id)
            if db_object:
                annotation_id = db_object.id
            else:
                db.add(db_annotation)
                db.commit()
                db.refresh(db_annotation)
                annotation_id = db_annotation.id
            db.execute(insert(timeseries_timeseries_annotations_table).values(timeseries_id=timeseries_id, annotation_id=annotation_id))
            db.commit()
        if not no_log:
            new_values['annotations'] = new_annotations
    # add patch to changelog table
    if not no_log:
        if new_values:
            timeseries_dict2.update(new_values)
        db_changelog = TimeseriesChangelog(description=description, timeseries_id=timeseries_id, author_id=author_id, type_of_change=type_of_change,
                                           old_value=str(old_values), new_value=str(timeseries_dict2))
        db.add(db_changelog)
        db.commit()
    status_code=200
    message = {"detail":{"message":"timeseries patched.","timeseries_id":db_timeseries.id}}
    return JSONResponse(status_code=status_code, content=message)
