# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
class Timeseries (Base)
=======================
"""
from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, text, String, \
                       Text, CheckConstraint, UniqueConstraint, PrimaryKeyConstraint, \
                       Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import JSONB
from toardb.stationmeta.models import StationmetaCore
from toardb.variables.models import Variable
from .models_programme import TimeseriesProgramme
from toardb.base import Base


TIMESERIES_ID_SEQ = Sequence('timeseries_id_seq')  # define sequence explicitly
class Timeseries(Base):
    """
    Table "public.timeseries"

    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    |       Column                 |           Type           | Collation | Nullable |                Default                 |
    +==============================+==========================+===========+==========+========================================+
    | id                           | integer                  |           | not null | nextval('timeseries_id_seq'::regclass) |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | label                        | character varying(128)   |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | order                        | integer                  |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | sampling_frequency           | integer                  |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | aggregation                  | integer                  |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_start_date              | timestamp with time zone |           | not null | '2050-01-01 00:00:00'                  |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_end_date                | timestamp with time zone |           | not null | '1900-01-01 00:00:00'                  |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_origin_type             | integer                  |           | not null | 0                                      |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_origin                  | integer                  |           | not null | 0                                      |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | sampling_height              | double precision         |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | additional_metadata          | jsonb                    |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | provider_version             | character varying(28)    |           | not null | 'NA'                                   |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | latest_version               | character varying(28)    |           | not null | '000001.000000.00000000000000'         |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | data_license_accepted        | timestamp with time zone |           | not null | '1900-01-01'                           |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | dataset_approved_by_provider | timestamp with time zone |           | not null | '1900-01-01'                           |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | station_id                   | integer                  |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | variable_id                  | integer                  |           | not null |                                        |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | programme_id                 | integer                  |           | not null | 0                                      |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | doi                          | character varying(64)    |           | not null | ''                                     |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+
    | coverage                     | double precision         |           | not null | -1.0                                   |
    +------------------------------+--------------------------+-----------+----------+----------------------------------------+

    Indexes:
        "timeseries_pkey" PRIMARY KEY, btree (id)
        "timeseries_station_id" btree (station_id)
        "timeseries_variable_id" btree (variable_id)
    Check constraints:
        "timeseries_aggregation_check" CHECK (aggregation >= 0)
        "timeseries_order_check" CHECK ("order" >= 0)
        "timeseries_sampling_frequency_check" CHECK (sampling_frequency >= 0)
    Foreign-key constraints:
        "timeseries_station_id_fk_stationmeta_core_id" FOREIGN KEY (station_id) REFERENCES stationmeta_core(id) DEFERRABLE INITIALLY DEFERRED
        "timeseries_variable_id_fk_variables_id" FOREIGN KEY (variable_id) REFERENCES variables(id) DEFERRABLE INITIALLY DEFERREDForeign-key constraints:
        "timeseries_aggregation_fk_at_vocabulary_enum_val" FOREIGN KEY (aggregation) REFERENCES at_vocabulary(enum_val)
        "timeseries_sampling_frequency_fk_sf_vocabulary_enum_val" FOREIGN KEY (sampling_frequency) REFERENCES sf_vocabulary(enum_val)
        "timeseries_data_origin_fk_do_vocabulary_enum_val" FOREIGN KEY (data_origin) REFERENCES do_vocabulary(enum_val)
    "timeseries_data_origin_type_fk_ot_vocabulary_enum_val" FOREIGN KEY (data_origin_type) REFERENCES ot_vocabulary(enum_val)
        "timeseries_programme_id_fk_timeseries_programmes_id" FOREIGN KEY (programme_id) REFERENCES timeseries_programmes(id)
    Referenced by:
        TABLE "data" CONSTRAINT "data_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_annotations" CONSTRAINT "timeseries_timeseries_annotations_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_roles" CONSTRAINT "timeseries_timeseries_roles_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
        TABLE "timeseries_timeseries_programmes" CONSTRAINT "timeseries_timeseries_programmes_timeseries_id_fk_timeseries_id" FOREIGN KEY (timeseries_id) REFERENCES timeseries(id) DEFERRABLE INITIALLY DEFERRED
    """

    __tablename__ = 'timeseries'
    __table_args__ = (
        CheckConstraint('"order" >= 0'),
        CheckConstraint('aggregation >= 0'),
        CheckConstraint('sampling_frequency >= 0'),
        UniqueConstraint('station_id', 'variable_id', 'label'),
    )

    id = Column(Integer, TIMESERIES_ID_SEQ, primary_key=True, server_default=TIMESERIES_ID_SEQ.next_value())
    label = Column(String(128), nullable=False)
    order = Column(Integer, nullable=False)
    sampling_frequency = Column(ForeignKey('sf_vocabulary.enum_val'), nullable=False)
    aggregation = Column(ForeignKey('at_vocabulary.enum_val'), nullable=False)
    data_start_date = Column(DateTime(True), nullable=False)
    data_end_date = Column(DateTime(True), nullable=False)
    data_origin_type = Column(ForeignKey('ot_vocabulary.enum_val'), nullable=False, server_default=text("0"))
    data_origin = Column(ForeignKey('do_vocabulary.enum_val'), nullable=False, server_default=text("0"))
    sampling_height = Column(Float(53), nullable=False)
    provider_version = Column(String(28), nullable=False, server_default=text("'NA'"))
    latest_version = Column(String(28), nullable=False, server_default=text("'000001.000000.00000000000000'"))
    data_license_accepted = Column(DateTime(True), nullable=False, server_default=text("'1900-01-01 00:00:00+00'::timestamp with time zone"))
    dataset_approved_by_provider = Column(DateTime(True), nullable=False, server_default=text("'1900-01-01 00:00:00+00'::timestamp with time zone"))
# do not use string declaration here (not working for pytest)
# use the explicit class name here,
# see: https://groups.google.com/forum/#!topic/sqlalchemy/YjGhE4d6K4U
    station_id = Column(ForeignKey(StationmetaCore.id, deferrable=True, initially='DEFERRED'), nullable=False, index=True)
    variable_id = Column(ForeignKey(Variable.id, deferrable=True, initially='DEFERRED'), nullable=False, index=True)
    programme_id = Column(ForeignKey(TimeseriesProgramme.id, deferrable=True, initially='DEFERRED'), nullable=False, server_default=text("0"))
    doi = Column(String(64), nullable=False, server_default=text("'000001.000000.00000000000000'"))
    coverage = Column(Float, nullable=False, server_default=text('-1.0'))

    # for the nested view
    # problems with station and coordinates in nested views!!!
    station = relationship('StationmetaCore')
    variable = relationship('Variable')
    programme = relationship('TimeseriesProgramme')
    changelog = relationship('TimeseriesChangelog', back_populates='timeseries')

    additional_metadata = Column(JSONB(astext_type=Text()), nullable=True)

