# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Pydantic schemas for TOAR database

"""

from typing import List, Any, Union

from pydantic import BaseModel, Json, validator, Field
import datetime as dt
from toardb.contacts.schemas import Contact
from toardb.variables.schemas import Variable 
from toardb.stationmeta.schemas import Stationmeta, Coordinates, get_full_description_from_abbreviation
from toardb.utils.utils import get_displaystr_from_value
import toardb

# ======== Timeseries =========

class TimeseriesCoreBase(BaseModel):
    id: int = None
    label: str = Field(None, description="a short string to distinguish this timeseries from others with the same combination of station and variable")
    order: int = Field(None, description="indicates position of this timeseries in a list when several timeseries share the same station and variable combination")
    sampling_frequency: str = Field(None, description="Sampling frequency of data in this timeseries (see controlled vocabulary: Sampling Frequency)")
    aggregation: str = Field(None, description="Aggregation type in this timeseries (see controlled vocabulary: Aggregation Type)")
    data_start_date: dt.datetime = Field(None, description="Start date of the variable data available for this station")
    data_end_date: dt.datetime = Field(None, description="End date of the variable data available for this station")
    data_origin: str = Field(None, description="origin of data (model name or instrument) (see controlled vocabulary: Data Origin)")
    data_origin_type: str = Field(None, description="type of data origin (see controlled vocabulary: Data Origin Type)")
    provider_version: str = Field(None, description="provider data version")
    sampling_height: float = Field(None, description="Height above the ground of the inlet/instrument/sampler (in m)")
    additional_metadata: Json = Field(None, description="Additional information about the timeseries as JSON structure.")
    #only for internal use
    data_license_accepted: dt.datetime = Field(None, description="date when provider accepted the data license agreement")
    dataset_approved_by_provider: dt.datetime = Field(None, description="date when provider approved the data")
    doi : str = Field(None, description="DOI of the data")
    coverage: float = Field(None, description="data coverage between the time series data_start_date and data_end_date")

# still missing: "Score values from automated data QA (5-star evaluation)"
# still missing: "detailed report of timeseries QA/QC evaluation"
# still missing: "arbitrary additional information about this timeseries"

    @validator('sampling_frequency')
    def check_sampling_frequency(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.SF_vocabulary))[0].display_str

    @validator('aggregation')
    def check_aggregation(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AT_vocabulary))[0].display_str

    @validator('data_origin_type')
    def check_data_origin_type(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.OT_vocabulary))[0].display_str

    @validator('data_origin')
    def check_data_origin(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.DO_vocabulary))[0].display_str

    @validator('additional_metadata')
    def check_additional_metadata(cls, v):
        if v:
            for key, value in v.items():
                if key == 'sampling_type':
                    v[key] = get_displaystr_from_value(toardb.toardb.KS_vocabulary,int(value))
                if key == 'absorption_cross_section':
                    v[key] = get_displaystr_from_value(toardb.toardb.CS_vocabulary,int(value))
                if key == 'calibration_type':
                    v[key] = get_displaystr_from_value(toardb.toardb.CT_vocabulary,int(value))
        return v


class TimeseriesCoreCreate(TimeseriesCoreBase):
    station_id: int
    variable_id: int
    programme_id: int = None

    @validator('sampling_frequency')
    def check_sampling_frequency(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.SF_vocabulary)):
            return v
        else:
            raise ValueError(f"sampling frequency not known: {v}")

    @validator('aggregation')
    def check_aggregation(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.AT_vocabulary)):
            return v
        else:
            raise ValueError(f"aggregation type not known: {v}")

    @validator('data_origin_type')
    def check_data_origin_type(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.OT_vocabulary)):
            return v
        else:
            raise ValueError(f"data origin type not known: {v}")

    @validator('data_origin')
    def check_data_origin(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.DO_vocabulary)):
            return v
        else:
            raise ValueError(f"data origin not known: {v}")

    # overwrite validator in creation process
    @validator('additional_metadata')
    def check_additional_metadata(cls, v):
        return v


class TimeseriesCore(TimeseriesCoreBase):
    id: int

    class Config:
        orm_mode = True

# ======== TimeseriesRole =========

class TimeseriesRoleBase(BaseModel):
    id: int = None
    role: str = Field(None, description="Role of contact (see controlled vocabulary: Role Code)")
    status: str = Field(None, description="Status of contact (see controlled vocabulary: Role Status)")
    contact: Contact = Field(None, description="Contact for this role")

    @validator('role')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.RC_vocabulary))[0].display_str

    @validator('status')
    def check_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.RS_vocabulary))[0].display_str

    @validator('contact')
    def check_contact_is_private(cls, v):
        if v.person:
            if (v.person.id != 0):
                if (v.person.isprivate):
                    v.person = None
        return v

    class Config:
        orm_mode = True


class TimeseriesRoleBaseFields(TimeseriesRoleBase):
    contact_id: int = None

    class Config:
        orm_mode = True


class TimeseriesRoleCreate(TimeseriesRoleBase):
    contact: Contact = None
    contact_id: int

    @validator('role')
    def check_role(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.RC_vocabulary)):
            return v
        else:
            raise ValueError(f"role code not known: {v}")

    @validator('status')
    def check_status(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.RS_vocabulary)):
            return v
        else:
            raise ValueError(f"role status not known: {v}")

    class Config:
        orm_mode = True

# do not create a role while creating a timeseries!
class TimeseriesRoleNoCreate(TimeseriesRoleCreate):
    contact_id: int
#   contact: Contact = None

    class Config:
        orm_mode = True


class TimeseriesRole(TimeseriesRoleBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

class TimeseriesRoleFields(TimeseriesRoleCreate):
    pass

    @validator('role')
    def check_role(cls, v):
        return v

    @validator('status')
    def check_status(cls, v):
        return v


# ======== TimeseriesAnnotation =========

class TimeseriesAnnotationBase(BaseModel):
    id: int = None
    kind: str = Field(..., description="kind of annotation (see controlled vocabulary: Kind Of Annotation)")
    text: str = Field(..., description="text of annotation")
    date_added: dt.datetime = Field(..., description="timestamp when annotation was added")
    approved: bool = Field(..., description="Flag indicating whether the annotation of a time-series has been verified")
    contributor_id: int = Field(..., description="ID of contributor who added the annotation")

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AK_vocabulary))[0].display_str


class TimeseriesAnnotationPatch(BaseModel):
    kind: int = None
    text: str = None
    date_added: dt.datetime = None
    approved: bool = None
    contributor_id: int = None

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AK_vocabulary))[0].display_str


class TimeseriesAnnotationCreate(TimeseriesAnnotationBase):

    @validator('kind')
    def check_kind(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.AK_vocabulary)):
            return v
        else:
            raise ValueError(f"kind of annotation code not known: {v}")



class TimeseriesAnnotation(TimeseriesAnnotationBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True


# ======== TimeseriesProgramme =========

class TimeseriesProgrammeBase(BaseModel):
    id: int = None
    name: str = Field(..., description="(Short) name of programme")
    longname: str = Field(..., description="(Long) name of programme")
    homepage: str = Field(..., description="Homepage (URL) of programme")
    description: str = Field(..., description="Description of programme")


class TimeseriesProgrammeCreate(TimeseriesProgrammeBase):
    pass


class TimeseriesProgramme(TimeseriesProgrammeBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True


# ======== TimeseriesChangelog =========

class TimeseriesChangelogBase(BaseModel):
    datetime: dt.datetime = Field(..., description="Date of change to the TOAR database")
    description: str = Field(..., description="Description of change")
    old_value: str = Field(..., description="value that has been changed in the TOAR database")
    new_value: str = Field(..., description="new, changed value")
    timeseries_id: int = Field(..., description="internal ID of time-series to which this change belongs")
    author_id: int = Field(..., description="internal ID of person who submitted the change to the TOAR database")
    type_of_change: int = Field(..., description="Type of change (see controlled vocabulary: Type Of Change)")
    period_start: dt.datetime = Field(None, description="for changes on data: start date from which changes were applied")
    period_end: dt.datetime = Field(None, description="for changes on data: end date to which changes were applied")
    version: str = Field(None, description="version that results from this change")


    @validator('type_of_change')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CL_vocabulary))[0].display_str


class TimeseriesChangelog(TimeseriesChangelogBase):

    class Config:
        orm_mode = True



# ======== for nested view/upload =========

class TimeseriesBase(TimeseriesCoreBase):
    station: Stationmeta = None
    variable: Variable = None
    programme: TimeseriesProgramme = None
    additional_metadata: Json = None
    roles: List[TimeseriesRole] = None
    annotations: List[TimeseriesAnnotation] = None
    changelog: List[TimeseriesChangelog] = None
#   station: StationmetaCoreCreate
#   station: StationmetaCore
#   station: StationmetaCreate
#   station: StationmetaBase
#   station: Stationmeta
# Try, which one of the above is wanted for station
# The next one works:

    class Config:
        orm_mode = True

    @validator('changelog')
    def order_changelog(cls, v):
        return sorted(v, key=lambda x: x.datetime)

    @validator('roles')
    def check_roles(cls, v):
        if v == []:
            return None
        else:
            return v

    @validator('annotations')
    def check_annotations(cls, v):
        if v == []:
            return None
        else:
            return v

    @validator('changelog')
    def check_changelog(cls, v):
        if v == []:
            return None
        else:
            return v


class TimeseriesCreate(TimeseriesCoreCreate):
    roles: List[TimeseriesRoleNoCreate] = None
    annotations: List[TimeseriesAnnotation] = None

    class Config:
        orm_mode = True


class TimeseriesPatch(TimeseriesCoreCreate):
    label: str = None
    order: int = None
    sampling_frequency: str = None
    aggregation: str = None
    data_origin_type: str = None
    data_origin: str = None
    data_start_date: dt.datetime = None
    data_end_date: dt.datetime = None
    sampling_height: float = None
#   roles: List[TimeseriesRole] = None
#   annotations: List[TimeseriesAnnotationPatch] = None
    # just to get things working
    roles: list = None
    annotations: list = None
#   variable: Variable = None
#   station: StationmetaCoreBase = None
#   programme: TimeseriesProgramme = None
    station_id: int = None
    variable_id: int = None
    programme_id: int = None
    additional_metadata: Json = None

    class Config:
        orm_mode = True


class Timeseries(TimeseriesBase):
    id: int = None
    # hot fix
    coordinates: Coordinates = None
    name: str = None
#   variable_id: int = None
#   station_id: int = None
    codes: List[str] = None
    station_country: str = None
    type: str = None
    type_of_area: str = None
    timezone: str = None
    state: str = None
    coordinate_validation_status: str = None
    mean_topography_srtm_alt_90m_year1994: float = None
    mean_topography_srtm_alt_1km_year1994: float = None
    max_topography_srtm_relative_alt_5km_year1994: float = None
    min_topography_srtm_relative_alt_5km_year1994: float = None
    stddev_topography_srtm_relative_alt_5km_year1994: float = None
    climatic_zone_year2016: str = None
    htap_region_tier1_year2010: str = None
    dominant_landcover_year2012: str = None
    landcover_description_25km_year2012: str = None
    dominant_ecoregion_year2017: str = None
    ecoregion_description_25km_year2017: str = None
    distance_to_major_road_year2020: float = None
    mean_stable_nightlights_1km_year2013: float = None
    mean_stable_nightlights_5km_year2013: float = None
    max_stable_nightlights_25km_year2013: float = None
    max_stable_nightlights_25km_year1992: float = None
    mean_population_density_250m_year2015: float = None
    mean_population_density_5km_year2015: float = None
    max_population_density_25km_year2015: float = None
    mean_population_density_250m_year1990: float = None
    mean_population_density_5km_year1990: float = None
    max_population_density_25km_year1990: float = None
    mean_nox_emissions_10km_year2015: float = None
    mean_nox_emissions_10km_year2000: float = None
    toar1_category: str = None

    class Config:
        orm_mode = True

    # hot fix
    @validator('station_country')
    def check_station_country(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CN_vocabulary))[0].display_str

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CV_vocabulary))[0].display_str

    @validator('type')
    def check_type(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.ST_vocabulary))[0].display_str

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TA_vocabulary))[0].display_str

    @validator('timezone')
    def check_timezone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TZ_vocabulary))[0].display_str

    @validator('climatic_zone_year2016')
    def check_climatic_zone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.CZ_vocabulary))[0].display_str

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TC_vocabulary))[0].display_str

    @validator('htap_region_tier1_year2010')
    def check_htap_region_tier1(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.TR_vocabulary))[0].display_str

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.LC_vocabulary))[0].display_str

    @validator('landcover_description_25km_year2012')
    def check_landcover_description_year2012(cls, v):
        return get_full_description_from_abbreviation(toardb.toardb.LC_vocabulary, v)

    @validator('ecoregion_description_25km_year2017')
    def check_ecoregion_description_25km_year2017(cls, v):
        return get_full_description_from_abbreviation(toardb.toardb.ER_vocabulary, v)

    @validator('dominant_ecoregion_year2017')
    def check_dominant_ecoregion_year2017(cls, v):
        return tuple(filter(lambda x: x.value == int(v), toardb.toardb.ER_vocabulary))[0].display_str

class TimeseriesFields(Timeseries):
    roles: TimeseriesRoleBaseFields = None
    pass


class TimeseriesWithCitation(TimeseriesBase):
    id: int
    citation: str
    attribution: str = None
    license: str

    class Config:
        orm_mode = True

    @validator('sampling_frequency')
    def check_sampling_frequency(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.SF_vocabulary)):
            return v
        elif tuple(filter(lambda x: x.value == int(v), toardb.toardb.SF_vocabulary)):
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.SF_vocabulary))[0].display_str
        else:
            raise ValueError(f"sampling frequency not known: {v}")

    @validator('aggregation')
    def check_aggregation(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.AT_vocabulary)):
            return v
        elif tuple(filter(lambda x: x.value == int(v), toardb.toardb.AT_vocabulary)):
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.AT_vocabulary))[0].display_str
        else:
            raise ValueError(f"aggregation type not known: {v}")

    @validator('data_origin_type')
    def check_data_origin_type(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.OT_vocabulary)):
            return v
        elif tuple(filter(lambda x: x.value == int(v), toardb.toardb.OT_vocabulary)):
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.OT_vocabulary))[0].display_str
        else:
            raise ValueError(f"data origin type not known: {v}")

    @validator('data_origin')
    def check_data_origin(cls, v):
        if tuple(filter(lambda x: x.string == v, toardb.toardb.DO_vocabulary)):
            return v
        elif tuple(filter(lambda x: x.value == int(v), toardb.toardb.DO_vocabulary)):
            return tuple(filter(lambda x: x.value == int(v), toardb.toardb.DO_vocabulary))[0].display_str
        else:
            raise ValueError(f"data origin not known: {v}")

    @validator('additional_metadata')
    def check_additional_metadata(cls, v):
#       return str(v).replace("'",'"')
        return v


Contributors = Union[TimeseriesProgramme, TimeseriesRole]
