# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Simple test API for timeseries management
"""

import re
import datetime as dt
from typing import List, Union
from fastapi import APIRouter, Depends, HTTPException, Body, Request, File, UploadFile
from fastapi.responses import JSONResponse
from sqlalchemy.orm import Session
from sqlalchemy.inspection import inspect
from starlette.datastructures import QueryParams
import urllib.parse
from . import crud, schemas
from toardb.utils.database import ToarDbSession, get_db
from toardb.stationmeta.crud import get_stationmeta_by_id
from toardb.variables.crud import get_variable
from toardb.contacts.crud import get_contact
from toardb.timeseries.models import Timeseries
from toardb.stationmeta.models import StationmetaCore, StationmetaGlobal
from toardb.data.models import Data
from toardb.utils.utils import (
        get_str_from_value,
        get_admin_access_rights,
        get_timeseries_md_change_access_rights,
        get_register_contributors_access_rights
)


router = APIRouter()


def get_query_params(raw_query_string):
    updated_query_string = raw_query_string.replace('+', '%2B')
    updated_params = dict(urllib.parse.parse_qsl(updated_query_string))
    return QueryParams(updated_params)

# CRUD: create, retrieve, update, delete
# 1. retrieve
#get all entries of table timeseries
@router.get('/search/', response_model=List[schemas.Timeseries]  | List[schemas.TimeseriesFields], response_model_exclude_none=True, response_model_exclude_unset=True)
def search_all_timeseries(request: Request, db: Session = Depends(get_db)):
    updated_query_params = get_query_params(request.url.query)
    return crud.search_all(db, path_params=request.path_params, query_params=updated_query_params)


# Does not make sense when only certain fields are selected in different requests needs different syntax
# order of concatenated statements is also unclear
@router.get('/search/a', response_model=List[schemas.Timeseries]  | List[schemas.TimeseriesFields], response_model_exclude_none=True, response_model_exclude_unset=True)
def search_all_timeseries_aggregations(request: Request, db: Session = Depends(get_db)):
    urls = re.split(r"(?=[+-]\D)", "+" + request.url.query)[1:]
    if urls:
        signs = [url.startswith("+") for url in urls]
        query_params = [get_query_params(url[1:]) for url in urls]
        return crud.search_all_aggregation(
            db, path_params=request.path_params, signs=signs, query_params_list=query_params
        )
    else:
        return search_all_timeseries(request, db)


#get all entries of table timeseries
@router.get('/timeseries/', response_model=List[schemas.Timeseries], response_model_exclude_none=True)
def get_all_timeseries(request: Request, db: Session = Depends(get_db)):
    raw_query_string = request.url.query
    updated_query_string = raw_query_string.replace('+', '%2B')
    updated_params = dict(urllib.parse.parse_qsl(updated_query_string))
    updated_query_params = QueryParams(updated_params)
    return crud.search_all(db, path_params=request.path_params, query_params=updated_query_params,lts=True)


#get all metadata of one timeseries (known its ID)
@router.get('/timeseries/{timeseries_id}', response_model=schemas.Timeseries, response_model_exclude_none=True, response_model_exclude_unset=True)
def get_timeseries(timeseries_id: int, fields: str = None, db: Session = Depends(get_db)):
    db_timeseries = crud.get_timeseries(db, timeseries_id=timeseries_id, fields=fields)
    if db_timeseries is None:
        raise HTTPException(status_code=404, detail="Timeseries not found.")
    return db_timeseries


#get all metadata of one timeseries (known its ID)
@router.get('/timeseries/id/{timeseries_id}', response_model=schemas.Timeseries, response_model_exclude_none=True, response_model_exclude_unset=True)
def get_timeseries2(timeseries_id: int, db: Session = Depends(get_db)):
    db_timeseries = crud.get_timeseries(db, timeseries_id=timeseries_id)
    if db_timeseries is None:
        raise HTTPException(status_code=404, detail="Timeseries not found.")
    return db_timeseries


#build citation text of one timeseries (known its ID)
@router.get('/timeseries/citation/{timeseries_id}')
def get_citation(timeseries_id: int, datetime: dt.datetime = None, db: Session = Depends(get_db)):
    citation = crud.get_citation(db, timeseries_id=timeseries_id, datetime=datetime)
    return citation


#get all metadata of one timeseries (known its unique label)
@router.get('/timeseries/unique/')
def get_timeseries(station_id: int, variable_id: int, resource_provider: str = None, 
                   sampling_frequency: str = None, provider_version: str = None, data_origin_type: str = None,
                   data_origin: str = None, sampling_height: float = None, label: str = None,
                   db: Session = Depends(get_db)):
    db_timeseries = crud.get_timeseries_by_unique_constraints(db, station_id=station_id,
                         variable_id=variable_id, resource_provider=resource_provider,
                         sampling_frequency=sampling_frequency, provider_version=provider_version, data_origin_type=data_origin_type,
                         data_origin= data_origin, sampling_height=sampling_height, label=label)
    if db_timeseries is None:
        raise HTTPException(status_code=404, detail="Timeseries not found.")
    return db_timeseries


@router.get('/timeseries_changelog/{timeseries_id}', response_model=List[schemas.TimeseriesChangelog])
def get_timeseries_changelog(timeseries_id: int, db: Session = Depends(get_db)):
    db_changelog = crud.get_timeseries_changelog(db, timeseries_id=timeseries_id)
    return db_changelog


@router.get('/timeseries_programme/{name}', response_model=List[schemas.TimeseriesProgramme])
def get_timeseries_programme(name: str, db: Session = Depends(get_db)):
    db_programme = crud.get_timeseries_programme(db, name=name)
    return db_programme


@router.get('/timeseries/request_timeseries_list_of_contributors/{rid}', response_model=Union[str,List[schemas.Contributors]])
def get_registered_request_contributors(rid: str, format: str = 'text', db: Session = Depends(get_db)):
    contributors = crud.get_registered_request_contributors(db, rid=rid, format=format)
    return contributors


@router.post('/timeseries/register_timeseries_list_of_contributors/{rid}')
def register_request_contributors(rid: str,
                                  ids: List[int],
                                  access: dict = Depends(get_register_contributors_access_rights),
                                  db: Session = Depends(get_db)):
    return crud.register_request_contributors(db, rid=rid, ids=ids)


@router.post('/timeseries/request_contributors', response_model=Union[str,List[schemas.Contributors]])
def get_request_contributors(format: str = 'text', file: UploadFile = File(...), db: Session = Depends(get_db)):
    contributors = crud.get_request_contributors(db, format=format, input_handle=file)
    return contributors


# post and patch only via authorization by Helmholtz AAI

# 2. create

#problems with Roles!!! prelimarily return TimeseriesPatch (instead of Timeseries!)
#@router.post('/timeseries/', response_model=schemas.TimeseriesPatch)
#@router.post('/timeseries/', response_model=schemas.Timeseries)
#@do not return object, just some information (see stationmeta)
@router.post('/timeseries/')
def create_timeseries(timeseries: schemas.TimeseriesCreate = Body(..., embed = True),
                      access: dict = Depends(get_admin_access_rights),
                      db: Session = Depends(get_db)):
    # check whether the post is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        return crud.create_timeseries(db=db, timeseries=timeseries,
                author_id=access['auth_user_id'])
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")


#@router.patch('/timeseries/id/{timeseries_id}', response_model=schemas.TimeseriesPatch)
@router.patch('/timeseries/id/{timeseries_id}')
def patch_timeseries(timeseries_id: int,
                     description: str = None,
                     timeseries: schemas.TimeseriesPatch = Body(..., embed = True),
                     access: dict = Depends(get_timeseries_md_change_access_rights),
                     db: Session = Depends(get_db)):
    # check whether the patch is authorized (401: Unauthorized)
    if access['status_code'] == 200:
        # check, if description has been given in query_params!
        if not description:
            raise HTTPException(status_code=404, detail="description text ist missing.")
        db_timeseries = crud.get_timeseries(db, timeseries_id=timeseries_id)
        if db_timeseries is None:
            raise HTTPException(status_code=404, detail="Time series for patching not found.")

        # for Max's bachelor thesis: extend description string to be more explicit
        # still to be done: GUI and REST should use the same user database!
        description = ": ".join([f"{access['user_name']} ({access['user_email']})", description])

        return crud.patch_timeseries(db=db, description=description,
                timeseries_id=db_timeseries.id, timeseries=timeseries,
                author_id=access['auth_user_id'])
    else:
        raise HTTPException(status_code=401, detail="Unauthorized.")

