# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
API for TOAR-II database
"""

from collections import namedtuple
from fastapi.responses import Response
import io
import functools
import json
import pandas as pd

from pydantic import BaseSettings
from typing import List
from fastapi import FastAPI, Depends, HTTPException, APIRouter, \
                    Request
from fastapi.responses import HTMLResponse, JSONResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from sqlalchemy.orm import Session
from starlette.responses import FileResponse 
from fastapi.templating import Jinja2Templates

from pyinstrument import Profiler
from pyinstrument.renderers.html import HTMLRenderer
from pyinstrument.renderers.speedscope import SpeedscopeRenderer
from pathlib import Path
import time

from toardb.utils.database import ToarDbSession, engine, get_db
from toardb.utils.settings import base_url
from toardb.utils.utils import normalize_metadata
from toardb.auth_user import auth_user
from toardb.variables import variables
from toardb.contacts import contacts
from toardb.stationmeta import stationmeta, models
from toardb.timeseries import timeseries
from toardb.data import data

RC_vocabulary = 0
RS_vocabulary = 0
AK_vocabulary = 0
OK_vocabulary = 0
SF_vocabulary = 0
AT_vocabulary = 0
OT_vocabulary = 0
DO_vocabulary = 0
CZ_vocabulary = 0
CV_vocabulary = 0
CN_vocabulary = 0
TZ_vocabulary = 0
ST_vocabulary = 0
TA_vocabulary = 0
TC_vocabulary = 0
TR_vocabulary = 0
LC_vocabulary = 0
ER_vocabulary = 0
RT_vocabulary = 0
DF_vocabulary = 0
CL_vocabulary = 0
CS_vocabulary = 0
KS_vocabulary = 0
CT_vocabulary = 0
controlled_fields = 0

app = FastAPI()
app.mount("/static", StaticFiles(directory="static"), name="_static")
templates = Jinja2Templates(directory="templates")


# check more on https://pyinstrument.readthedocs.io/en/latest/guide.html#profile-a-web-request-in-fastapi
@app.middleware("http")
async def profile_request(request: Request, call_next):
    profile_type_to_ext = {"html": "html", "json": "json"}
    profile_type_to_renderer = {
        "html": HTMLRenderer,
        "json": SpeedscopeRenderer,
    }
    if request.query_params.get("profile", False): # pragma: no cover
        current_dir = Path(__file__).parent
        profile_type = request.query_params.get("profile_format", "html")
        with Profiler(interval=0.001, async_mode="enabled") as profiler:
            response = await call_next(request)
        ext = profile_type_to_ext[profile_type]
        renderer = profile_type_to_renderer[profile_type]()
        with open(current_dir / f"../profile.{ext}", "a") as outfile:
            outfile.write(profiler.output(renderer=renderer))
        return response
    return await call_next(request)


# check more on https://fastapi.tiangolo.com/tutorial/middleware/
@app.middleware("http")
async def add_process_time_header(request: Request, call_next):
    if request.query_params.get("timing", False): # pragma: no cover
        current_dir = Path(__file__).parent
        start_time = time.time()
        response = await call_next(request)
        with open(current_dir / f"../timing.txt", "a") as outfile:
            outfile.write("{} s: {}\n".format(time.time() - start_time, request.url))
        return response
    return await call_next(request)


@app.middleware("http")
async def response_to_csv(request: Request, call_next):
    response = await call_next(request)
    if ((response.status_code != 200) or
        (request['path'].startswith(('/data/timeseries/','/data/timeseries_with_staging/','/ontology/','/controlled_vocabulary/'))) or
        (request.query_params.get('format') != 'csv')):
        return response
    # from: https://stackoverflow.com/a/71883126
    response_body = b""
    async for chunk in response.body_iterator:
        response_body += chunk
    list_response = json.loads(response_body)
    if isinstance(list_response,dict):
        list_response = [ list_response ]
    metadata = pd.DataFrame([ normalize_metadata(resp_object) for resp_object in list_response ])
    return Response(content=metadata.to_csv(index=False), status_code=response.status_code, media_type="text/csv")

# add endpoints
# additional yaml version of openapi.json
@app.get('/openapi.yaml', include_in_schema=False)
@functools.lru_cache()
def read_openapi_yaml() -> Response:
    openapi_json= app.openapi()
    yaml_s = io.StringIO()
    yaml.dump(openapi_json, yaml_s)
    return Response(yaml_s.getvalue(), media_type='text/yaml')


@app.get('/', response_class=HTMLResponse)
def show_api(request: Request):
    message = None
    return templates.TemplateResponse("TOARDB_FASTAPI_Rest.html", {"request": request,
        'message': message})


@app.get("/ontology")
def read_onto(format: str = 'xml'):
    if format == 'owldoc':
        return RedirectResponse(base_url + "/documentation/ontologies/v1.0/index.html")
    else:
        return FileResponse('static/ontology.xml')


@app.get("/controlled_vocabulary/")
def info():
    controlled_vocabulary = {
        "Role Code": RC_vocabulary,
        "Role Status": RS_vocabulary,
        "Kind Of Annotation": AK_vocabulary,
        "Kind Of Organization": OK_vocabulary,
        "Sampling Frequency": SF_vocabulary,
        "Aggregation Type": AT_vocabulary,
        "Data Origin Type": OT_vocabulary,
        "Data Origin": DO_vocabulary,
        "Climatic Zone 2019": CZ_vocabulary,
        "Country Code": CN_vocabulary,
        "Timezone": TZ_vocabulary,
        "Station Coordinate Validity": CV_vocabulary,
        "Station Type": ST_vocabulary,
        "Station Type Of Area": TA_vocabulary,
        "Station TOAR Category": TC_vocabulary,
        "Station HTAP Region": TR_vocabulary,
        "Station Landcover Type": LC_vocabulary,
        "Station ECO Region Type": ER_vocabulary,
        "Result Type": RT_vocabulary,
        "Data Flag": DF_vocabulary,
        "Type Of Change": CL_vocabulary,
        "Absorption Cross Section": CS_vocabulary,
        "Sampling Type": KS_vocabulary,
        "Calibration Type": CT_vocabulary,
    }
    return controlled_vocabulary

@app.get("/controlled_vocabulary/{name}")
def info(name: str):
    controlled_vocabulary = {
        "role code": RC_vocabulary,
        "role status": RS_vocabulary,
        "kind of annotation": AK_vocabulary,
        "kind of organization": OK_vocabulary,
        "sampling frequency": SF_vocabulary,
        "aggregation type": AT_vocabulary,
        "data origin": DO_vocabulary,
        "data origin type": OT_vocabulary,
        "climatic zone 2019": CZ_vocabulary,
        "coordinate validity": CV_vocabulary,
        "country code": CN_vocabulary,
        "timezone": TZ_vocabulary,
        "station coordinate validity": CV_vocabulary,
        "station type": ST_vocabulary,
        "station type of area": TA_vocabulary,
        "station toar category": TC_vocabulary,
        "station htap region": TR_vocabulary,
        "station landcover type": LC_vocabulary,
        "station eco region type": ER_vocabulary,
        "result type": RT_vocabulary,
        "data flag": DF_vocabulary,
        "type of change": CL_vocabulary,
        "absorption cross section": CS_vocabulary,
        "sampling type": KS_vocabulary,
        "calibration type": CT_vocabulary,
    }
    if name.lower() in controlled_vocabulary.keys():
        return controlled_vocabulary[name.lower()]
    else:
        status_code=200
        message = f"No controlled vocabulary found for '{name}'"
        return JSONResponse(status_code=status_code, content=message)

@app.get("/database_statistics/")
def stats_info():
    with open("static/db_statistics.json") as f:
        db_stats=json.load(f)
    return db_stats

@app.get('/database_statistics/{name}')
def stats_info(name: str):
    with open("static/db_statistics.json") as f:
        db_stats=json.load(f)
    return db_stats[name]

@app.get("/geopeas_urls/")
def geo_info(db: Session = Depends(get_db)):
    res = db.query(models.StationmetaGlobalService).all()
    return res


# Dependency
def get_db():
    try:
        db = ToarDbSession()
        yield db
    finally:
        db.close()

app.include_router(auth_user.router)
app.include_router(variables.router)
app.include_router(contacts.router)
app.include_router(stationmeta.router)
app.include_router(timeseries.router)
app.include_router(data.router)

# get the controlled vocabulary from table
def __get_enum_dict(connection, table_name):
    res = connection.execute("select * from "+table_name+" order by enum_val")
    Enumdict=namedtuple("Dict",["value","string","display_str"])
    enum_dict = []
    for entry in res:
        enum_dict.append(Enumdict(*entry))
    return enum_dict

@app.on_event("startup")
# will be executed before application *starts*
# ==> again: at this point no database connection available!
#     (and also all tables are unknown)
def startup_event():

    global RC_vocabulary
    global RS_vocabulary
    global AK_vocabulary
    global OK_vocabulary
    global SF_vocabulary
    global AT_vocabulary
    global OT_vocabulary
    global DO_vocabulary
    global CZ_vocabulary
    global CV_vocabulary
    global CN_vocabulary
    global TZ_vocabulary
    global ST_vocabulary
    global TA_vocabulary
    global TC_vocabulary
    global TR_vocabulary
    global LC_vocabulary
    global ER_vocabulary
    global RT_vocabulary
    global DF_vocabulary
    global CL_vocabulary
    global CS_vocabulary
    global KS_vocabulary
    global CT_vocabulary
    global controlled_fields

    with engine.begin() as connection:
        RC_vocabulary = __get_enum_dict(connection, "rc_vocabulary")
        RS_vocabulary = __get_enum_dict(connection, "rs_vocabulary")
        AK_vocabulary = __get_enum_dict(connection, "ak_vocabulary")
        OK_vocabulary = __get_enum_dict(connection, "ok_vocabulary")
        SF_vocabulary = __get_enum_dict(connection, "sf_vocabulary")
        AT_vocabulary = __get_enum_dict(connection, "at_vocabulary")
        OT_vocabulary = __get_enum_dict(connection, "ot_vocabulary")
        DO_vocabulary = __get_enum_dict(connection, "do_vocabulary")
        CZ_vocabulary = __get_enum_dict(connection, "cz_vocabulary")
        CV_vocabulary = __get_enum_dict(connection, "cv_vocabulary")
        CN_vocabulary = __get_enum_dict(connection, "cn_vocabulary")
        TZ_vocabulary = __get_enum_dict(connection, "tz_vocabulary")
        ST_vocabulary = __get_enum_dict(connection, "st_vocabulary")
        TA_vocabulary = __get_enum_dict(connection, "ta_vocabulary")
        TC_vocabulary = __get_enum_dict(connection, "tc_vocabulary")
        TR_vocabulary = __get_enum_dict(connection, "tr_vocabulary")
        LC_vocabulary = __get_enum_dict(connection, "lc_vocabulary")
        ER_vocabulary = __get_enum_dict(connection, "er_vocabulary")
        RT_vocabulary = __get_enum_dict(connection, "rt_vocabulary")
        DF_vocabulary = __get_enum_dict(connection, "df_vocabulary")
        CL_vocabulary = __get_enum_dict(connection, "cl_vocabulary")
        CS_vocabulary = __get_enum_dict(connection, "cs_vocabulary")
        KS_vocabulary = __get_enum_dict(connection, "ks_vocabulary")
        CT_vocabulary = __get_enum_dict(connection, "ct_vocabulary")

        # also get information, which database field is using which controlled vocabulary

        controlled_fields = {}
        res = connection.execute("SELECT c.conname, c1.conname FROM pg_catalog.pg_constraint c " +
                           "INNER JOIN pg_catalog.pg_constraint c1 ON c.confrelid=c1.conrelid " +
                           "AND c1.conname LIKE '%%enum_val%%' AND c.conname LIKE '%%_fk_%%_voc%%'")
        for line in res:
            field, voc = (line[0].split('_fk_'))
            voc = voc.split('_')[0].upper()
            controlled_fields[field] = f"{voc}_vocabulary"
    engine.dispose()

