"""
Max Lensing 22.03.2022
-------------------------
the script compares the content of the TOARII ontology and the controlled vocabulary

requirements:
- rdflib, rdflib-jsonld, requests, argparse (python packages)
- VPN to the Forschungszentrum to be able to open the link to the controlled vocabulary and ontology
- 'TOARII_ontology.py' class
-------------------------

methods:

similar(word, list):
    checks if a word is similar to a word contained in a list
    returns the matching key

parse_args():
    parses the given arguments from the command line
    returns the arguments
"""

__author__ = 'Max Lensing'
__email__ = 'm.lensing@fz-juelich.de'

TOAR_SERVICE_URL='https://toar-data-dev.fz-juelich.de/api/v2/'
controlled_vocabulary_link = TOAR_SERVICE_URL + 'controlled_vocabulary/'
ontology_link = TOAR_SERVICE_URL + 'ontology'

import TOARII_ontology
import argparse
import requests
import json
import os.path

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--onto', type=str, help='allows to specify a path to an ontology file to be used (in .jsonld format)')
    parser.add_argument('-o', action='store_true', help='outputs all comparisons of all classes and their entries')
    parser.add_argument('-fo', action='store_true', help='outputs only the classes with incorrect comparisons')
    args = parser.parse_args()

    if not args.onto or not (os.path.isfile(args.onto) and args.onto.endswith(".jsonld")):
        args.onto = ontology_link

    if args.o and args.fo:
        args.o = False

    return args



def similar(word_x, list):
    keywords = ['unknown','urban','unclassified','undefined','notchecked']
    word = word_x.lower()
    for key in keywords:
        if word.startswith(key):
            try:
                followed_by_num = word[len(key) + 1].isdigit()
            except IndexError:
                pass
            else:
                if followed_by_num:
                    for entry in list:
                        b = entry.lower()
                        if b == key:
                            return key
    return None

if __name__ == "__main__":
    mistakes = 0
    missing_classes = 0

    args = parse_args()
    output = args.o
    false_output = args.fo

    try:
        onto = TOARII_ontology.TOAR_Ontology(args.onto)
    except:
        print(f"Connection to {args.onto} failed")
        exit(-1)

    try:
        controlled_voc = requests.get(controlled_vocabulary_link, timeout=10)
    except:
        print(f"Connection to {controlled_vocabulary_link} failed")
        exit(-1)

    print(f"Ontology: {args.onto}")
    print(f"Controlled vocabulary: {controlled_vocabulary_link}")

    convoc = json.loads(controlled_voc.text)
    transl_dict = onto.get_class_translation()

    spacing = 56 * "-"
    print(spacing + '\n--- comparison of controlled vocabulary and ontology ---\n' + spacing)

    for onto_class in sorted(transl_dict, key=lambda transl_dict: transl_dict[1]):
        classmistakes = 0
        #onto_class_title = onto_class[1]
        #onto_class_id = onto_class[0]
        ontology_class_str = f"'{onto_class[1]}' (id: {onto_class[0][-6:]})"
        onto_class_entries = onto.get_entries(onto_class[0])
        if transl_dict[onto_class] in convoc.keys():
            convoc_class_entries = []
            for convoc_entry in convoc[transl_dict[onto_class]]:
                convoc_class_entries.append(convoc_entry[1])
        else:
            print(f"Onto: {ontology_class_str} -> Convoc: '{transl_dict[onto_class]}' not included in controlled vocabulary")
            print(spacing)
            missing_classes += 1
            continue

        if output:
            print(f"Onto: {ontology_class_str}, Convoc: '{transl_dict[onto_class]}'")

        #print(sorted(onto_class_entries))
        #print(sorted(convoc_class_entries))
        for entry in sorted(onto_class_entries):
            if entry in convoc_class_entries:
                convoc_class_entries.remove(entry)
                onto_class_entries.remove(entry)
                if output:
                    print(entry + ': True')

            elif similar(entry, convoc_class_entries):
                sim = similar(entry, convoc_class_entries)
                onto_class_entries.remove(entry)
                convoc_class_entries.remove(sim)
                if output:
                    print(f'{entry} and {sim} similar')
            else:
                classmistakes += 1
                if output:
                    print(f'{entry}: False')
        classmistakes += len(convoc_class_entries)
        if false_output and classmistakes > 0 or not (output or false_output):
            print(f"Onto: {ontology_class_str}, Convoc: '{transl_dict[onto_class]}'")
        if len(onto_class_entries) > 0:# and (false_output or output):
            print(f"left from Onto: {sorted(onto_class_entries)}")
        if len(convoc_class_entries) > 0:# and (false_output or output):
            print(f"left from Convoc: {sorted(convoc_class_entries)}")
        if (classmistakes > 0 and false_output) or output or not(output or false_output):
            print(spacing)
        mistakes += classmistakes

    print(f"{mistakes} mistakes")
    if missing_classes > 0:
        print(f"{missing_classes} missing classes")
