"""
class TOAR_Ontology
===================
"""

__author__ = "Sabine Schroeder"
__version__ = "0.9.5"
__maintainer__ = "Sabine Schroeder"
__email__ = "s.schroeder@fz-juelich.de"

import rdflib
from rdflib import URIRef, Literal, BNode
from rdflib.namespace import RDF, RDFS, SKOS, OWL
from rdflib.collection import Collection

class TOAR_Ontology:
    """
    A class to represent the TOAR onthology.

    ...

    Attributes
    ----------
    filename : str
        filename of file containing onthology (in json-ld-Format)

    graph : Graph
        onthology graph

    Methods
    -------
    get_classes():
        show all classes of onthology

    get_class(title, superclass):
        find the onthology class for the given title

    get_class_by_id(id):
        find the onthology class for the given id

    show_class_definition(ont_class):
        show information on given ontology class

    get_class_definition(ont_class):
        get information on given ontology class

    get_class_translation():
        get a dictionary with tuples of ids and names of the ontology classes as keys
        the referring values are the classnames of the controlled vocabulary

    get_entries(ont_class):
        searches all entries matching with the given ontology class
        and returns their labels

    get_entries_with_definition(ont_class):
        searches all entries matching with the given ontology class
        and returns their labels and definitions

    show_entries(ont_class):
        show information on given ontology class

    add_entry(ont_class, label, definition, idcount):
        add entry to given ontology class

    delete_entry(entry, ont_class):
        delete given entry from given ontology class

    check_entry(entry, ont_class):
        check whether entry is a valid value of the ontology

    print(format=""):
        Prints the onthology in the given format

    save(filename, format):
        save the onthology in given format to the given file

    get_ids():
        get a list with tuples of id and title of all ontology classes

    def get_next_id():
        determine next free ID to build a NamedIndividual
    """

    def __init__(self, filename):
        """
        initialize the onthology

        Parameters
        ----------
        filename: str
            filename containing the ontology in json-ld format

        Returns
        -------
        TOAR_Ontology
        """

        self.filename = filename
        g = rdflib.Graph()
        self.root = "https://toar-data.fz-juelich.de/documentation/ontologies/v1.0"
        if filename.startswith('https') or filename.endswith('xml'):
            self.graph = g.parse(self.filename, format='application/rdf+xml')
        else:
            self.graph = g.parse(self.filename, format='json-ld')

    def get_classes(self):
        """
        show all classes of onthology

        Parameters
        ----------
        None

        Returns
        -------
        [str]: list of class titles
        """
        list = []
        for s, p, o in self.graph.triples( (None,  RDF.type, OWL.Class) ):
            title = self.graph.value(subject=s,predicate=RDFS.label)
            if title:
                list.append(title.strip())
        return list

    def get_class(self, title, superclass=None):
        """
        find the onthology class for the given title

        Parameters
        ----------
        title: str
            title of class
        superclass: str
            ID of the superclass

        Returns
        -------
        OWL.Class: ontology class matching the given title
        """
        for s, p, o in self.graph.triples( (None,  RDF.type, OWL.Class) ):
            value = self.graph.value(subject=s,predicate=RDFS.label)
            cl_superclass = None
            if superclass:
                cl_superclass = self.graph.value(subject=s,predicate=RDFS.subClassOf)
            if value and value.strip() == title.strip() and cl_superclass == superclass:
                ont_class = s
                break
        return ont_class

    def get_class_by_id(self, id):
        """
        find the onthology class for the given id

        Parameters
        ----------
        id: str
            id of class

        Returns
        -------
        OWL.Class: ontology class matching the given id
        """
        for s, p, o in self.graph.triples( (None,  RDF.type, OWL.Class) ):
            if s.strip() == id.strip():
                ont_class = s
        return ont_class

    def show_class_definition(self, ont_class):
        """
        show information on given ontology class

        Parameters
        ----------
        ont_class: OWL.class
            onthology class

        Returns
        -------
        None
        """
        for s, p, o in self.graph.triples( (ont_class,  SKOS.definition, None) ):
            print(f"definition of {ont_class}: {o}")

    def get_class_definition(self, ont_class):
        """
        get information on given ontology class

        Parameters
        ----------
	ont_class: OWL.class
            onthology class

        Returns
        -------
        class_definition: str
            definition string of given class
        """
        #return self.graph.objects( (ont_class,  SKOS.definition, None) )
        for s, p, o in self.graph.triples( (ont_class,  SKOS.definition, None) ):
            definition = f"{o}"
        return(definition)

    def get_class_translation(self):
        """
        get a dictionary with tuples of ids and names of the ontology classes as keys
        the referring values are the classnames of the controlled vocabulary

        Returns
        -------
        dict
        """
        translation_dict = {}
        id_title = self.get_ids()
        for i_t in id_title:
            for s, p, o in self.graph.triples((i_t[0], SKOS.definition, None)):
                start = o.find("see controlled vocabulary")
                if start != -1:
                    convoc_str = o[start:]
                    end = convoc_str.find(")")
                    convoc_class = convoc_str[:end].split(":")
                    try:
                        translation_dict[i_t] = convoc_class[1].strip()
                    except IndexError:
                        pass
        return translation_dict

    def get_entries(self, ont_class):
        """
        searches all entries matching with the given ontology class
        and returns their labels

        Parameters
        ----------
        ont_class: str
            ontology classname

        Returns
        -------
        list of entry labels
        """
        entries = []
        for s, p, o in self.graph.triples((None, RDF.type, ont_class)):
            entries.append(self.graph.value(subject=s, predicate=RDFS.label).strip())
        return entries

    def get_entries_with_definition(self, ont_class):
        """
        searches all entries matching with the given ontology class
        and returns their labels and definitions

        Parameters
        ----------
        ont_class: str
            ontology classname

        Returns
        -------
        dictionary of entry labels containing their definitions
        """
        entries = {}
        for s, p, o in self.graph.triples( (None,  RDF.type, ont_class ) ):
            label=self.graph.value(subject=s,predicate=RDFS.label)
            for s, p, o in self.graph.triples( (s,  SKOS.definition, None) ):
                entries[f"{label}"] = f"{o}"
        return entries

    def show_entries(self, ont_class):
        """
        show information on given ontology class

        Parameters
        ----------
        ont_class: OWL.class
            onthology class

        Returns
        -------
        None
        """
        for s, p, o in self.graph.triples( (None,  RDF.type, ont_class ) ):
            label=self.graph.value(subject=s,predicate=RDFS.label)
            print(f"{s} is {label}")
            for s, p, o in self.graph.triples( (s,  SKOS.definition, None) ):
                print(f"definition of {label}: {o}")

    def add_entry(self, ont_class, label, definition, idcount = None):
        """
        add entry to given ontology class

        Parameters
        ----------
        ont_class: OWL.class
            onthology class
        label: str
            title of entry to be inserted
        definition: str
            definition of entry to be inserted
        idcount: int
            index to form a unique string for a NamedIndividual

        Returns
        -------
        None
        """
        if idcount == None:
            idcount = self.get_next_id()
        subject=URIRef(f"{self.root}#OWLNamedIndividual_{idcount:06d}")
        p=RDF.type
        o=URIRef(ont_class)
        self.graph.add((subject,p,o))
        o=OWL.NamedIndividual
        self.graph.add((subject,p,o))
        self.graph.add((subject,RDFS.label,Literal(label,lang="en")))
        self.graph.add((subject,SKOS.definition,Literal(definition)))
        # if there is an equivalent class: add the entry also there!
        for s, p, o in self.graph.triples( (ont_class,  OWL.equivalentClass, None) ):
            equivalentClass = self.get_class_by_id(o)
            # get oneOf-List of equivalentClass
            for s, p, o in self.graph.triples( (equivalentClass,  OWL.oneOf, None) ):
                c = Collection(self.graph, BNode(o))
                c.append(subject)
# The following is just an example how to iterate through a linked list
#               # The obtained list is a linked list,
#               # consisting of the first element and the rest (which is a linked list again)
#               newSubject = o
#               while newSubject != RDF.nil:
#                   for s, p, o in self.graph.triples( (newSubject,  RDF.first, None) ):
#                       print(f"first {newSubject}: s={s}, p={p}, o={o}")
#                   for s, p, o in self.graph.triples( (newSubject,  RDF.rest, None) ):
#                       print(f"rest {newSubject}: s={s}, p={p}, o={o}")
#                   newSubject=o


    def delete_entry(self, entry, ont_class):
        """
        delete entry with given label from given ontology class

        Parameters
        ----------
        entry: str
            title of entry to be deleted
        ont_class: OWL.class
            onthology class

        Returns
        -------
        None
        """
        for s, p, o in self.graph.triples( (None,  RDF.type, ont_class ) ):
            if self.graph.value(subject=s,predicate=RDFS.label).strip() == entry:
                self.graph.remove((s,p,o))


    def check_entry(self, entry, ont_class):
        """
        check whether entry is a valid value of the ontology

        Parameters
        ----------
        entry: str
            title of entry to be checked
        ont_class: OWL.class
            onthology class

        Returns
        -------
        Boolean: True, if entry was found in class
        """
        found=False
        for s, p, o in self.graph.triples( (None,  RDF.type, ont_class ) ):
            if self.graph.value(subject=s,predicate=RDFS.label).strip() == entry:
                found=True
        return found

    def print(self, format='json-ld'):
        """
        print the onthology in the given format

        Parameters
        ----------
        format: str, optional
            output format (default: json-ld)

        Returns
        -------
        None
        """
        print(self.graph.serialize(format=format).decode("utf-8"))

    def save(self, filename, format='json-ld'):
        """
        save the onthology in json-ld format to the given file

        Parameters
        ----------
        filename: str
            output file name
        format: str, optional
            output format (default: json-ld)

        Returns
        -------
        None
        """
        outfile = open(filename,'w')
        outfile.write(self.graph.serialize(format=format))

    def get_ids(self):
        """
        get a list with tuples of id and title of all ontology classes

        Returns
        -------
        [(str,str)]: list of class ids, [0] id, [1] title
        """
        list = []
        for s, p, o in self.graph.triples((None, RDF.type, OWL.Class)):
            title = self.graph.value(subject=s, predicate=RDFS.label)
            if title:
                list.append((s, title))
        return list

    def get_next_id(self):
        """
        determine next free ID to build a NamedIndividual

        Parameters
        ----------
        None

        Returns
        -------
        int: next free index (NamedIndividual) in ontology
        """
        nid=-1
        for s, p, o in self.graph.triples( (None,  RDF.type, OWL.NamedIndividual) ):
            try:
                sid = int(s[-6:])
                if sid > nid:
                    nid=sid
            except:
                continue
        return nid+1

#       #check_entry would have been nicer, if something like this worked:
#       subj2 = rdflib.URIRef(f"{self.root}#OWLNamedIndividual_000141")
#       pred2 = RDFS.label
#       obj2 = rdflib.Literal("Croplands")
#       passed = ((subj2, pred2, obj2) in g)
#       if not passed:
#           print("sub: ", subj2, "\npred: ", pred2, "\nobj: ", obj2, "\nnot found in ontology!\n")
#       else:
#           print("sub: ", subj2, "\npred: ", pred2, "\nobj: ", obj2, "\nfound in ontology!\n")

#       # getting the literals directly here:
#       subj2 = rdflib.URIRef(f"{self.root}#OWLNamedIndividual_000141")
#       generator = g.objects(subj2, RDFS.label)
#       for lit in generator:
#           print(lit.value)

