# SPDX-FileCopyrightText: 2021 Forschungszentrum Jülich GmbH
# SPDX-License-Identifier: MIT

"""
Helper functions for TOAR database

"""
from sqlalchemy import Table, and_
from sqlalchemy.orm import Session
from sqlalchemy.inspection import inspect
from sqlalchemy.dialects import postgresql
from fastapi import HTTPException, Request, Header
from collections import namedtuple
from copy import copy
import requests
import datetime as dt
from typing import List

from toardb.utils.settings import base_geopeas_url, userinfo_endpoint
from toardb.utils.deployment_settings import dashboard_token

# the following statement only if not in testing (pytest) mode!
from toardb.utils.database import get_db
from toardb.contacts.models import Contact, Organisation, Person
from toardb.timeseries.models import Timeseries, TimeseriesRole, timeseries_timeseries_roles_table
from toardb.stationmeta.models import StationmetaCore, StationmetaGlobal
from toardb.data.models import Data
from toardb.variables.models import Variable
from toardb.auth_user.models import AuthUser
from toardb.auth_user.crud import get_eduperson_and_roles
import toardb

roles_params = {column.name for column in inspect(TimeseriesRole).c} - {"id"}


def get_access_rights(request: Request, access_right: str = 'admin', incr: List[int] = [0, 1]):
    # Do not use underscores; they are not valid in header attributes!
    user_name = ''
    user_email = ''
    auth_user_id = -1
    role = 'anonymous'
    db = get_db()
    personinfo = get_eduperson_and_roles(request=request, DoIncr=incr)
    status_code = personinfo['status_code']
    if status_code != 401:
        ltoken = True
        userinfo = personinfo['userinfo']
        role = personinfo['role']
        if ("eduperson_entitlement" in userinfo and \
           f"urn:geant:helmholtz.de:res:toar-data{access_right}#login.helmholtz.de" \
           in userinfo["eduperson_entitlement"]) or access_right in [":data-download", ":map-data-download"]:
            user_name = userinfo["name"]
            user_email = userinfo["email"]
            db = next(get_db())
            db_user = db.query(AuthUser).filter(AuthUser.email == user_email).first()
            if db_user:
                auth_user_id = db_user.id
        else:
            status_code = 401
        if access_right == ":data-download" and (personinfo["max_timeseries"] is not None):
            if personinfo["num_timeseries"] > personinfo["max_timeseries"]:
                status_code = 401
        if access_right == ":map-data-download" and (personinfo["max_gridded"] is not None):
            if personinfo["num_gridded"] > personinfo["max_gridded"]:
                status_code = 401
    else:
        role = 'unauthorized'
    access_dict = { "status_code": status_code,
                    "user_name": user_name,
                    "user_email": user_email,
                    "auth_user_id": auth_user_id,
                    "role": role }
    return access_dict


def get_admin_access_rights(request: Request):
    return get_access_rights(request, ':admin')


def get_station_md_change_access_rights(request: Request):
    return get_access_rights(request, ':station-md-change')


def get_timeseries_md_change_access_rights(request: Request):
    return get_access_rights(request, ':timeseries-md-change')


def get_data_change_access_rights(request: Request):
    return get_access_rights(request, ':data-change')


def get_register_contributors_access_rights(request: Request):
    return get_access_rights(request, ':contributors-register')


def get_data_download_access_rights(request: Request):
    lfromdashboard = request.headers.get('DashboardToken') == dashboard_token
    access = get_access_rights(request, ':data-download', [1, 1])
    access['lfromdashboard'] = lfromdashboard
    return access


def get_map_data_download_access_rights(request: Request):
    return get_access_rights(request, ':map-data-download', [2, 1])


# function to return code for given value
def get_str_from_value(enum_dict, value) -> str:
    return tuple(filter(lambda x: x.value == value, enum_dict))[0].string

def get_displaystr_from_value(enum_dict, value) -> str:
    return tuple(filter(lambda x: x.value == value, enum_dict))[0].display_str

# function to return value for given code
def get_value_from_str(enum_dict, string) -> int:
    try:
        return tuple(filter(lambda x: x.string == string, enum_dict))[0].value
    except:
        raise ValueError(f"value not known: {string}")

# function to return value for given display string
def get_value_from_display_str(enum_dict, string) -> int:
    return tuple(filter(lambda x: x.display_str == string, enum_dict))[0].value


# get human readable fields, if database field is controlled (by vocabulary)
def get_hr_value(table_str,field,value):
    index = None
    for mid in ['core', 'global', 'glob', 'annotations', 'roles', 'changelog']:
        if table_str + f'_{mid}_' + field in toardb.toardb.controlled_fields:
            index = table_str + f'_{mid}_' + field
    if index:
        vocabulary = getattr(toardb.toardb, toardb.toardb.controlled_fields[index])
        value = get_str_from_value(vocabulary,int(value))
    return value


# translate filters that contain controlled vocabulary
def translate_convoc_list(values, table, display_name):
    try:
        return [get_value_from_str(table,v) for v in values]
    except ValueError:
        raise HTTPException(status_code=470, detail=f"{display_name} not known: {values}")


# expand subdicts except for additional_metadata
def normalize_metadata(metadata):
    normalized_metadata = {}
    for key, val in metadata.items():
        if isinstance(val, dict) and val and key != "additional_metadata":
            normalized_metadata.update(
                {
                    f"{key}_{sub_key}": sub_val
                    for sub_key, sub_val in normalize_metadata(val).items()
                }
            )
        else:
            normalized_metadata[key] = val
    return normalized_metadata


#
def create_filter(query_params, endpoint):

    # for ideas on how to create filter on special roles see:
    # https://gitlab.jsc.fz-juelich.de/esde/toar-data/toardb_fastapi/-/issues/95#note_144292

    # determine allowed query parameters (first *only* from Timeseries)
    timeseries_params = {column.name for column in inspect(Timeseries).c} | {"has_role"}
    timeseries_params = timeseries_params | {"additional_metadata-"}
    gis_params = {"bounding_box", "altitude_range"}
    if endpoint in ['search', 'timeseries']:
        core_params = {column.name for column in inspect(StationmetaCore).c if column.name not in ['id']} 
    else:
        core_params = {column.name for column in inspect(StationmetaCore).c} 
    core_params |= {"globalmeta", "station_additional_metadata-"}
    global_params = {column.name for column in inspect(StationmetaGlobal).c if column.name not in ['id','station_id']}
    data_params = {column.name for column in inspect(Data).c} | {"daterange", "format"}
    ambig_params = {"station_id", "station_changelog", "station_country", "station_additional_metadata"}
    variable_params = {column.name for column in inspect(Variable).c}
    person_params = {column.name for column in inspect(Person).c}
    allrel_params = {"limit", "offset", "fields", "format"}
    profiling_params = {"profile", "profile_format", "timing"}

    # pagination
    offset= int(query_params.get("offset", 0))
    try:
        limit = int(query_params.get("limit", 10))
    except:
        limit = query_params.get("limit")
        if limit == "None":
            limit = None
        else:
            raise ValueError(f"Wrong value for limit given: {limit}")

    # fields and format are no filter options
    fields = query_params.get("fields", "")
    format = query_params.get("format", 'json')

    allowed_params = allrel_params | profiling_params
    endpoint_params_map = {
        'stationmeta': gis_params | core_params | global_params,
        'timeseries': timeseries_params | roles_params,
        'search': gis_params | core_params | global_params | timeseries_params | roles_params | ambig_params,
        'data': data_params | profiling_params,
        'variables': variable_params,
        'persons': person_params,
    }
    try:
        allowed_params |= endpoint_params_map[endpoint]
    except KeyError:
        raise ValueError(f"Wrong endpoint given: {endpoint}")


    if fields:
        for field in fields.split(','):
            if field not in allowed_params:
                raise ValueError(f"Wrong field given: {field}")
        if fields.find("globalmeta") >= 0:
            fields = fields.replace("globalmeta","")
            fields += ','.join(global_params)

    t_filter = []
    t_r_filter = []
    s_c_filter = []
    s_g_filter = []
    d_filter = []
    v_filter = []
    p_filter = []
    # query_params is a multi-dict!
    for param_long in query_params:
        param = param_long.split('>')[0]
        if param not in allowed_params: #inform user, that an unknown parameter name was used (this could be a typo and falsify the result!)
            raise KeyError(f"An unknown argument was received: {param}.")
        if param in allrel_params or param in profiling_params:
            continue
        values = [item.strip() for v in query_params.getlist(param_long) for item in v.split(',')]
        # make sure ids are ints
        if param.endswith("id"):
            try:
                int_values = [ int(v) for v in values ]
                values = int_values
            except:
                raise ValueError(f"Wrong value (not int) given: {param}")
        # allow '+' in datestring (for adding timezone information)
        if param.endswith("date"):
            try:
                value = dt.datetime.fromisoformat(values[0])
            except:
                raise ValueError(f"Wrong value for time given: {values[0]}")
        if endpoint in ["stationmeta", "timeseries", "search"] and param in core_params:
            #check for parameters of the controlled vocabulary
            if param == "timezone":
                values = translate_convoc_list(values, toardb.toardb.TZ_vocabulary, "timezone")
            elif param == "coordinate_validation_status":
                values = translate_convoc_list(values, toardb.toardb.CV_vocabulary, "coordinate validation status")
            elif param == "country":
                values = translate_convoc_list(values, toardb.toardb.CN_vocabulary, "country")
            elif param == "type":
                values = translate_convoc_list(values, toardb.toardb.ST_vocabulary, "type")
            elif param == "type_of_area":
                values = translate_convoc_list(values, toardb.toardb.TA_vocabulary, "type of area")
            elif param == "station_additional_metadata-":
                param = f"{param_long[8:]}"
            # exceptions for special fields (codes, name)
            if param == 'codes':
                tmp_filter = []
                for v in values:
                    tmp_filter.append(f"('{v}'=ANY(stationmeta_core.codes))")
                tmp_filter = " OR ".join(tmp_filter)
                s_c_filter.append(f"({tmp_filter})")
            elif param == 'name':
                s_c_filter.append(f"LOWER(stationmeta_core.name) LIKE '%{values[0].lower()}%'")
            elif param_long.split('>')[0] == "station_additional_metadata-":
                val_mod = [ f"'\"{val}\"'::text" for val in values ]
                values = ",".join(val_mod)
                s_c_filter.append(f"to_json(stationmeta_core.{param})::text IN ({values})")
            else:
                s_c_filter.append(f"stationmeta_core.{param} IN {values}")
        elif endpoint in ["stationmeta", "search"] and param in global_params:
            if param == "climatic_zone_year2016":
                values = translate_convoc_list(values, toardb.toardb.CZ_vocabulary, "climatic zone year2016")
            elif param == "toar1_category":
                values = translate_convoc_list(values, toardb.toardb.TC_vocabulary, "TOAR-I category")
            elif param == "htap_region_tier1_year2010":
                values = translate_convoc_list(values, toardb.toardb.TR_vocabulary, "HTAP region TIER1 year2010")
            elif param == "dominant_landcover_year2012":
                values = translate_convoc_list(values, toardb.toardb.LC_vocabulary, "landcover type")
            elif param == "dominant_ecoregion_year2017":
                values = translate_convoc_list(values, toardb.toardb.ER_vocabulary, "ECO region type")
            s_g_filter.append(f"stationmeta_global.{param} IN {values}")
        elif endpoint in ["stationmeta", "search"] and param in gis_params:
            if param == "bounding_box":
                min_lat, min_lon, max_lat, max_lon = values
                bbox= f'SRID=4326;POLYGON (({min_lon} {min_lat}, {min_lon} {max_lat}, {max_lon} {max_lat}, {max_lon} {min_lat}, {min_lon} {min_lat}))'
                s_c_filter.append(f"ST_CONTAINS(ST_GeomFromEWKT('{bbox}'), coordinates)")
            else:
                s_c_filter.append(f"ST_Z(coordinates) BETWEEN {values[0]} AND {values[1]}")
        elif endpoint in ["timeseries", "search"]:
            #check for parameters of the controlled vocabulary
            if param == "sampling_frequency":
                values = translate_convoc_list(values, toardb.toardb.SF_vocabulary, "sampling_frequency")
            elif param == "aggregation":
                values = translate_convoc_list(values, toardb.toardb.AT_vocabulary, "aggregation")
            elif param == "data_origin_type":
                values = translate_convoc_list(values, toardb.toardb.OT_vocabulary, "data origin type")
            elif param == "data_origin":
                values = translate_convoc_list(values, toardb.toardb.DO_vocabulary, "data origin")
            elif param == "additional_metadata-":
                param = param_long
                if param == "additional_metadata->'absorption_cross_section'":
                    trlist = translate_convoc_list(values, toardb.toardb.CS_vocabulary, "absorption_cross_section")
                    values = [ str(val) for val in trlist ]
                    param = f"timeseries.{param}"
                elif param == "additional_metadata->'sampling_type'":
                    trlist = translate_convoc_list(values, toardb.toardb.KS_vocabulary, "sampling_type")
                    values = [ str(val) for val in trlist ]
                    param = f"timeseries.{param}"
                elif param == "additional_metadata->'calibration_type'":
                    trlist = translate_convoc_list(values, toardb.toardb.CT_vocabulary, "calibration_type")
                    values = [ str(val) for val in trlist ]
                    param = f"timeseries.{param}"
                else:
                    val_mod = [ f"'\"{val}\"'::text" for val in values ]
                    values = "(" + ",".join(val_mod) + ")"
                    param = f"to_json(timeseries.{param})::text"
            if param == "has_role":
                operator = "IN"
                join_operator = "OR"
                if (values[0][0] == '~'):
                    operator = "NOT IN"
                    join_operator = "AND"
                    values[0] = values[0][1:]
                t_r_filter.append(f"organisations.longname {operator} {values}")
                t_r_filter.append(f"organisations.name {operator} {values}")
                t_r_filter.append(f"organisations.city {operator} {values}")
                t_r_filter.append(f"organisations.homepage {operator} {values}")
                t_r_filter.append(f"organisations.contact_url {operator} {values}")
                t_r_filter.append(f"persons.email {operator} {values}")
                t_r_filter.append(f"persons.name {operator} {values}")
                t_r_filter.append(f"persons.orcid {operator} {values}")
                t_r_filter = f" {join_operator} ".join(t_r_filter)
            elif param_long.split('>')[0] == "additional_metadata-":
                t_filter.append(f"{param} IN {values}")
            else:
                t_filter.append(f"timeseries.{param} IN {values}")
        elif param in data_params:
            if param == "daterange":
                start_date = dt.datetime.fromisoformat(values[0])
                stop_date = dt.datetime.fromisoformat(values[1])
                d_filter.append(f"datetime BETWEEN '{start_date}' AND '{stop_date}'")
            elif param in ("flags", "format"):
                # translation of flags should be done in data.crud
                continue
            else:
                d_filter.append(f"data.{param} IN {values}")
        elif endpoint == "variables":
            v_filter.append(f"variables.{param} IN {values}")
        elif param in person_params:
            p_filter.append(f"persons.{param} IN {values}")


    t_filter = " AND ".join(t_filter).replace('[','(').replace(']',')')
    t_r_filter = '(' + "".join(t_r_filter).replace('[','(').replace(']',')') + ')'
    if t_r_filter == '()':
        t_r_filter = ''
    s_c_filter = " AND ".join(s_c_filter).replace('[','(').replace(']',')')
    s_g_filter = " AND ".join(s_g_filter).replace('[','(').replace(']',')')
    d_filter = " AND ".join(d_filter).replace('[','(').replace(']',')')
    v_filter = " AND ".join(v_filter).replace('[','(').replace(']',')')
    p_filter = " AND ".join(p_filter).replace('[','(').replace(']',')')
    filters = {
        "t_filter": t_filter,
        "t_r_filter":t_r_filter,
        "s_c_filter": s_c_filter,
        "s_g_filter": s_g_filter,
        "d_filter": d_filter,
        "v_filter": v_filter,
        "p_filter": p_filter,
    }
    return limit, offset, fields, format, filters


###
# Rasdaman does not run stable!
# --> since GEO PEAS runs in DEBUG mode, its messages flood the error log file!
#
# only activate the following lines if you want to update pages!
###
# also get provenance information
geopeas_services = [ 'topography_srtm', 'ecoregion', 'stable_nightlights',
                     'climatic_zone', 'nox_emissions', 'landcover',
                     'major_road', 'population_density', 'htap_region_tier1' ]
provenance = {}
#for service in geopeas_services:
#    result = requests.get(f"{base_geopeas_url}/{service}/").json()
#    provenance[service] = result['provenance']
#    # major_road has different dict entries
#    tmp_provenance = copy(provenance[service])
#    # htap_region_tier1 does not provide a dict
#    try:
#        keys = [x for x in result['provenance'].keys()]
#        for key in keys:
#            if key not in ['units', 'data_source', 'citation', 'doi']:
#                result['provenance'].pop(key)
#        if len(result['provenance']) == 0:
#            provenance[service] = tmp_provenance
#    except:
#        continue
for service in geopeas_services:
    provenance[service] = "dummy text"

